/* crypto/aes/aes_wrap.c */
/*
 * Written by Dr Stephen N Henson (steve@openssl.org) for the OpenSSL
 * project. Modified by Attaullah Baig (abaig@paypal.com) to wrap/unwrap any
 * size keys (Implemented RFC5649)
 */
/* ====================================================================
 * Copyright (c) 2008 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    licensing@OpenSSL.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#include "cryptlib.h"
#include <openssl/aes.h>
#include <openssl/modes.h>
#include <string.h>
#ifdef OPENSSL_SYS_WIN32
# include <winsock.h>
#else
# include "../../e_os.h"
# if defined(OPENSSL_SYS_SE)
#  include <endian.h>
# else
#  include <netinet/in.h>
# endif
#endif

int AES_wrap_key(AES_KEY *key, const unsigned char *iv,
                 unsigned char *out,
                 const unsigned char *in, unsigned int inlen)
{
    return CRYPTO_128_wrap(key, iv, out, in, inlen, (block128_f) AES_encrypt);
}

#ifdef CISCOSSL_DEPRECATED_UNWRAP_API
int AES_unwrap_key(AES_KEY *key, const unsigned char *iv,
                   unsigned char *out,
                   const unsigned char *in, unsigned int inlen,
                   unsigned char *a_iv)
{
    return CRYPTO_128_unwrap_aiv(key, iv, out, in, inlen,
                             (block128_f) AES_decrypt, a_iv);
}
#else
int AES_unwrap_key(AES_KEY *key, const unsigned char *iv,
                   unsigned char *out,
                   const unsigned char *in, unsigned int inlen)
{
    return CRYPTO_128_unwrap(key, iv, out, in, inlen,
                             (block128_f) AES_decrypt);
}
#endif

static const unsigned char alternate_iv[] = {
    0xA6, 0x59, 0x59, 0xA6,
};

int AES_wrap_key_withpad(AES_KEY *key, unsigned char *out,
                         const unsigned char *in, unsigned int inlen)
{
    int len, nlen, ret = -1;
    unsigned char *input, iv[8];

    if (!inlen)
        return -1;
    len = inlen + (inlen % 8 == 0 ? 0 : (8 - inlen % 8));
    nlen = htonl(inlen);

    input = OPENSSL_malloc(len + 8);
    if (!input)
        return -1;
    memset(input, 0, len + 8);
    memcpy(iv, alternate_iv, 4);
    memcpy(iv + 4, ((unsigned char *)&nlen), 4);

    if (len == 8) {
        memcpy(input, iv, 8);
        memcpy(input + 8, in, inlen);
        AES_encrypt(input, out, key);
        ret = 8 + 8;
    } else {
        memcpy(input, in, inlen);
        ret = AES_wrap_key(key, iv, out, input, len);
    }
    OPENSSL_free(input);
    return ret;
}

int AES_unwrap_key_withpad(AES_KEY *key, unsigned char *out,
                           const unsigned char *in, unsigned int inlen)
{
    unsigned int len, padlen;
    unsigned char a_iv[8], zero_iv[8];

    memset(zero_iv, 0, 8);
    if (inlen & 0x7)
        return -1;
    if (inlen < 16)
        return -1;

    if (inlen == 16) {
        AES_decrypt(in, out, key);
        memcpy(a_iv, out, 8);
        memmove(out, out + 8, 8);
    } else {
        if (CRYPTO_128_unwrap_aiv
            (key, NULL, out, in, inlen, (block128_f) AES_decrypt,
             a_iv) <= 0) {
            return -1;
        }
    }

    if (memcmp(a_iv, alternate_iv, 4))
        return -1;

    memcpy((unsigned char *)&len, a_iv + 4, 4);
    len = ntohl(len);
    inlen -= 8;

    if (len > inlen || len <= (inlen - 8))
        return -1;

    padlen = inlen - len;

    if (padlen && memcmp(zero_iv, out + len, padlen))
        return -1;

    return len;
}
