/* -----------------------------------------------------------------
 * crypto/aes/asm/aes_cbc-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#ifndef AES_DEBUG
# ifndef NDEBUG
#  define NDEBUG
# endif
#endif
#include <assert.h>
#include <openssl/crypto.h>

#include <openssl/aes.h>
#include "aes_locl.h"
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif

/*
 * AES_cbc_encrypt -  encrypt/decrypt using the cipher-block-chaining (CBC) mode of AES
 */
void
AESOCT_cbc_encrypt(const unsigned char *in, unsigned char *out,
                   size_t length, const AES_KEY *key,
                   unsigned char *ivec, const int enc)
{

    uint64_t *iv;
    uint64_t *inp, *outp;
    uint64_t i0, i1, r0 = 0, r1 = 0;
    unsigned long len = length;

    inp = (uint64_t *) in;
    outp = (uint64_t *) out;
    iv = (uint64_t *) ivec;
    CVMX_MT_AES_IV (cvmx_htobe64(iv[0]), 0);
    CVMX_MT_AES_IV (cvmx_htobe64(iv[1]), 1);

    /* Initialise the keys */
    CVMX_MT_AES_KEY (cvmx_htobe64(key->cvmkey[0]), 0);
    CVMX_MT_AES_KEY (cvmx_htobe64(key->cvmkey[1]), 1);
    CVMX_MT_AES_KEY (cvmx_htobe64(key->cvmkey[2]), 2);
    CVMX_MT_AES_KEY (cvmx_htobe64(key->cvmkey[3]), 3);

    CVMX_MT_AES_KEYLENGTH(key->cvm_keylen / 64 - 1);
    i0 = inp[0];
    i1 = inp[1];
    if (AES_ENCRYPT == enc) {
        if (len >= 32) {
	    CVMX_MT_AES_ENC_CBC0 (cvmx_htobe64(i0));
	    CVMX_MT_AES_ENC_CBC1 (cvmx_htobe64(i1));
            // unrolled the loop. The first iteration doesn't store data
            /* The crypto takes 24 cycles so do some stuffs in the gap */
            len -= 16;
            inp += 2;
            outp += 2;
            if (len >= 16) {
                i0 = inp[0];
                i1 = inp[1];
                CVMX_MF_AES_RESULT(r0, 0);
                CVMX_MF_AES_RESULT(r1, 1);
		CVMX_MT_AES_ENC_CBC0 (cvmx_htobe64(i0));
		CVMX_MT_AES_ENC_CBC1 (cvmx_htobe64(i1));

                while (1) {
		    outp[-2] = cvmx_betoh64(r0);
		    outp[-1] = cvmx_betoh64(r1);
                    len -= 16;
                    inp += 2;
                    outp += 2;
                    if (len < 16) {
                        break;
                    }
                    i0 = inp[0];
                    i1 = inp[1];
                    CVMX_PREFETCH(inp, 64);

                    CVMX_MF_AES_RESULT(r0, 0);
                    CVMX_MF_AES_RESULT(r1, 1);
		    CVMX_MT_AES_ENC_CBC0 (cvmx_htobe64(i0));
		    CVMX_MT_AES_ENC_CBC1 (cvmx_htobe64(i1));
                }
            }
            CVMX_MF_AES_RESULT(r0, 0);
            CVMX_MF_AES_RESULT(r1, 1);
	    outp[-2] = cvmx_betoh64(r0);
	    outp[-1] = cvmx_betoh64(r1);
        }

        if (len) {
            if (len <= 16) {
                uint64_t in64[2] = { 0, 0 };
                memcpy(&(in64[0]), inp, len);
		CVMX_MT_AES_ENC_CBC0 (cvmx_htobe64(in64[0]));
		CVMX_MT_AES_ENC_CBC1 (cvmx_htobe64(in64[1]));
                CVMX_MF_AES_RESULT(*(outp), 0);
                CVMX_MF_AES_RESULT(*(outp + 1), 1);
		outp[0]  =  cvmx_betoh64(outp[0]);
		outp[1]  =  cvmx_betoh64(outp[1]);
                memcpy(iv, outp, 16);
            } else {
                uint64_t in64[2] = { 0, 0 };
		CVMX_MT_AES_ENC_CBC0 (cvmx_htobe64(i0));
		CVMX_MT_AES_ENC_CBC1 (cvmx_htobe64(i1));
                CVMX_MF_AES_RESULT(*(outp), 0);
                CVMX_MF_AES_RESULT(*(outp + 1), 1);
		outp[0]  =  cvmx_betoh64(outp[0]);
		outp[1]  =  cvmx_betoh64(outp[1]);
                inp += 2;
                outp += 2;
                len -= 16;
                memcpy(&(in64[0]), inp, len);
		CVMX_MT_AES_ENC_CBC0 (cvmx_htobe64(in64[0]));
		CVMX_MT_AES_ENC_CBC1 (cvmx_htobe64(in64[1]));
                CVMX_MF_AES_RESULT(*(outp), 0);
                CVMX_MF_AES_RESULT(*(outp + 1), 1);
		outp[0]  =  cvmx_betoh64(outp[0]);
		outp[1]  =  cvmx_betoh64(outp[1]);
                memcpy(iv, outp, 16);
            }
        } else {
            memcpy(iv, (outp - 2), 16);
        }
    } else {
        /* Decrypt */
        if (len >= 32) {
	    CVMX_MT_AES_DEC_CBC0 (cvmx_htobe64(i0));
	    CVMX_MT_AES_DEC_CBC1 (cvmx_htobe64(i1));

            len -= 16;
            outp += 2;
            inp += 2;

            if (len >= 16) {
                /* Load ahead */
                i0 = inp[0];
                i1 = inp[1];
                CVMX_MF_AES_RESULT(r0, 0);
                CVMX_MF_AES_RESULT(r1, 1);

		CVMX_MT_AES_DEC_CBC0 (cvmx_htobe64(i0));
		CVMX_MT_AES_DEC_CBC1 (cvmx_htobe64(i1));
                while (1) {
		    outp[-2] = cvmx_betoh64(r0);
		    outp[-1] = cvmx_betoh64(r1);

                    len -= 16;
                    outp += 2;
                    inp += 2;
                    if (len < 16) {
                        break;
                    }
                    i0 = inp[0];
                    i1 = inp[1];
                    CVMX_PREFETCH(inp, 64);
                    CVMX_MF_AES_RESULT(r0, 0);
                    CVMX_MF_AES_RESULT(r1, 1);
		    CVMX_MT_AES_DEC_CBC0 (cvmx_htobe64(i0));
		    CVMX_MT_AES_DEC_CBC1 (cvmx_htobe64(i1));
                }
            }
            /* Fetch the result of the last 16B in the 16B stuff */
            CVMX_MF_AES_RESULT(r0, 0);
            CVMX_MF_AES_RESULT(r1, 1);
            memcpy(iv, (inp - 2), 16);
	    outp[-2] = cvmx_betoh64(r0);
	    outp[-1] = cvmx_betoh64(r1);
        }
        if (len) {
            if (len <= 16) {
                /*
                 * To avoid len>16 tat is when len=17-31 to enter into this
                 * loop
                 */
                uint64_t in64[2] = { 0, 0 };
                memcpy(iv, inp, 16);
                memcpy(&(in64[0]), inp, len);
		CVMX_MT_AES_DEC_CBC0 (cvmx_htobe64(in64[0]));
		CVMX_MT_AES_DEC_CBC1 (cvmx_htobe64(in64[1]));
		CVMX_MF_AES_RESULT (r0, 0);
		CVMX_MF_AES_RESULT (r1, 1);
		outp[0]  =  cvmx_betoh64(r0);
		outp[1]  =  cvmx_betoh64(r1);
            }
        }
    }
}

void AESOCT_ecb_encrypt(const unsigned char *in, unsigned char *out,
                        const AES_KEY *key, const int enc)
{

    assert(in && out && key);
    assert((AES_ENCRYPT == enc) || (AES_DECRYPT == enc));

    if (AES_ENCRYPT == enc)
        AESOCT_encrypt(in, out, key);
    else
        AESOCT_decrypt(in, out, key);
}
