/* -----------------------------------------------------------------
 * crypto/aes/asm/aes_ccm-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <openssl/crypto.h>
#include <openssl/aes.h>
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif
#include "modes_lcl.h"
#include "cvmx-aes.h"

/*
 * octeon_ccm_set_iv()
 */
static
void octeon_ccm_set_iv(uint64_t iv1, uint64_t iv2)
{
    CVMX_MT_AES_IV(iv1, 0);
    CVMX_MT_AES_IV(iv2, 1);
}

/*
 * octeon_ccm_set_key() - set key based on keylength
 */
int octeon_ccm_set_key(EVP_AES_OCT_CCM_CTX * cctx, uint64_t * K, int klen)
{
    uint64_t s0, s1;

    CVMX_MT_AES_KEY(K[0], 0);
    CVMX_MT_AES_KEY(K[1], 1);
    switch (klen) {
    case 128:
        CVMX_MT_AES_KEY(0, 2);
        CVMX_MT_AES_KEY(0, 3);
        break;
    case 192:
        CVMX_MT_AES_KEY(K[2], 2);
        CVMX_MT_AES_KEY(0, 3);
        break;
    case 256:
        CVMX_MT_AES_KEY(K[2], 2);
        CVMX_MT_AES_KEY(K[3], 3);
        break;
    default:
        return -1;
        break;
    }
    CVMX_MT_AES_KEYLENGTH(klen / 64 - 1);

    octeon_ccm_set_iv(0, 0);

    CVMX_MT_AES_ENC_CBC0(cctx->b0high);
    CVMX_MT_AES_ENC_CBC1(cctx->b0low);
    CVMX_MF_AES_RESULT(s0, 0);
    CVMX_MF_AES_RESULT(s1, 1);
    if (!cctx->no_alen) {
        CVMX_MT_AES_ENC_CBC0(cctx->block64_t.val[0]);
        CVMX_MT_AES_ENC_CBC1(cctx->block64_t.val[1]);
        CVMX_MF_AES_RESULT(cctx->r0, 0);
        CVMX_MF_AES_RESULT(cctx->r1, 1);
    } else {                    /* If there is no Associated data */
        cctx->r0 = s0;
        cctx->r1 = s1;
    }
    CVMX_MT_AES_ENC0(cctx->a0high);
    CVMX_MT_AES_ENC1(cctx->a0low);
    CVMX_MF_AES_RESULT(cctx->s0, 0);
    CVMX_MF_AES_RESULT(cctx->s1, 1);

    return 1;
}

/*
 * octeon_ccm_setup_blocks() - setup blocks using AAD and nonce
 */
int octeon_ccm_setup_blocks(uint8_t m, uint8_t l,
                            const uint8_t * ain, uint64_t alen,
                            uint64_t plen, uint64_t * nonce_val,
                            EVP_AES_OCT_CCM_CTX * cctx)
{
    cctx->mac_length = m;
    cctx->nonce.val[0] = nonce_val[0];
    cctx->nonce.val[1] = nonce_val[1];
    cctx->b0_flags.val = 0;
    cctx->b0_flags.s.adata = (alen ? 1 : 0);
    cctx->b0_flags.s.m = (m - 2) / 2;
    cctx->b0_flags.s.l = (l - 1);
    /* A0  = [Flags, Nonce, Counter] */
    cctx->a0_flags.val = 0;
    cctx->a0_flags.s.l = (l - 1);

    /* Make b0, a0 */
    cctx->a0high = (((uint64_t) (cctx->a0_flags.val)) << 56);
    cctx->a0low = 0;
    if (l < 8) {
        int i;
        for (i = 0; i < 7; i++) {
            cctx->a0high |=
                (((uint64_t) cctx->nonce.byte[i]) << (48 - 8 * i));
        }
        for (i = 7; i < 15 - l; i++) {
            cctx->a0low |=
                (((uint64_t) cctx->nonce.byte[i]) << (56 - (i - 7) * 8));
        }
    } else {                    /* l == 8 */
        cctx->a0high |= cctx->nonce.val[0];
    }
    cctx->b0high = (((uint64_t) cctx->b0_flags.val) << 56);
    cctx->b0low = 0;
    if (l < 8) {
        int i;
        for (i = 0; i < 7; i++) {
            cctx->b0high |=
                (((uint64_t) cctx->nonce.byte[i]) << (48 - 8 * i));
        }
        for (i = 7; i < 15 - l; i++) {
            cctx->b0low |=
                (((uint64_t) cctx->nonce.byte[i]) << (56 - (i - 7) * 8));
        }
    } else {                    /* l == 8 */
        cctx->b0high |= cctx->nonce.val[0];
    }
    cctx->b0low |= plen;
    cctx->no_alen = (alen ? 0 : 1);

#define MIN(a, b) (a < b ? (a) : (b))
    /* Make first authentication block */
    if ((alen < 256 * 255)) {
        /* Length (1, (2^16 - 2^8)) Encode value = 0x0001 ... 0xFEFF */
        int i;
        cctx->block64_t.val[0] = 0;
        cctx->block64_t.val[1] = 0;
        cctx->block64_t.s.two_octets = alen;
        cctx->offset = MIN(alen, 14);
        for (i = 0; i < cctx->offset; i++) {
            cctx->block64_t.s.adata[i] = ain[i];
        }
    } else if (((alen >= 0xFF00) && (alen < 0x100000000ULL))) {
        /*
         * Length Encode value 0xFF , 0xFE and the four bytes encoding length
         * of adata
         */
        int i;
        cctx->block64_t.val[0] = 0;
        cctx->block64_t.val[1] = 0;
        cctx->block64_t.s1.two_octets[0] = 0xff;
        cctx->block64_t.s1.two_octets[1] = 0xfe;
	*(uint32_t *)(cctx->block64_t.s1.four_octets) = (uint32_t)alen;	
        cctx->offset = 10;
        for (i = 0; i < cctx->offset; i++) {
            cctx->block64_t.s1.adata[i] = ain[i];
        }
    } else if (((alen >= 0x100000000ULL))) {
        /*
         * Length Encode value 0xFF , 0xFF and the eight bytes encoding
         * length of adata
         */
        int i;
        cctx->block64_t.val[0] = 0;
        cctx->block64_t.val[1] = 0;
        cctx->block64_t.s2.two_octets[0] = 0xff;
        cctx->block64_t.s2.two_octets[1] = 0xff;
	*(uint64_t *)(cctx->block64_t.s2.eight_octets) = alen;	
        cctx->offset = 6;
        for (i = 0; i < cctx->offset; i++) {
            cctx->block64_t.s2.adata[i] = ain[i];
        }
    } else {
        return AES_CCM_INVALID_AUTH_DATA;
    }
    return 1;
}

/*
 * octeon_ccm_encrypt() - perform the CCM encryption
 */
int octeon_ccm_encrypt(const uint8_t * pin, uint64_t plen,
                       const uint8_t * ain, uint64_t alen,
                       uint8_t * cout, uint8_t * aout,
                       EVP_AES_OCT_CCM_CTX * cctx)
{
    uint64_t *aptr, *pptr, *cptr;
    int loop = 0;
    uint64_t data0, data1, r0, r1;
    AES_KEY *key_st;
    int ret;

    if (!cctx->key_set) {
        return -1;
    }

    key_st = &cctx->ks;
    ret =
        octeon_ccm_set_key(cctx, (uint64_t *) key_st->cvmkey,
                           key_st->cvm_keylen);
    if (ret != 1) {
        return ret;
    }

    octeon_ccm_set_iv(cctx->r0, cctx->r1);
    if ((!plen)) {
        loop = 1;
    }
    if ((!alen)) {
        aptr = (uint64_t *) pin;
        alen = plen;
        pptr = (uint64_t *) pin;
        cptr = (uint64_t *) cout;
        loop = 1;
        goto no_auth_data;
    }

    aptr = (uint64_t *) ain;
    pptr = (uint64_t *) pin;
    cptr = (uint64_t *) cout;

    /*
     * We need to offset alen and aptr by the size of adata written in the
     * first authentication block in AES_CCM_setup_blocks()
     */
    alen -= (uint64_t) cctx->offset;
    aptr = (uint64_t *) ((uint8_t *) aptr + cctx->offset);

 no_auth_data:
    do {
        uint64_t data0, data1;
        while (alen > 16) {
            data0 = aptr[0];
            data1 = aptr[1];
            CVMX_PREFETCH(aptr, 64);
            aptr += 2;

            CVMX_MF_AES_RESULT(cctx->r0, 0);
            CVMX_MF_AES_RESULT(cctx->r1, 1);

            CVMX_MT_AES_ENC_CBC0(data0);
            CVMX_MT_AES_ENC_CBC1(data1);

            alen -= 16;
        }

        /* alen is < 16 */
        if (alen) {
            data0 = 0;
            data1 = 0;
            if (alen > 8) {
                data0 = aptr[0];
                aptr++;
                alen -= 8;
                memcpy(&data1, aptr, alen);
            } else {
                memcpy(&data0, aptr, alen);
            }
            CVMX_MF_AES_RESULT(cctx->r0, 0);
            CVMX_MF_AES_RESULT(cctx->r1, 1);

            CVMX_MT_AES_ENC_CBC0(data0);
            CVMX_MT_AES_ENC_CBC1(data1);
        }

        if (alen || plen) {
            CVMX_MF_AES_RESULT(cctx->r0, 0);
            CVMX_MF_AES_RESULT(cctx->r1, 1);
        }

        aptr = (uint64_t *) pin;
        alen = plen;
        loop++;
    } while (loop < 2);

    /* Encryption (CTR mode) */
    data0 = pptr[0];
    data1 = pptr[1];
    pptr += 2;
    cctx->a0low++;

    if ((!plen)) {
        goto no_encrypt_data;
    }
    CVMX_MT_AES_ENC0(cctx->a0high);
    CVMX_MT_AES_ENC1(cctx->a0low);

    while (plen > 16) {
        cctx->a0low++;
        if ((cctx->a0low == 0)) {
            cctx->a0low--;
            cctx->a0high++;
        }

        CVMX_MF_AES_RESULT(r0, 0);
        CVMX_MF_AES_RESULT(r1, 1);

        CVMX_MT_AES_ENC0(cctx->a0high);
        CVMX_MT_AES_ENC1(cctx->a0low);

        r0 ^= data0;
        r1 ^= data1;
        plen -= 16;

        cptr[0] = r0;
        cptr[1] = r1;
        data0 = pptr[0];
        data1 = pptr[1];
        pptr += 2;
        cptr += 2;
    }

    if (plen) {
        pptr -= 2;
        data0 = 0;
        data1 = 0;
        if (plen > 8) {
            data0 = pptr[0];
            pptr++;
            memcpy(&data1, pptr, plen - 8);
        } else {
            memcpy(&data0, pptr, plen);
        }
        CVMX_MF_AES_RESULT(r0, 0);
        CVMX_MF_AES_RESULT(r1, 1);
        r0 ^= data0;
        r1 ^= data1;
        if (plen > 8) {
            cptr[0] = r0;
            cptr++;
            memcpy(cptr, &r1, plen - 8);
        } else {
            memcpy(cptr, &r0, plen);
        }
    }
 no_encrypt_data:
    r0 = (cctx->r0) ^ (cctx->s0);
    r1 = (cctx->r1) ^ (cctx->s1);
    if (cctx->mac_length > 8) {
        *(uint64_t *) aout = r0;
        memcpy(aout + 8, &r1, cctx->mac_length - 8);
    } else {
        memcpy(aout, &r0, cctx->mac_length);
    }
    return AES_CCM_ENCRYPT_SUCCESS;
}

/*
 * octeon_ccm_decrypt() - perform the CCM decryption
 */
int octeon_ccm_decrypt(const uint8_t * cin, uint64_t clen,
                       const uint8_t * ain, uint64_t alen,
                       uint8_t * pout, uint8_t * adata,
                       EVP_AES_OCT_CCM_CTX * cctx)
{
    uint64_t *aptr, *pptr, *cptr;
    int loop = 0;
    uint64_t data0, data1, r0, r1, orig_clen;
    AES_KEY *key_st;
    int ret;

    if (!cctx->key_set) {
        return -1;
    }
    key_st = &cctx->ks;
    ret =
        octeon_ccm_set_key(cctx, (uint64_t *) key_st->cvmkey,
                           key_st->cvm_keylen);
    if (ret != 1) {
        return ret;
    }

    orig_clen = clen;
    pptr = (uint64_t *) pout;
    cptr = (uint64_t *) cin;

    /* Encryption (CTR mode) */
    data0 = cptr[0];
    data1 = cptr[1];
    cptr += 2;
    cctx->a0low++;

    if ((!clen)) {
        loop = 1;
        goto no_decrypt_data;
    }
    CVMX_MT_AES_ENC0(cctx->a0high);
    CVMX_MT_AES_ENC1(cctx->a0low);

    while (clen > 16) {
        cctx->a0low++;
        if ((cctx->a0low == 0)) {
            cctx->a0low--;
            cctx->a0high++;
        }

        CVMX_MF_AES_RESULT(r0, 0);
        CVMX_MF_AES_RESULT(r1, 1);

        CVMX_MT_AES_ENC0(cctx->a0high);
        CVMX_MT_AES_ENC1(cctx->a0low);

        r0 ^= data0;
        r1 ^= data1;
        clen -= 16;

        pptr[0] = r0;
        pptr[1] = r1;
        data0 = cptr[0];
        data1 = cptr[1];
        cptr += 2;
        pptr += 2;
    }

    if (clen) {
        cptr -= 2;
        data0 = 0;
        data1 = 0;
        if (clen > 8) {
            data0 = cptr[0];
            cptr++;
            memcpy(&data1, cptr, clen - 8);
        } else {
            memcpy(&data0, cptr, clen);
        }
        CVMX_MF_AES_RESULT(r0, 0);
        CVMX_MF_AES_RESULT(r1, 1);
        r0 ^= data0;
        r1 ^= data1;
        if (clen > 8) {
            pptr[0] = r0;
            pptr++;
            memcpy(pptr, &r1, clen - 8);
        } else {
            memcpy(pptr, &r0, clen);
        }
    }

    /* Authentication Work */
 no_decrypt_data:
    clen = orig_clen;
    octeon_ccm_set_iv(cctx->r0, cctx->r1);
    if ((!alen)) {
        aptr = (uint64_t *) pout; /* Now, it contains plain data */
        alen = clen;
        loop = 1;
        goto no_auth_data;
    }

    aptr = (uint64_t *) ain;
    pptr = (uint64_t *) pout;

    /*
     * We need to offset alen and aptr by the size of adata written in the
     * first authentication block in AES_CCM_setup_blocks()
     */

    alen -= (uint64_t) cctx->offset;
    aptr = (uint64_t *) ((uint8_t *) aptr + cctx->offset);

 no_auth_data:
    do {
        uint64_t data0, data1;
        while (alen > 16) {
            data0 = aptr[0];
            data1 = aptr[1];
            CVMX_PREFETCH(aptr, 64);
            aptr += 2;

            CVMX_MF_AES_RESULT(cctx->r0, 0);
            CVMX_MF_AES_RESULT(cctx->r1, 1);

            CVMX_MT_AES_ENC_CBC0(data0);
            CVMX_MT_AES_ENC_CBC1(data1);

            alen -= 16;
        }

        /* alen is <= 16 */
        if (alen) {
            data0 = 0;
            data1 = 0;
            if (alen > 8) {
                data0 = aptr[0];
                aptr++;
                alen -= 8;
                memcpy(&data1, aptr, alen);
            } else {
                memcpy(&data0, aptr, alen);
            }
            CVMX_MF_AES_RESULT(cctx->r0, 0);
            CVMX_MF_AES_RESULT(cctx->r1, 1);

            CVMX_MT_AES_ENC_CBC0(data0);
            CVMX_MT_AES_ENC_CBC1(data1);
        }

        if (alen || clen) {
            CVMX_MF_AES_RESULT(cctx->r0, 0);
            CVMX_MF_AES_RESULT(cctx->r1, 1);
        }
        aptr = (uint64_t *) pout;
        alen = clen;
        loop++;
    } while (loop < 2);

    r0 = cctx->s0 ^ cctx->r0;
    r1 = cctx->s1 ^ cctx->r1;
    if (cctx->mac_length > 8) {
        if (r0 != *(uint64_t *) adata) {
            return AES_CCM_AUTH_CHECK_FAILED;
        }
        if (memcmp(&r1, (adata + 8), (cctx->mac_length - 8))) {
            return AES_CCM_AUTH_CHECK_FAILED;
        }
        return AES_CCM_DECRYPT_SUCCESS;
    }
    if (memcmp(&r0, adata, cctx->mac_length)) {
        return AES_CCM_AUTH_CHECK_FAILED;
    }
    return AES_CCM_DECRYPT_SUCCESS;
}
