/* -----------------------------------------------------------------
 * crypto/aes/asm/aes_gcm-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <openssl/crypto.h>
#include <openssl/aes.h>
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif
#include "modes_lcl.h"
#include "cvmx-aes.h"

/*
 * The output registers need to be set in clobbered mode, so as to indicate
 * compiler about not sharing the same input and output registers
 */
#define CVM_AES_RD_RESULT_WR_DATA(in1, in2, out1, out2) asm volatile\
                                              (".set noreorder    \n" \
                                               "dmfc2 %[r1],0x0100\n" \
                                               "dmfc2 %[r2],0x0101\n" \
                                               "dmtc2 %[r3],0x010a\n" \
                                               "dmtc2 %[r4],0x310b\n" \
                                               ".set reorder      \n" \
                                               : [r1] "=&d" (in1) , [r2] "=&d" (in2) \
                                               : [r3] "d" (out1),  [r4] "d" (out2) )

/*
 * GHASH_init() - init the GHASH registers
 */
static inline void GHASH_init(uint16_t polynomial, void *multiplier)
{
    // Init Galois multiplier
    CVMX_MT_GFM_POLY((uint64_t) polynomial);

    // Init multiplier
    if (multiplier) {
        uint64_t *poly = (uint64_t *) multiplier;
        CVMX_MT_GFM_MUL(poly[0], 0);
        CVMX_MT_GFM_MUL(poly[1], 1);
    }
    // Multiply by 0 to clear result
    CVMX_MT_GFM_RESINP(0, 0);
    CVMX_MT_GFM_RESINP(0, 1);
}

/*
 * GHASH_update() - perform the XOR/MUL
 */
static inline void GHASH_update(uint64_t * data)
{
    // Feed data to the hash
    CVMX_MT_GFM_XOR0 (cvmx_htobe64(data[0]));
    CVMX_MT_GFM_XORMUL1 (cvmx_htobe64(data[1]));
}

/*
 * GHASH_finish() - last XOR/MUL and read result of GHASH
 */
static inline void GHASH_finish(uint64_t alen, uint64_t clen, void *res)
{
    block16_t *result = (block16_t *) res;

    // Feed lengths into the hash
    CVMX_MT_GFM_XOR0(alen);
    CVMX_MT_GFM_XORMUL1(clen);

    // Read the result (note stalls here until MPY is finished)
    CVMX_MF_GFM_RESINP(result->val64[0], 0);
    CVMX_MF_GFM_RESINP(result->val64[1], 1);

    result->val64[0]  =  cvmx_betoh64(result->val64[0]);
    result->val64[1]  =  cvmx_betoh64(result->val64[1]);
}

/*
 * octeon_gcm_set_iv() - setup the IV and save initial counter
 */
int
octeon_gcm_set_iv(const uint8_t * iv, uint32_t ivlen,
                  EVP_AES_OCT_GCM_CTX * gctx)
{
    int i;

    if (!gctx->key_set) {
        return AES_GCM_KEY_NOT_SET;
    }
    // Generate Y_0 as follows:
    //
    // / IV || 0^31 || 1
    // Y_0 = |
    // \ GHASH(H,{},IV)

    if (ivlen == (96 / 8)) {

        // Y_O = IV || 0^31 || 1
	gctx->Y_i.val64[0] = cvmx_htobe64(*((uint64_t *)iv));
	gctx->Y_i.val32.w2 = cvmx_htobe32(*(((uint32_t *)iv + 2)));
        gctx->Y_i.val32.w3 = 1;
    } else {
        int len = ivlen;
        block16_t *ivb_ptr = (block16_t *) iv;
        block16_t last_block;

        // Init GHASH
        GHASH_init(0xe100, &gctx->H.val64[0]);

        // Run GHASH for blocks 1 .. n-1
        for (i = 0; i < (len - 16); i += 16) {
            GHASH_update((uint64_t *) ivb_ptr);
            ivb_ptr++;
        }

        len = len - i;

        // Run GHASH for the last block
        last_block.val64[0] = 0;
        last_block.val64[1] = 0;
        for (i = 0; i < len; i++) {
            last_block.val8[i] = ivb_ptr->val8[i];
        }
        GHASH_update(last_block.val64);

        // Finish GHASH
        GHASH_finish(0, ivlen * 8, &gctx->Y_i.val64[0]);
	gctx->Y_i.val64[0] = cvmx_htobe64(gctx->Y_i.val64[0]);
	gctx->Y_i.val64[1] = cvmx_htobe64(gctx->Y_i.val64[1]);
    }
    gctx->Y_0 = gctx->Y_i.val32.w3;

    // Y_1
    gctx->Y_i.val32.w3++;

    GHASH_init(0xe100, &gctx->H.val64[0]);
    return AES_GCM_SUCCESS;
}

/*
 * octeon_gmac_set_key() - guts of the key setup
 */

static void octeon_gcm_gmac_set_key(EVP_AES_OCT_GCM_CTX * gctx)
{
    // Init key

    CVMX_MT_AES_KEY(cvmx_htobe64(gctx->K.val64[3]), 3);
    CVMX_MT_AES_KEY(cvmx_htobe64(gctx->K.val64[2]), 2);
    CVMX_MT_AES_KEY(cvmx_htobe64(gctx->K.val64[1]), 1);
    CVMX_MT_AES_KEY(cvmx_htobe64(gctx->K.val64[0]), 0);

    CVMX_MT_AES_KEYLENGTH(gctx->keylen - 1);
}

/**
 * octeon_gcm_encrypt() - perform the encrypt
 * @param *ain pointer to auth-only data
 * @param alen size of auth-only data in bytes
 * @param *pin pointer to plain-text data (to be encrypted)
 * @param plen size of plain-text data in bytes
 * @param *out pointer to encrypted data (output)
 * @param *gctx pointer AES-GCM context
 *
 */
int
octeon_gcm_encrypt(const uint8_t * pin, uint32_t plen,
                   const uint8_t * ain, uint32_t alen,
                   uint8_t * out, EVP_AES_OCT_GCM_CTX * gctx)
{
    int len, i;
    block16_t *iptr, *optr;
    block16_t input, result, mask;

    // Pre-fetch first cache line
    CVMX_PREFETCH0(pin);

    // Start encrypting 2nd counter block
    // (to be used to XOR the first input data block)
    CVMX_MT_AES_ENC0(gctx->Y_i.val64[0]);
    CVMX_MT_AES_ENC1(gctx->Y_i.val64[1]);

    if (alen == 0) {
        goto encrypt;
    }
    // ==============
    // auth-only loop
    // ==============

    iptr = (block16_t *) ain;
    len = alen;

    // Run GHASH for auth blocks 1 .. n-1
    for (i = 0; i < (len - 16); i += 16) {
        // Read auth data block
        CVMX_LOADUNA_INT64(input.val64[0], iptr, 0);
        CVMX_LOADUNA_INT64(input.val64[1], iptr++, 8);

        // GHASH Update
        CVMX_MT_GFM_XOR0 (cvmx_htobe64(input.val64[0]));
        CVMX_MT_GFM_XORMUL1 (cvmx_htobe64(input.val64[1]));
    }

    len = alen - i;

    // GHASH Update for the last auth block
    input.val64[0] = 0;
    input.val64[1] = 0;
    for (i = 0; i < len; i++) {
        input.val8[i] = iptr->val8[i];
    }
    CVMX_MT_GFM_XOR0 (cvmx_htobe64(input.val64[0]));
    CVMX_MT_GFM_XORMUL1 (cvmx_htobe64(input.val64[1]));


 encrypt:
    // =================
    // encrypt-auth loop
    // =================

    iptr = (block16_t *) pin;
    optr = (block16_t *) out;
    len = plen;

    if (len < 16) {
        goto encrypt_loop_done;
    }
    do {
        // Pre-fetch next cache-line
        CVMX_PREFETCH128(iptr);

        // Update remaining length
        len -= 16;

        // Increment counter value
        gctx->Y_i.val32.w3++;

        // Read input data block
        CVMX_LOADUNA_INT64(input.val64[0], iptr, 0);
        CVMX_LOADUNA_INT64(input.val64[1], iptr++, 8);

        // Read previous result & start encrypting next counter block
        CVM_AES_RD_RESULT_WR_DATA(result.val64[0], result.val64[1],
                                  gctx->Y_i.val64[0], gctx->Y_i.val64[1]);
        // XOR input with AES result
	result.val64[0] ^= cvmx_htobe64(input.val64[0]);
	result.val64[1] ^= cvmx_htobe64(input.val64[1]);

        // Feed XOR result to GHASH
        CVMX_MT_GFM_XOR0(result.val64[0]);
        CVMX_MT_GFM_XORMUL1(result.val64[1]);

        // Write output
	CVMX_STOREUNA_INT64 (cvmx_betoh64(result.val64[0]), optr, 0);
	CVMX_STOREUNA_INT64 (cvmx_betoh64(result.val64[1]), optr++, 8);

    } while (len >= 16);

    // ====================
    // encrypt-auth trailer
    // ====================
 encrypt_loop_done:

    if (len == 0) {
        goto encrypt_done;
    }
    mask.val64[0] = 0;
    mask.val64[1] = 0;

    // Get last input block
    for (i = 0; i < len; i++) {
        input.val8[i] = iptr->val8[i];
        mask.val8[i] = 0xff;
    }

    // Read last AES result
    CVMX_MF_AES_RESULT(result.val64[0], 0);
    CVMX_MF_AES_RESULT(result.val64[1], 1);

    // XOR input with last AES result
    result.val64[0] ^= cvmx_htobe64(input.val64[0]);
    result.val64[1] ^= cvmx_htobe64(input.val64[1]);

    // Mask last XOR result
    result.val64[0] &= cvmx_htobe64(mask.val64[0]);
    result.val64[1] &= cvmx_htobe64(mask.val64[1]);

    // Feed last XOR result to GHASH
    CVMX_MT_GFM_XOR0(result.val64[0]);
    CVMX_MT_GFM_XORMUL1(result.val64[1]);

    // Write out last result
#ifdef B_ENDIAN
    for (i = 0; i < len; i++)
      optr->val8[i] = result.val8[i];
#else
  result.val64[0] = cvmx_betoh64(result.val64[0]);
  result.val64[1] = cvmx_betoh64(result.val64[1]);
  for (i = 0; i < len; i++)
    optr->val8[i] = result.val8[i];
#endif
 encrypt_done:
    return AES_GCM_SUCCESS;
}

static int
octeon_gmac_ctx_tag(const uint8_t * ain, uint32_t alen, uint8_t * tag,
                    EVP_AES_OCT_GCM_CTX * aes_ctx)
{
    block16_t *iptr;
    block16_t input, result;
    uint32_t len;
    int i;

    /* set up AES key */
    octeon_gcm_gmac_set_key(aes_ctx);

    GHASH_init(0xe100, &aes_ctx->H.val64[0]);

    if (alen == 0)
        goto auth_done;

    iptr = (block16_t *) ain;
    len = alen;

    // Run GHASH for auth blocks 1 .. n-1
    for (i = 0; i < (len - 16); i += 16) {
        // Read auth data block
        CVMX_LOADUNA_INT64(input.val64[0], iptr, 0);
        CVMX_LOADUNA_INT64(input.val64[1], iptr++, 8);

        // GHASH Update
        CVMX_MT_GFM_XOR0 (cvmx_htobe64(input.val64[0]));
        CVMX_MT_GFM_XORMUL1 (cvmx_htobe64(input.val64[1]));
    }

    len = alen - i;

    // GHASH Update for the last auth block
    input.val64[0] = 0;
    input.val64[1] = 0;
    for (i = 0; i < len; i++) {
        input.val8[i] = iptr->val8[i];
    }
    CVMX_MT_GFM_XOR0 (cvmx_htobe64(input.val64[0]));
    CVMX_MT_GFM_XORMUL1 (cvmx_htobe64(input.val64[1]));

 auth_done:
    // Feed lengths to GHASH
    CVMX_MT_GFM_XOR0((uint64_t) alen * 8);
    CVMX_MT_GFM_XORMUL1((uint64_t) 0x0ull);

    aes_ctx->Y_i.val32.w3 = aes_ctx->Y_0;

    // Encrypt first counter block (Y_0)
    CVMX_MT_AES_ENC0(aes_ctx->Y_i.val64[0]);
    CVMX_MT_AES_ENC1(aes_ctx->Y_i.val64[1]);

    CVMX_MF_GFM_RESINP(result.val64[0], 0);
    CVMX_MF_GFM_RESINP(result.val64[1], 1);

    // Read AES result
    CVMX_MF_AES_RESULT(input.val64[0], 0);
    CVMX_MF_AES_RESULT(input.val64[1], 1);

    // Construct tag
    result.val64[0] ^= input.val64[0];
    result.val64[1] ^= input.val64[1];

    // Write out tag
    CVMX_STOREUNA_INT64 (cvmx_betoh64(result.val64[0]), tag, 0);
    CVMX_STOREUNA_INT64 (cvmx_betoh64(result.val64[1]), tag, 8);

    return 1;
}

static int octeon_gmac_tag(EVP_AES_OCT_GCM_CTX * gctx, uint8_t * tag)
{
    int ret;
    AES_KEY *key_st;

    key_st = &gctx->ks;
    ret =
        octeon_gcm_set_key(gctx, key_st,
                           (const unsigned char *)key_st->cvmkey);
    if (ret <= 0) {
        return ret;
    }
    ret = octeon_gcm_set_iv(gctx->iv, gctx->ivlen, gctx);
    if (ret <= 0) {
        return ret;
    }
    return octeon_gmac_ctx_tag(gctx->ain, gctx->alen, tag, gctx);
}

/**
 * octeon_gcm_final() - calculate and return the tag
 * @param plen size of plain-text data in bytes
 * @param alen size of auth-only data in bytes
 * @param *tag pointer to 16-byte tag value (output)
 * @param *gctx pointer AES-GCM context
 *
 */
int
octeon_gcm_final(uint32_t plen, uint32_t alen, uint8_t * tag,
                 EVP_AES_OCT_GCM_CTX * gctx)
{
    block16_t input, result;

    /* if no plaintext, rebuild tag from scratch */
    if (!plen) {
        return (octeon_gmac_tag(gctx, tag));
    } else {

        uint32_t Y_t;

        // Restore 1st counter value
        Y_t = gctx->Y_i.val32.w3;
        gctx->Y_i.val32.w3 = gctx->Y_0;

        // Encrypt first counter block (Y_0)
        CVMX_MT_AES_ENC0(gctx->Y_i.val64[0]);
        CVMX_MT_AES_ENC1(gctx->Y_i.val64[1]);

        // Feed lengths to GHASH
        CVMX_MT_GFM_XOR0((uint64_t) alen * 8);
        CVMX_MT_GFM_XORMUL1((uint64_t) plen * 8);

        gctx->Y_i.val32.w3 = Y_t;

        // Read AES result
        CVMX_MF_AES_RESULT(input.val64[0], 0);
        CVMX_MF_AES_RESULT(input.val64[1], 1);

        // Read Galois result (GHASH_finish)
        // (have to stall here -- there is
        // nothing else we can do)
        CVMX_MF_GFM_RESINP(result.val64[0], 0);
        CVMX_MF_GFM_RESINP(result.val64[1], 1);

        // Construct tag
        result.val64[0] ^= input.val64[0];
        result.val64[1] ^= input.val64[1];

        // Write out tag
	CVMX_STOREUNA_INT64 (cvmx_betoh64(result.val64[0]), tag, 0);
	CVMX_STOREUNA_INT64 (cvmx_betoh64(result.val64[1]), tag, 8);
    }

    return AES_GCM_SUCCESS;
}

/**
 * octeon_gcm_decrypt() - perform the decrypt
 * @param *ain pointer to auth-only data
 * @param alen size of auth-only data in bytes
 * @param *cin pointer to encrypted-text data (to be decrypted)
 * @param clen size of plain-text data in bytes
 * @param *out pointer to decrypted data (output)
 * @param *gctx pointer AES-GCM context
 *
 */
int
octeon_gcm_decrypt(const uint8_t * cin, uint32_t clen,
                   const uint8_t * ain, uint32_t alen,
                   uint8_t * out, EVP_AES_OCT_GCM_CTX * gctx)
{
    int len, i;
    block16_t *iptr, *optr;
    block16_t input, result, mask;

    // Pre-fetch first cache line
    CVMX_PREFETCH0(cin);

    // Start encrypting block
    // (to be used to XOR the first input data block)
    CVMX_MT_AES_ENC0(gctx->Y_i.val64[0]);
    CVMX_MT_AES_ENC1(gctx->Y_i.val64[1]);

    if (alen == 0) {
        goto decrypt;
    }
    // ==============
    // auth-only loop
    // ==============

    iptr = (block16_t *) ain;
    len = alen;

    // Run GHASH for auth blocks 1 .. n-1
    for (i = 0; i < (len - 16); i += 16) {
        // Read auth data block
        CVMX_LOADUNA_INT64(input.val64[0], iptr, 0);
        CVMX_LOADUNA_INT64(input.val64[1], iptr++, 8);

        // Send to GHASH
        CVMX_MT_GFM_XOR0(cvmx_htobe64(input.val64[0]));
        CVMX_MT_GFM_XORMUL1(cvmx_htobe64(input.val64[1]));
    }

    len = alen - i;

    // Run GHASH for the last auth block
    input.val64[0] = 0;
    input.val64[1] = 0;
    for (i = 0; i < len; i++) {
        input.val8[i] = iptr->val8[i];
    }
    CVMX_MT_GFM_XOR0(cvmx_htobe64(input.val64[0]));
    CVMX_MT_GFM_XORMUL1(cvmx_htobe64(input.val64[1]));

 decrypt:
    // =================
    // decrypt-auth loop
    // =================

    iptr = (block16_t *) cin;
    optr = (block16_t *) out;
    len = clen;

    if (len < 16) {
        goto decrypt_loop_done;
    }
    do {

        // Pre-fetch next cache-line
        CVMX_PREFETCH128(iptr);

        // Update remaining length
        len -= 16;

        // Increment counter value
        gctx->Y_i.val32.w3++;

        // Read input data block
        CVMX_LOADUNA_INT64(input.val64[0], iptr, 0);
        CVMX_LOADUNA_INT64(input.val64[1], iptr++, 8);

        // Read previous result & start encrypting next counter block
        CVM_AES_RD_RESULT_WR_DATA(result.val64[0], result.val64[1],
                                  gctx->Y_i.val64[0], gctx->Y_i.val64[1]);
        // Feed XOR result to GHASH
	CVMX_MT_GFM_XOR0 (cvmx_htobe64(input.val64[0]));
	CVMX_MT_GFM_XORMUL1 (cvmx_htobe64(input.val64[1]));

	// XOR input with AES result
	result.val64[0] ^= cvmx_htobe64(input.val64[0]);
	result.val64[1] ^= cvmx_htobe64(input.val64[1]);

	// Write output
	CVMX_STOREUNA_INT64 (cvmx_betoh64(result.val64[0]), optr, 0);
	CVMX_STOREUNA_INT64 (cvmx_betoh64(result.val64[1]), optr++, 8);

    } while (len >= 16);

    // ====================
    // decrypt-auth trailer
    // ====================
 decrypt_loop_done:

    if (len == 0) {
        goto decrypt_done;
    }
    mask.val64[0] = 0;
    mask.val64[1] = 0;

    input.val64[0] = 0;
    input.val64[1] = 0;

    // Get last input block
    for (i = 0; i < len; i++) {
        input.val8[i] = iptr->val8[i];
        mask.val8[i] = 0xff;
    }

    // Feed last XOR result to GHASH
    CVMX_MT_GFM_XOR0 (cvmx_htobe64(input.val64[0]));
    CVMX_MT_GFM_XORMUL1 (cvmx_htobe64(input.val64[1]));

    // Read last AES result
    CVMX_MF_AES_RESULT(result.val64[0], 0);
    CVMX_MF_AES_RESULT(result.val64[1], 1);

    // XOR input with last AES result
    result.val64[0] ^= cvmx_htobe64(input.val64[0]);
    result.val64[1] ^= cvmx_htobe64(input.val64[1]);

    // Mask last XOR result
    result.val64[0] &= cvmx_htobe64(mask.val64[0]);
    result.val64[1] &= cvmx_htobe64(mask.val64[1]);

    // Write out last result
#ifdef B_ENDIAN
    for (i = 0; i < len; i++)
        optr->val8[i] = result.val8[i];
#else
  result.val64[0] = cvmx_betoh64(result.val64[0]);
  result.val64[1] = cvmx_betoh64(result.val64[1]);
  for (i = 0; i < len; i++)
    optr->val8[i] = result.val8[i];
#endif

 decrypt_done:

    return AES_GCM_SUCCESS;
}

/*
 * octeon_gcm_set_key() - setup the key parms in gctx call set_key and calc H.
 */
int octeon_gcm_set_key(EVP_AES_OCT_GCM_CTX * gctx, AES_KEY *key_st,
                       const unsigned char *key)
{

    uint64_t *kptr = (uint64_t *) key;

    gctx->K.val64[0] = kptr[0];
    gctx->K.val64[1] = kptr[1];

    switch (key_st->cvm_keylen) {
    case 128:
        gctx->K.val64[2] = 0;
        gctx->K.val64[3] = 0;
        break;
    case 192:
        gctx->K.val64[2] = kptr[2];
        gctx->K.val64[3] = 0;
        break;
    case 256:
        gctx->K.val64[2] = kptr[2];
        gctx->K.val64[3] = kptr[3];
        break;
    default:
        return (AES_GCM_INVALID_KEYLENGTH);
    }

    gctx->keylen = key_st->cvm_keylen / 64;
    octeon_gcm_gmac_set_key(gctx);

    // Run key schedule and get H
    CVMX_MT_AES_ENC0(0);
    CVMX_MT_AES_ENC1(0);

    CVMX_MF_AES_RESULT(gctx->H.val64[0], 0);
    CVMX_MF_AES_RESULT(gctx->H.val64[1], 1);
    return 1;
}
