/*------------------------------------------------------------------
 * include/cavium/cvmx-aes.h - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * February 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <openssl/aes.h>
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif
#include <openssl/ossl_typ.h>
#include <openssl/evp.h>

typedef struct {
    AES_KEY ks;		/* AES key schedule to use */
    uint64_t ctrblk[2];
    uint64_t iv[2];
    uint64_t reserved;
    uint64_t key_set;
  } EVP_AES_OCT_CTR_CTX;

/*
 * XTS AES APIs
 * Based on IEEE Draft P1619/D11 December 2006
 */
typedef struct{
    AES_KEY ks1;		/* AES key schedule to use */
    AES_KEY ks2;		/* AES key schedule to use */
    uint64_t key1[4];
    uint64_t key2[4];
    /* last tweak */
    uint64_t tweak[2];
    uint32_t keylen;
    uint32_t state;
    uint32_t key_set;
} EVP_AES_OCT_XTS_CTX;

#define XTS_AES_SUCCESS 0
#define XTS_AES_NULL_POINTER_ARGUMENT -1
#define XTS_AES_INVALID_KEYLENGTH -2
#define XTS_AES_INVALID_CTX -3
#define XTS_AES_INVALID_DATALENGTH -4

//return codes
#define AES_CCM_ENCRYPT_SUCCESS    1
#define AES_CCM_DECRYPT_SUCCESS    1
#define AES_CCM_INVALID_AUTH_DATA -1
#define AES_CCM_AUTH_CHECK_FAILED -2

// Represents First Byte of the Authentication Block and Encryptio Block 
typedef union {
   uint8_t val;
#ifdef L_ENDIAN
   struct {
      uint8_t l        : 3;
      uint8_t m        : 3;
      uint8_t adata    : 1;
      uint8_t reserved : 1;
   } s;
#else
   struct {
      uint8_t reserved : 1;
      uint8_t adata    : 1;
      uint8_t m        : 3;
      uint8_t l        : 3;
   } s;
#endif
} flags_t;

//Represents Authentication length Encode Block
typedef union {
   uint64_t val[2];
   struct {
      uint16_t two_octets;
      uint8_t adata[14];
   } s;
   struct {
      uint8_t two_octets[2];
      uint8_t four_octets[4];
      uint8_t  adata[10];
   }s1;
   struct {
      uint8_t two_octets[2];
      uint8_t eight_octets[8];
      uint8_t  adata[6];
   }s2;
   uint8_t byte[16];
} auth_len_encode1_t;

typedef auth_len_encode1_t nonce_t;

  typedef union {
    uint64_t val64[2];
#ifdef L_ENDIAN
    struct {
        uint32_t  w1;
        uint32_t  w0;
        uint32_t  w3;
        uint32_t  w2;
    } val32;
#else
    struct {
        uint32_t  w0;
        uint32_t  w1;
        uint32_t  w2;
        uint32_t  w3;
    } val32;
#endif
    uint8_t val8[16];
  } block16_t;

  typedef union {
    uint64_t val64[4];
    uint32_t val32[8];
    uint8_t val8[32];
  } block32_t;

  typedef struct {
    AES_KEY ks;		/* AES key schedule to use */
    // Counter value Y_i (128 bits)
    block16_t Y_i;

    // AES Key (128, 192, or 256 bits)
    block32_t K;

    // H (128 bits)
    block16_t H;

    // (used at the end to XOR with GHASH output to form auth tag)
    uint32_t Y_0;

    // AES key length
    uint32_t keylen;

    // state
      uint64_t plen;
      const uint8_t *ain;
      uint64_t alen;
      uint64_t total_len;
      int key_set;		/* Set if key initialised */
      int key_init;		/* Set if key initialised */
      int iv_set;		/* Set if an iv is set */
      int ivlen;		/* IV length */
      int taglen;
      int iv_gen;		/* It is OK to generate IVs */
      int tls_aad_len;	/* TLS AAD length */
      int block_iga5;
      uint64_t iv_95ctr;
      uint64_t iv_100ctr;
      unsigned char *iv;	/* Temporary IV store */
      unsigned char aad[EVP_MAX_BLOCK_LENGTH];
      unsigned char buf[EVP_MAX_BLOCK_LENGTH];/* saved partial block */
      GCM128_CONTEXT gcm;
      void (*evp_gcm_ctr_status_cb_fn) (EVP_CIPHER_CTX *ctx,  gcm_ctr_status_t status);
      void *ctr_context;
  } EVP_AES_OCT_GCM_CTX;


#define AES_GCM_KEY_DONE         0xAC
#define AES_GCM_IV_DONE          0xBD

  // Return codes
#define AES_GCM_SUCCESS              1
#define AES_GCM_INVALID_KEYLENGTH   -4
#define AES_GCM_KEY_NOT_SET         -2
#define AES_GMAC_SUCCESS             1

//AES CCM Context Structure
typedef struct  {
      AES_KEY ks;		/* AES key schedule to use */
      uint64_t b0high;
      uint64_t b0low;
      uint64_t a0high;
      uint64_t a0low;
      uint64_t s0;
      uint64_t s1;
      uint64_t r0;
      uint64_t r1;
      flags_t b0_flags;
      flags_t a0_flags;
      auth_len_encode1_t block64_t;
      nonce_t nonce;	
      uint32_t offset;
      uint32_t mac_length;
      uint8_t no_alen;
      uint64_t iv[2];
      uint64_t plen;
      const uint8_t *ain;
      uint64_t alen;
      uint32_t ivlen;
      int key_set;		/* Set if key initialised */
      int iv_set;		/* Set if an iv is set */
      int tag_set;		/* Set if tag is valid */
      int len_set;		/* Set if message length set */
      int L, M;		/* L and M parameters from RFC3610 */
      CCM128_CONTEXT ccm;
} EVP_AES_OCT_CCM_CTX;

int octeon_ccm_set_key(EVP_AES_OCT_CCM_CTX *cctx, uint64_t *K, int klen);
int octeon_ccm_setup_blocks(uint8_t m, uint8_t l,
                         const uint8_t *ain, uint64_t alen, 
			 uint64_t plen, uint64_t *nonce_val, 
			 EVP_AES_OCT_CCM_CTX * aes_ctx);
int octeon_ccm_decrypt(const uint8_t *cin, uint64_t clen,
                    const uint8_t *ain, uint64_t alen,
		    uint8_t *pout, uint8_t *adata, 
		    EVP_AES_OCT_CCM_CTX *aes_ctx);
int octeon_ccm_encrypt(const uint8_t *pin, uint64_t plen,
                        const uint8_t *ain, uint64_t alen,
			uint8_t *cout, uint8_t *aout, 
			EVP_AES_OCT_CCM_CTX * aes_ctx);
void octeon_ctr_initialize(uint64_t * key, uint32_t key_len);
void octeon_ctr_encrypt(EVP_AES_OCT_CTR_CTX *cctx, uint64_t *out,
	 	        uint64_t *in, size_t len);
int octeon_gcm_set_iv (const uint8_t * iv, uint32_t ivlen, 
                       EVP_AES_OCT_GCM_CTX * gctx);
int octeon_gcm_init_key(EVP_CIPHER_CTX *ctx, const unsigned char *key,
                        const unsigned char *iv, int enc);
int octeon_gcm_set_key (EVP_AES_OCT_GCM_CTX *gctx, AES_KEY *key_st, 
                        const unsigned char *key);
int octeon_gcm_encrypt(const uint8_t * pin, uint32_t plen,
                       const uint8_t * ain, uint32_t alen,
		       uint8_t * out, EVP_AES_OCT_GCM_CTX * gctx);
int octeon_gcm_decrypt(const uint8_t * cin, uint32_t clen,
                       const uint8_t * ain, uint32_t alen,
		       uint8_t * out, EVP_AES_OCT_GCM_CTX * gctx);
int octeon_gcm_final(uint32_t plen, uint32_t alen, uint8_t * tag, 
                     EVP_AES_OCT_GCM_CTX * gctx);
void octeon_xts_set_key(EVP_AES_OCT_XTS_CTX *xctx);
int octeon_xts_decrypt (const uint8_t *ct,uint32_t len,uint8_t *pt,
                        EVP_AES_OCT_XTS_CTX *ctx);
int octeon_xts_encrypt(const uint8_t *pt,uint32_t len,uint8_t *ct,
                       EVP_AES_OCT_XTS_CTX *ctx);


