/* -----------------------------------------------------------------
 * crypto/ec/ec_donna_meth.c - Implements EC_METHOD for curve25519 
 *
 * November, 2015
 *
 * Copyright (c) 2015 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <string.h>
#include <openssl/err.h>

#ifdef OPENSSL_FIPS
# include <openssl/fips.h>
#endif

#include "ec_lcl.h"

#ifndef OPENSSL_NO_C25519
/*
 * This is the length of the curve25519 key in bytes
 */
#define C25519_R_LEN  32 

const EC_METHOD *EC_c25519_method(void)
{
    static const EC_METHOD ret = {
	0,
        NID_x25519,
        ec_c25519_init,
        0,
        0,
        ec_c25519_copy,
        0,
        0,
        ec_c25519_get_degree,
        0,
        ec_c25519_point_init,
        ec_c25519_point_finish,
        0,
        ec_c25519_point_copy,
        0,
        0,
        0,
        0,
        0,
        0, 
	ec_c25519_point2oct, 
	ec_c25519_oct2point, 
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0 
    };

    return &ret;
}

int ec_c25519_init(EC_GROUP *group)
{
    return 1;
}

int ec_c25519_point_init(EC_POINT *point)
{
    BN_init(&point->X);
    return 1;
}

void ec_c25519_point_finish(EC_POINT *point)
{
    BN_free(&point->X);
}

int ec_c25519_point_copy(EC_POINT *dest, const EC_POINT *src)
{
    if (!BN_copy(&dest->X, &src->X)) return 0;
    return 1;
}

int ec_c25519_copy(EC_GROUP *dest, const EC_GROUP *src)
{
    return 1;
}

int ec_c25519_get_degree(const EC_GROUP *group)
{
    return 255; 
}

#define C25519_MAX_KEYGEN_ATTEMPTS  10 
/*
 * This function does the work of generating a curve25519
 * key (both public & private).  The values are stored on
 * the EC_KEY struct under the c25519_pub and c25519_prv
 * members. 
 *
 * Returns 1 on success, 0 on failure.
 */
int ec_c25519_generate_key(EC_KEY *key)
{
    EC_POINT *point = NULL;
    unsigned char secret[C25519_R_LEN];
    unsigned char public[C25519_R_LEN];
    static const char basepoint[C25519_R_LEN] = {9};
    int len = 0;
    int pub_len = 0;
    int attempts = 0;

#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
        ECerr(EC_F_EC_C25519_GENERATE_KEY, EC_R_NON_FIPS_METHOD);
	return 0;
    }
#endif

    if (key->c25519_prv) {
	BN_clear_free(key->c25519_prv);
    }
    key->c25519_prv = BN_new();
    if (!key->c25519_prv) {
        ECerr(EC_F_EC_C25519_GENERATE_KEY, EC_R_MALLOC_FAILED);
        return 0;
    }

    if (key->c25519_pub) {
	BN_clear_free(key->c25519_pub);
    }
    key->c25519_pub = BN_new();
    if (!key->c25519_pub) {
        ECerr(EC_F_EC_C25519_GENERATE_KEY, EC_R_MALLOC_FAILED);
        return 0;
    }

    /*
     * Too many leading zeros in the random number generator will 
     * result in a key size that's possibly too small.  This causes
     * the curve25519 algorithm to fail.  We'll attempt up to
     * 10 times to generate a key that fills all 32 bytes for
     * both the public and private key.
     */
    while ((pub_len < C25519_R_LEN) && (attempts < C25519_MAX_KEYGEN_ATTEMPTS)) {
	len = 0;

	/* For curve 25519, the private key is a random 256 bit number */
	while ((len != C25519_R_LEN) && (attempts < C25519_MAX_KEYGEN_ATTEMPTS)) {
	    if (!BN_rand(key->c25519_prv, 256, -1, 0)) {
		ECerr(EC_F_EC_C25519_GENERATE_KEY, ERR_R_INTERNAL_ERROR);
		return 0;
	    }
	    len = BN_num_bytes(key->c25519_prv);
	    attempts++;
	}

	/*
	 * If we exceed the max attempts, the key wouldn't be long
	 * enough.  We also want to make sure the secret key isn't
	 * larger than 32 bytes to prevent a buffer overflow
	 * on the stack in the upcoming BN_bn2bin() call.
	 */
	if (len != C25519_R_LEN) {
	    ECerr(EC_F_EC_C25519_GENERATE_KEY, ERR_R_INTERNAL_ERROR);
	    BN_clear_free(key->c25519_prv);
	    BN_clear_free(key->c25519_pub);
	    return 0;
	}

	/*
	 * Convert the private key to a char array
	 */
	memset(secret, 0, C25519_R_LEN);
	len = BN_bn2bin(key->c25519_prv, secret);
	if (len != C25519_R_LEN) {
	    ECerr(EC_F_EC_C25519_GENERATE_KEY, EC_R_INVALID_LEN);
	    BN_clear_free(key->c25519_prv);
	    BN_clear_free(key->c25519_pub);
	    return 0;
	}

	/* 
	 * Now generate the public key and stuff it into
	 * the bignum on the EC_KEY struct.
	 */
	curve25519_donna((char*)public, (const char*)secret, (const char*)basepoint);
	BN_bin2bn((const unsigned char *)public, C25519_R_LEN, key->c25519_pub);
    
	/*
	 * Make sure the public key is long enough
	 */
	pub_len = BN_num_bytes(key->c25519_pub);
	attempts++;
    }

    /*
     * Make sure we didn't exceed max attempts by checking
     * if the public key was too small
     */
    if (pub_len != C25519_R_LEN) {
	ECerr(EC_F_EC_C25519_GENERATE_KEY, ERR_R_INTERNAL_ERROR);
	BN_clear_free(key->c25519_prv);
	BN_clear_free(key->c25519_pub);
	return 0;
    }

    /*
     * The public key also needs to go on the point
     * We overload the X coordinate on the point to store the public key
     */
    if (key->pub_key == NULL) {
        point = EC_POINT_new(key->group);
        if (point == NULL) {
	    ECerr(EC_F_EC_C25519_GENERATE_KEY, EC_R_MALLOC_FAILED);
	    BN_clear_free(key->c25519_prv);
	    BN_clear_free(key->c25519_pub);
	    return 0;
	}
	key->pub_key = point;
    } else {
        point = key->pub_key;
    }
    BN_init(&point->X);
    BN_copy(&point->X, key->c25519_pub);

    return 1;
}

/*
 * This routine performs ECDH shared secret computation for
 * curve25519.  The prototype for this function is identical
 * to ECDH_compute_key().
 *
 * Returns the length of the computed key, or -1 if an error occurs.
 */
int ec_c25519_compute_key(void *out, size_t outlen, const EC_POINT *pub_key, EC_KEY *eckey,
                          void *(*KDF) (const void *in, size_t inlen, void *out, size_t *outlen))
{
    unsigned char c519_priv[C25519_R_LEN];
    unsigned char c519_pub[C25519_R_LEN];
    unsigned char tmpbuf[C25519_R_LEN];
    int len;
    int retval = -1;

    if (out == NULL || pub_key == NULL || eckey == NULL) {
        ECerr(EC_F_EC_C25519_COMPUTE_KEY, EC_R_NULL_PARMS_PASSED_IN);
	return -1;
    }

#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
        ECerr(EC_F_EC_C25519_COMPUTE_KEY, EC_R_NON_FIPS_METHOD);
	return -1;
    }
#endif
    /*
     * Sanity check the key lengths
     */
    if (BN_num_bytes(eckey->c25519_prv) != C25519_R_LEN) {
        ECerr(EC_F_EC_C25519_COMPUTE_KEY, EC_R_INVALID_LEN);
	return -1;
    }

    /*
     * Convert private to binary
     */
    memset(c519_priv, 0, C25519_R_LEN);
    len = BN_bn2bin(eckey->c25519_prv, c519_priv);
    if (len != C25519_R_LEN) {
        ECerr(EC_F_EC_C25519_COMPUTE_KEY, EC_R_INVALID_LEN);
	return -1;
    }

    /*
     * Convert public key to binary
     */
    memset(c519_pub, 0, C25519_R_LEN);
    len = BN_bn2bin(&pub_key->X, c519_pub);
    if (len != C25519_R_LEN) {
        ECerr(EC_F_EC_C25519_COMPUTE_KEY, EC_R_INVALID_LEN);
	return -1;
    }

    /*
     * Now compute the shared secret
     */
    memset(tmpbuf, 0, C25519_R_LEN);
    curve25519_donna((char*)tmpbuf, (char*)c519_priv, (char*)c519_pub);

    /* run the result through the KDF */
    if (KDF != NULL) {
        /* call the KDF */
        KDF(tmpbuf, C25519_R_LEN, out, &outlen);
        retval = outlen;
    } else {
        /* no KDF, copy as much as we can */
        if (outlen > C25519_R_LEN) outlen = C25519_R_LEN;
        memcpy(out, tmpbuf, outlen);
        retval = outlen;
    }
    return retval;
}


/*
 * Converts the public key to an octet string.  Only the
 * uncompressed format is suppoted.  Big endian encoding
 * is used when generating the octet string.
 *
 * Returns the length of the octet string, or zero if an error
 */
size_t ec_c25519_point2oct(const EC_GROUP *group, const EC_POINT *point,
                           point_conversion_form_t form, unsigned char *buf,
                           size_t len, BN_CTX *ctx)
{
    int ret_len;
    int i;

    if (form != POINT_CONVERSION_UNCOMPRESSED) {
        ECerr(EC_F_EC_C25519_POINT2OCT, EC_R_INVALID_FORM);
	return 0;
    }

    /*
     * We only support curve25519, which is always 32 bytes
     */
    ret_len = C25519_R_LEN;

    /*
     * If the buffer isn't provided by the caller, then all we
     * do is return the length.
     */
    if (!buf) {
	return (ret_len);
    }
    
    /*
     * Make sure caller gave us enough memory to hold the key
     */
    if (len < ret_len) {
        ECerr(EC_F_EC_C25519_POINT2OCT, EC_R_BUFFER_TOO_SMALL);
	return 0;
    }
 
    memset(buf, 0x00, ret_len);
    i = BN_bn2bin(&point->X, buf);
    if (i != C25519_R_LEN) {
        ECerr(EC_F_EC_C25519_POINT2OCT, EC_R_INVALID_FORM);
	return 0;
    }

    return (ret_len);
}


/*
 * Converts a public key stored in an octet string to an EC_POINT.
 * The octet string should be big endian format.
 *
 * Returns 1 for success, or zero on error.
 */
int ec_c25519_oct2point(const EC_GROUP *group, EC_POINT *point,
                        const unsigned char *buf, size_t len, BN_CTX *ctx)
{
    point_conversion_form_t form;

    /*
     * Make sure the length is at least as large as a public key 
     */
    if (len < C25519_R_LEN) {
        ECerr(EC_F_EC_C25519_OCT2POINT, EC_R_BUFFER_TOO_SMALL);
        return 0;
    }

    /*
     * Finally, decode the octet string
     */
    if (!BN_bin2bn(buf, C25519_R_LEN, &point->X)) {
        ECerr(EC_F_EC_C25519_OCT2POINT, EC_R_INVALID_ENCODING);
	return 0;
    }

    return (1);
}
#endif
