/* crypto/ec/ec_mult.c */
/*
 * Originally written by Bodo Moeller and Nils Larsch for the OpenSSL project.
 */
/* ====================================================================
 * Copyright (c) 1998-2007 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */
/* ====================================================================
 * Copyright 2002 Sun Microsystems, Inc. ALL RIGHTS RESERVED.
 * Portions of this software developed by SUN MICROSYSTEMS, INC.,
 * and contributed to the OpenSSL project.
 */

#include <string.h>

#include <openssl/err.h>

#include "ec_lcl.h"

/*
 * This file implements the wNAF-based interleaving multi-exponentation method
 * (<URL:http://www.informatik.tu-darmstadt.de/TI/Mitarbeiter/moeller.html#multiexp>);
 * for multiplication with precomputation, we use wNAF splitting
 * (<URL:http://www.informatik.tu-darmstadt.de/TI/Mitarbeiter/moeller.html#fastexp>).
 */


/*
 * The OPENSSL_NO_PFSCM knob can be used to disable Scott Fluhrer's 
 * ECDSA side-channel mitigation logic for ECDSA verify.  This was implemented
 * due to the risks described in https://eprint.iacr.org/2015/036 for
 * VM environments.  This mitigation logic does not work in conjunction with
 * the OpenSSL precomputation support for point multiplication, which is
 * enabled using EC_GROUP_precompute_mult() or EC_KEY_precompute_mult().
 * If you are certain that side-channel is not a risk for your
 * deployment and would like to take advantage of the OpenSSL precomputation
 * performance improvements, use OPENSSL_NO_PFSCM when compiling CiscoSSL.
 */
#ifndef OPENSSL_NO_PFSCM 
static int ec_single_point_multiply(const EC_GROUP *group, EC_POINT *R,
                                    const BIGNUM *k, const EC_POINT *H,
                                    BN_CTX *ctx );
#endif

/* structure for precomputed multiples of the generator */
typedef struct ec_pre_comp_st {
    const EC_GROUP *group;      /* parent EC_GROUP object */
    size_t blocksize;           /* block size for wNAF splitting */
    size_t numblocks;           /* max. number of blocks for which we have
                                 * precomputation */
    size_t w;                   /* window size */
    EC_POINT **points;          /* array with pre-calculated multiples of
                                 * generator: 'num' pointers to EC_POINT
                                 * objects followed by a NULL */
    size_t num;                 /* numblocks * 2^(w-1) */
    int references;
} EC_PRE_COMP;

/* functions to manage EC_PRE_COMP within the EC_GROUP extra_data framework */
static void *ec_pre_comp_dup(void *);
static void ec_pre_comp_free(void *);
static void ec_pre_comp_clear_free(void *);

static EC_PRE_COMP *ec_pre_comp_new(const EC_GROUP *group)
{
    EC_PRE_COMP *ret = NULL;

    if (!group)
        return NULL;

    ret = (EC_PRE_COMP *)OPENSSL_malloc(sizeof(EC_PRE_COMP));
    if (!ret) {
        ECerr(EC_F_EC_PRE_COMP_NEW, ERR_R_MALLOC_FAILURE);
        return ret;
    }
    ret->group = group;
    ret->blocksize = 8;         /* default */
    ret->numblocks = 0;
    ret->w = 4;                 /* default */
    ret->points = NULL;
    ret->num = 0;
    ret->references = 1;
    return ret;
}

static void *ec_pre_comp_dup(void *src_)
{
    EC_PRE_COMP *src = src_;

    /* no need to actually copy, these objects never change! */

    CRYPTO_add(&src->references, 1, CRYPTO_LOCK_EC_PRE_COMP);

    return src_;
}

static void ec_pre_comp_free(void *pre_)
{
    int i;
    EC_PRE_COMP *pre = pre_;

    if (!pre)
        return;

    i = CRYPTO_add(&pre->references, -1, CRYPTO_LOCK_EC_PRE_COMP);
    if (i > 0)
        return;

    if (pre->points) {
        EC_POINT **p;

        for (p = pre->points; *p != NULL; p++)
            EC_POINT_free(*p);
        OPENSSL_free(pre->points);
    }
    OPENSSL_free(pre);
}

static void ec_pre_comp_clear_free(void *pre_)
{
    int i;
    EC_PRE_COMP *pre = pre_;

    if (!pre)
        return;

    i = CRYPTO_add(&pre->references, -1, CRYPTO_LOCK_EC_PRE_COMP);
    if (i > 0)
        return;

    if (pre->points) {
        EC_POINT **p;

        for (p = pre->points; *p != NULL; p++) {
            EC_POINT_clear_free(*p);
            OPENSSL_cleanse(p, sizeof *p);
        }
        OPENSSL_free(pre->points);
    }
    OPENSSL_cleanse(pre, sizeof *pre);
    OPENSSL_free(pre);
}

/*-
 * Determine the modified width-(w+1) Non-Adjacent Form (wNAF) of 'scalar'.
 * This is an array  r[]  of values that are either zero or odd with an
 * absolute value less than  2^w  satisfying
 *     scalar = \sum_j r[j]*2^j
 * where at most one of any  w+1  consecutive digits is non-zero
 * with the exception that the most significant digit may be only
 * w-1 zeros away from that next non-zero digit.
 */
static signed char *compute_wNAF(const BIGNUM *scalar, int w, size_t *ret_len)
{
    int window_val;
    int ok = 0;
    signed char *r = NULL;
    int sign = 1;
    int bit, next_bit, mask;
    size_t len = 0, j;

    if (BN_is_zero(scalar)) {
        r = OPENSSL_malloc(1);
        if (!r) {
            ECerr(EC_F_COMPUTE_WNAF, ERR_R_MALLOC_FAILURE);
            goto err;
        }
        r[0] = 0;
        *ret_len = 1;
        return r;
    }

    if (w <= 0 || w > 7) {      /* 'signed char' can represent integers with
                                 * absolute values less than 2^7 */
        ECerr(EC_F_COMPUTE_WNAF, ERR_R_INTERNAL_ERROR);
        goto err;
    }
    bit = 1 << w;               /* at most 128 */
    next_bit = bit << 1;        /* at most 256 */
    mask = next_bit - 1;        /* at most 255 */

    if (BN_is_negative(scalar)) {
        sign = -1;
    }

    if (scalar->d == NULL || scalar->top == 0) {
        ECerr(EC_F_COMPUTE_WNAF, ERR_R_INTERNAL_ERROR);
        goto err;
    }

    len = BN_num_bits(scalar);
    r = OPENSSL_malloc(len + 1); /* modified wNAF may be one digit longer
                                  * than binary representation (*ret_len will
                                  * be set to the actual length, i.e. at most
                                  * BN_num_bits(scalar) + 1) */
    if (r == NULL) {
        ECerr(EC_F_COMPUTE_WNAF, ERR_R_MALLOC_FAILURE);
        goto err;
    }
    window_val = scalar->d[0] & mask;
    j = 0;
    while ((window_val != 0) || (j + w + 1 < len)) { /* if j+w+1 >= len,
                                                      * window_val will not
                                                      * increase */
        int digit = 0;

        /* 0 <= window_val <= 2^(w+1) */

        if (window_val & 1) {
            /* 0 < window_val < 2^(w+1) */

            if (window_val & bit) {
                digit = window_val - next_bit; /* -2^w < digit < 0 */

#if 1                           /* modified wNAF */
                if (j + w + 1 >= len) {
                    /*
                     * special case for generating modified wNAFs: no new
                     * bits will be added into window_val, so using a
                     * positive digit here will decrease the total length of
                     * the representation
                     */

                    digit = window_val & (mask >> 1); /* 0 < digit < 2^w */
                }
#endif
            } else {
                digit = window_val; /* 0 < digit < 2^w */
            }

            if (digit <= -bit || digit >= bit || !(digit & 1)) {
                ECerr(EC_F_COMPUTE_WNAF, ERR_R_INTERNAL_ERROR);
                goto err;
            }

            window_val -= digit;

            /*
             * now window_val is 0 or 2^(w+1) in standard wNAF generation;
             * for modified window NAFs, it may also be 2^w
             */
            if (window_val != 0 && window_val != next_bit
                && window_val != bit) {
                ECerr(EC_F_COMPUTE_WNAF, ERR_R_INTERNAL_ERROR);
                goto err;
            }
        }

        r[j++] = sign * digit;

        window_val >>= 1;
        window_val += bit * BN_is_bit_set(scalar, j + w);

        if (window_val > next_bit) {
            ECerr(EC_F_COMPUTE_WNAF, ERR_R_INTERNAL_ERROR);
            goto err;
        }
    }

    if (j > len + 1) {
        ECerr(EC_F_COMPUTE_WNAF, ERR_R_INTERNAL_ERROR);
        goto err;
    }
    len = j;
    ok = 1;

 err:
    if (!ok) {
        OPENSSL_free(r);
        r = NULL;
    }
    if (ok)
        *ret_len = len;
    return r;
}

/*
 * TODO: table should be optimised for the wNAF-based implementation,
 * sometimes smaller windows will give better performance (thus the
 * boundaries should be increased)
 */
#define EC_window_bits_for_scalar_size(b) \
                ((size_t) \
                 ((b) >= 2000 ? 6 : \
                  (b) >=  800 ? 5 : \
                  (b) >=  300 ? 4 : \
                  (b) >=   70 ? 3 : \
                  (b) >=   20 ? 2 : \
                  1))

/*-
 * Compute
 *      \sum scalars[i]*points[i],
 * also including
 *      scalar*generator
 * in the addition if scalar != NULL
 */
int ec_wNAF_mul(const EC_GROUP *group, EC_POINT *r, const BIGNUM *scalar,
                size_t num, const EC_POINT *points[], const BIGNUM *scalars[],
                BN_CTX *ctx)
{
    BN_CTX *new_ctx = NULL;
    const EC_POINT *generator = NULL;
    EC_POINT *tmp = NULL;
    size_t totalnum;
    size_t blocksize = 0, numblocks = 0; /* for wNAF splitting */
    size_t pre_points_per_block = 0;
    size_t i, j;
    int k;
    int r_is_inverted = 0;
    int r_is_at_infinity = 1;
    size_t *wsize = NULL;       /* individual window sizes */
    signed char **wNAF = NULL;  /* individual wNAFs */
    size_t *wNAF_len = NULL;
    size_t max_len = 0;
    size_t num_val;
    EC_POINT **val = NULL;      /* precomputation */
    EC_POINT **v;
    EC_POINT ***val_sub = NULL; /* pointers to sub-arrays of 'val' or
                                 * 'pre_comp->points' */
    const EC_PRE_COMP *pre_comp = NULL;
    int num_scalar = 0;         /* flag: will be set to 1 if 'scalar' must be
                                 * treated like other scalars, i.e.
                                 * precomputation is not available */
    int ret = 0;

    if (group->meth != r->meth) {
        ECerr(EC_F_EC_WNAF_MUL, EC_R_INCOMPATIBLE_OBJECTS);
        return 0;
    }

    if ((scalar == NULL) && (num == 0)) {
        return EC_POINT_set_to_infinity(group, r);
    }

    for (i = 0; i < num; i++) {
        if (group->meth != points[i]->meth) {
            ECerr(EC_F_EC_WNAF_MUL, EC_R_INCOMPATIBLE_OBJECTS);
            return 0;
        }
    }

    if (ctx == NULL) {
        ctx = new_ctx = BN_CTX_new();
        if (ctx == NULL)
            goto err;
    }

    if (scalar != NULL) {
        generator = EC_GROUP_get0_generator(group);
        if (generator == NULL) {
            ECerr(EC_F_EC_WNAF_MUL, EC_R_UNDEFINED_GENERATOR);
            goto err;
        }

        /* look if we can use precomputed multiples of generator */

        pre_comp =
            EC_EX_DATA_get_data(group->extra_data, ec_pre_comp_dup,
                                ec_pre_comp_free, ec_pre_comp_clear_free);

        if (pre_comp && pre_comp->numblocks
            && (EC_POINT_cmp(group, generator, pre_comp->points[0], ctx) ==
                0)) {
            blocksize = pre_comp->blocksize;

            /*
             * determine maximum number of blocks that wNAF splitting may
             * yield (NB: maximum wNAF length is bit length plus one)
             */
            numblocks = (BN_num_bits(scalar) / blocksize) + 1;

            /*
             * we cannot use more blocks than we have precomputation for
             */
            if (numblocks > pre_comp->numblocks)
                numblocks = pre_comp->numblocks;

            pre_points_per_block = (size_t)1 << (pre_comp->w - 1);

            /* check that pre_comp looks sane */
            if (pre_comp->num != (pre_comp->numblocks * pre_points_per_block)) {
                ECerr(EC_F_EC_WNAF_MUL, ERR_R_INTERNAL_ERROR);
                goto err;
            }
        } else {
            /* can't use precomputation */
            pre_comp = NULL;
            numblocks = 1;
            num_scalar = 1;     /* treat 'scalar' like 'num'-th element of
                                 * 'scalars' */
        }
    }

    totalnum = num + numblocks;

    wsize = OPENSSL_malloc(totalnum * sizeof wsize[0]);
    wNAF_len = OPENSSL_malloc(totalnum * sizeof wNAF_len[0]);
    wNAF = OPENSSL_malloc((totalnum + 1) * sizeof wNAF[0]); /* includes space
                                                             * for pivot */
    val_sub = OPENSSL_malloc(totalnum * sizeof val_sub[0]);

    /* Ensure wNAF is initialised in case we end up going to err */
    if (wNAF)
        wNAF[0] = NULL;         /* preliminary pivot */

    if (!wsize || !wNAF_len || !wNAF || !val_sub) {
        ECerr(EC_F_EC_WNAF_MUL, ERR_R_MALLOC_FAILURE);
        goto err;
    }

#ifndef OPENSSL_NO_PFSCM 
    if ((NID_X9_62_prime_field == EC_METHOD_get_field_type(EC_GROUP_method_of(group))) && 
	(num + num_scalar == 1)) {
        /*
         * We're doing a simple point multiply; switch to an alternative
         * algorithm that has constant memory accesses
         */
        const BIGNUM *k = num ? scalars[0] : scalar;
        const EC_POINT *H = num ? points[0] : generator;
        ret = ec_single_point_multiply( group, r, k, H, ctx );
	/* Branch to error handler to release any resources used */
	goto err;
    }
#endif

    /*
     * num_val will be the total number of temporarily precomputed points
     */
    num_val = 0;

    for (i = 0; i < num + num_scalar; i++) {
        size_t bits;

        bits = i < num ? BN_num_bits(scalars[i]) : BN_num_bits(scalar);
        wsize[i] = EC_window_bits_for_scalar_size(bits);
        num_val += (size_t)1 << (wsize[i] - 1);
        wNAF[i + 1] = NULL;     /* make sure we always have a pivot */
        wNAF[i] =
            compute_wNAF((i < num ? scalars[i] : scalar), wsize[i],
                         &wNAF_len[i]);
        if (wNAF[i] == NULL)
            goto err;
        if (wNAF_len[i] > max_len)
            max_len = wNAF_len[i];
    }

    if (numblocks) {
        /* we go here iff scalar != NULL */

        if (pre_comp == NULL) {
            if (num_scalar != 1) {
                ECerr(EC_F_EC_WNAF_MUL, ERR_R_INTERNAL_ERROR);
                goto err;
            }
            /* we have already generated a wNAF for 'scalar' */
        } else {
            signed char *tmp_wNAF = NULL;
            size_t tmp_len = 0;

            if (num_scalar != 0) {
                ECerr(EC_F_EC_WNAF_MUL, ERR_R_INTERNAL_ERROR);
                goto err;
            }

            /*
             * use the window size for which we have precomputation
             */
            wsize[num] = pre_comp->w;
            tmp_wNAF = compute_wNAF(scalar, wsize[num], &tmp_len);
            if (!tmp_wNAF)
                goto err;

            if (tmp_len <= max_len) {
                /*
                 * One of the other wNAFs is at least as long as the wNAF
                 * belonging to the generator, so wNAF splitting will not buy
                 * us anything.
                 */

                numblocks = 1;
                totalnum = num + 1; /* don't use wNAF splitting */
                wNAF[num] = tmp_wNAF;
                wNAF[num + 1] = NULL;
                wNAF_len[num] = tmp_len;
                if (tmp_len > max_len)
                    max_len = tmp_len;
                /*
                 * pre_comp->points starts with the points that we need here:
                 */
                val_sub[num] = pre_comp->points;
            } else {
                /*
                 * don't include tmp_wNAF directly into wNAF array - use wNAF
                 * splitting and include the blocks
                 */

                signed char *pp;
                EC_POINT **tmp_points;

                if (tmp_len < numblocks * blocksize) {
                    /*
                     * possibly we can do with fewer blocks than estimated
                     */
                    numblocks = (tmp_len + blocksize - 1) / blocksize;
                    if (numblocks > pre_comp->numblocks) {
                        ECerr(EC_F_EC_WNAF_MUL, ERR_R_INTERNAL_ERROR);
                        goto err;
                    }
                    totalnum = num + numblocks;
                }

                /* split wNAF in 'numblocks' parts */
                pp = tmp_wNAF;
                tmp_points = pre_comp->points;

                for (i = num; i < totalnum; i++) {
                    if (i < totalnum - 1) {
                        wNAF_len[i] = blocksize;
                        if (tmp_len < blocksize) {
                            ECerr(EC_F_EC_WNAF_MUL, ERR_R_INTERNAL_ERROR);
                            goto err;
                        }
                        tmp_len -= blocksize;
                    } else
                        /*
                         * last block gets whatever is left (this could be
                         * more or less than 'blocksize'!)
                         */
                        wNAF_len[i] = tmp_len;

                    wNAF[i + 1] = NULL;
                    wNAF[i] = OPENSSL_malloc(wNAF_len[i]);
                    if (wNAF[i] == NULL) {
                        ECerr(EC_F_EC_WNAF_MUL, ERR_R_MALLOC_FAILURE);
                        OPENSSL_free(tmp_wNAF);
                        goto err;
                    }
                    memcpy(wNAF[i], pp, wNAF_len[i]);
                    if (wNAF_len[i] > max_len)
                        max_len = wNAF_len[i];

                    if (*tmp_points == NULL) {
                        ECerr(EC_F_EC_WNAF_MUL, ERR_R_INTERNAL_ERROR);
                        OPENSSL_free(tmp_wNAF);
                        goto err;
                    }
                    val_sub[i] = tmp_points;
                    tmp_points += pre_points_per_block;
                    pp += blocksize;
                }
                OPENSSL_free(tmp_wNAF);
            }
        }
    }

    /*
     * All points we precompute now go into a single array 'val'.
     * 'val_sub[i]' is a pointer to the subarray for the i-th point, or to a
     * subarray of 'pre_comp->points' if we already have precomputation.
     */
    val = OPENSSL_malloc((num_val + 1) * sizeof val[0]);
    if (val == NULL) {
        ECerr(EC_F_EC_WNAF_MUL, ERR_R_MALLOC_FAILURE);
        goto err;
    }
    val[num_val] = NULL;        /* pivot element */

    /* allocate points for precomputation */
    v = val;
    for (i = 0; i < num + num_scalar; i++) {
        val_sub[i] = v;
        for (j = 0; j < ((size_t)1 << (wsize[i] - 1)); j++) {
            *v = EC_POINT_new(group);
            if (*v == NULL)
                goto err;
            v++;
        }
    }
    if (!(v == val + num_val)) {
        ECerr(EC_F_EC_WNAF_MUL, ERR_R_INTERNAL_ERROR);
        goto err;
    }

    if (!(tmp = EC_POINT_new(group)))
        goto err;

    /*-
     * prepare precomputed values:
     *    val_sub[i][0] :=     points[i]
     *    val_sub[i][1] := 3 * points[i]
     *    val_sub[i][2] := 5 * points[i]
     *    ...
     */
    for (i = 0; i < num + num_scalar; i++) {
        if (i < num) {
            if (!EC_POINT_copy(val_sub[i][0], points[i]))
                goto err;
        } else {
            if (!EC_POINT_copy(val_sub[i][0], generator))
                goto err;
        }

        if (wsize[i] > 1) {
            if (!EC_POINT_dbl(group, tmp, val_sub[i][0], ctx))
                goto err;
            for (j = 1; j < ((size_t)1 << (wsize[i] - 1)); j++) {
                if (!EC_POINT_add
                    (group, val_sub[i][j], val_sub[i][j - 1], tmp, ctx))
                    goto err;
            }
        }
    }

#if 1                           /* optional; EC_window_bits_for_scalar_size
                                 * assumes we do this step */
    if (!EC_POINTs_make_affine(group, num_val, val, ctx))
        goto err;
#endif

    r_is_at_infinity = 1;

    for (k = max_len - 1; k >= 0; k--) {
        if (!r_is_at_infinity) {
            if (!EC_POINT_dbl(group, r, r, ctx))
                goto err;
        }

        for (i = 0; i < totalnum; i++) {
            if (wNAF_len[i] > (size_t)k) {
                int digit = wNAF[i][k];
                int is_neg;

                if (digit) {
                    is_neg = digit < 0;

                    if (is_neg)
                        digit = -digit;

                    if (is_neg != r_is_inverted) {
                        if (!r_is_at_infinity) {
                            if (!EC_POINT_invert(group, r, ctx))
                                goto err;
                        }
                        r_is_inverted = !r_is_inverted;
                    }

                    /* digit > 0 */

                    if (r_is_at_infinity) {
                        if (!EC_POINT_copy(r, val_sub[i][digit >> 1]))
                            goto err;
                        r_is_at_infinity = 0;
                    } else {
                        if (!EC_POINT_add
                            (group, r, r, val_sub[i][digit >> 1], ctx))
                            goto err;
                    }
                }
            }
        }
    }

    if (r_is_at_infinity) {
        if (!EC_POINT_set_to_infinity(group, r))
            goto err;
    } else {
        if (r_is_inverted)
            if (!EC_POINT_invert(group, r, ctx))
                goto err;
    }

    ret = 1;

 err:
    if (new_ctx != NULL)
        BN_CTX_free(new_ctx);
    if (tmp != NULL)
        EC_POINT_free(tmp);
    if (wsize != NULL)
        OPENSSL_free(wsize);
    if (wNAF_len != NULL)
        OPENSSL_free(wNAF_len);
    if (wNAF != NULL) {
        signed char **w;

        for (w = wNAF; *w != NULL; w++)
            OPENSSL_free(*w);

        OPENSSL_free(wNAF);
    }
    if (val != NULL) {
        for (v = val; *v != NULL; v++)
            EC_POINT_clear_free(*v);

        OPENSSL_free(val);
    }
    if (val_sub != NULL) {
        OPENSSL_free(val_sub);
    }
    return ret;
}

/*-
 * ec_wNAF_precompute_mult()
 * creates an EC_PRE_COMP object with preprecomputed multiples of the generator
 * for use with wNAF splitting as implemented in ec_wNAF_mul().
 *
 * 'pre_comp->points' is an array of multiples of the generator
 * of the following form:
 * points[0] =     generator;
 * points[1] = 3 * generator;
 * ...
 * points[2^(w-1)-1] =     (2^(w-1)-1) * generator;
 * points[2^(w-1)]   =     2^blocksize * generator;
 * points[2^(w-1)+1] = 3 * 2^blocksize * generator;
 * ...
 * points[2^(w-1)*(numblocks-1)-1] = (2^(w-1)) *  2^(blocksize*(numblocks-2)) * generator
 * points[2^(w-1)*(numblocks-1)]   =              2^(blocksize*(numblocks-1)) * generator
 * ...
 * points[2^(w-1)*numblocks-1]     = (2^(w-1)) *  2^(blocksize*(numblocks-1)) * generator
 * points[2^(w-1)*numblocks]       = NULL
 */
int ec_wNAF_precompute_mult(EC_GROUP *group, BN_CTX *ctx)
{
    const EC_POINT *generator;
    EC_POINT *tmp_point = NULL, *base = NULL, **var;
    BN_CTX *new_ctx = NULL;
    BIGNUM *order;
    size_t i, bits, w, pre_points_per_block, blocksize, numblocks, num;
    EC_POINT **points = NULL;
    EC_PRE_COMP *pre_comp;
    int ret = 0;

    /* if there is an old EC_PRE_COMP object, throw it away */
    EC_EX_DATA_free_data(&group->extra_data, ec_pre_comp_dup,
                         ec_pre_comp_free, ec_pre_comp_clear_free);

    if ((pre_comp = ec_pre_comp_new(group)) == NULL)
        return 0;

    generator = EC_GROUP_get0_generator(group);
    if (generator == NULL) {
        ECerr(EC_F_EC_WNAF_PRECOMPUTE_MULT, EC_R_UNDEFINED_GENERATOR);
        goto err;
    }

    if (ctx == NULL) {
        ctx = new_ctx = BN_CTX_new();
        if (ctx == NULL)
            goto err;
    }

    BN_CTX_start(ctx);
    order = BN_CTX_get(ctx);
    if (order == NULL)
        goto err;

    if (!EC_GROUP_get_order(group, order, ctx))
        goto err;
    if (BN_is_zero(order)) {
        ECerr(EC_F_EC_WNAF_PRECOMPUTE_MULT, EC_R_UNKNOWN_ORDER);
        goto err;
    }

    bits = BN_num_bits(order);
    /*
     * The following parameters mean we precompute (approximately) one point
     * per bit. TBD: The combination 8, 4 is perfect for 160 bits; for other
     * bit lengths, other parameter combinations might provide better
     * efficiency.
     */
    blocksize = 8;
    w = 4;
    if (EC_window_bits_for_scalar_size(bits) > w) {
        /* let's not make the window too small ... */
        w = EC_window_bits_for_scalar_size(bits);
    }

    numblocks = (bits + blocksize - 1) / blocksize; /* max. number of blocks
                                                     * to use for wNAF
                                                     * splitting */

    pre_points_per_block = (size_t)1 << (w - 1);
    num = pre_points_per_block * numblocks; /* number of points to compute
                                             * and store */

    points = OPENSSL_malloc(sizeof(EC_POINT *) * (num + 1));
    if (!points) {
        ECerr(EC_F_EC_WNAF_PRECOMPUTE_MULT, ERR_R_MALLOC_FAILURE);
        goto err;
    }

    var = points;
    var[num] = NULL;            /* pivot */
    for (i = 0; i < num; i++) {
        if ((var[i] = EC_POINT_new(group)) == NULL) {
            ECerr(EC_F_EC_WNAF_PRECOMPUTE_MULT, ERR_R_MALLOC_FAILURE);
            goto err;
        }
    }

    if (!(tmp_point = EC_POINT_new(group)) || !(base = EC_POINT_new(group))) {
        ECerr(EC_F_EC_WNAF_PRECOMPUTE_MULT, ERR_R_MALLOC_FAILURE);
        goto err;
    }

    if (!EC_POINT_copy(base, generator))
        goto err;

    /* do the precomputation */
    for (i = 0; i < numblocks; i++) {
        size_t j;

        if (!EC_POINT_dbl(group, tmp_point, base, ctx))
            goto err;

        if (!EC_POINT_copy(*var++, base))
            goto err;

        for (j = 1; j < pre_points_per_block; j++, var++) {
            /*
             * calculate odd multiples of the current base point
             */
            if (!EC_POINT_add(group, *var, tmp_point, *(var - 1), ctx))
                goto err;
        }

        if (i < numblocks - 1) {
            /*
             * get the next base (multiply current one by 2^blocksize)
             */
            size_t k;

            if (blocksize <= 2) {
                ECerr(EC_F_EC_WNAF_PRECOMPUTE_MULT, ERR_R_INTERNAL_ERROR);
                goto err;
            }

            if (!EC_POINT_dbl(group, base, tmp_point, ctx))
                goto err;
            for (k = 2; k < blocksize; k++) {
                if (!EC_POINT_dbl(group, base, base, ctx))
                    goto err;
            }
        }
    }

    if (!EC_POINTs_make_affine(group, num, points, ctx))
        goto err;

    pre_comp->group = group;
    pre_comp->blocksize = blocksize;
    pre_comp->numblocks = numblocks;
    pre_comp->w = w;
    pre_comp->points = points;
    points = NULL;
    pre_comp->num = num;

    if (!EC_EX_DATA_set_data(&group->extra_data, pre_comp,
                             ec_pre_comp_dup, ec_pre_comp_free,
                             ec_pre_comp_clear_free))
        goto err;
    pre_comp = NULL;

    ret = 1;
 err:
    if (ctx != NULL)
        BN_CTX_end(ctx);
    if (new_ctx != NULL)
        BN_CTX_free(new_ctx);
    if (pre_comp)
        ec_pre_comp_free(pre_comp);
    if (points) {
        EC_POINT **p;

        for (p = points; *p != NULL; p++)
            EC_POINT_free(*p);
        OPENSSL_free(points);
    }
    if (tmp_point)
        EC_POINT_free(tmp_point);
    if (base)
        EC_POINT_free(base);
    return ret;
}

int ec_wNAF_have_precompute_mult(const EC_GROUP *group)
{
#ifdef OPENSSL_NO_PFSCM 
    if (EC_EX_DATA_get_data
        (group->extra_data, ec_pre_comp_dup, ec_pre_comp_free,
         ec_pre_comp_clear_free) != NULL)
        return 1;
    else
#endif
        return 0;
}

#ifndef OPENSSL_NO_PFSCM 
/*
 * This copies the bignum n into the fixed-sized buffer
 */
static int ec_BN_copy_out(unsigned char *dest, size_t len_dest, const BIGNUM *n) {
    size_t num_bytes;
    memset( dest, 0, len_dest );
    num_bytes = BN_num_bytes(n);

    if (num_bytes > len_dest) {
        ECerr(EC_F_EC_BN_COPY_OUT, ERR_R_INTERNAL_ERROR);
	return 0; /* OVERFLOW */
    }

    BN_bn2bin(n, dest + (len_dest - num_bytes));
    return 1;
}

/* The below selection code can deal with multiple bytes at once */
/* Once we feel we can take advantage of this (that is, it won't cause */
/* problems on some obscure CPU we build for), we can modify the below */
/* typedef to be a larger unsigned type to get a bit more efficiency */
typedef unsigned char word;

/*
 * The goal of this routine is to allow us to select one of eight possible
 * precomputed points from the table, and return either that point, or the
 * inverse (negation) of that elliptic curve point, and do it all in (close
 * to) constant time and memory acceses
 *
 * Parameters:
 * point - The preallocated point structure we're supposed to place the
 *         selected point into
 * index - The value 0-7 which indicates which precomputed point to return
 * sign  - If 0, this actually returns the inverse of the precomputed point;
 *         if 1, this is the origianl point
 * table - The precomputed table of the 8 points
 * temp  - A temp buffer that's word_size words long
 * p     - The prime that the curve is defined over (as a bigendian integer
 *         in a buffer that's word_size*sizeof(word) bytes long.  We need
 *         this to compute the inverse of the points
 * word_size - The size of a field element, in word's
 */  
static void ec_select_point( EC_POINT *point, int index, int sign,
         word *table, word *temp, unsigned char *p, int word_size ) {
    int entry_size = 3 * word_size; /* Total size of a precomputed point */
                                    /* (in words) */
    unsigned num_bytes = word_size * sizeof(word); /* Number of bytes */
                                    /* within a single field element */
    int i;
    word m0 = ((word)(index   )&1)-1;
    word m1 = ((word)(index>>1)&1)-1;
    word m2 = ((word)(index>>2)&1)-1;

    memset( temp, 0, entry_size );
    for (i=0; i<8; i++) {
        word mask = m0&m1&m2;   /* 0 if this is not the entry we're */
                                /* interested in, 0xff if it is */
        int j;
        for (j=0; j<entry_size; j++) {
            temp[j] ^= *table++ & mask;
        }
        m2 ^= m0 & m1;          /* Step m0, m1, m2 to the next entry */
        m1 ^= m0;
        m0 = ~m0;
    }
    /* At this point, temp contains the encoding of the selected point */

    /*
     * Now, if sign==0, invert the Y coordinate, that is, replace Y with P-Y,
     * where P is the prime this curve is over; this will invert the point
     * On the other hand, if sign==1; leave it alone 
     *
     * This is made much more interesting by the requirement to do this in
     * constant time.  If we didn't have that restriction, well, we've got this
     * nice Bignum library sitting here; it'd be quite willing to do this
     * calculation; just not in constant time.
     *
     * If sign==1, this effectively does y = y + 0
     * If sign==0, this effectively does y = ~y + p + 1 = p - y
     * Note: this doesn't work if y is initially 0 (as, in that case, we're
     * supposed to map it to 0, not p.  However, y=0 never happens for the
     * curves we support (as that would be a point of order 2, and our curves
     * never has those)
     *
     * Also note that, unlike the above logic, it's not as straightforward to
     * process multiple bytes at once.  That's because these entries are in
     * strict bigendian order; if the CPU happens to be little-endian, trying
     * to do adds on multiple bytes will to do the Wrong Thing (unless we
     * throw in ntoh/hton calls...)
     */
    {
        unsigned char *y = (unsigned char *)(temp + word_size);
        unsigned carry = sign^1;  /* 1 if sign==0, 0 if sign==1 */
            /* If sign==0, we need to set carry to offset the fact that */
            /* ~y = -y - 1; that is, bitwise complementing y also implicitly */
            /* decrements it */

        unsigned invert = (sign-1) & 0xff; /* 0xff if sign==0, 0x00 if sign==1 */
        unsigned mask = invert;    /* ditto */
        for (i=num_bytes-1; i>=0; i--) {
            unsigned next = (invert ^ y[i]) + (mask & p[i]) + carry;
            y[i] = next;
            carry = next >> 8;
        }
    }

    /*
     * We have selected the point, and optimially inverted it.  Now, export
     * the point.  This part is constant time if each of the three
     * coordinates are 'full size' (no leading 0's)
     */
    BN_bin2bn( (unsigned char*)(temp              ), num_bytes, &point->X );
    BN_bin2bn( (unsigned char*)(temp +   word_size), num_bytes, &point->Y );
    BN_bin2bn( (unsigned char*)(temp + 2*word_size), num_bytes, &point->Z );
    point->Z_is_one = 0;
}

/*
 * This takes a point P, and randomizes its representation.  That is, for any
 * elliptic curve point, there are a lot of different (X, Y, Z) values that
 * mean that point. What this routine does is modify the X, Y, Z value within
 * P so that still means the same point, but it is randomly selected over all
 * the possible (X, Y, Z) ways of saying it
 */
static void ec_randomize_representation( const EC_GROUP *group, EC_POINT *P,
                                      BN_CTX *ctx ) {
    BIGNUM *r = BN_CTX_get(ctx);
    const BIGNUM *p = &group->field;
    BIGNUM *s = BN_CTX_get(ctx);
    if (!r || !s) {
        /* Allocation failed.  In this case, I suppose it's Ok to stay with */
        /* the original point (the only other option is to signal failure) */ 
        ECerr(EC_F_EC_RANDOMIZE_REPRESENTATION, ERR_R_MALLOC_FAILURE);
        return;
    }

    /* Select a random value r between 1 and p-1 */
    BN_copy( s, p );
    BN_sub_word( s, 1 ); /* s == group->p - 1 */
    if (!BN_rand_range( r, s )) {
        ECerr(EC_F_EC_RANDOMIZE_REPRESENTATION, ERR_R_INTERNAL_ERROR);
        return;
    }
    BN_add_word( r, 1 );

    /*
     * Now, set x to be xr^3, y to be yr^2 and z to be zr.  Because of how
     * Jacobian coordinates work, this'll be the same point
     */
    BN_mod_mul( &P->Z, &P->Z, r, p, ctx );  /* z *= r */
    BN_mod_sqr( s, r, p, ctx );             /* s = r^2 */
    BN_mod_mul( &P->X, &P->X, s, p, ctx );  /* x *= r^2 */
    BN_mod_mul( s, s, r, p, ctx );          /* s = r^3 */
    BN_mod_mul( &P->Y, &P->Y, s, p, ctx );  /* y *= r^3 */

    P->Z_is_one = 0;
    /* p1, r will be freed once we BN_CTX_end this ctx */
}

/* This gives a/b, however rounding up, rather than down */
#define DIV_ROUND_UP(a, b) (((a) + (b) - 1) / (b))

/*
 * This computes the point mulitiplication R = kH
 * It runs with constant time/memory access, for side channel resistance
 *
 * It uses a variant of Radix 16 multiplication.  That is, we represent
 * the value k as the 4 bit digits a[n], a[n-1], ..., a[0], and perform the
 * computation:
 *     (16**n a[n]) H + (16**(n-1) a[n-1]) H + ... + (16**0 a[0]) H
 * (by computing the possible values of jH, for small j, and then do repeated
 * iterations of multiplying by 16, and adding jH)
 *
 * However, we deviate from the straightforward implementation in two ways:
 * - The addition of 0H isn't even approximately constant time; we need to
 *   avoid that
 * - It happens to be cheap to compute -J given the point J; we'd like to take
 *   advantage of that
 * So, what we do is actually compute the eight values
 * 1H, 3H, 5H, 7H, 9H, 11H, 13H, 15H; and, we logically extend the table by
 * assuming we also have the inverses of those values -1H, -3H, -5H, ..., -15H.
 * When time comes to add kH, we select the appropriate precomputed value
 * from the table (reading all 8 values to gain constant memory access), and
 * if necessary, invert it (logic is in ec_select_point).
 *
 * One glitch is that all our jH values have j odd, which means that we can
 * only multiply by odd values (because the last operations we'll do is
 * multiply by 16, and then add in the last jH value).  We get around that
 * by taking advantage of the group order being odd; if necessary, we'll
 * add in the group order to the scalar, to make the value we're actually
 * multiplying by odd.
 *
 * You might ask: what's the big deal about side channel resistance?  The
 * issue really is ECDSA signature generation; it's fragile, in the sense that
 * if an attacker gets side channel of the random 'k' we select, they can
 * (with effort) recover the private key.  To prevent this, we try fairly hard
 * not to give him a side channel
 */
static int ec_single_point_multiply(const EC_GROUP *group, EC_POINT *R,
                                    const BIGNUM *k, const EC_POINT *H,
                                    BN_CTX *ctx ) {
    int retval = 0, i;
    size_t scratchpad_size;
    const BIGNUM *order = &group->order;
    const BIGNUM *p = &group->field;
    int scalar_size = DIV_ROUND_UP(BN_num_bits(order), 8) + 1; /* Bytes we */
        /* allocate for the scalar (the integer we're multiplying by */
        /* We deliberately set it to one more byte than what we need to */
        /* express the order */
    int word_size = DIV_ROUND_UP(BN_num_bits(&group->field), 8*sizeof(word));
                /* The size of a field element, in words */
    EC_POINT *H2 = 0;  /* 2*H */
    EC_POINT *Hn = R;  /* n*H, for various values of n.  The result will */
                       /* end up here */
    EC_POINT *Hi = 0;  /* Intermediate value from the table */
    word          *scratchpad = 0;
    word          *precompute;   /* Precomputed 1H, ..., 15H values */
    word          *select_temp;  /* Scratchpad space needed by ec_select_point */
    unsigned char *b_p;          /* The curve prime p, in bigendian binary */
    unsigned char *b_order;      /* The curve order, in bidendian binary */
    unsigned char *b_k;          /* The scalar we're multiplying by */
                                 /* in bigendian binary */
    int index;
    unsigned prev_lsbit;
    int idx;

    BN_CTX_start(ctx);

    /* Allocate the scratchpad we'll need for intermediate values */
    scratchpad_size = scalar_size       /* curve order */
                    + scalar_size       /* k */
                    + 3*8*sizeof(word)*word_size /* precompute table */
                    + 3*sizeof(word)*word_size  /* temp space */
                    + sizeof(word)*word_size  /* p */
                    ;
    scratchpad = OPENSSL_malloc( scratchpad_size );
    if (!scratchpad) goto err;
    precompute = scratchpad;
    select_temp = scratchpad + 3*8*word_size;;
    b_p = (unsigned char*)(select_temp + 3*word_size);
    b_order = (unsigned char*)(b_p + word_size * sizeof(word));
    b_k = b_order + scalar_size;

    /*
     * Check if k is out of the expected range (negative, or large enough that
     * it (plus perhaps a few multiples of the order) won't fit in our temp
     * registers).  If it is out of range, reduce it into range.  This isn't
     * constant time, however we never really need to do this in practice
     */
    if (BN_num_bits(k) >= 8*scalar_size-4 || BN_is_negative(k)) {
        BIGNUM *tmp_scalar = BN_CTX_get(ctx);
        if (!tmp_scalar || !BN_nnmod(tmp_scalar, k, &group->order, ctx)) {
            ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
            goto err;
        }
        k = tmp_scalar;
    }

    /*
     * Create binary copies of the group order, the scalar, and the prime the
     * curve is defined over.  Our constant time logic will need all three
     */
    if (!ec_BN_copy_out( b_order, scalar_size, order ) ||
        !ec_BN_copy_out( b_k,     scalar_size, k ) ||
        !ec_BN_copy_out( b_p,     word_size*sizeof(word), p)) {
        ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
	goto err;
    }

    /*
     * The multiplication algorithm we'll use requires us to pick a scalar
     * for which:
     * - The 4 highest bits are in the pattern 0001
     * - The lsbit is set
     * The scalar k that the caller passed us might not (probably won't)
     * happen to meet these critera; however we can fix that.  We can add
     * m*order (where order is the order of the curve, and m is any integer)
     * to that scalar, and it won't change the result.  So, compute the value
     * m such that m*order + k meets both criteria
     * BTW: this requirement is why we allocated one more byte for the scalar;
     * if we allocated the minimum required, they'd be no guarrantee that such
     * an m would exist
     *
     * This assumes that the curve order is odd (which it will be for any
     * prime curve we care about).  If even-order curves do come up (e.g.
     * someone implements a Montgomery curve, but still insists on the
     * generic point multiplication logic), we could special case it by
     * incrementing k by 1 or 2, and then subtracting either H or 2H at the
     * end; possible (idea to make this constant time: place 2H immediately
     * before the precompute table, and then use the ec_select_point logic to
     * select either -2H or -H), but unless we have a case where we can test
     * it out, I don't want to write untested code
     */
    {
        unsigned order_msb = 65536*b_order[0] + 256*b_order[1] + b_order[2];
        /* Order_msb >= 256 */
        unsigned k_msb     = 65536*b_k    [0] + 256*b_k    [1] + b_k    [2];
        /* k < 0x80000 */
        unsigned m = DIV_ROUND_UP(0x100000 - k_msb, order_msb);
        /* This gives us the smallest m with m*order_msb + k_msb > 0x100000 */
        /* That guarrantees that the high order bits of m*order + k have */
        /* the pattern we're looking for */
        unsigned carry;

        m += 1 & ~(m ^ b_k[scalar_size-1]); /* Make sure m*order + k is odd */
            /* Just adding 1 won't disrupt the higher order bit pattern */

        /* b_k += m * b_order, in constant time (naturally) */
        carry = 0;
        for (i=scalar_size-1; i>=0; i--) {
            carry = carry + m * b_order[i] + b_k[i];
            b_k[i] = carry & 0xff;
            carry >>= 8;
        }
    }

    if (!(H2 = EC_POINT_new(group)) ||
        !(Hi = EC_POINT_new(group))) {
        ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_MALLOC_FAILURE);
	goto err;
    }

    /*
     * We now fill out the precompute table with the values
     *     1H, 3H, 5H, 7H, 9H, 11H, 13H, 15H
     *
     * One glitch; the value 1H does have a subtle issue. The input point
     * sometimes really is in affine coordinates (that is, has Z=1); one common
     * case is when H really is the curve generator. That would cause an issue
     * because the Point Addition code does have sidechannels for very small
     * coordinates (such as Z=1), as the Bignum library they're based on
     * doesn't attempt to be constant time.  To avoid this being a potential
     * leakage, we randomize the representation of the point we fill in for 1H
     * (which will also happens to randomize all the other computations we do;
     * not exactly a bad thing)
     */
    if (!EC_POINT_copy(Hn, H)) {
        ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
	goto err;
    }
    ec_randomize_representation( group, Hn, ctx );
    if (!EC_POINT_dbl(group, H2, H, ctx)) {
        ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
	goto err;   /* H2 = 2H */
    }

    for (i=0;; i++) {
        if (!ec_BN_copy_out( (unsigned char*)(precompute + (0 + 3*i) * word_size),
                          sizeof(word) * word_size, &Hn->X ) ||
            !ec_BN_copy_out( (unsigned char*)(precompute + (1 + 3*i) * word_size),
                          sizeof(word) * word_size, &Hn->Y ) ||
            !ec_BN_copy_out( (unsigned char*)(precompute + (2 + 3*i) * word_size),
                          sizeof(word) * word_size, &Hn->Z )) {
            ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
            goto err;
        }
        if (i == 7) break;  /* Stop after we've generated all 8 */
            /* Hn += 2H */
        if (!EC_POINT_add(group, Hn, Hn, H2, ctx)) {
            ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
	    goto err;
	}
    }

    /*
     * Now, we need to set Hn = 16H (to start off the chain).  We could go
     * back from scratch, and double H four times to get 16H.  However, Hn
     * just happens to be 15H here; just add one more H, and we're there
     */
    if (!EC_POINT_add(group, Hn, Hn, H, ctx)) {
        ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
	goto err;
    }

    /* Add in the first 4 bits (which will always be a positive) */
    index = b_k[0] & 0xf;
    prev_lsbit = index & 1;
    index >>= 1;
    ec_select_point( Hi, index, 1, precompute, select_temp, b_p, word_size );
    if (!EC_POINT_add(group, Hn, Hn, Hi, ctx)) {
        ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
	goto err; /* Hn += Hi */
    }

    /*
     * Here, the value of Hn is (1 a b c 1)H, where (1 a b c 1) is the binary
     * representation, and the msbyte of b_k is (in binary) 0 0 0 1 a b c x
     * (the lsbit x hasn't been stirred in yet)
     */
    for (idx=1; idx<scalar_size; idx++) {
        int shift;
        int nybble_lsbit;
        for (shift=4; shift>=0; shift-=4) {
            /*
             * Here, the value of Hn is (... s s s 1)H, where s are the bits
             * we've already processed
             */
            int k;
            for (k=0; k<4; k++) {
                if (!EC_POINT_dbl(group, Hn, Hn, ctx)) {
                    ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
		    goto err;
		}
            }
            /*
             * Now, the value of Hn is (... s s s 1 0 0 0 0)H
             */

            index = (b_k[idx] >> shift) & 0x0f;
            nybble_lsbit = index&1;
            index >>= 1;   /* We're interested in bits 3-1 */

            /*
             * The next four bits we want to stir in are prev_lsbit,
             * index[3], index[2], index[1] (where prev_lsbit is actually
             * the lsbit from the previous nybble, and index[3-1] are bits 3-1
             * from this nybble (and we won't stir in the nybble lsbit this
             * iteration)
             *
             * If prev_lsbit is 1, we want to add (2*index+1)H to the point; 
             * we have that value precomputed in our table
             *
             * If prev_lsbit is 0, we want to add (-16 + 2*index+1)H =
             * -(2*(7-index)+1)H to it; the -16 to turn the bit which is set
             * clear, and the 2*index+1 to set the bits we want to set.
             * We can find this in the table by converting index into 7-index,
             * and then inverting the entry we find in the table
             *
             * So, if prev_lsbit is 0, replace index with 7-index
             */
            index ^= (prev_lsbit-1) & 0x07;   /* If sign is zero, we'll flip */
                /* the bits of the index we look up (in constant time) */

            /*
             * Get the precomputed index; if prev_lsbit is 0, we want the
             * inverse of the point
             */
            ec_select_point( Hi, index, prev_lsbit, precompute, select_temp,
                          b_p, word_size );
            /* Hn += Hi */
            if (!EC_POINT_add(group, Hn, Hn, Hi, ctx)) {
                ECerr(EC_F_EC_SINGLE_POINT_MULTIPLY, ERR_R_INTERNAL_ERROR);
		goto err;
	    }
            /*
             * Now, the value of Hn is (... s s s d c b a 1)H,
             * where d, c, b, a are the bits we processed this iteration
             */

            prev_lsbit = nybble_lsbit; /* Preserve the lsbit from this nybble */
                                  /* for the next one */
        }
    }
    /*
     * Now, the value of Hn is (... s s s 1 )H, where 's' are now all the bits
     * from k (except for the last; we made sure that lsbit is 1 
     */

    retval = 1;   /* Success */
err:
    if (H2 != NULL) EC_POINT_free(H2);
    if (Hi != NULL) EC_POINT_free(Hi);
    BN_CTX_end(ctx);
    if (scratchpad) {
        OPENSSL_cleanse(scratchpad, scratchpad_size);
        OPENSSL_free(scratchpad);
    }
     
    return retval;
}
#endif
