#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "../e_os.h"

#if defined(OPENSSL_NO_EC) || defined(OPENSSL_NO_C25519)
int main(int argc, char *argv[])
{
    puts("Elliptic curves are disabled or curve25519 is disabled.");
    return 0;
}
#else

#include <openssl/ec.h>
#include <openssl/ecdh.h>
#include <openssl/err.h>
#ifdef OPENSSL_FIPS
# include <openssl/fips.h>
#endif

#include <openssl/evp.h>
#include <openssl/sha.h>
static const int KDF1_SHA1_len = 20;
static void *KDF1_SHA1(const void *in, size_t inlen, void *out, size_t *outlen)
{
	const EVP_MD *md = EVP_sha1();
	EVP_MD_CTX *mdctx;

#ifndef OPENSSL_NO_SHA
	if (*outlen < SHA_DIGEST_LENGTH)
		return NULL;
	else
		*outlen = SHA_DIGEST_LENGTH;

	mdctx = EVP_MD_CTX_create();
	EVP_DigestInit_ex(mdctx, md, NULL);
	EVP_DigestUpdate(mdctx, in, inlen);
	EVP_DigestFinal_ex(mdctx, out, (unsigned int*)outlen);
	EVP_MD_CTX_destroy(mdctx);
	return (out);
#else
	return NULL;
#endif
}


/*
 * Performs a simple public/private pair key generation
 * using curve25519.
 *
 * Returns 0 on success.
 */
static int test_keygen()
{
    int rv = 1;
    EC_GROUP *group = NULL;
    EC_KEY *K = NULL;

    printf("testing curve25519 keygen\n");

#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
	if (!FIPS_mode_set(0)) {
	    ERR_load_crypto_strings();
	    ERR_print_errors_fp(stderr);
	    exit(1);
	}
    }
#endif
    ERR_clear_error();

    /*
     * First try to create a key using EC_GROUP_new method
     */
    group = EC_GROUP_new(EC_c25519_method());
    if (!group) {
	printf("\nFailed to create EC_GROUP for donna curve.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto keygen_done;
    }

    K = EC_KEY_new();
    if (!K) {
	printf("\nFailed to create EC_KEY for donna curve.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto keygen_done;
    }

    EC_KEY_set_group(K, group);

    if (!EC_KEY_generate_key(K)) {
	printf("\nFailed to generate key for donna curve(1).\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto keygen_done;
    }
    if (K) EC_KEY_free(K);

    /*
     * Now lets create a key using the EC_KEY_new_by_curve_name method
     */
    K = EC_KEY_new_by_curve_name(NID_x25519);
    if (!K) {
	printf("\nFailed to create EC_KEY using _by_curve_name method.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto keygen_done;
    }

    if (!EC_KEY_generate_key(K)) {
	printf("\nFailed to generate key for donna curve(2).\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto keygen_done;
    }

#ifdef OPENSSL_FIPS
    if (!FIPS_mode_set(1)) {
        ERR_load_crypto_strings();
        ERR_print_errors_fp(stderr);
	exit(1);
    }

    printf("testing curve25519 keygen with FIPS enabled\n");
    if (EC_KEY_generate_key(K)) {
	printf("\nFailed to block curve25519 keygen while FIPS is enabled.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto keygen_done;
    }
#endif

    printf("Curve25519 key generation finished w/o errors.\n");
    rv = 0;

keygen_done:

    if (K) EC_KEY_free(K);
    if (group) EC_GROUP_free(group);
    return rv;
}

/*
 * Performs a simple pairwise shared secret computation using
 * two curve25519 key pairs. 
 *
 * Returns 0 on success.
 */
static int test_pairwise()
{
    int rv = 1;
    EC_KEY *K1 = NULL;
    EC_KEY *K2 = NULL;
    int alen, aout;
    unsigned char *abuf1 = NULL;
    unsigned char *abuf2 = NULL;

    printf("testing curve25519 shared secret computation\n");

#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
	if (!FIPS_mode_set(0)) {
	    ERR_load_crypto_strings();
	    ERR_print_errors_fp(stderr);
	    exit(1);
	}
    }
#endif
    ERR_clear_error();

    /*
     * Create the first key pair 
     */
    K1 = EC_KEY_new_by_curve_name(NID_x25519);
    if (!K1) {
	printf("\nFailed to create EC_KEY using _by_curve_name method.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto pairwise_done;
    }
    if (!EC_KEY_generate_key(K1)) {
	printf("\nFailed to generate key for donna curve(2).\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto pairwise_done;
    }

    /*
     * Create the second key pair 
     */
    K2 = EC_KEY_new_by_curve_name(NID_x25519);
    if (!K2) {
	printf("\nFailed to create EC_KEY using _by_curve_name method.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto pairwise_done;
    }
    if (!EC_KEY_generate_key(K2)) {
	printf("\nFailed to generate key for donna curve(2).\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto pairwise_done;
    }

    alen = KDF1_SHA1_len;
    abuf1 = (unsigned char *)OPENSSL_malloc(alen);
    aout = ECDH_compute_key(abuf1,alen,EC_KEY_get0_public_key(K2),K1,KDF1_SHA1);
    if (aout <= 0) {
        printf("ECDH computation 1 failed!!!\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto pairwise_done;
    }

    abuf2 = (unsigned char *)OPENSSL_malloc(alen);
    aout = ECDH_compute_key(abuf2,alen,EC_KEY_get0_public_key(K1),K2,KDF1_SHA1);
    if (aout <= 0) {
        printf("ECDH computation 2 failed!!!\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto pairwise_done;
    }

    if (memcmp(abuf1, abuf2, alen)) {
	printf("ECDH computation mismatch!!!\n");
	goto pairwise_done;
    }

#ifdef OPENSSL_FIPS
    if (!FIPS_mode_set(1)) {
        ERR_load_crypto_strings();
        ERR_print_errors_fp(stderr);
	exit(1);
    }

    printf("testing curve25519 shared secret compuation with FIPS enabled\n");
    aout = ECDH_compute_key(abuf2,alen,EC_KEY_get0_public_key(K1),K2,KDF1_SHA1);
    if (aout > 0) {
        printf("ECDH computation worked in FIPS mode!!!\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto pairwise_done;
    }
#endif

    printf("shared secret computation finished w/o errors.\n");
    rv = 0;

pairwise_done:
    if (K1) EC_KEY_free(K1);
    if (K2) EC_KEY_free(K2);
    if (abuf1) OPENSSL_free(abuf1);
    if (abuf2) OPENSSL_free(abuf2);
    return rv;
}


static int test_serialize_key(EC_KEY *key, unsigned char **buf, int *len)
{
    int i;
    const EC_GROUP *group;

    group = EC_KEY_get0_group(key);

    *len = EC_POINT_point2oct(group, EC_KEY_get0_public_key(key), 
	    POINT_CONVERSION_UNCOMPRESSED, NULL, 0, NULL);

    *buf = (unsigned char *) OPENSSL_malloc(*len * sizeof(unsigned char));

    i = EC_POINT_point2oct(group, EC_KEY_get0_public_key(key), POINT_CONVERSION_UNCOMPRESSED,
                                       *buf, *len, NULL);
    return i;
}

/*
 * Performs a simple pairwise shared secret computation using
 * two curve25519 key pairs.  However, we use EC_POINT_oct2point
 * and EC_POINT_point2oct to serialize the public keys.  This
 * is how the TLS stack would use curve25519 for shared secret
 * computation.
 *
 * Returns 0 on success.
 */
static int test_serialize()
{
    int rv = 1;
    EC_KEY *K1 = NULL;
    EC_KEY *K2 = NULL;
    EC_POINT *K1_pub = NULL;
    EC_POINT *K2_pub = NULL;
    int alen, aout;
    unsigned char *abuf1 = NULL;
    unsigned char *abuf2 = NULL;
    unsigned char *pub1 = NULL;
    unsigned char *pub2 = NULL;
    int pub1_len = 0;
    int pub2_len = 0;
    int i;

    printf("testing curve25519 pub key serialization\n");

#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
	if (!FIPS_mode_set(0)) {
	    ERR_load_crypto_strings();
	    ERR_print_errors_fp(stderr);
	    exit(1);
	}
    }
#endif
    ERR_clear_error();

    /*
     * Create the first key pair 
     */
    K1 = EC_KEY_new_by_curve_name(NID_x25519);
    if (!K1) {
	printf("\nFailed to create EC_KEY using _by_curve_name method.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }
    if (!EC_KEY_generate_key(K1)) {
	printf("\nFailed to generate key for donna curve(2).\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }

    /*
     * Create the second key pair 
     */
    K2 = EC_KEY_new_by_curve_name(NID_x25519);
    if (!K2) {
	printf("\nFailed to create EC_KEY using _by_curve_name method.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }
    if (!EC_KEY_generate_key(K2)) {
	printf("\nFailed to generate key for donna curve(2).\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }

    /*
     * Serialize the public key of K1
     */
    i = test_serialize_key(K1, &pub1, &pub1_len);
    if (i != pub1_len) {
	printf("\nFailed to serialize K1.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }

    /*
     * Serialize the public key of K2
     */
    i = test_serialize_key(K2, &pub2, &pub2_len);
    if (i != pub2_len) {
	printf("\nFailed to serialize K2.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }

    /*
     * Rebuild K1 pub from it's octet stream
     */
    K1_pub = EC_POINT_new(EC_KEY_get0_group(K1));
    i = EC_POINT_oct2point(EC_KEY_get0_group(K1), K1_pub, pub1, pub1_len, NULL);
    if (!i) {
	printf("\nFailed to de-serialize K1.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }

    /*
     * Rebuild K2 pub from it's octet stream
     */
    K2_pub = EC_POINT_new(EC_KEY_get0_group(K2));
    i = EC_POINT_oct2point(EC_KEY_get0_group(K2), K2_pub, pub2, pub2_len, NULL);
    if (!i) {
	printf("\nFailed to de-serialize K2.\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }

    /*
     * Perform the pairwise computations
     */
    alen = KDF1_SHA1_len;
    abuf1 = (unsigned char *)OPENSSL_malloc(alen);
    aout = ECDH_compute_key(abuf1,alen,K2_pub,K1,KDF1_SHA1);
    if (aout <= 0) {
        printf("ECDH computation 1 failed!!!\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }

    abuf2 = (unsigned char *)OPENSSL_malloc(alen);
    aout = ECDH_compute_key(abuf2,alen,K1_pub,K2,KDF1_SHA1);
    if (aout <= 0) {
        printf("ECDH computation 2 failed!!!\n");
	ERR_load_crypto_strings();
	ERR_print_errors_fp(stderr);
	goto serialize_done;
    }

    if (memcmp(abuf1, abuf2, alen)) {
	printf("ECDH computation mismatch!!!\n");
	goto serialize_done;
    }

    printf("shared pub key serialization finished w/o errors.\n");
    rv = 0;

serialize_done:
    if (K1) EC_KEY_free(K1);
    if (K2) EC_KEY_free(K2);
    if (K1_pub) EC_POINT_free(K1_pub);
    if (K2_pub) EC_POINT_free(K2_pub);
    if (abuf1) OPENSSL_free(abuf1);
    if (abuf2) OPENSSL_free(abuf2);
    if (pub1) OPENSSL_free(pub1);
    if (pub2) OPENSSL_free(pub2);
    return rv;
}

int main(int argc, char *argv[])
{
    int rv;
    int result = -1;
    int i;

    /* memory leak debugging */
    CRYPTO_malloc_debug_init();
    CRYPTO_set_mem_debug_options(V_CRYPTO_MDEBUG_ALL);
    CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);
    ERR_load_crypto_strings();

    rv = test_keygen();
    if (rv) {
	printf("curve25519 keygen tests failed.\n");
	goto main_done;
    }

    for (i=0; i<1000; i++) {
	rv = test_pairwise();
	if (rv) {
	    printf("curve25519 pairwise tests failed.\n");
	    goto main_done;
	}
	rv = test_serialize();
	if (rv) {
	    printf("curve25519 serialization tests failed.\n");
	    goto main_done;
	}
    }

    result = 0;
main_done:

    /* memory debugging cleanup */
    CRYPTO_cleanup_all_ex_data();
    ERR_free_strings();
    ERR_remove_state(0);
    CRYPTO_mem_leaks_fp(stderr);

    return result;
}
#endif


