/* crypto/ecdsa/ecdsa_sign.c */
/* ====================================================================
 * Copyright (c) 1998-2002 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@OpenSSL.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include "ecs_locl.h"
#include "../ec/ec_lcl.h"
#ifndef OPENSSL_NO_ENGINE
# include <openssl/engine.h>
#endif
#include <openssl/rand.h>
#include <openssl/evp.h>
#include <openssl/ec.h>

/* Externally declared buffer used for DRBG adin stir */
extern unsigned char global_drbg_adin_buf[48];

/* ecdsa_drbg_stir
 * Description: Create unique data for a particular ECDSA message to be used
 *      as additional input during the generation of 800-90A DRBG bytes.
 *      Use the dgst as input to an EVP_Digest routine cycle which
 *      then creates a SHA256 hash digest as output. The output, of size 32 bytes,
 *      is temporarily stored in lcl_hash_buf[] and then exclusive-ored (XOR) with
 *      the first 32 bytes of the global_drbg_adin_buf[].
 * Parameters:
 *      dgst - const unsigned char pointer to an array of bytes that represent
 *             a currently exisiting digest of the message being signed.
 *      dlen - The length of dgst, representing the number of bytes because
 *             dgst is of type unsigned char.
 * Returns:
 *       0 - the function encountered an error.
 *       1 - the function was successful.
 * */
static int ecdsa_drbg_stir(const unsigned char *dgst, int dlen)
{
    int rv = 1;
    int j = 0;
    unsigned int gen_len = 0;
    const EVP_MD *hash_md = NULL;
    EVP_MD_CTX *hash_ctx = NULL;
    unsigned char lcl_hash_buf[32] = {0};

    /* Setup the hash method for DRBG adin stir */
    if (!(hash_ctx = EVP_MD_CTX_create()) || !(hash_md = EVP_sha256())) {
        rv = 0;
        ECDSAerr(ECDSA_F_ECDSA_DRBG_STIR, ECDSA_R_DIGEST_MALLOC_FAILED);
        goto end;
    }

    /* SHA256 hash the message digest, and XOR results in global buf for DRBG adin */
    if (EVP_DigestInit_ex(hash_ctx, hash_md, NULL) == 0) {
        rv = 0;
        goto end;
    }
    if (EVP_DigestUpdate(hash_ctx, dgst, (size_t)dlen) == 0) {
        rv = 0;
        goto end;
    }
    if (EVP_DigestFinal(hash_ctx, lcl_hash_buf, &gen_len) == 0) {
        rv = 0;
        goto end;
    }
    if (gen_len != 32) {
        rv = 0;
        goto end;
    }

    for (j = 0; j < 32; j++) {
        global_drbg_adin_buf[j] ^= lcl_hash_buf[j];
    }

    end:
        if (hash_ctx) {
            EVP_MD_CTX_destroy(hash_ctx);
        }
        return rv;
}

ECDSA_SIG *ECDSA_do_sign(const unsigned char *dgst, int dlen, EC_KEY *eckey)
{
    return ECDSA_do_sign_ex(dgst, dlen, NULL, NULL, eckey);
}

ECDSA_SIG *ECDSA_do_sign_ex(const unsigned char *dgst, int dlen,
                            const BIGNUM *kinv, const BIGNUM *rp,
                            EC_KEY *eckey)
{
    ECDSA_DATA *ecdsa = ecdsa_check(eckey);
    if (ecdsa == NULL)
        return NULL;

    if (!ecdsa_drbg_stir(dgst, dlen))
        return NULL;

    return ecdsa->meth->ecdsa_do_sign(dgst, dlen, kinv, rp, eckey);
}

int ECDSA_sign(int type, const unsigned char *dgst, int dlen, unsigned char
               *sig, unsigned int *siglen, EC_KEY *eckey)
{
    return ECDSA_sign_ex(type, dgst, dlen, sig, siglen, NULL, NULL, eckey);
}

int ECDSA_sign_ex(int type, const unsigned char *dgst, int dlen, unsigned char
                  *sig, unsigned int *siglen, const BIGNUM *kinv,
                  const BIGNUM *r, EC_KEY *eckey)
{
    ECDSA_SIG *s;

#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
	if (!eckey->group) {
	    ECDSAerr(ECDSA_F_ECDSA_SIGN_EX, EC_R_INVALID_CURVE);
	    return 0;
	}
	/*
	 * Verify that this curve is allowed in FIPS mode
	 */
	if (!EC_curve_nid2nist(eckey->group->curve_name)) {
	    ECDSAerr(ECDSA_F_ECDSA_SIGN_EX, EC_R_NON_FIPS_METHOD);
	    return 0;
	}
    }
#endif

    RAND_seed(dgst, dlen);
    s = ECDSA_do_sign_ex(dgst, dlen, kinv, r, eckey);
    if (s == NULL) {
        *siglen = 0;
        return 0;
    }
    *siglen = i2d_ECDSA_SIG(s, &sig);
    ECDSA_SIG_free(s);
    return 1;
}

int ECDSA_sign_setup(EC_KEY *eckey, BN_CTX *ctx_in, BIGNUM **kinvp,
                     BIGNUM **rp)
{
    ECDSA_DATA *ecdsa = ecdsa_check(eckey);
    if (ecdsa == NULL)
        return 0;
    return ecdsa->meth->ecdsa_sign_setup(eckey, ctx_in, kinvp, rp);
}
