/* -----------------------------------------------------------------
 * crypto/fips/sp800131_test.c - Test application to test FIPS mode
 * complies with key size and algorithm constraints defined in 
 * SP-800-131. 
 *
 * May, 2016
 *
 * Copyright (c) 2016 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <openssl/opensslconf.h>
#include <openssl/err.h>

#include <openssl/hmac.h>

#ifndef OPENSSL_FIPS
int main(int argc, char *argv[])
{
    puts("This test suite is not used for non-FIPS builds.");
    return 0;
}

#else

#include <openssl/evp.h>
#include <openssl/dh.h>
#include <openssl/rand.h>
#include <openssl/pem.h>

static const char rnd_seed[] = "string to make the random number generator think it has entropy";
static BIO *bio_err = NULL;

#define SP800131_DISABLE_FIPS     if (FIPS_mode()) { \
       if (!FIPS_mode_set(0)) { \
          fprintf(stderr, "Failed to disable FIPS mode\n"); \
          return 1; \
       } \
    }

#define SP800131_ENABLE_FIPS     if (!FIPS_mode()) { \
       if (!FIPS_mode_set(1)) { \
          fprintf(stderr, "Failed to enable FIPS mode\n"); \
          return 1; \
       } \
    }


#define DH_PARMS_1024	    "../apps/dh1024.pem"
#define DH_PARMS_2048	    "../apps/dh2048.pem"


static DH * sp800131_dh_parms_load(char *dh_file)
{
    DH *dh = NULL;
    BIO *in;

    in = BIO_new(BIO_s_file());
    if (in == NULL) {
        ERR_print_errors(bio_err);
        return NULL;
    }
    if (BIO_read_filename(in, dh_file) <= 0) {
        ERR_print_errors(bio_err);
	return NULL;
    }

    dh = PEM_read_bio_DHparams(in, NULL, NULL, NULL);
    if (!dh) {
        ERR_print_errors(bio_err);
	return NULL;
    }

    BIO_free_all(in);
    return dh;
}


/*
 * This routine will preform various DH test cases to ensure
 * key size limitations are properly enforced when FIPS is
 * enabled (as defined in SP-800-131.
 *
 * Returns zero on success, non-zero on failure.
 */
static int sp800131_dh_test()
{
    int rv = 1;
    DH *dh_a = NULL, *dh_b = NULL;
    unsigned char *abuf = NULL, *bbuf = NULL;
    int alen, blen;
    int aout, bout;

    /*
     * Allocate a couple of DH key objects 
     */
    dh_a = sp800131_dh_parms_load(DH_PARMS_1024);
    dh_b = sp800131_dh_parms_load(DH_PARMS_1024);
    if (!dh_a || !dh_b) {
	fprintf(stderr, "Unable to read DH parms 1024 file\n"); 
	goto dh_err;
    }

    /* 
     * Set a to run with normal modexp and b to use constant time 
     */
    dh_a->flags &= ~DH_FLAG_NO_EXP_CONSTTIME;
    dh_b->flags |= DH_FLAG_NO_EXP_CONSTTIME;

    /*
     * Generate the two key pairs 
     */
    if (!DH_generate_key(dh_a)) {
	fprintf(stderr, "DH_generate_key failed for dh_a (FIPS=%d)\n", FIPS_mode()); 
        goto dh_err;
    }
    if (!DH_generate_key(dh_b)) {
	fprintf(stderr, "DH_generate_key failed for dh_b (FIPS=%d)\n", FIPS_mode()); 
        goto dh_err;
    }

    /*
     * Compute shared secret using A private and B public
     */
    alen = DH_size(dh_a);
    abuf = (unsigned char *)OPENSSL_malloc(alen);
    aout = DH_compute_key(abuf, dh_b->pub_key, dh_a);
    if (aout < 4) { 
	fprintf(stderr, "DH_compute_key failed using dh_a public (FIPS=%d)\n", FIPS_mode()); 
	goto dh_err;
    }

    /*
     * Compute shared secret using B private and A public
     */
    blen = DH_size(dh_b);
    bbuf = (unsigned char *)OPENSSL_malloc(blen);
    bout = DH_compute_key(bbuf, dh_a->pub_key, dh_b);
    if (bout < 4) { 
	fprintf(stderr, "DH_compute_key failed using dh_b public (FIPS=%d)\n", FIPS_mode()); 
	goto dh_err;
    }

    /*
     * Compare the two secrets
     */
    if (memcmp(abuf, bbuf, aout)) {
	fprintf(stderr, "DH shared secrets do not match (FIPS=%d)\n", FIPS_mode()); 
	goto dh_err;
    }

    DH_free(dh_a);
    dh_a = NULL;
    DH_free(dh_b);
    dh_b = NULL;
    OPENSSL_free(abuf);
    abuf = NULL;
    OPENSSL_free(bbuf);
    bbuf = NULL;

    /*
     * The rest of the test cases require FIPS to be enabled 
     */
    if (!FIPS_mode_set(1)) {
	fprintf(stderr, "Failed to enable FIPS mode\n");
	goto dh_err;
    }

    /*
     * Allocate a couple of DH key objects 
     */
    dh_a = sp800131_dh_parms_load(DH_PARMS_1024);
    dh_b = sp800131_dh_parms_load(DH_PARMS_1024);
    if (!dh_a || !dh_b) {
	fprintf(stderr, "Unable to read DH parms 1024 file\n"); 
	goto dh_err;
    }

    /*
     * Generate private key for B 
     */
    dh_a->flags |= DH_FLAG_NON_FIPS_ALLOW; //Need this since key is 1024 bits
    if (!DH_generate_key(dh_a)) {
	fprintf(stderr, "DH_generate_key failed for dh_a (FIPS=%d)\n", FIPS_mode()); 
        goto dh_err;
    }
    dh_b->flags |= DH_FLAG_NON_FIPS_ALLOW; //Need this since key is 1024 bits
    if (!DH_generate_key(dh_b)) {
	fprintf(stderr, "DH_generate_key failed for dh_b (FIPS=%d)\n", FIPS_mode()); 
        goto dh_err;
    }

    /*
     * Attempt to use a 1024 bit key, this should fail with FIPS enabled
     */
    dh_b->flags &= ~DH_FLAG_NON_FIPS_ALLOW;
    blen = DH_size(dh_b);
    bbuf = (unsigned char *)OPENSSL_malloc(blen);
    bout = DH_compute_key(bbuf, dh_a->pub_key, dh_b);
    if (bout >= 0) { 
	fprintf(stderr, "DH_compute_key did not fail with key size 1024 (FIPS=%d, rv=%d)\n", 
		FIPS_mode(), bout); 
	goto dh_err;
    }
    ERR_clear_error();

    /*
     * Now try the NON_FIPS_ALLOW flag on the A key.  The A key was
     * created before FIPS was enabled, so it's not using the
     * FIPS method.
     */
    dh_b->flags |= DH_FLAG_NON_FIPS_ALLOW;
    bout = DH_compute_key(bbuf, dh_a->pub_key, dh_b);
    if (bout <= 0) { 
	fprintf(stderr, "DH_compute_key with NON_FIPS_ALLOW failed with key size 1024 (FIPS=%d, rv=%d)\n", 
		FIPS_mode(), bout); 
	goto dh_err;
    }

    DH_free(dh_a);
    dh_a = NULL;
    DH_free(dh_b);
    dh_b = NULL;
    OPENSSL_free(abuf);
    abuf = NULL;
    OPENSSL_free(bbuf);
    bbuf = NULL;

    /*
     * Allocate a couple of 2048 bit DH key objects 
     */
    dh_a = sp800131_dh_parms_load(DH_PARMS_2048);
    dh_b = sp800131_dh_parms_load(DH_PARMS_2048);
    if (!dh_a || !dh_b) {
	fprintf(stderr, "Unable to read DH parms 2048 file\n"); 
	goto dh_err;
    }

    /* 
     * Set a to run with normal modexp and b to use constant time 
     */
    dh_a->flags &= ~DH_FLAG_NO_EXP_CONSTTIME;
    dh_b->flags |= DH_FLAG_NO_EXP_CONSTTIME;

    /*
     * Generate the two key pairs 
     */
    if (!DH_generate_key(dh_a)) {
	fprintf(stderr, "DH_generate_key failed for dh_a (FIPS=%d)\n", FIPS_mode()); 
        goto dh_err;
    }
    if (!DH_generate_key(dh_b)) {
	fprintf(stderr, "DH_generate_key failed for dh_b (FIPS=%d)\n", FIPS_mode()); 
        goto dh_err;
    }

    /*
     * Compute shared secret using A private and B public
     */
    alen = DH_size(dh_a);
    abuf = (unsigned char *)OPENSSL_malloc(alen);
    aout = DH_compute_key(abuf, dh_b->pub_key, dh_a);
    if (aout < 4) { 
	fprintf(stderr, "DH_compute_key failed using dh_a public (FIPS=%d)\n", FIPS_mode()); 
	goto dh_err;
    }

    /*
     * Compute shared secret using B private and A public
     */
    blen = DH_size(dh_b);
    bbuf = (unsigned char *)OPENSSL_malloc(blen);
    bout = DH_compute_key(bbuf, dh_a->pub_key, dh_b);
    if (bout < 4) { 
	fprintf(stderr, "DH_compute_key failed using dh_b public (FIPS=%d)\n", FIPS_mode()); 
	goto dh_err;
    }

    /*
     * Compare the two secrets
     */
    if (memcmp(abuf, bbuf, aout)) {
	fprintf(stderr, "DH shared secrets do not match (FIPS=%d)\n", FIPS_mode()); 
	goto dh_err;
    }

    rv = 0;
dh_err:
    if (abuf) OPENSSL_free(abuf);
    if (bbuf) OPENSSL_free(bbuf);
    if (dh_a) DH_free(dh_a);
    if (dh_b) DH_free(dh_b);
    return rv;
}


/*
 * This routine will invoke various HMAC functions to ensure
 * key size limitations are properly enforced when FIPS is
 * enabled (as defined in SP-800-131).
 *
 * key_len  - the key_length to use
 * set_flag - when 1, set the EVP_MD_CTX_FLAG_NON_FIPS_ALLOW flag
 * 
 * Returns zero on success, non-zero on failure.
 */
static int sp800131_hmac_invoke(int key_len, int set_flag)
{
    int rv = 1;
    int i;
    unsigned char *result = NULL;
    const char key[] = "etaonrishd1234";  /* 14 * 8 = 112 bits */
    char data[] = "Sample text";
    unsigned int hash_len = 20;
    HMAC_CTX ctx;

    HMAC_CTX_init(&ctx);

    if (set_flag) {
        HMAC_CTX_set_flags(&ctx, EVP_MD_CTX_FLAG_NON_FIPS_ALLOW);
    }
 
    // Using sha1 hash engine here.
    // You may use other hash engines. e.g EVP_md5(), EVP_sha224, EVP_sha512, etc
    result = OPENSSL_malloc(sizeof(char) * hash_len);

    if (! HMAC_Init_ex(&ctx, key, key_len, EVP_sha1(), NULL)) {
       fprintf(stderr, "Failed to initialize  HMAC\n");
       goto hmac_err;
    }

    if (!HMAC_Update(&ctx, (unsigned char*)&data, strlen(data))) {
       fprintf(stderr, "Failed to update HMAC\n");
       goto hmac_err;
    }

    if (!HMAC_Final(&ctx, result, &hash_len)) {
       fprintf(stderr, "Failed HMAC final\n");
       goto hmac_err;
    }

    printf("HMAC digest: ");
    for (i = 0; i != hash_len; i++) {
        printf("%02x", (unsigned int)result[i]);
    }
    printf("\n");
 
    rv = 0;
    ERR_clear_error();

hmac_err:
    HMAC_CTX_cleanup(&ctx);
    if (result != NULL) {
        OPENSSL_free(result);
    }
 
    return rv;
}


/*
 * This routine will preform various HMAC test cases to ensure
 * key size limitations are properly enforced when FIPS is
 * enabled (as defined in SP-800-131).
 *
 * Returns zero on success, non-zero on failure.
 */
static int sp800131_hmac_test(void)
{
    /* Disable FIPS */
    if (FIPS_mode()) {
       if (!FIPS_mode_set(0)) {
          fprintf(stderr, "Failed to disable FIPS mode in HMAC\n");
          return 1;
       }
    }

    if (sp800131_hmac_invoke(14, 0)) {
        /* expect everything to work - length is good */
	return 1;
    }
    if (sp800131_hmac_invoke(10, 1)) {
        /* expect everything to work - length not important */
	return 1;
    }
    if (sp800131_hmac_invoke(10, 0)) {
        /* expect everything to work - length not important */
	return 1;
    }

    /* Enable FIPS */
    if (!FIPS_mode()) {
       if (!FIPS_mode_set(1)) {
          fprintf(stderr, "Failed to enable FIPS mode in HMAC\n");
          return 1;
       }
    }

    if (sp800131_hmac_invoke(14, 0)) {
        /* expect everything to work - length is good */
	return 1;
    }
    if (sp800131_hmac_invoke(10, 1)) {
        /* expect everything to work - length too short, but flag is set */
	return 1;
    }
    if (!sp800131_hmac_invoke(10, 0)) {
        /* expect failure - length too short (and no flag) */
	return 1;
    }

    /* Disable Fips */
    if(FIPS_mode()) {
        if (!FIPS_mode_set(0)) {
	         fprintf(stderr, "Failed to disable FIPS mode\n");
            return 1;
        }
    }

    ERR_clear_error();
    return 0;
}

/*
 * This routine will invoke EVP encrypt and decrypt functions
 * to ensure the limitations of using des ciphers when FIPS mode is
 * enabled is properly enforced (as defined in SP-800-131).
 *
 * type - cipher to use
 * key - key to use for encryption/decryption
 * iv - iv for encryption/decryption
 * out - output for encrypted/decrypted in
 * outl - output for number of bytes encrypted/decrypted
 * in - plaintext or ciphertext input
 * inl - length of in
 * check - expected plaintext output
 * encrypt - 1 for encryption, 0 for decryption
 * set_flag - 1 set EVP_CIPH_NON_FIPS_ALLOW
 *
 * returns zero on success and one on failure
 */
static int sp800131_encrypt_des(const EVP_CIPHER *type, unsigned char *key, unsigned char *iv,
                                unsigned char *out, int *outl, unsigned char *in , int inl,
                                unsigned char *check, int encrypt, int set_flag)
{

    int rv = 1;
    EVP_CIPHER_CTX *ctx = NULL;
    /* initialize context */
    ctx = EVP_CIPHER_CTX_new();
    EVP_CIPHER_CTX_init(ctx);

    if(set_flag) {
        ctx->flags |= EVP_CIPH_FLAG_NON_FIPS_ALLOW;
    }

    if(encrypt) {
        if(EVP_EncryptInit_ex(ctx, type, NULL, key, iv) == 0) {
            fprintf(stderr, "Failed to initialize DES encrypt (FIPS=%d) and (FLAG=%d)\n",
                    FIPS_mode(), set_flag);
            goto des_err;
        }


        if(EVP_EncryptUpdate(ctx, out, outl, in, inl) ==0) {
            fprintf(stderr,"Failed to update DES encrypt (FIPS=%d) and (FLAG=%d)\n",
                    FIPS_mode(), set_flag);
            goto des_err;
        }

        EVP_CIPHER_CTX_set_padding(ctx, 0);

        if(EVP_EncryptFinal_ex(ctx, out + *outl, outl) ==0) {
            fprintf(stderr,"Failed DES encrypt final (FIPS=%d) and (FLAG=%d)\n",
                    FIPS_mode(), set_flag);
            goto des_err;
        }

        check = out;
    }
    else {

        if(EVP_DecryptInit_ex(ctx, type, NULL, key, iv) == 0) {
            fprintf(stderr, "Failed to initialize DES decrypt (FIPS=%d) and (FLAG=%d)\n",
                    FIPS_mode(), set_flag);
            goto des_err;
        }

        if(EVP_DecryptUpdate(ctx, out, outl, in, inl) ==0) {
            fprintf(stderr,"Failed to update DES decrypt (FIPS=%d) and (FLAG=%d)\n",
                    FIPS_mode(), set_flag);
            goto des_err;
        }

        EVP_CIPHER_CTX_set_padding(ctx, 0);

        if(EVP_DecryptFinal_ex(ctx, out + *outl, outl) ==0) {
            fprintf(stderr,"Failed DES decrypt final (FIPS=%d) and (FLAG=%d)\n",
                    FIPS_mode(), set_flag);
            goto des_err;
        }

    }

    if(memcmp(out, check, inl)!=0) {
            fprintf(stderr,"input and output did not match\n");
            goto des_err;
    }

    rv = 0;
    ERR_clear_error();

des_err:
    if(ctx != NULL) {
        EVP_CIPHER_CTX_free(ctx);
    }
    return rv;

}

/*
 * This routine will test encrypt and decrypt using two key and three key des
 * ciphers to ensure the limitations of using des ciphers when FIPS mode is
 * enabled is properly enforced (as defined in SP-800-131).
 *
 * returns zero on success and one on failure
 */
static int sp800131_des_test()
{
    /* two key des variables */
    unsigned char key[16] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
    unsigned char plaintext[8] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
    int size=8;
    unsigned char two_key_ciphertext[16];
    /* EVP_CIPHER pointer because functions in array return EVP_CIPHER pointers */
    const EVP_CIPHER* cipher_types[4] =
        {EVP_des_ede_cbc(), EVP_des_ede_cfb(), EVP_des_ede_ofb(), EVP_des_ede()};
    int cipher_types_size = sizeof(cipher_types)/sizeof(cipher_types[0]);
    int x = 0;

    /* three key des variables */
    unsigned char key_3des[24] =
        {0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xf1, 0xe0, 0xd3, 0xc2,
         0xb5, 0xa4, 0x97, 0x86, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10};
    unsigned char iv_3des[8] ={0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10};
    unsigned char plaintext_3des[32]=
        {0x37, 0x36, 0x35, 0x34, 0x33, 0x32, 0x31, 0x20, 0x4E, 0x6F, 0x77, 0x20,
         0x69, 0x73, 0x20, 0x74, 0x68, 0x65, 0x20, 0x74, 0x69, 0x6D, 0x65, 0x20,
         0x66, 0x6F, 0x72, 0x20, 0x00, 0x00, 0x00, 0x00};
    int size_3des=32;

    /* output for encrypt and decrypt */
    unsigned char out[64];
    int outl;

    /* Disable Fips */
    if(FIPS_mode()) {
        if (!FIPS_mode_set(0)) {
	         fprintf(stderr, "Failed to disable FIPS mode\n");
            return 1;
        }
    }

    /* Test two key des encryption and decryption works with FIPS off */
    for(x=0;x<cipher_types_size;x++) {
        if(sp800131_encrypt_des(cipher_types[x], key, key, out,
           &outl, plaintext, size, NULL, 1, 0)) {
            fprintf(stderr, "two key DES encrypt failed for FIPS off\n");
            return 1;
        }

        /* save ciphertext to be used for all other two key des decrypt tests */
        if(x == 0) {
            memcpy(two_key_ciphertext, out,16);
        }
        if(sp800131_encrypt_des(cipher_types[x], key, key, out,
           &outl, two_key_ciphertext, size, plaintext, 0, 0)) {
            fprintf(stderr, "two key DES decrypt failed for FIPS off\n");
            return 1;
        }
    }

    /* Test three key des encryption and decryption works with FIPS off */
    if(sp800131_encrypt_des(EVP_des_ede3_cbc(), key_3des, iv_3des, out,
       &outl, plaintext_3des, size_3des, NULL, 1, 0)) {
        fprintf(stderr, "three key DES encrypt failed for FIPS off\n");
        return 1;
    }
    if(sp800131_encrypt_des(EVP_des_ede3_cbc(), key_3des, iv_3des, out,
       &outl, out, size_3des, plaintext_3des, 0, 0)) {
        fprintf(stderr, "three key DES decrypt failed for FIPS off\n");
        return 1;
    }

    /* Turn fips Mode for rest of tests */
    if (!FIPS_mode_set(1)) {
       fprintf(stderr, "Failed to enable FIPS mode\n");
       return 1;
    }

    /* Test two key des fails for encryption and decryption and expect failure */
    for(x=0;x<cipher_types_size;x++) {
        if(!sp800131_encrypt_des(cipher_types[x], key, key, out,
           &outl, plaintext, size, NULL, 1, 0)) {
            fprintf(stderr, "two key des encrypt didn't fail with FIPS on\n");
            return 1;
        }
        if(!sp800131_encrypt_des(cipher_types[x], key, key, out,
           &outl, two_key_ciphertext, size, plaintext, 0, 0)) {
            fprintf(stderr, "two key des decrypt didn't fail with FIPS on\n");
            return 1;
        }
    }

    /* Test three key des FIPS on and expect success */
    if(sp800131_encrypt_des(EVP_des_ede3_cbc(), key_3des, iv_3des, out,
       &outl, plaintext_3des, size_3des, NULL, 1, 0)) {
        fprintf(stderr, "three key DES encrypt failed for FIPS on\n");
        return 1;
    }
    if(sp800131_encrypt_des(EVP_des_ede3_cbc(), key_3des, iv_3des, out,
       &outl, out, size_3des, plaintext_3des, 0, 0)) {
        fprintf(stderr, "three key DES decrypt failed for FIPS on\n");
        return 1;
    }

    /* test that encrypt still fails with the flag on fails but passes for decrypt */
    for(x=0;x<cipher_types_size;x++) {
        if(!sp800131_encrypt_des(cipher_types[x], key, key, out,
           &outl, plaintext, size, NULL, 1, 1)) {
            fprintf(stderr, "two key des encrypt didn't fail with FIPS on and flag set\n");
            return 1;
        }
        if(sp800131_encrypt_des(cipher_types[x], key, key, out,
           &outl, two_key_ciphertext, size, plaintext, 0, 1)) {
            fprintf(stderr, "two key des decrypt failed with FIPS on and flag set\n");
            return 1;
        }
    }

    /* Disable Fips */
    if(FIPS_mode()){
        if (!FIPS_mode_set(0)) {
	         fprintf(stderr, "Failed to disable FIPS mode\n");
            return 1;
        }
    }

    ERR_clear_error();
    return 0;
}

/*
 * Helper function for sp800131_sha1_test to generate
 * a rsa key for use.
 *
 * pkey - handle for the rsa key data
 * bits - the requested length of rsa key in bits
 *
 * returns 0 on success and 1 on failure
 */
static int gen_rsa_key(EVP_PKEY **pkey,
                       int bits)
{
    int rv = 1;
    EVP_PKEY_CTX *ctx = NULL;

    if (!(ctx = EVP_PKEY_CTX_new_id(EVP_PKEY_RSA, NULL))) {
        fprintf(stderr, "EVP_PKEY_CTX_new_id failed (FIPS=%d)\n", FIPS_mode());
        goto err;
    }

    if (EVP_PKEY_keygen_init(ctx) <= 0) {
        fprintf(stderr, "EVP_PKEY_keygen_init failed (FIPS=%d)\n", FIPS_mode());
        goto err;
    }

    if (EVP_PKEY_CTX_set_rsa_keygen_bits(ctx, bits) <= 0) {
        fprintf(stderr, "EVP_PKEY_CTX_set_rsa_keygen_bits (FIPS=%d)\n", FIPS_mode());
        goto err;
    }

    if (EVP_PKEY_keygen(ctx, pkey) <= 0) {
        fprintf(stderr, "EVP_PKEY_keygen (FIPS=%d)\n", FIPS_mode());
        goto err;
    }

    rv = 0;

err:
    if (ctx) EVP_PKEY_CTX_free(ctx);
    return rv;
}


/*
 * Perform a traditional digest sequence.
 *
 * msg - data upon which digest is calculated
 * mlen - the length of msg array in bytes
 *
 * returns 0 on success and 1 on failure
 */
static int sha1_digest(const unsigned char *msg,
                       size_t mlen)
{
    int rv = 1;
    EVP_MD_CTX mctx;
    unsigned char digest[EVP_MAX_MD_SIZE];
    unsigned int dlen = 0;
    const EVP_MD *md = EVP_sha1();

    EVP_MD_CTX_init(&mctx);

    if (!EVP_DigestInit_ex(&mctx, md, NULL)) {
        fprintf(stderr, "EVP_DigestInit_ex failed for mctx (FIPS=%d)\n", FIPS_mode());
        goto err;
    }

    if (!EVP_DigestUpdate(&mctx, msg, mlen)) {
        fprintf(stderr, "EVP_DigestInit_ex failed for mctx (FIPS=%d)\n", FIPS_mode());
        goto err;
    }

    if (!EVP_DigestFinal_ex(&mctx, digest, &dlen)) {
        fprintf(stderr, "EVP_DigestInit_ex failed for mctx (FIPS=%d)\n", FIPS_mode());
        goto err;
    }

    ERR_clear_error();
    rv = 0;

err:
    EVP_MD_CTX_cleanup(&mctx);
    return rv;
}


/*
 * Perform digest sign and verify using deprecated api
 * EVP_Sign/Verify.
 *
 * pkey - keypair to use
 * md - sha hash to use (EVP_sha1(), EVP_sha256(), etc)
 * sig - handle for calculated signature
 * slen - handle for the length of signature
 * msg - data upon which signature is made
 * mlen - length of the msg array in bytes
 * verify - 0: perform sign, 1: perform verify
 * non_fips_allow - 1: set flag 0: do nothing
 * non_ex - 1: use EVP_SignInit 0: use EVP_SignInit_ex
 *
 * returns 0 on success and 1 on failure
 */
static int sha_sign_or_verify(EVP_PKEY *pkey,
			      const EVP_MD *md,
			      unsigned char **sig,
			      unsigned int *slen,
			      const unsigned char *msg,
			      unsigned int mlen,
			      int verify,
			      int non_fips_allow,
			      int non_ex)
{
    int rv = 1;
    EVP_MD_CTX mctx;

    EVP_MD_CTX_init(&mctx);

    if (non_fips_allow) {
        mctx.flags |= EVP_MD_CTX_FLAG_NON_FIPS_ALLOW;
    }

    if (!verify) {
        /*
         * Sign operation is taking place.
         */
        if (non_ex) {
            if (!EVP_SignInit(&mctx, md)) {
                fprintf(stderr, "EVP_SignInit failed for mctx (fips=%d)\n", FIPS_mode());
                goto err;
            }
        } else {
            if (!EVP_SignInit_ex(&mctx, md, NULL)) {
                fprintf(stderr, "EVP_SignInit_ex failed for mctx (fips=%d)\n", FIPS_mode());
                goto err;
            }
        }

        if (!EVP_SignUpdate(&mctx, msg, mlen)) {
            fprintf(stderr, "EVP_SignUpdate failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }

        if (*sig == NULL) {
            if (!(*sig = OPENSSL_malloc(EVP_PKEY_size(pkey)))) {
                fprintf(stderr, "OPENSSL_malloc failed for sig (FIPS=%d)\n", FIPS_mode());
                goto err;
            }
        }

        if (!EVP_SignFinal(&mctx, *sig, slen, pkey)) {
            fprintf(stderr, "EVP_SignFinal failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }
    } else {
        /*
         * Verify operation is taking place
         */
        if (!EVP_VerifyInit_ex(&mctx, md, NULL)) {
            fprintf(stderr, "EVP_VerifyInit failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }
        if (!EVP_VerifyUpdate(&mctx, msg, mlen)) {
            fprintf(stderr, "EVP_VerifyUpdate failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }
        if (!EVP_VerifyFinal(&mctx, *sig, *slen, pkey)) {
            fprintf(stderr, "EVP_VerifyFinal failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }
    }

    ERR_clear_error();
    rv = 0;

err:
    EVP_MD_CTX_cleanup(&mctx);
    return rv;
}

/*
 * Perform digest sign and verify using
 * EVP_DigestSign/Verify.
 *
 * pkey - keypair to use
 * md - sha hash to use such as EVP_sha1(), EVP_sha256(), etc.
 * sig - handle for calculated signature
 * slen - handle for the length of signature
 * msg - data upon which signature is made
 * mlen - length of the msg array in bytes
 * verify - 0: perform sign, 1: perform verify
 * non_fips_allow - 1: set flag 0: do nothing
 *
 * returns 0 on success and 1 on failure
 */
static int sha_digest_sign_or_verify(EVP_PKEY *pkey,
				      const EVP_MD *md,
                                      unsigned char **sig,
                                      size_t *slen,
                                      const unsigned char *msg,
                                      size_t mlen,
                                      int verify,
                                      int non_fips_allow)
{
    int rv = 1;
    EVP_MD_CTX mctx;

    EVP_MD_CTX_init(&mctx);

    if (non_fips_allow) {
        mctx.flags |= EVP_MD_CTX_FLAG_NON_FIPS_ALLOW;
    }

    if (!verify) {
        /*
         * Sign operation is taking place.
         */

        if (!EVP_DigestSignInit(&mctx, NULL, md, NULL, pkey)) {
            fprintf(stderr, "EVP_DigestSignInit failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }

        if (!EVP_DigestSignUpdate(&mctx, msg, mlen)) {
            fprintf(stderr, "EVP_DigestSignUpdate failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }

        if (!EVP_DigestSignFinal(&mctx, NULL, slen)) {
            fprintf(stderr, "EVP_DigestSignFinal failed (1) for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }

        if (*sig == NULL) {
            if (!(*sig = OPENSSL_malloc(*slen))) {
                fprintf(stderr, "OPENSSL_malloc failed for sig (FIPS=%d)\n", FIPS_mode());
                goto err;
            }
        }

        if (!EVP_DigestSignFinal(&mctx, *sig, slen)) {
            fprintf(stderr, "EVP_DigestSignFinal failed (2) for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }
    } else {
        /*
         * Verify operation is taking place
         */
        if (!EVP_DigestVerifyInit(&mctx, NULL, md, NULL, pkey)) {
            fprintf(stderr, "EVP_DigestVerifyInit failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }
        if (!EVP_DigestVerifyUpdate(&mctx, msg, mlen)) {
            fprintf(stderr, "EVP_DigestVerifyUpdate failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }
        if (!EVP_DigestVerifyFinal(&mctx, *sig, *slen)) {
            fprintf(stderr, "EVP_DigestVerifyFinal failed for mctx (FIPS=%d)\n", FIPS_mode());
            goto err;
        }
    }

    ERR_clear_error();
    rv = 0;

err:
    EVP_MD_CTX_cleanup(&mctx);
    return rv;
}


/*
 * This routine will test EVP_Digestxxx, EVP_DigestSign/Verifyxxx, EVP_Sign/Verifyxxx
 * api with sha-1 to ensure the limitations when FIPS mode is enabled are properly enforced
 * (as defined in SP-800-131).
 *
 * returns zero on success and one on failure
 */
static int sp800131_sha1_test() {
    int rv = 1;
    EVP_PKEY *pkey = NULL;
    unsigned char *sig_a = NULL, *sig_b = NULL;
    const unsigned char msg[] = "12345678901234567890";
    size_t mlen = 20, slen = 0;
    unsigned int i_mlen = 20, i_slen = 0;

    /* Disable FIPS */
    if (FIPS_mode()) {
       if (!FIPS_mode_set(0)) {
          fprintf(stderr, "Failed to disable FIPS mode in SHA-1\n");
          return 1;
       }
    }

    /*
     * Init pkey for signing
     */
    if (!(pkey = EVP_PKEY_new())) {
	fprintf(stderr, "EVP_PKEY_new failed for pkey (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    if (gen_rsa_key(&pkey, 2048)) {
	fprintf(stderr, "gen_rsa_key failed for pkey (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /* *********************************
     * sha1_digest
     * *********************************
     */

    /*
     * EVP Digest operation
     * The digest operation should pass using sha1.
     */
    if (sha1_digest(msg, mlen)) {
	fprintf(stderr, "sha1_digest failed (1) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /* ******************************************
     * sha_sign_or_verify
     * ******************************************
     */

    /*
     * EVP Sign operation
     * The sign operation should pass using sha1.
     */
    if (sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 0, 0, 0)) {
	fprintf(stderr, "sha_sign_or_verify failed (1) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /*
     * EVP Verify operation
     * The verify operation should pass using sha1.
     */
    if (sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 1, 0, 0)) {
	fprintf(stderr, "sha_sign_or_verify failed (2) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /* ******************************************
     * sha_sign_or_verify non-extended init
     * ******************************************
     */

    /*
     * EVP Sign operation
     * The sign operation should pass using sha1.
     */
    if (sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 0, 0, 1)) {
	fprintf(stderr, "sha_sign_or_verify failed (1) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /*
     * EVP Verify operation
     * The verify operation should pass using sha1.
     */
    if (sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 1, 0, 1)) {
	fprintf(stderr, "sha_sign_or_verify failed (2) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /* ******************************************
     * sha_digest_sign_or_verify
     * ******************************************
     */

    /*
     * EVP Digest Sign operation
     * The digest sign operation should pass using sha1.
     */
    if (sha_digest_sign_or_verify(pkey, EVP_sha1(), &sig_b, &slen, msg, mlen, 0, 0)) {
	fprintf(stderr, "sha_digest_sign_or_verify failed (1) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /*
     * EVP Digest Verify operation
     * The digest verify operation should pass using sha1.
     */
    if (sha_digest_sign_or_verify(pkey, EVP_sha1(), &sig_b, &slen, msg, mlen, 1, 0)) {
	fprintf(stderr, "sha_digest_sign_or_verify failed (2) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }


    /*
     * Reset values for use in FIPS mode
     */
    slen = 0;
    i_slen = 0;
    if (pkey) {
        EVP_PKEY_free(pkey);
        pkey = NULL;
    }
    if (sig_a) {
        OPENSSL_free(sig_a);
        sig_a = NULL;
    }
    if (sig_b) {
        OPENSSL_free(sig_b);
        sig_b = NULL;
    }


    /*
     * The rest of the test cases require FIPS to be enabled
     */
    if (!FIPS_mode_set(1)) {
	fprintf(stderr, "Failed to enable FIPS mode in SHA-1\n");
	goto sha1_err;
    }

    /*
     * Init pkey for signing
     */
    if (!(pkey = EVP_PKEY_new())) {
	fprintf(stderr, "EVP_PKEY_new failed for pkey (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    if (gen_rsa_key(&pkey, 2048)) {
	fprintf(stderr, "gen_rsa_key failed for pkey (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }


    /* *********************************
     * sha1_digest FIPS ON!
     * *********************************
     */

    /*
     * EVP Digest operation, FIPS on
     * The digest operation should pass using sha1 with FIPS enabled.
     */
    if (sha1_digest(msg, mlen)) {
	fprintf(stderr, "sha1_digest failed (1) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /* ******************************************
     * sha_sign_or_verify FIPS ON!
     * ******************************************
     */

    /*
     * EVP Sign operation, FIPS on
     * The sign operation should fail using sha1 with FIPS enabled.
     */
    if (!sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 0, 0, 0)) {
	fprintf(stderr, "sha_sign_or_verify failed (1) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /*
     * EVP Sign operation, FIPS on
     * Set the non-fips-allow flag so that sha1 will pass with FIPS enabled.
     */
    if (sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 0, 1, 0)) {
	fprintf(stderr, "sha_sign_or_verify failed (2) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /*
     * EVP Verify operation, FIPS on
     * The verify operation should pass using sha1 with FIPS enabled.
     */
    if (sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 1, 0, 0)) {
	fprintf(stderr, "sha_sign_or_verify failed (3) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /* ***********************************************
     * sha_sign_or_verify FIPS ON! non-extended init
     * ***********************************************
     */

    /*
     * EVP Sign operation, FIPS on
     * The sign operation should fail using sha1 with FIPS enabled.
     */
    if (!sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 0, 0, 1)) {
	fprintf(stderr, "sha_sign_or_verify failed (4) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /*
     * EVP Sign operation, FIPS on
     * Set the non-fips-allow flag so that sha1 will pass with FIPS enabled.
     */
    if (sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 0, 1, 1)) {
	fprintf(stderr, "sha_sign_or_verify failed (5) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /*
     * EVP Verify operation, FIPS on
     * The verify operation should pass using sha1 with FIPS enabled.
     */
    if (sha_sign_or_verify(pkey, EVP_sha1(), &sig_a, &i_slen, msg, i_mlen, 1, 0, 1)) {
	fprintf(stderr, "sha_sign_or_verify failed (6) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /* ******************************************
     * sha_digest_sign_or_verify FIPS ON!
     * ******************************************
     */

    /*
     * EVP Digest Sign operation, FIPS on
     * The signature should fail here using sha1 with FIPS enabled.
     */
    if (!sha_digest_sign_or_verify(pkey, EVP_sha1(), &sig_b, &slen, msg, mlen, 0, 0)) {
	fprintf(stderr, "sha_digest_sign_or_verify failed (1) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /*
     * EVP Digest Sign operation, FIPS on
     * Set the non-fips-allow flag so that sha1 will pass with FIPS enabled.
     */
    if (sha_digest_sign_or_verify(pkey, EVP_sha1(), &sig_b, &slen, msg, mlen, 0, 1)) {
	fprintf(stderr, "sha_digest_sign_or_verify failed (2) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    /*
     * EVP Digest Verify operation, FIPS on
     * The verify should pass here using sha1 with FIPS enabled.
     */
    if (sha_digest_sign_or_verify(pkey, EVP_sha1(), &sig_b, &slen, msg, mlen, 1, 0)) {
	fprintf(stderr, "sha_digest_sign_or_verify failed (3) (FIPS=%d)\n", FIPS_mode());
        goto sha1_err;
    }

    ERR_clear_error();
    rv = 0;

sha1_err:
    /* Disable FIPS */
    if (FIPS_mode()) {
       if (!FIPS_mode_set(0)) {
          fprintf(stderr, "Failed to disable FIPS mode in SHA-1\n");
          return 1;
       }
    }

    if (pkey) EVP_PKEY_free(pkey);
    if (sig_a != NULL) {
        OPENSSL_free(sig_a);
    }
    if (sig_b != NULL) {
        OPENSSL_free(sig_b);
    }
    return rv;
}

#define SP800131_EC_MIN_KEY_LEN	224
#define SP800131_CURVES_LEN 16
/*
 * Contains the NID values of the curves that will be tested
 */
static int ec_curves_list[SP800131_CURVES_LEN] = {
    NID_sect163r2,
    NID_sect233r1,
    NID_sect283r1,
    NID_sect409r1,
    NID_sect571r1,
    NID_sect163k1,
    NID_sect233k1,
    NID_sect283k1,
    NID_sect409k1,
    NID_sect571k1,
    NID_X9_62_prime192v1,
    NID_secp224r1,
    NID_X9_62_prime256v1,
    NID_secp384r1,
    NID_secp521r1,
    NID_brainpoolP256r1
};

static const int KDF1_SHA256_len = 32;
static void *KDF1_SHA256(const void *in, size_t inlen, void *out, size_t *outlen)
{
	const EVP_MD *md = EVP_sha256();
	EVP_MD_CTX *mdctx;

#ifndef OPENSSL_NO_SHA
	if (*outlen < KDF1_SHA256_len)
		return NULL;
	else
		*outlen = KDF1_SHA256_len;

	mdctx = EVP_MD_CTX_create();
	EVP_DigestInit_ex(mdctx, md, NULL);
	EVP_DigestUpdate(mdctx, in, inlen);
	EVP_DigestFinal_ex(mdctx, out, (unsigned int*)outlen);
	EVP_MD_CTX_destroy(mdctx);
	return (out);
#else
	return NULL;
#endif
}

/*
 * Returns the size in bits of the curve key length
 */
static int sp800131_ec_curve_size(int nid)
{
    EC_KEY *key = NULL;
    const EC_GROUP *group = NULL;
    BIGNUM *order = NULL;
    int keysize = 0;

    /*
     * Make sure FIPS is disabled to ensure the curve can be
     * instantiated.
     */
    SP800131_DISABLE_FIPS

    /* 
     * Instantiate a curve of type NID
     */
    key = EC_KEY_new_by_curve_name(nid);
    if (!key) {
        fprintf(stderr, "Failed to create key1 for NID %d\n", nid);
	return 1;
    }
    /*
     * Get the group for this curve, then get the group order
     */
    group = EC_KEY_get0_group(key);
    order = BN_new();
    if (!EC_GROUP_get_order(group, order, NULL)) {
	ERR_print_errors_fp(stderr);
	fprintf(stderr, "Failed to get order\n");
	return 1;
    }

    /*
     * Key size is the length of the order
     */
    keysize = BN_num_bits(order);

    if (key) EC_KEY_free(key);
    if (order) BN_free(order);
    return keysize;
}

/*
 * Used by the EC test cases to geneate a new EC keypair.
 */
static EC_KEY *sp800131_ec_create_key(int nid)
{
    EC_KEY *key = NULL;

    key = EC_KEY_new_by_curve_name(nid);
    if (!key) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "Failed to create EC key for NID %d\n", nid);
        goto ec_create_done;
    }
    if (!EC_KEY_generate_key(key)) {
	ERR_print_errors_fp(stderr);
	fprintf(stderr, "Failed to generate EC key for NID %d\n", nid);
	EC_KEY_free(key);
	key = NULL;
	goto ec_create_done;
    }
ec_create_done:
    return key;
}

/*
 * Performs the ECDH portion of the EC tests
 *
 * returns zero on success and one on failure
 */
static int sp800131_ecdh_test(int nid, int fips_on) {
    int rv = 1;
    EC_KEY *key1 = NULL;
    EC_KEY *key2 = NULL;
    int alen, aout;
    unsigned char *abuf1 = NULL;
    unsigned char *abuf2 = NULL;
    int key_len;
    const char *nist;

    SP800131_DISABLE_FIPS

    printf("Testing ECDH with FIPS mode %d using NID %d (%s)\n", fips_on, nid, EC_curve_nid2nist(nid));

    /*
     * Get the key size, which we need to know whether FIPS is allowed.
     */
    key_len = sp800131_ec_curve_size(nid);

    /*
     * Returns null if the curve isn't allowed by NIST (e.g. brainpool)
     */
    nist = EC_curve_nid2nist(nid);

    /*
     * Enable FIPS if this test requires it
     */
    if (fips_on) {
	SP800131_ENABLE_FIPS
    }

    /*
     * Create the first key pair
     */
    key1 = sp800131_ec_create_key(nid);

    /*
     * Make sure we were able to create the key 
     */
    if (!key1) {
        fprintf(stderr, "Failed to create key1 for NID %d\n", nid);
        goto ecdh_err;
    }

    /*
     * Create the second key pair
     */
    key2= sp800131_ec_create_key(nid);
    if (!key2) {
	goto ecdh_err;
    }

    /*
     * Compute the first shared secret 
     */
    alen = KDF1_SHA256_len;
    abuf1 = (unsigned char *)OPENSSL_malloc(alen);
    memset(abuf1, 0x0, alen);
    aout = ECDH_compute_key(abuf1,alen,EC_KEY_get0_public_key(key2),key1,KDF1_SHA256);

    /* 
     * First check if the computation worked when it shouldn't have
     */
    if ((aout > 0) && FIPS_mode() && (!nist || (key_len < SP800131_EC_MIN_KEY_LEN))) { 
        fprintf(stderr, "ECDH computation worked for non-NIST curve in FIPS mode\n");
        goto ecdh_err;
    }

    /*
     * The computation should always work when FIPS is disabled, or when
     * FIPS is enabled and the minimum SP800-131 key size is met
     */
    if (aout <= 0) {
	if (!FIPS_mode()) {
	    /*
	     * All curves should work when not in FIPS mode
	     */
	    ERR_print_errors_fp(stderr);
	    fprintf(stderr, "ECDH computation 1 failed with FIPS disabled!!!\n");
            goto ecdh_err;
	} else {
	    if (key_len >= SP800131_EC_MIN_KEY_LEN && nist) {
		/*
		 * It's a NIST approved curve, computation should not have failed
		 */
		ERR_print_errors_fp(stderr);
		fprintf(stderr, "ECDH computation 1 failed for NIST curve with FIPS enabled!!!\n");
		goto ecdh_err;
	    } else {
		/*
		 * It's not a NIST SP800-131 compliant curve, so we expected
		 * the computation to fail.
		 */
		fprintf(stderr, "   ...computation for non-NIST curve failed as expected, test passed\n");
                ERR_clear_error();
		rv = 0;
		goto ecdh_err;
	    }
	}
    }

    /*
     * Compute the second shared secret 
     */
    abuf2 = (unsigned char *)OPENSSL_malloc(alen);
    memset(abuf2, 0x0, alen);
    aout = ECDH_compute_key(abuf2,alen,EC_KEY_get0_public_key(key1),key2,KDF1_SHA256);
    if (aout <= 0) {
        ERR_print_errors_fp(stderr);
        fprintf(stderr, "ECDH computation 2 failed!!!\n");
        goto ecdh_err;
    }

    /*
     * Compare the two shared secrets, they should match
     */
    if (memcmp(abuf1, abuf2, alen)) {
        fprintf(stderr, "ECDH computation mismatch!!!\n");
        goto ecdh_err;
    }
    printf("   ...shared secrets match, test passed\n");

    /* Release the memory before processing the next NID */
    if (key1) EC_KEY_free(key1);
    if (key2) EC_KEY_free(key2);
    if (abuf1) OPENSSL_free(abuf1);
    if (abuf2) OPENSSL_free(abuf2);
    key1 = key2 = NULL;
    abuf1 = abuf2 = NULL;

    rv = 0;
    ERR_clear_error();
ecdh_err:
    if (key1) EC_KEY_free(key1);
    if (key2) EC_KEY_free(key2);
    if (abuf1) OPENSSL_free(abuf1);
    if (abuf2) OPENSSL_free(abuf2);
    return rv;
}


/*
 * Generates an EC EVP_PKEY
 *
 * pkey - handle for the key data
 * nid - NID value for the curve to use
 * genkey - set to non-zero to generate a key pair
 *
 * returns 0 on success and 1 on failure
 */
static int gen_evp_pkey_ec(EVP_PKEY **pkey, int nid, int genkey)
{
    int rv = 1;
    int rc;
    EVP_PKEY_CTX *ctx = NULL;

    if (!(ctx = EVP_PKEY_CTX_new_id(EVP_PKEY_EC, NULL))) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "EVP_PKEY_CTX_new_id failed (FIPS=%d)\n", FIPS_mode());
        goto genec_err;
    }

    if (EVP_PKEY_keygen_init(ctx) <= 0) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "EVP_PKEY_keygen_init failed (FIPS=%d)\n", FIPS_mode());
        goto genec_err;
    }

    if ((rc = EVP_PKEY_CTX_set_ec_paramgen_curve_nid(ctx, nid)) <= 0) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "EVP_PKEY_CTX_set_ec_paramgen_curve_nid (rc=%d)\n", rc);
        goto genec_err;
    }

    if (genkey) {
	if (EVP_PKEY_keygen(ctx, pkey) <= 0) {
	    ERR_print_errors_fp(stderr);
	    fprintf(stderr, "EVP_PKEY_keygen (FIPS=%d)\n", FIPS_mode());
	    goto genec_err;
	}
    }

    rv = 0;
    ERR_clear_error();

genec_err:
    if (ctx) EVP_PKEY_CTX_free(ctx);
    return rv;
}



/*
 * Performs the ECDSA signature generation portion of the EC tests
 *
 * returns zero on success and one on failure
 */
static int sp800131_ecdsa_sign_test(int nid, int fips_on) {
    int rv = 1;
    EC_KEY *key = NULL;
    EVP_PKEY *pkey = NULL;
    EVP_MD_CTX mctx;
    const EVP_MD *md = EVP_sha256();
    unsigned char *sig = NULL;
    size_t slen;
    const unsigned char msg[] = "12345678901234567890";
    size_t mlen = 20;
    int key_len;
    const char *nist;

    printf("Testing ECDSA sign with FIPS mode %d using NID %d (%s)\n", fips_on, nid, EC_curve_nid2nist(nid));

    SP800131_DISABLE_FIPS

    /*
     * Returns null if the curve isn't allowed by NIST (e.g. brainpool)
     */
    nist = EC_curve_nid2nist(nid);

    /*
     * Get the key size, which we need to know whether FIPS is allowed.
     */
    key_len = sp800131_ec_curve_size(nid);

    /*
     * Create a key pair
     */
    key = sp800131_ec_create_key(nid);
    if (!key) {
	goto ecdsa_sign_err;
    }

    /*
     * Allocate space for the signature
     */
    slen = ECDSA_size(key);
    sig = OPENSSL_malloc(slen);
    if (!sig) {
	fprintf(stderr, "Malloc failed %s\n", __FUNCTION__);
        goto ecdsa_sign_err;
    }

    /*
     * Set the desired FIPS enabled state
     */
    if (fips_on) {
	SP800131_ENABLE_FIPS
    } else {
	SP800131_DISABLE_FIPS
    }

    EVP_MD_CTX_init(&mctx);

    /*
     * Init pkey for signing
     */
    if (!(pkey = EVP_PKEY_new())) {
	ERR_print_errors_fp(stderr);
	fprintf(stderr, "EVP_PKEY_new failed for pkey (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_sign_err;
    }
    if (gen_evp_pkey_ec(&pkey, nid, 1)) {
	fprintf(stderr, "gen_evp_pkey_ec failed (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_sign_err;
    }

    /*
     * Now generate a signature using the key
     */
    if (EVP_DigestSignInit(&mctx, NULL, md, NULL, pkey) <= 0) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "EVP_DigestSignInit failed for mctx (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_sign_err;
    }
    if (EVP_DigestSignUpdate(&mctx, msg, mlen) <= 0) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "EVP_DigestSignUpdate failed for mctx (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_sign_err;
    }
    if (EVP_DigestSignFinal(&mctx, sig, &slen) <= 0) {
	if (FIPS_mode()) {
	    if (key_len >= SP800131_EC_MIN_KEY_LEN && nist) {
		/*
		 * It's a NIST approved curve, computation should not have failed
		 */
		ERR_print_errors_fp(stderr);
		fprintf(stderr, "EVP_DigestSignFinal failed with FIPS enabled\n");
		goto ecdsa_sign_err;
	    } else {
		/*
		 * It's not a NIST SP800-131 compliant curve, so we expected
		 * the computation to fail.
		 */
		fprintf(stderr, "   ...computation for non-NIST curve failed as expected, test passed\n");
                ERR_clear_error();
		rv = 0;
		goto ecdsa_sign_err;
	    }
	} else {
	    ERR_print_errors_fp(stderr);
	    fprintf(stderr, "EVP_DigestSignFinal failed with FIPS disabled, all curves should work w/o FIPS\n");
	    goto ecdsa_sign_err;
	}
    }

    EVP_MD_CTX_cleanup(&mctx);

    /*
     * We made it, the test has passed
     */
    rv = 0;
    ERR_clear_error();

ecdsa_sign_err:
    EVP_MD_CTX_cleanup(&mctx);
    if (key) EC_KEY_free(key);
    if (pkey) EVP_PKEY_free(pkey);
    if (sig) OPENSSL_free(sig);
    return rv;
}

/*
 * Performs the ECDSA verify portion of the EC tests
 *
 * returns zero on success and one on failure
 */
static int sp800131_ecdsa_verify_test(int nid, int fips_on, int legacy_use) {
    int rv = 1;
    EC_KEY *key = NULL;
    EVP_PKEY *pkey = NULL;
    EVP_MD_CTX mctx;
    const EVP_MD *md = EVP_sha256();
    unsigned char *sig = NULL;
    size_t slen;
    const unsigned char msg[] = "12345678901234567890";
    size_t mlen = 20;
    int key_len;
    const char *nist;
    EC_KEY *ec1, *ec2;

    printf("Testing ECDSA verify with FIPS mode=%d, legacy_use=%d, and using NID %d (%s)\n", fips_on, legacy_use, nid, EC_curve_nid2nist(nid));

    SP800131_DISABLE_FIPS

    /*
     * Returns null if the curve isn't allowed by NIST (e.g. brainpool)
     */
    nist = EC_curve_nid2nist(nid);

    /*
     * Get the key size, which we need to know whether FIPS is allowed.
     */
    key_len = sp800131_ec_curve_size(nid);

    /*
     * Create a key pair
     */
    key = sp800131_ec_create_key(nid);
    if (!key) {
	goto ecdsa_vfy_err;
    }

    /*
     * Allocate space for the signature
     */
    slen = ECDSA_size(key);
    sig = OPENSSL_malloc(slen);
    if (!sig) {
	fprintf(stderr, "Malloc failed %s\n", __FUNCTION__);
        goto ecdsa_vfy_err;
    }

    EVP_MD_CTX_init(&mctx);

    /*
     * Init pkey for signing
     */
    if (!(pkey = EVP_PKEY_new())) {
	ERR_print_errors_fp(stderr);
	fprintf(stderr, "EVP_PKEY_new failed for pkey (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_vfy_err;
    }
    if (gen_evp_pkey_ec(&pkey, nid, 1)) {
	fprintf(stderr, "gen_evp_pkey_ec failed (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_vfy_err;
    }

    /*
     * Now generate a signature using the key
     */
    if (EVP_DigestSignInit(&mctx, NULL, md, NULL, pkey) <= 0) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "EVP_DigestSignInit failed for mctx (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_vfy_err;
    }
    if (EVP_DigestSignUpdate(&mctx, msg, mlen) <= 0) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "EVP_DigestSignUpdate failed for mctx (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_vfy_err;
    }
    if (EVP_DigestSignFinal(&mctx, sig, &slen) <= 0) {
	ERR_print_errors_fp(stderr);
	fprintf(stderr, "EVP_DigestSignFinal failed with FIPS disabled, all curves should work w/o FIPS\n");
	goto ecdsa_vfy_err;
    }

    EVP_MD_CTX_cleanup(&mctx);

    /*
     * Now that we have a signature that we can verify, set the desired FIPS enabled state
     */
    if (fips_on) {
	SP800131_ENABLE_FIPS
    } else {
	SP800131_DISABLE_FIPS
    }

    /*
     * This is a hack and should never be used as an example in production
     * code on how to use the OpenSSL API.  The problem is the key generated
     * above was done with FIPS disabled.  The method associated with the key
     * is set when the key is created.  Therefore, we can't use this key when
     * FIPS is enabled because it has the non-FIPS method attached to the key.
     * Now that the FIPS state has been set, we duplicate the key so the duplicate
     * will have the FIPS method attached to it.  We then stuff this duplicate
     * on the EVP_PKEY and free the original EC_KEY.
     */
    ec1 = pkey->pkey.ec;
    ec2 = EC_KEY_dup(ec1);
    pkey->pkey.ec = ec2;
    EC_KEY_free(ec1);

    /*
     * Check if we need to enable SP800-131 legacy use for ECDSA verify
     */
    if (fips_on && legacy_use) {
	/*
	 * Set the non_fips_allow flag for SP800-131 legacy use if needed,
	 * otherwise the verify will fail.
	 */
	if (key_len < SP800131_EC_MIN_KEY_LEN || !nist) {
	    printf("   ...setting NON_FIPS_ALLOW flag\n");
	    EC_KEY_set_flags(pkey->pkey.ec, EC_KEY_get_flags(pkey->pkey.ec) | EC_FLAG_NON_FIPS_ALLOW);
	}
    }

    EVP_MD_CTX_init(&mctx);

    /*
     * Finally we can do the EVP init/update/final sequence
     */
    if (EVP_DigestVerifyInit(&mctx, NULL, md, NULL, pkey) <= 0) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "EVP_DigestVerifyInit failed for mctx (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_vfy_err;
    }
    if (EVP_DigestVerifyUpdate(&mctx, msg, mlen) <= 0) {
	ERR_print_errors_fp(stderr);
        fprintf(stderr, "EVP_DigestVerifyUpdate failed for mctx (FIPS=%d)\n", FIPS_mode());
        goto ecdsa_vfy_err;
    }
    if (EVP_DigestVerifyFinal(&mctx, sig, slen) <= 0) {
	if (!FIPS_mode()) {
	    ERR_print_errors_fp(stderr);
	    fprintf(stderr, "EVP_DigestVerifyFinal failed with FIPS disabled, this should never happen\n");
	    goto ecdsa_vfy_err;
	} else {
	    if (legacy_use && key_len < SP800131_EC_MIN_KEY_LEN) {
		ERR_print_errors_fp(stderr);
		fprintf(stderr, "EVP_DigestVerifyFinal failed with FIPS enabled and legacy use enabled, this should never happen\n");
		goto ecdsa_vfy_err;
	    } else if (!legacy_use && (key_len < SP800131_EC_MIN_KEY_LEN || !nist)) {
		fprintf(stderr, "   ...computation for non-NIST curve failed as expected, test passed\n");
                ERR_clear_error();
		rv = 0;
		goto ecdsa_vfy_err;
	    } else {
		ERR_print_errors_fp(stderr);
		fprintf(stderr, "EVP_DigestVerifyFinal failed with FIPS enabled using NIST curve, this should never happen\n");
		goto ecdsa_vfy_err;
	    }
	}
    }
    /*
     * We made it, the test has passed
     */
    rv = 0;
    ERR_clear_error();

ecdsa_vfy_err:
    SP800131_DISABLE_FIPS
    EVP_MD_CTX_cleanup(&mctx);
    if (key) EC_KEY_free(key);
    if (pkey) EVP_PKEY_free(pkey);
    if (sig) OPENSSL_free(sig);
    return rv;
}


/*
 * This routine will test EC key size limitations when FIPS mode is enabled
 * (as defined in SP-800-131).
 *
 * returns zero on success and one on failure
 */
static int sp800131_ec_test() {
    int rv = 1;
    int i, nid;

    for (i=0; i<SP800131_CURVES_LEN; i++) {
	nid = ec_curves_list[i]; 
	if (sp800131_ecdh_test(nid, 0)) {
	    fprintf(stderr, "ECDH FIPS disabled tests failed\n");
	    goto ec_err;
	}
	if (sp800131_ecdh_test(nid, 1)) {
	    fprintf(stderr, "ECDH FIPS disabled tests failed\n");
	    goto ec_err;
	}
	if (sp800131_ecdsa_sign_test(nid, 0)) {
	    fprintf(stderr, "ECDSA sign FIPS disabled tests failed\n");
	    goto ec_err;
	}
	if (sp800131_ecdsa_sign_test(nid, 1)) {
	    fprintf(stderr, "ECDSA sign FIPS enabled tests failed\n");
	    goto ec_err;
	}
	if (sp800131_ecdsa_verify_test(nid, 0, 0)) {
	    fprintf(stderr, "ECDSA verify FIPS disabled tests failed\n");
	    goto ec_err;
	}
	if (sp800131_ecdsa_verify_test(nid, 1, 0)) {
	    fprintf(stderr, "ECDSA verify FIPS enabled tests failed\n");
	    goto ec_err;
	}
	/*
	 * One more test with FIPS and legacy use enabled 
	 */
	if (sp800131_ecdsa_verify_test(nid, 1, 1)) {
	    fprintf(stderr, "ECDSA verify FIPS enabled tests failed\n");
	    goto ec_err;
	}
    }

    rv = 0;
ec_err:
    return rv;
}

/*
 * This routine will test EVP_DigestSign/Verify, EVP_Sign/Verify, and RSA_sign/verify
 * APIs with RSA to ensure the limitations when FIPS mode is enabled are properly enforced
 * (as defined in SP-800-131).  Work was performed under TASM-315.
 *
 * returns zero on success and one on failure
 */
static int sp800131_rsa_test() {
    printf("Starting RSA Sign/Verify tests for SP-800-131 Compliance...\n");
    /* print some additional debugging info when rsa_debug is set to 1 */
    int rsa_debug = 0;
    int rv = 1;
    EVP_PKEY *rsa_2048_pkey = NULL;
    EVP_PKEY *rsa_1024_pkey = NULL;
    
    unsigned char *sig_a = NULL, *sig_b = NULL;
    const unsigned char msg[] = "Here is a message to use as data for signing.";
    size_t mlen = strlen((char *)msg);
    size_t slen = 0;
    unsigned int i_mlen = (unsigned int)mlen;
    unsigned int i_slen = 0;

    /* define some variables to allow us to interpret the sign and verify parameters */
    #define SIGN 0
    #define VERIFY 1
    #define ALLOW_OFF 0
    #define USE_EVP 1
    #define USE_EVP_EX 0

    /* Negative tests need a 1024 key created in FIPS mode (which takes some work) ****/
    const int SSL_SIG_LEN = 36;
    RSA *rsa1024 = NULL;
    RSA *rsa1016 = NULL;
    /* message_digest string for use with signs and verifies  */
    unsigned char *md2 = NULL;
    /* string to hold rsa signature */
    unsigned char *sig = NULL;
    unsigned char *sig1024 = NULL;
    unsigned char *sig1016 = NULL;
    /* rsa signature length  */
    unsigned int siglen = 0;
    BIGNUM *bn = NULL;
    int ret = 0;

    /* create an int variable to hold the rsa flags of a key - we will save the value, manipulate the value and restore the value  */
    int rsaflags = 0;

    /* ******************* Disable FIPS ************************* */
    SP800131_DISABLE_FIPS

    /*
     * Init rsa_2048_pkey for sign and verify
     */
    if (!(rsa_2048_pkey = EVP_PKEY_new())) {
	fprintf(stderr, "EVP_PKEY_new failed for rsa_2048_pkey (FIPS=%d)\n", FIPS_mode());
        goto rsa_err;
    }

    if (gen_rsa_key(&rsa_2048_pkey, 2048)) {
	fprintf(stderr, "gen_rsa_key failed for rsa_2048_pkey (FIPS=%d)\n", FIPS_mode());
        goto rsa_err;
    } else {
      if(rsa_debug) printf("Success creating 2048 bit rsa key with FIPS mode=%d \n", FIPS_mode());
    }

    /*
     * Init rsa_1024_pkey for sign and verify
     */
    if (!(rsa_1024_pkey = EVP_PKEY_new())) {
	fprintf(stderr, "EVP_PKEY_new failed for rsa_1024_pkey (FIPS=%d)\n", FIPS_mode());
        goto rsa_err;
    }

    if (gen_rsa_key(&rsa_1024_pkey, 1024)) {
	fprintf(stderr, "gen_rsa_key failed for rsa_1024_pkey (FIPS=%d)\n", FIPS_mode());
        goto rsa_err;
    } else {
      if(rsa_debug) printf("Success creating 1024 bit rsa key with FIPS mode=%d \n", FIPS_mode());
    }
    
   /* ******************************************
     * Tests using 2048 bit keys
     * ******************************************
     */

    /* **********************************************
     * rsa_sign_or_verify extended 
     * **********************************************
     */

    /*
     * EVP Sign Ex operation:  AC2
     * The sign operation should pass using rsa.
     */
    if (sha_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, SIGN, ALLOW_OFF, USE_EVP_EX)) {
        fprintf(stderr, "( 1) evp_sign_ex failed with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
	goto rsa_err;
    } else {
      printf("( 1) evp_sign_ex succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * EVP Verify Ex operation:  AC6
     * The verify operation should pass using rsa.
     */
    if (sha_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, VERIFY, ALLOW_OFF, USE_EVP_EX)) {
	fprintf(stderr, "( 2) evp_verify_ex failed with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("( 2) evp_verify_ex succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /* ******************************************
     * rsa_sign_or_verify non-extended init
     * ******************************************
     */

    /*
     * EVP Sign operation:  AC2
     * The sign operation should pass using rsa. 
     */
    if (sha_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, SIGN, ALLOW_OFF, USE_EVP)) {
	fprintf(stderr, "( 3) evp sign failed with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("( 3) evp sign succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * EVP Verify operation:  AC6
     * The verify operation should pass using rsa.
     */
    if (sha_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, VERIFY, ALLOW_OFF, USE_EVP)) {
	fprintf(stderr, "( 4) evp verify failed with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("( 4) evp verify succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /* ******************************************
     * rsa_digest_sign_or_verify
     * ******************************************
     */

    /*
     * EVP Digest Sign operation:  AC2
     * The digest sign operation should pass using rsa.
     */
    if (sha_digest_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_b, &slen, msg, mlen, SIGN, ALLOW_OFF)) {
	fprintf(stderr, "( 5) digest sign failed with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("( 5) digest sign succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * EVP Digest Verify operation:  AC6
     * The digest verify operation should pass using rsa. 
     */
    if (sha_digest_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_b, &slen, msg, mlen, VERIFY, ALLOW_OFF)) {
	fprintf(stderr, "( 6) digest verify failed with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("( 6) digest verify succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /********************************
     *  Tests using 1024 bit rsa key
     ********************************
     */

    /*
     * EVP Sign Ex operation:  AC4
     * The sign operation should pass using rsa.
     */
    if (sha_sign_or_verify(rsa_1024_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, SIGN, ALLOW_OFF, USE_EVP_EX)) {
        fprintf(stderr, "( 7) evp_sign_ex failed with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
	goto rsa_err;
    } else {
      printf("( 7) evp_sign_ex succeeded with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * EVP Verify Ex operation:  AC8
     * The verify operation should pass using rsa.
     */
    if (sha_sign_or_verify(rsa_1024_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, VERIFY, ALLOW_OFF, USE_EVP_EX)) {
	fprintf(stderr, "( 8) evp_verify_ex failed with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("( 8) evp_verify_ex succeeded with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /* ******************************************
     * rsa_sign_or_verify non-extended init
     * ******************************************
     */

    /*
     * EVP Sign operation:  AC4
     * The sign operation should pass using rsa. 
     */
    if (sha_sign_or_verify(rsa_1024_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, SIGN, ALLOW_OFF, USE_EVP)) {
	fprintf(stderr, "( 9) evp sign failed with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("( 9) evp sign succeeded with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * EVP Verify operation:  AC8
     * The verify operation should pass using rsa.
     */
    if (sha_sign_or_verify(rsa_1024_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, VERIFY, ALLOW_OFF, USE_EVP)) {
	fprintf(stderr, "(10) evp verify failed with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("(10) evp verify succeeded with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /* ******************************************
     * rsa_digest_sign_or_verify
     * ******************************************
     */

    /*
     * EVP Digest Sign operation:  AC4
     * The digest sign operation should pass using rsa.
     */
    if (sha_digest_sign_or_verify(rsa_1024_pkey, EVP_sha384(), &sig_b, &slen, msg, mlen, SIGN, ALLOW_OFF)) {
	fprintf(stderr, "(11) digest sign failed with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("(11) digest sign succeeded with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * EVP Digest Verify operation:  AC8
     * The digest verify operation should pass using rsa. 
     */
    if (sha_digest_sign_or_verify(rsa_1024_pkey, EVP_sha384(), &sig_b, &slen, msg, mlen, VERIFY, ALLOW_OFF)) {
	fprintf(stderr, "(12) digest verify failed with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
        goto rsa_err;
    } else {
      printf("(12) digest verify succeeded with 1024 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * Reset values for use in FIPS mode
     */
    slen = 0;
    i_slen = 0;
    if (rsa_2048_pkey) {
        EVP_PKEY_free(rsa_2048_pkey);
        rsa_2048_pkey = NULL;
    }
    if (rsa_1024_pkey) {
        EVP_PKEY_free(rsa_1024_pkey);
        rsa_1024_pkey = NULL;
    }
    if (sig_a) {
        OPENSSL_free(sig_a);
        sig_a = NULL;
    }
    if (sig_b) {
        OPENSSL_free(sig_b);
        sig_b = NULL;
    }

      
    /* get a random string to act as our message digest of length 36 to sign  */
    /* apparently 16 bytes for the MD5, 20 bytes for the SHA-1  */
    md2  = malloc(SSL_SIG_LEN); 
    RAND_pseudo_bytes(md2, SSL_SIG_LEN);


    /* Now create our keys again for use in FIPS mode  */
    rsa1024 = RSA_new();
    rsa1016 = RSA_new();
    bn = BN_new();

    /* if we didn't get one of the RSA structs successfully then bail out */
    if (rsa1024 == NULL || rsa1016 == NULL ) {
	BIO_printf(bio_err, "Problem using RSA_new() ...\n");
	goto rsa_err;
    }
    /* if we didn't get a BN or RSA struct successfully then bail out */
    if (bn == NULL) {
	BIO_printf(bio_err, "Problem creating bn...\n");
	goto rsa_err;
    }
    /* OK we should be able to create RSA keys now */
    /* call BN_set_word to set the exponent to 65537 before trying key gen */
    BN_set_word(bn, 0x10001);
     
    if(rsa_debug) BIO_printf(bio_err, "Generating temp (%d bit) RSA key ... \n", 1024);
    if (!RSA_generate_key_ex(rsa1024, 1024, bn, NULL)) {
	BIO_printf(bio_err, "Error generating RSA %d bit key...\n", 1024);
	goto rsa_err;
    } 
    if (!RSA_generate_key_ex(rsa1016, 1016, bn, NULL)) {
	BIO_printf(bio_err, "Error generating RSA %d bit key...\n", 1016);
	goto rsa_err;
    } 

    BIO *mem1024 = BIO_new(BIO_s_mem());
    if (!PEM_write_bio_RSAPrivateKey(mem1024, rsa1024, NULL, NULL, 0, NULL, NULL)) {
	printf("failed to write 1024 bit RSA key in PEM format\n");
	ERR_print_errors(bio_err);
	goto rsa_err;
    }
    BIO *mem1016 = BIO_new(BIO_s_mem());
    if (!PEM_write_bio_RSAPrivateKey(mem1016, rsa1016, NULL, NULL, 0, NULL, NULL)) {
	printf("failed to write 1016 bit RSA key in PEM format\n");
	ERR_print_errors(bio_err);
	goto rsa_err;
    }

    /*
     * Create a signature to be used for 1024 & 1016 verify tests
     */
    sig1024 = malloc(RSA_size(rsa1024));
    ret = RSA_sign(NID_md5_sha1, md2, SSL_SIG_LEN, sig1024, &siglen, rsa1024);
    if (ret <= 0) {
	printf("rsa verify failed with key length of 1024 and FIPS off\n");
	ERR_print_errors(bio_err);
	goto rsa_err;
    }
    sig1016 = malloc(RSA_size(rsa1016));
    ret = RSA_sign(NID_md5_sha1, md2, SSL_SIG_LEN, sig1016, &siglen, rsa1016);
    if (ret <= 0) {
	printf("rsa verify failed with key length of 1016 and FIPS off\n");
	ERR_print_errors(bio_err);
	goto rsa_err;
    }

    RSA_free(rsa1024);
    RSA_free(rsa1016);
    rsa1024 = NULL;
    rsa1016 = NULL;


    /***************************** Enable FIPS Mode **************************/
    SP800131_ENABLE_FIPS

    rsa1024 = PEM_read_bio_RSAPrivateKey(mem1024, NULL, NULL, NULL); 
    BIO_free_all(mem1024);
    if (!rsa1024) {
	printf("failed to read 1024 bit RSA key in PEM format\n");
	ERR_print_errors(bio_err);
	goto rsa_err;
    }
    rsa1016 = PEM_read_bio_RSAPrivateKey(mem1016, NULL, NULL, NULL); 
    BIO_free_all(mem1016);
    if (!rsa1016) {
	printf("failed to read 1016 bit RSA key in PEM format\n");
	ERR_print_errors(bio_err);
	goto rsa_err;
    }


    /*
     * Init rsa_2048_pkey for signing
     */
    if (!(rsa_2048_pkey = EVP_PKEY_new())) {
	fprintf(stderr, "EVP_PKEY_new failed for rsa_2048_pkey (FIPS=%d)\n", FIPS_mode());
        goto rsa_err;
    }

    if (gen_rsa_key(&rsa_2048_pkey, 2048)) {
	fprintf(stderr, "gen_rsa_key failed for rsa_2048_pkey (FIPS=%d)\n", FIPS_mode());
        goto rsa_err;
    } else {
        if(rsa_debug) printf("Success creating 2048 bit rsa key with FIPS mode=%d \n", FIPS_mode());
    }

    /* restore the value of the flags for this key  */
    if(rsa_debug) printf("value of flags for rsa1024 key after key creation is 0x%X \n", rsa1024->flags);
    rsa1024->flags = rsaflags;
    if(rsa_debug) printf("value of flags for rsa1024 key after restoring is 0x%X \n", rsa1024->flags);
    
    /* ******************************************
     * rsa_sign_or_verify FIPS ON!
     * ******************************************
     */

    /*
     * EVP Sign Ex operation, FIPS on:  AC1
     * The sign operation should pass in FIPS mode with 2048 bit rsa key size
     */
    if (sha_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, SIGN, ALLOW_OFF, USE_EVP_EX)) {
	fprintf(stderr, "(13) evp_sign_ex failed with 2048 bit rsa key, sha384, and FIPS mode =%d\n", FIPS_mode());
        goto rsa_err;
    } else {
        printf("(13) evp_sign_ex succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * EVP Verify Ex operation, FIPS on:  AC5
     * The verify operation should pass using rsa with FIPS enabled.
     */
    if (sha_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, VERIFY, ALLOW_OFF, USE_EVP_EX)) {
	fprintf(stderr, "(14) evp_verify_ex failed with 2048 bit rsa key, sha384, and FIPS mode =%d\n", FIPS_mode());
        goto rsa_err;
    } else {
        printf("(14) evp_verify_ex succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /* ***********************************************
     * rsa_sign_or_verify FIPS ON! non-extended init
     * ***********************************************
     */

    /*
     * EVP Sign operation, FIPS on:  AC1
     * The sign operation should pass using rsa with FIPS enabled.
     */
    if (sha_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, SIGN, ALLOW_OFF, USE_EVP)) {
	fprintf(stderr, "(15) evp sign failed with 2048 bit rsa key, sha384, and FIPS mode =%d\n", FIPS_mode());
        goto rsa_err;
    } else {
        printf("(15) evp sign succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * EVP Verify operation, FIPS on:  AC5
     * The verify operation should pass using rsa with FIPS enabled.
     */
    if (sha_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_a, &i_slen, msg, i_mlen, VERIFY, ALLOW_OFF, USE_EVP)) {
	fprintf(stderr, "(16) evp verify failed with 2048 bit rsa key, sha384, and FIPS mode =%d\n", FIPS_mode());
        goto rsa_err;
    } else {
        printf("(16) evp verify succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /* ******************************************
     * rsa_digest_sign_or_verify FIPS ON!
     * ******************************************
     */

    /*
     * EVP Digest Sign operation, FIPS on:  AC1
     * The signature should still succeed here because 2048 is allowed in FIPS mode
     */
    if (sha_digest_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_b, &slen, msg, mlen, SIGN, ALLOW_OFF)) {
	fprintf(stderr, "(17) evp digest sign failed with 2048 bit rsa key, sha384, and FIPS mode =%d\n", FIPS_mode());
        goto rsa_err;
    } else {
        printf("(17) evp digest sign succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /*
     * EVP Digest Verify operation, FIPS on:  AC5
     * The verify should pass here using rsa with FIPS enabled.
     */
    if (sha_digest_sign_or_verify(rsa_2048_pkey, EVP_sha384(), &sig_b, &slen, msg, mlen, VERIFY, ALLOW_OFF)) {
	fprintf(stderr, "(18) evp digest verify failed with 2048 bit rsa key, sha384, and FIPS mode =%d\n", FIPS_mode());
        goto rsa_err;
    } else {
        printf("(18) evp digest verify succeeded with 2048 bit rsa key, sha384, and FIPS mode = %d\n", FIPS_mode());
    }

    /* AC3 RSA sign fails when key size less than 2048 with FIPS enabled.  */
    sig = malloc(RSA_size(rsa1024));
    if(rsa_debug) printf("Attempt RSA sign and verify in FIPS mode with RSA keylength 1024 ... should fail due to key size... \n");
    ret = RSA_sign(NID_md5_sha1, md2, SSL_SIG_LEN, sig, &siglen, rsa1024);
    if (ret == 0) {
	BIO_printf(bio_err, "(19) rsa sign failed AS EXPECTED with key length of 1024 and FIPS mode %d \n", FIPS_mode());
	if(rsa_debug) ERR_print_errors(bio_err);
    } else {
	printf("(19) rsa sign succeeded (SHOULD HAVE FAILED) with RSA key length of 1024 and FIPS mode %d \n", FIPS_mode());
	goto rsa_err;
    }

    /*  
     * Now in order to test verify for key length and not fail on a bad signature we need
     *  a valid signature to pass in ... so we will create one with RSA_sign by temporarily
     *  turning on the RSA_FLAG_NON_FIPS_ALLOW flag
     */
    
    /* Save flags settings for each key before manipulating it to allow key creation  */
    rsaflags = rsa1024->flags;
    if(rsa_debug) printf("value of rsaflags for rsa1024 key before manipulation is 0x%X \n",rsaflags);
    
    /* set the RSA_FLAG_NON_FIPS_ALLOW to allow RSA_sign with this key to succeed in FIPS mode */
    rsa1024->flags |= RSA_FLAG_NON_FIPS_ALLOW;
    if(rsa_debug) printf("value of flags for rsa1024 key after manipulation is 0x%X \n", rsa1024->flags);

    ret = RSA_sign(NID_md5_sha1, md2, SSL_SIG_LEN, sig, &siglen, rsa1024);
    if (ret > 0) {
	BIO_printf(bio_err, "(20) rsa sign succeeded (SHOULD FAIL) with key length of 1024 and FIPS mode %d even with RSA_FLAG_NON_FIPS_ALLOW set\n", FIPS_mode());
	ERR_print_errors(bio_err);
	goto rsa_err;
    } else {
	printf("(20) rsa sign failed as expected with key length of 1024 and FIPS mode %d with RSA_FLAG_NON_FIPS_ALLOW set for testing verify...\n", FIPS_mode());
    }

    /* AC9 RSA verify succeeds when key size less than 2048 with FIPS enabled but RSA_FLAG_NON_FIPS_ALLOW set  */
    ret = RSA_verify(NID_md5_sha1, md2, 36, sig1024, RSA_size(rsa1024), rsa1024);
    if (ret == 0) {
	BIO_printf(bio_err, "(21) rsa verify failed with key length of 1024 and FIPS mode %d even with RSA_FLAG_NON_FIPS_ALLOW set \n", FIPS_mode());
	ERR_print_errors(bio_err);
	goto rsa_err;
    } else {
	printf("(21) rsa verify succeeded with RSA key length of 1024 and FIPS mode %d and RSA_FLAG_NON_FIPS_ALLOW set \n", FIPS_mode());
    }

    /* Hopefully we have a good signature now to verify - unset the RSA_FLAG_NON_FIPS_ALLOW for the next verify test */
    rsa1024->flags = rsaflags;
    if(rsa_debug) printf("value of flags for rsa1024 key after restoring is 0x%X \n", rsa1024->flags);
    
    /* AC7 RSA verify fails when key size less than 2048 with FIPS enabled.  */
    ret = RSA_verify(NID_md5_sha1, md2, 36, sig1024, RSA_size(rsa1024), rsa1024);
    if (ret <= 0) {
	BIO_printf(bio_err, "(22) rsa verify failed AS EXPECTED with key length of 1024 and FIPS mode %d \n", FIPS_mode());
	if(rsa_debug) ERR_print_errors(bio_err);
    } else {
	printf("(22) rsa verify succeeded (SHOULD HAVE FAILED) with RSA key length of 1024 and FIPS mode %d \n", FIPS_mode());
	goto rsa_err;
    }

    /* RSA verify fails when key size less than 1024 with FIPS enabled but RSA_FLAG_NON_FIPS_ALLOW set  */
    ret = RSA_verify(NID_md5_sha1, md2, 36, sig1016, RSA_size(rsa1016), rsa1016);
    if (ret > 0) {
	BIO_printf(bio_err, "(23) rsa verify succeeded with key length of 1016 and FIPS mode %d with RSA_FLAG_NON_FIPS_ALLOW set \n", FIPS_mode());
	ERR_print_errors(bio_err);
	goto rsa_err;
    } else {
	printf("(23) rsa verify failed (AS EXPECTED) with RSA key length of 1016 and FIPS mode %d and RSA_FLAG_NON_FIPS_ALLOW set \n", FIPS_mode());
    }


    ERR_clear_error();
    rv = 0;

rsa_err:
    
 
    if (rsa_2048_pkey) {
      EVP_PKEY_free(rsa_2048_pkey);
    }
    if (rsa_1024_pkey) {
      EVP_PKEY_free(rsa_1024_pkey);
    }
    if (sig_a != NULL) {
        OPENSSL_free(sig_a);
    }
    if (sig_b != NULL) {
        OPENSSL_free(sig_b);
    }
    if (bn) 
	BN_free(bn);
    if (md2)
	free(md2);
    if (sig)
	free(sig);
    if (sig1024)
	free(sig1024);
    if (sig1016)
	free(sig1016);
    if (rsa1024)
        RSA_free(rsa1024);
    if (rsa1016)
        RSA_free(rsa1016);
   return rv;
}

int main(int argc, char *argv[])
{
    int any_failures = 0;

    /* memory leak debugging */
    CRYPTO_malloc_debug_init();
    CRYPTO_set_mem_debug_options(V_CRYPTO_MDEBUG_ALL);
    CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);
    ERR_load_crypto_strings();

    RAND_seed(rnd_seed, sizeof rnd_seed);

    if ((bio_err = BIO_new(BIO_s_file())) != NULL) {
        BIO_set_fp(bio_err, stderr, BIO_NOCLOSE | BIO_FP_TEXT);
    } else {
	fprintf(stderr, "bio_err creation failed\n");
	exit(1);
    }

    /* Add all the digests so they are available when needed  */
    OpenSSL_add_all_digests();
    
    if (sp800131_dh_test()) {
    	any_failures = 1;
    }

    if (sp800131_ec_test())  {
        any_failures = 1;
    }

    if (sp800131_hmac_test()) {
    	any_failures = 1;
    }

    if (sp800131_des_test())  {
        any_failures = 1;
    }
    
    if (sp800131_sha1_test()) {
    	any_failures = 1;
    }

    if (sp800131_rsa_test()) {
    	any_failures = 1;
    }

    if (!any_failures) {
	printf("All SP-800-131 tests passed\n");
    }
    ERR_print_errors_fp(stderr);
    BIO_free_all(bio_err);
    
    /* EVP_cleanup cleans up after OpenSSL_add_all_digests() */
    EVP_cleanup();

    /* memory debugging cleanup */
    CRYPTO_cleanup_all_ex_data();
    ERR_free_strings();
    ERR_remove_state(0);
    CRYPTO_mem_leaks_fp(stderr);

    return any_failures;
}
#endif
