/* -----------------------------------------------------------------
 * iv_gen.c - generic IV generator to be used with protocols
 * requiring an initial vector.
 *
 * June, 2011
 *
 * Copyright (c) 2011,2012 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/opensslconf.h>
#include <openssl/crypto.h>
#include <openssl/iv_gen.h>

/*
 * @brief Increments the counter for the given IV Generator context @param
 * iv_gen_ctx the context to use @retval 1 on success @retval 0 on failure
 */
static int iv_generator_increment_current_iv(iv_generator_t * iv_gen_ctx)
{
    int unsigned i = 0;
    int cy = 0;

    /* increment the counter */
    for (i = iv_gen_ctx->iv_counter_length; i > 0; i--) {
        if (0xff == iv_gen_ctx->iv_counter[i - 1]) {
            cy = 1;
        }
        iv_gen_ctx->iv_counter[i - 1]++;
        if (!cy)
            break;
        else
            cy = 0;
    }

    /* see if we have used up all of the IVs for this context */
    for (i = 0; i < iv_gen_ctx->iv_counter_length; i++) {
        if (iv_gen_ctx->iv_counter[i] != 0x00)
            break;
    }

    /* if all the bytes for the counter are zero, we have no more IVs left */
    if (i == iv_gen_ctx->iv_counter_length)
        return (0);
    else
        return (1);
}

/** @brief Creates a new IV Generator context
    @param none
    @retval pointer to new context on success
    @retval NULL on failure
*/
iv_generator_t *iv_generator_new(void)
{
    iv_generator_t *iv_gen_ctx = NULL;

    iv_gen_ctx = OPENSSL_malloc(sizeof(iv_generator_t));
    if (iv_gen_ctx != NULL) {
        iv_gen_ctx->initialized = IV_GEN_ALLOCATED;
    }
    return (iv_gen_ctx);
}

/** @brief Disposes of an IV Generator context
    @param iv_gen_ctx the context to remove
    @retval none
*/
void iv_generator_free(iv_generator_t * iv_gen_ctx)
{
    if (iv_gen_ctx != NULL) {
        memset(iv_gen_ctx, 0x00, sizeof(iv_generator_t));
        OPENSSL_free(iv_gen_ctx);
    }
}

/** @brief Initializes an IV Generator context with the
    desired characteristics
    @param iv_gen_ctx the context to initialize
    @param bytes_in_iv number of bytes for the IV
    @param fixed_common pointer to the fixed common portion
    @param bytes_in_fixed_common length of fixed common
    @param fixed_distinct pointer to the fixed distinct portion
    @param bytes_in_fixed_distinct length of fixed distinct
    @param salt pointer to the salt if needed
    @param bytes_in_salt length of the salt
    @retval 1 on success
    @retval 0 on failure
*/
IV_GEN_RC iv_generator_init(iv_generator_t * iv_gen_ctx,
                            unsigned long bytes_in_iv,
                            const unsigned char *fixed_common,
                            unsigned long bytes_in_fixed_common,
                            const unsigned char *fixed_distinct,
                            unsigned long bytes_in_fixed_distinct,
                            const unsigned char *salt,
                            unsigned long bytes_in_salt)
{
    int unsigned i = 0;

    /* make sure the context is there */
    if (iv_gen_ctx == NULL) {
        return (IV_NULL_ERR);
    }

    /* make sure the context has been properly allocated */
    if (iv_gen_ctx->initialized != IV_GEN_ALLOCATED) {
        return (IV_NOT_ALLOCATED_ERR);
    }

    /* make sure we are not trying to use an IV that is too big */
    if (bytes_in_iv > IV_GEN_MAX_NUM_BYTES) {
        return (IV_LENGTH_ERR);
    }

    /*
     * make sure that the fixed common and fixed distinct
     * portions of the IV leave room for a counter
     */
    if (bytes_in_iv <= (bytes_in_fixed_common + bytes_in_fixed_distinct)) {
        return (IV_LENGTH_ERR);
    }

    /* setup the IV lengths */
    iv_gen_ctx->iv_length = bytes_in_iv;
    iv_gen_ctx->iv_fixed_part_length = bytes_in_fixed_common +
        bytes_in_fixed_distinct;
    iv_gen_ctx->iv_counter_length = iv_gen_ctx->iv_length -
        iv_gen_ctx->iv_fixed_part_length;
    iv_gen_ctx->iv_implicit_part_len = bytes_in_fixed_common;

    /* intiialize and copy in the fixed components */
    memset(iv_gen_ctx->current_iv, 0x00, IV_GEN_MAX_NUM_BYTES);
    memset(iv_gen_ctx->iv_counter, 0x00, IV_GEN_MAX_NUM_BYTES);
    memcpy(iv_gen_ctx->current_iv, fixed_common, bytes_in_fixed_common);
    memcpy((iv_gen_ctx->current_iv + bytes_in_fixed_common), fixed_distinct,
           bytes_in_fixed_distinct);

    /* see if we need to xor in the salt */
    if ((salt != NULL) && (bytes_in_salt > 0)
        && (bytes_in_salt <= iv_gen_ctx->iv_length)) {
        for (i = 0; i < bytes_in_salt; i++) {
            iv_gen_ctx->current_iv[i] ^= salt[i];
        }
    }

    /*
     * set the IV context to initialized
     */
    iv_gen_ctx->initialized = IV_GEN_INITIALIZED;
    return (IV_OK);
}

/** @brief Gets the next available IV from the IV Generator context
    @param iv_gen_ctx the context to use
    @param location_to_write_iv place to store the IV that is generated
    @param bytes_in_iv length of the IV requested
    @retval 1 on success, location_to_write_iv contains new IV
    @retval 0 on failure
*/
IV_GEN_RC iv_generator_output_next_iv(iv_generator_t * iv_gen_ctx,
                                      unsigned char *location_to_write_iv,
                                      unsigned long bytes_in_iv)
{
    /* make sure the context is there */
    if (iv_gen_ctx == NULL) {
        return (IV_NULL_ERR);
    }

    /* make sure the context has been properly initialized */
    if (iv_gen_ctx->initialized != IV_GEN_INITIALIZED) {
        return (IV_NOT_INITIALIZED_ERR);
    }

    /* make sure they are requesting the correct IV length */
    if (bytes_in_iv != iv_gen_ctx->iv_length) {
        return (IV_LENGTH_ERR);
    }

    /* make sure we have a place to write the IV */
    if (location_to_write_iv == NULL) {
        return (IV_NULL_ERR);
    }

    /* make sure we can generate another IV */
    if (!iv_generator_increment_current_iv(iv_gen_ctx)) {
        return (IV_EXHAUSTED_ERR);
    }

    /* write the new IV out */
    memcpy((iv_gen_ctx->current_iv + iv_gen_ctx->iv_fixed_part_length),
           iv_gen_ctx->iv_counter, iv_gen_ctx->iv_counter_length);
    memcpy(location_to_write_iv, iv_gen_ctx->current_iv,
           iv_gen_ctx->iv_length);

    return (IV_OK);
}

/** @brief Gets the length of the IV that the given IV Generator
    context is configured to use
    @param iv_gen_ctx the context to use
    @retval value of IV length on success
    @retval -1 on failure
*/
int iv_generator_get_iv_length(iv_generator_t * iv_gen_ctx)
{
    if (iv_gen_ctx != NULL) {
        if (iv_gen_ctx->initialized == IV_GEN_INITIALIZED) {
            return (iv_gen_ctx->iv_length);
        } else {
            return (-1);
        }
    } else {
        return (-1);
    }
}

/** @brief Gets the length of the fixed portion that the given
    IV Generator context is configured to use
    @param iv_gen_ctx the context to use
    @retval value of fixed part length on success
    @retval -1 on failure
*/
int iv_generator_get_fixed_part_length(iv_generator_t * iv_gen_ctx)
{
    if (iv_gen_ctx != NULL) {
        if (iv_gen_ctx->initialized == IV_GEN_INITIALIZED) {
            return (iv_gen_ctx->iv_fixed_part_length);
        } else {
            return (-1);
        }
    } else {
        return (-1);
    }
}

/** @brief Writes the implicit part of the IV (fixed common field) and its length
    @param iv_gen_ctx the context to use
    @retval 1 on success
    @retval -1 on failure
*/
int iv_generator_get_implicit_part(iv_generator_t * iv_gen_ctx, unsigned char
                                   *location_to_write_implicit_part, unsigned long
                                   *location_to_write_implicit_part_len)
{
    if (iv_gen_ctx != NULL) {
        if (iv_gen_ctx->initialized == IV_GEN_INITIALIZED) {
            memcpy(location_to_write_implicit_part,
                   iv_gen_ctx->current_iv, iv_gen_ctx->iv_implicit_part_len);
            *location_to_write_implicit_part_len =
                iv_gen_ctx->iv_implicit_part_len;
            return 1;
        } else {
            return (-1);
        }
    } else {
        return (-1);
    }
}
