/* -----------------------------------------------------------------
 * iv_gen_test.c - Test application for generic IV generator to be used with
 * protocols requiring an initial vector.
 *
 * June, 2011
 *
 * Copyright (c) 2011,2012 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/opensslconf.h>
#include <openssl/crypto.h>
#include <openssl/iv_gen.h>
#include <openssl/err.h>

/** @brief Fixed Common part to use for IV Generator test routines */
const unsigned char fixed_common[4] = { 0xFA, 0xCE, 0xEC, 0xAF };

/** @brief Fixed Common part to use for IV Generator limit test routines */
const unsigned char fixed_common_limit[8] = {
    0xFA, 0xCE, 0xEC, 0xAF, 0xFA, 0xCE, 0xEC, 0xAF
};

/** @brief Fixed Distinct part to use for IV Generator test routines */
const unsigned char fixed_distinct[4] = { 0xDE, 0xAD, 0xBE, 0xEF };

/** @brief Fixed Distinct part to use for IV Generator limit test routines */
const unsigned char fixed_distinct_limit[3] = { 0x00, 0x00, 0x00 };

/** @brief Prints out the data for a given IV Generator context
    @param iv_gen_ctx the context to use
    @retval none
*/
void print_iv_gen_info(iv_generator_t * iv_ctx)
{
    int unsigned i = 0;

    printf("IV Generator Context:\n");
    printf("   IV Length      : %ld\n", iv_ctx->iv_length);
    printf("   Fixed Length   : %ld\n", iv_ctx->iv_fixed_part_length);
    printf("   Counter Length : %ld\n", iv_ctx->iv_counter_length);
    printf("   Counter        :");
    for (i = 0; i < iv_ctx->iv_counter_length; i++)
        printf("%.2x", iv_ctx->iv_counter[i]);
    printf("\n");
    printf("   Current IV     :");
    for (i = 0; i < iv_ctx->iv_length; i++)
        printf("%.2x", iv_ctx->current_iv[i]);
    printf("\n");
}

/** @brief Main test routine for the IV Generator tests.
    No arguments are required to execute the tests
    @param argc argument count
    @param argv argument values
    @retval 0 on success
    @retval prints out errors messages on failure
*/
int main(int argc, char *argv[])
{
    IV_GEN_RC rc = 0;
    int i = 0;
    unsigned char next_iv[12];
    iv_generator_t *iv_ctx = NULL;

    /* memory leak debugging */
    CRYPTO_malloc_debug_init();
    CRYPTO_set_mem_debug_options(V_CRYPTO_MDEBUG_ALL);
    CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);
    ERR_load_crypto_strings();

    iv_ctx = iv_generator_new();
    rc = iv_generator_init(iv_ctx, 12,
                           fixed_common, 4, fixed_distinct, 4, NULL, 0);
    if (rc != IV_OK) {
        printf("iv_generator_init failed\n");
        goto cleanup;
    }
    print_iv_gen_info(iv_ctx);

    rc = iv_generator_output_next_iv(iv_ctx, next_iv, 12);
    if (rc != IV_OK) {
        printf("iv_generator_output_next_iv failed\n");
        goto cleanup;
    }
    print_iv_gen_info(iv_ctx);
    iv_generator_free(iv_ctx);

    printf("\n\nIV Generator Limit Test\n");
    printf("+++++++++++++++++++++++\n");
    /* now test the generator limits with a 1 byte counter */
    iv_ctx = iv_generator_new();
    rc = iv_generator_init(iv_ctx, 12,
                           fixed_common_limit, 8,
                           fixed_distinct_limit, 3, NULL, 0);
    if (rc != IV_OK) {
        printf("iv_generator_init failed\n");
        goto cleanup;
    }
    print_iv_gen_info(iv_ctx);
    printf("\nExecuting the counter limit test.....\n");

    for (i = 0; i < 300; ++i) {
        rc = iv_generator_output_next_iv(iv_ctx, next_iv, 12);
        switch (rc) {
        case IV_OK:
            break;
        case IV_EXHAUSTED_ERR:
            printf("Generator produced %d IVs with a 1-byte counter\n", i);
            printf("Test Passed\n\n");
            goto cleanup;
            break;
        default:
            printf("Counter limit test failed (rc=%d)\n\n", rc);
            goto cleanup;
            break;
        }
    }

    if (i == 300) {
        printf("Generator produced too many IVs for this context\n");
        printf("Test Failed\n\n");
    }

 cleanup:
    iv_generator_free(iv_ctx);

    /* memory debugging cleanup */
    CRYPTO_cleanup_all_ex_data();
    ERR_free_strings();
    ERR_remove_state(0);
    CRYPTO_mem_leaks_fp(stderr);
    ERR_print_errors_fp(stdout);

    return 0;
}
