/*------------------------------------------------------------------
 * kdf/kdf_ikev2.c - Key Derivation Functions for IKEv2
 *
 * April 2015
 *
 * Copyright (c) 2015 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cryptlib.h"
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/kdf.h>
#include <openssl/sha.h>
#include <openssl/hmac.h>
#ifdef OPENSSL_FIPS
# include <openssl/fips.h>
#endif


/*! @brief kdf_ikev2_gen - KDF in compliance with SP800-135 for IKEv2,
                           generate the seedkey.
  
@latexonly
\begin{verbatim}
      nonce = Ni || Nr
      shared_secret = g^ir

\end{verbatim}
@endlatexonly

      @param seedkey - pointer to output for seedkey
      @param evp_md - SHA digest
      @param nonce - pointer to nonce input
      @param nonce_len - nonce legnth(in bytes)
      @param shared_secret - pointer to secret input
      @param shared_secret_len - secret length(in bytes)

      @return - 0 pass, -1 fail
 */
int kdf_ikev2_gen(unsigned char *seedkey, const EVP_MD *evp_md, const void *nonce, 
                  unsigned int nonce_len, const void *shared_secret, 
		  unsigned int shared_secret_len)
{
    HMAC_CTX hctx;
    unsigned int len;
    int ret = -1;

    if (!nonce || !nonce_len || !shared_secret || 
        !shared_secret_len || !evp_md || !seedkey) {
        KDFerr(KDF_F_KDF_IKEV2_GEN, KDF_R_INPUT_PARAMETER_ERROR);
	return ret;
    }

#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
        /* since the digest is already chosen, need to convert for FIPS */
        evp_md = FIPS_get_digestbynid(evp_md->type);
	if (!evp_md) {
            KDFerr(KDF_F_KDF_IKEV2_GEN, KDF_R_BAD_DIGEST);
            return ret;
        }
        return FIPS_kdf_ikev2_gen(seedkey, evp_md, nonce, 
				  nonce_len, shared_secret, 
				  shared_secret_len);
    }
#endif

    HMAC_CTX_init(&hctx);
    if (!HMAC_Init_ex(&hctx, nonce, nonce_len, evp_md, NULL)) {
        goto err;
    }
    if (!HMAC_Update(&hctx, shared_secret, shared_secret_len)) {
        goto err;
    }
    if (!HMAC_Final(&hctx, seedkey, &len)) {
        goto err;
    }

    ret = 0;
err:
    HMAC_CTX_cleanup(&hctx);
    return ret;
}

/*! @brief kdf_ikev2_rekey - KDF in compliance with SP800-135 for IKEv2,
                             re-generate the seedkey.
  
@latexonly
\begin{verbatim}
      nonce = Ni || Nr
      shared_secret = g^ir(new)
      sk_d = DKM (len = md_size)

\end{verbatim}
@endlatexonly

      @param seedkey - pointer to output for seedkey
      @param evp_md - SHA digest
      @param nonce - pointer to nonce input
      @param nonce_len - nonce legnth(in bytes)
      @param shared_secret - pointer to secret input
      @param shared_secret_len - secret length(in bytes)
      @param dh - 0/1 whether shared secret included
      @param sk_d - pointer to DKM
      @param skd_len - length of DKM(in bytes)

      @return - 0 pass, -1 fail

 */
int kdf_ikev2_rekey(unsigned char *seedkey, const EVP_MD *evp_md, const void *nonce, 
                    unsigned int nonce_len, const void *shared_secret, 
		    unsigned int shared_secret_len, int dh,
		    const void *sk_d, unsigned int skd_len)
{
    HMAC_CTX hctx;
    unsigned int len;
    int ret = -1;

    if (!nonce || !nonce_len || !shared_secret || 
        !shared_secret_len || !evp_md || !seedkey ||
	!sk_d || !skd_len) {
        KDFerr(KDF_F_KDF_IKEV2_REKEY, KDF_R_INPUT_PARAMETER_ERROR);
	return ret;
    }

#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
        /* since the digest is already chosen, need to convert for FIPS */
        evp_md = FIPS_get_digestbynid(evp_md->type);
	if (!evp_md) {
            KDFerr(KDF_F_KDF_IKEV2_REKEY, KDF_R_BAD_DIGEST);
            return ret;
        }
        return FIPS_kdf_ikev2_rekey(seedkey, evp_md, nonce, 
				    nonce_len, shared_secret, 
				    shared_secret_len, dh,
				    sk_d, skd_len);
    }
#endif

    HMAC_CTX_init(&hctx);
    if (!HMAC_Init_ex(&hctx, sk_d, skd_len, evp_md, NULL)) {
        goto err;
    }

    /* If used for DH then also include secret */
    if (dh) {
        if (!HMAC_Update(&hctx, shared_secret, shared_secret_len)) {
            goto err;
        }
    }

    if (!HMAC_Update(&hctx, nonce, nonce_len)) {
        goto err;
    }

    if (!HMAC_Final(&hctx, seedkey, &len)) {
        goto err;
    }

    ret = 0;
err:
    HMAC_CTX_cleanup(&hctx);
    return ret;
}


/*! @brief kdf_ikev2_dkm - KDF in compliance with SP800-135 for IKEv2,
                           generate the Derived Keying Material. For DKM,
  		           DKM(Child SA) and DKM(Child SA D-H).
  
@latexonly
\begin{verbatim}
      seedkey(DKM) = SEEDKEY
      seedkey(DKM SA) = DKM
      seedkey(DKM SA DH) = DKM
      nonce(DKM) = Ni || Nr || SPIi || SPIr
      nonce(DKM SA) = Ni || Nr
      nonce(DKM SA DH) = Ni || Nr
      shared_secret(DKM) = Null
      shared_secret(DKM SA) = Null
      shared_secret(DKM SA DH) = g^ir (new)

\end{verbatim}
@endlatexonly

      @param dkm - pointer to output dkm
      @param len_out - output length(in bytes)
      @param evp_md - SHA digest
      @param seedkey - pointer to seedkey input
      @param seedkey_len - lebgth of seedkey(in bytes)
      @param nonce - pointer to nonce input
      @param nonce_len - nonce legnth(in bytes)
      @param shared_secret - pointer to secret input
      @param shared_secret_len - secret length(in bytes)

      @return - 0 pass, -1 fail

 */
int kdf_ikev2_dkm(unsigned char *dkm, unsigned int len_out, const EVP_MD *evp_md, 
	          const void *seedkey, unsigned int seedkey_len,
		  unsigned char *nonce, unsigned int nonce_len,
		  unsigned char *shared_secret, unsigned int shared_secret_len)
{
    HMAC_CTX hctx;
    unsigned int len = 0;
    int i, value_len, ret = -1;
    unsigned char *value;

    if (!nonce || !nonce_len || !evp_md || !seedkey || 
        !seedkey_len || !dkm || !len_out) {
        KDFerr(KDF_F_KDF_IKEV2_DKM, KDF_R_INPUT_PARAMETER_ERROR);
	return ret;
    }

    /* if len is defined, ss must be defined */
    if (shared_secret_len && !shared_secret) {
        KDFerr(KDF_F_KDF_IKEV2_DKM, KDF_R_INPUT_PARAMETER_ERROR);
	return ret;
    }

#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
        /* since the digest is already chosen, need to convert for FIPS */
        evp_md = FIPS_get_digestbynid(evp_md->type);
	if (!evp_md) {
            KDFerr(KDF_F_KDF_IKEV2_DKM, KDF_R_BAD_DIGEST);
            return ret;
        }
        return FIPS_kdf_ikev2_dkm(dkm, len_out, evp_md, 
				  seedkey, seedkey_len,
				  nonce, nonce_len,
				  shared_secret, shared_secret_len);
    }
#endif

    value_len = shared_secret_len + nonce_len + 1;
    value = OPENSSL_malloc(value_len);
    if (!value) {
        KDFerr(KDF_F_KDF_IKEV2_DKM, KDF_R_MALLOC);
        return ret;
    }

    /* concatenate and add counter to end */
    value[value_len - 1] = 01;
    if (shared_secret_len) {
        memcpy(value, shared_secret, shared_secret_len);
    }
    memcpy(value + shared_secret_len, nonce, nonce_len);

    HMAC_CTX_init(&hctx);
    
    for (i=0; i<len_out; i+=evp_md->md_size) {

        if (!HMAC_Init_ex(&hctx, seedkey, seedkey_len, evp_md, NULL)) {
	    goto err;
        }
	/* after first pass hash in the dkm */
	if (i != 0) {
            if (!HMAC_Update(&hctx, &dkm[i - evp_md->md_size], len)) {
	        goto err;
            }
        }

	if (!HMAC_Update(&hctx, value, value_len)) {
	    goto err;
        }

        if (!HMAC_Final(&hctx, &dkm[i], &len)) {
            goto err;
        }
	value[value_len - 1]++; 
    }

    ret = 0;
err:
    OPENSSL_cleanse(value, value_len);
    OPENSSL_free(value);
    HMAC_CTX_cleanup(&hctx);
    return ret;
}

