/* -----------------------------------------------------------------
 * crypto/md5/asm/md5-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is not in the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <stdio.h>
#define OCTEON_MD5
#include "md5_locl.h"
#include <openssl/opensslv.h>
#include <openssl/crypto.h>
#ifndef _MIPS_ARCH_OCTEON3
# include "ciscossl-cvmx.h"
#else
# include "cvmx.h"
#endif

/*
 * Implemented from RFC1321 The MD5 Message-Digest Algorithm
 */

static inline uint64_t swap64(uint64_t v)
{
    return ((v >> 56) |
            (((v >> 48) & 0xfful) << 8) |
            (((v >> 40) & 0xfful) << 16) |
            (((v >> 32) & 0xfful) << 24) |
            (((v >> 24) & 0xfful) << 32) |
            (((v >> 16) & 0xfful) << 40) |
            (((v >> 8) & 0xfful) << 48) | (((v >> 0) & 0xfful) << 56));
}

int MD5OCT_Init(MD5_CTX *c)
{
    memset(c, 0, sizeof(*c));
    c->E = 0x0123456789abcdefull;
    c->F = 0xfedcba9876543210ull;
    c->num = 0;
    return 1;
}

/*
 * MD5OCT_Update - can be called repeatedly with chunks of the
 * message to be hashed (len bytes at data).
 */
int MD5OCT_Update(MD5_CTX *c, const void *data_, size_t len)
{
    unsigned long remaining = 0, totlen = 0, copied = 0;
    const uint64_t *ptr = (const uint64_t *)data_;

    if (!len)
        return 1;

    totlen = len;
    if (c->num) {
        /* Adjust the previous data */
        /* This is not the first update call */
        totlen += remaining = c->num % MD5_CBLOCK;
        if (remaining) {
            memcpy((void *)&(c->data) + remaining, data_,
                   MD5_CBLOCK - remaining);
            copied = 1;
        }
    }
    if (totlen >= MD5_CBLOCK) {
        /* Initialise the running IV */
        CVMX_MT_HSH_IV(c->E, 0);
        CVMX_MT_HSH_IV(c->F, 1);

        while (totlen >= MD5_CBLOCK) {
            if (copied) {
                copied = 0;
            }
            if (remaining) {
                ptr = (uint64_t *) & c->data;
            }
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 0);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 1);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 2);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 3);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 4);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 5);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 6);
	    CVMX_MT_HSH_STARTMD5 (cvmx_htobe64(*ptr++));

            totlen -= MD5_CBLOCK;
            if (remaining) {
                ptr = data_ + (MD5_CBLOCK - remaining);
                remaining = 0;
            }
        }
        CVMX_MF_HSH_IV(c->E, 0);
        CVMX_MF_HSH_IV(c->F, 1);
    }
    /*
     * if (len > MD5_CBLOCK)
     */
    c->num += (len);
    /*
     * Copy the remaining stuffs in to a buffer and do the hash in the next
     * update or final.
     */
    if (!copied)
        memcpy(&(c->data), ptr, totlen);
    return 1;
}

/*
 * MD5OCT_Final - places the message digest in md
 */
int MD5OCT_Final(unsigned char *md, MD5_CTX *c)
{
    unsigned long len;
    uint8_t chunk[MD5_CBLOCK];
    uint64_t bits;
    const uint64_t *ptr;

    len = c->num % MD5_CBLOCK;
#ifdef L_ENDIAN
    bits = (c->num * 8);
#else
    bits = swap64 (c->num * 8);
#endif

    memset(chunk, 0, MD5_CBLOCK);
    /* The rest of the data will need to be copied into a chunk */
    if (len > 0) {
        memcpy(chunk, c->data, len);
    }
    chunk[len] = 0x80;
    /* Initialise the running IV */
    CVMX_MT_HSH_IV(c->E, 0);
    CVMX_MT_HSH_IV(c->F, 1);

    ptr = (const uint64_t *)chunk;

    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 0);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 1);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 2);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 3);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 4);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 5);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 6);


    /* Check to see if there is room for the bit count */
    if (len < 56) {
        CVMX_MT_HSH_STARTMD5 (cvmx_bswap64(bits));
    } else {
        CVMX_MT_HSH_STARTMD5 (cvmx_htobe64(*ptr));
        /* Another block was needed */
        CVMX_MT_HSH_DATZ(0);
        CVMX_MT_HSH_DATZ(1);
        CVMX_MT_HSH_DATZ(2);
        CVMX_MT_HSH_DATZ(3);
        CVMX_MT_HSH_DATZ(4);
        CVMX_MT_HSH_DATZ(5);
        CVMX_MT_HSH_DATZ(6);

        /*
         * Workaround for issue described in section "2.2.3.2 MD5 Hash Calculation
         * Restart" of the CN63XX Pass 1 Known Issues document.
         */
        CVMX_MF_HSH_IV(c->E, 0);
        CVMX_MF_HSH_IV(c->F, 1);

	CVMX_MT_HSH_STARTMD5 (cvmx_bswap64(bits));
    }
    CVMX_MF_HSH_IV(c->E, 0);
    CVMX_MF_HSH_IV(c->F, 1);

#ifdef L_ENDIAN
    c->E  =  cvmx_betoh64(c->E);
    c->F  =  cvmx_betoh64(c->F);
#endif
    memcpy(md, (void *)&c->E, 8);
    memcpy(md + 8, (void *)&c->F, 8);
    c->num = 0;
    return 1;
}
