/* crypto/rand/drbg_rand.c */
/* Written by Dr Stephen N Henson (steve@openssl.org) for the OpenSSL
 * project.
 */
/* ====================================================================
 * Copyright (c) 2011 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    licensing@OpenSSL.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#include <string.h>
#include <openssl/crypto.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include "rand_lcl.h"
#include "drbg_rand_lcl.h"

#if !(defined(OPENSSL_SYS_WIN32) || defined(OPENSSL_SYS_VXWORKS))
# include <sys/time.h>
#endif

#if defined(OPENSSL_SYS_VXWORKS)
# include <time.h>
#endif

#ifndef OPENSSL_SYS_WIN32
# ifdef OPENSSL_UNISTD
#  include OPENSSL_UNISTD
# else
#  include <unistd.h>
# endif
#endif

/* Mapping of SP800-90 DRBGs to OpenSSL RAND_METHOD */

/* Since we only have one global PRNG used at any time in OpenSSL use a global
 * variable to store context.
 */

static DRBG_CTX ossl_dctx;
static int drbg_initialized = 0;
/* Global buffer for Deterministic ECDSA stir */
unsigned char global_drbg_adin_buf[48];

DRBG_CTX *drbg_get_default_ctx(void)
{
    return &ossl_dctx;
}

static int drbg_bytes(unsigned char *out, int count)
{
    DRBG_CTX *dctx = &ossl_dctx;
    int rv = 0;
    unsigned char *adin = NULL;
    size_t adinlen = 0;
    CRYPTO_w_lock(CRYPTO_LOCK_RAND);
    do {
        size_t rcnt;
        if (count > (int)dctx->max_request)
            rcnt = dctx->max_request;
        else
            rcnt = count;
        if (dctx->get_adin) {
            adinlen = dctx->get_adin(dctx, &adin);
            if (adinlen && !adin) {
                RANDerr(RAND_F_DRBG_BYTES,
                        RAND_R_ERROR_RETRIEVING_ADDITIONAL_INPUT);
                goto err;
            }
        }
        rv = drbg_generate(dctx, out, rcnt, 0, adin, adinlen);
        if (adin) {
            if (dctx->cleanup_adin)
                dctx->cleanup_adin(dctx, adin, adinlen);
            adin = NULL;
        }
        if (!rv)
            goto err;
        out += rcnt;
        count -= (int)rcnt;
    }
    while (count);
    rv = 1;
 err:
    CRYPTO_w_unlock(CRYPTO_LOCK_RAND);
    return rv;
}

static int drbg_pseudo(unsigned char *out, int count)
{
    if (drbg_bytes(out, count) <= 0)
        return -1;
    return 1;
}

static int drbg_status(void)
{
    DRBG_CTX *dctx = &ossl_dctx;
    int rv;
    CRYPTO_r_lock(CRYPTO_LOCK_RAND);
    rv = dctx->status == DRBG_STATUS_READY ? 1 : 0;
    CRYPTO_r_unlock(CRYPTO_LOCK_RAND);
    return rv;
}

static void drbg_cleanup(void)
{
    DRBG_CTX *dctx = &ossl_dctx;
    CRYPTO_w_lock(CRYPTO_LOCK_RAND);
    drbg_uninstantiate(dctx);
    CRYPTO_w_unlock(CRYPTO_LOCK_RAND);
}

static void drbg_seed(const void *seed, int seedlen)
{
    DRBG_CTX *dctx = &ossl_dctx;
    if (dctx->rand_seed_cb)
        dctx->rand_seed_cb(dctx, seed, seedlen);
}

static void drbg_add(const void *seed, int seedlen, double add_entropy)
{
    DRBG_CTX *dctx = &ossl_dctx;
    if (dctx->rand_add_cb)
        dctx->rand_add_cb(dctx, seed, seedlen, add_entropy);
}

static const RAND_METHOD rand_drbg_meth = {
    drbg_seed,
    drbg_bytes,
    drbg_cleanup,
    drbg_add,
    drbg_pseudo,
    drbg_status
};

static size_t drbg_get_entropy(DRBG_CTX *ctx, unsigned char **pout,
                               int entropy, size_t min_len, size_t max_len)
{
    /* Round up request to multiple of block size */
    min_len = ((min_len + 19) / 20) * 20;
    *pout = OPENSSL_malloc(min_len);
    if (!*pout)
        return 0;
    if (ssleay_rand_bytes(*pout, min_len, 0, 0) <= 0) {
        OPENSSL_free(*pout);
        *pout = NULL;
        return 0;
    }
    return min_len;
}

static void drbg_free_entropy(DRBG_CTX *ctx, unsigned char *out, size_t olen)
{
    if (out) {
        OPENSSL_cleanse(out, olen);
        OPENSSL_free(out);
    }
}

/*
 * Set "additional input" when generating random data. This uses the current
 * PID, a time value and a counter.
 */
static void drbg_get_timevec(unsigned char *buf, unsigned long *pctr)
{
#ifdef OPENSSL_SYS_WIN32
    FILETIME ft;
#elif defined(OPENSSL_SYS_VXWORKS)
    struct timespec ts;
#else
    struct timeval tv;
#endif

#ifndef GETPID_IS_MEANINGLESS
    unsigned long pid;
#endif

#ifdef OPENSSL_SYS_WIN32
    GetSystemTimeAsFileTime(&ft);
    buf[0] = (unsigned char)(ft.dwHighDateTime & 0xff);
    buf[1] = (unsigned char)((ft.dwHighDateTime >> 8) & 0xff);
    buf[2] = (unsigned char)((ft.dwHighDateTime >> 16) & 0xff);
    buf[3] = (unsigned char)((ft.dwHighDateTime >> 24) & 0xff);
    buf[4] = (unsigned char)(ft.dwLowDateTime & 0xff);
    buf[5] = (unsigned char)((ft.dwLowDateTime >> 8) & 0xff);
    buf[6] = (unsigned char)((ft.dwLowDateTime >> 16) & 0xff);
    buf[7] = (unsigned char)((ft.dwLowDateTime >> 24) & 0xff);
#elif defined(OPENSSL_SYS_VXWORKS)
    clock_gettime(CLOCK_REALTIME, &ts);
    buf[0] = (unsigned char)(ts.tv_sec & 0xff);
    buf[1] = (unsigned char)((ts.tv_sec >> 8) & 0xff);
    buf[2] = (unsigned char)((ts.tv_sec >> 16) & 0xff);
    buf[3] = (unsigned char)((ts.tv_sec >> 24) & 0xff);
    buf[4] = (unsigned char)(ts.tv_nsec & 0xff);
    buf[5] = (unsigned char)((ts.tv_nsec >> 8) & 0xff);
    buf[6] = (unsigned char)((ts.tv_nsec >> 16) & 0xff);
    buf[7] = (unsigned char)((ts.tv_nsec >> 24) & 0xff);
#else
    gettimeofday(&tv, NULL);
    buf[0] = (unsigned char)(tv.tv_sec & 0xff);
    buf[1] = (unsigned char)((tv.tv_sec >> 8) & 0xff);
    buf[2] = (unsigned char)((tv.tv_sec >> 16) & 0xff);
    buf[3] = (unsigned char)((tv.tv_sec >> 24) & 0xff);
    buf[4] = (unsigned char)(tv.tv_usec & 0xff);
    buf[5] = (unsigned char)((tv.tv_usec >> 8) & 0xff);
    buf[6] = (unsigned char)((tv.tv_usec >> 16) & 0xff);
    buf[7] = (unsigned char)((tv.tv_usec >> 24) & 0xff);
#endif
    buf[8] = (unsigned char)(*pctr & 0xff);
    buf[9] = (unsigned char)((*pctr >> 8) & 0xff);
    buf[10] = (unsigned char)((*pctr >> 16) & 0xff);
    buf[11] = (unsigned char)((*pctr >> 24) & 0xff);

    (*pctr)++;

#ifndef GETPID_IS_MEANINGLESS
    pid = (unsigned long)getpid();
    buf[12] = (unsigned char)(pid & 0xff);
    buf[13] = (unsigned char)((pid >> 8) & 0xff);
    buf[14] = (unsigned char)((pid >> 16) & 0xff);
    buf[15] = (unsigned char)((pid >> 24) & 0xff);
#endif
}

static size_t drbg_get_adin(DRBG_CTX *ctx, unsigned char **pout)
{
    /* Use of static variables is OK as this happens under a lock */
    static unsigned char buf[16];
    static unsigned long counter;
    drbg_get_timevec(buf, &counter);

    /* Copy the time data into last 16B of global_drbg_adin_buf.
     * buf size and offset is dep. upon values chosen in
     * crypto/ecdsa/ecs_sign.c */
    memcpy(global_drbg_adin_buf+32, buf, 16);

    *pout = global_drbg_adin_buf;
    return sizeof(global_drbg_adin_buf);
}

/*
 * RAND_add() and RAND_seed() pass through to OpenSSL PRNG so it is
 * correctly seeded by RAND_poll().
 */

static int drbg_rand_add(DRBG_CTX *ctx, const void *in, int inlen,
                         double entropy)
{
    RAND_SSLeay()->add(in, inlen, entropy);
    return 1;
}

static int drbg_rand_seed(DRBG_CTX *ctx, const void *in, int inlen)
{
    RAND_SSLeay()->seed(in, inlen);
    return 1;
}

const RAND_METHOD *RAND_get_drbg_method(void)
{
    size_t plen;
    unsigned char pers[64], *p;
    DRBG_CTX *dctx = &ossl_dctx;

    if (!drbg_initialized) {
        if (drbg_init(dctx, NID_aes_256_ctr, DRBG_FLAG_CTR_USE_DF <= 0)) {
	    return 0;
	}
        drbg_set_callbacks(dctx,
                           drbg_get_entropy, drbg_free_entropy, 20,
                           drbg_get_entropy, drbg_free_entropy);
        drbg_set_rand_callbacks(dctx, drbg_get_adin, 0, drbg_rand_seed,
                                drbg_rand_add);
        /* Personalisation string: a string followed by date time vector */
        strcpy((char *)pers, "CiscoSSL DRBG60");
        plen = drbg_get_adin(dctx, &p);
        memcpy(pers + 16, p, plen);

        if (drbg_instantiate(dctx, pers, sizeof(pers)) <= 0) {
            RANDerr(RAND_F_RAND_GET_DRBG_METHOD,
                    RAND_R_ERROR_INSTANTIATING_DRBG);
            fprintf(stderr, "drbg_instantiate failed\n");
            return 0;
        }
        //rand_set_method(FIPS_drbg_method());
        drbg_initialized = 1;
    }
    return &rand_drbg_meth;
}
