/* -----------------------------------------------------------------
 * crypto/rand/rand_oct.c - Cavium Octeon specific hardware assist
 * functionality for random number generation.
 *
 * February 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#ifdef OCTEON_OPENSSL
# include <stdint.h>
# ifndef _MIPS_ARCH_OCTEON3
#  include "ciscossl-cvmx.h"
# else
#  include "cvmx.h"
# endif


# ifdef NO_CRNG_TEST
#  undef CRNG_TEST
# endif

static CVMX_SHARED uint8_t oct_rng_enabled = 0;
# ifdef CRNG_TEST
static CVMX_SHARED uint8_t oct_fipsrand_start = 0;
static CVMX_SHARED uint8_t oct_prev_crng_rand[16];
static CVMX_SHARED uint64_t oct_prev_hwrng_rand = 0;
# endif
# ifdef CRNG_ERROR_FREQ
static CVMX_SHARED int oct_rng_calls = 0;
static CVMX_SHARED int oct_is_hwrng_fail = 1; /* 1 for hw rng, 0 for rand */
# endif

# ifdef _MIPS_ARCH_OCTEON3
#  define CVMX_ADD_SEG(segment, add)          ((((uint64_t)segment) << 62) | (add))
#  define CVMX_ADD_IO_SEG(add)                CVMX_ADD_SEG(CVMX_IO_SEG, (add))
#  define CVMX_OCT_DID_RNG 8ULL
#  define CVMX_RNG_LOAD_ADDRESS   CVMX_ADD_IO_SEG(cvmx_build_io_address(CVMX_OCT_DID_RNG, 0))
# endif
/**
 * Enables the random number generator. Must be called before RNG is used
 */
static inline void cvmx_rng_enable(void)
{
    cvmx_rnm_ctl_status_t rnm_ctl_status;
    rnm_ctl_status.u64 = cvmx_read_csr(CVMX_RNM_CTL_STATUS);
    rnm_ctl_status.s.ent_en = 1;
    rnm_ctl_status.s.rng_en = 1;
    cvmx_write_csr(CVMX_RNM_CTL_STATUS, rnm_ctl_status.u64);
}

static inline uint64_t cvmx_rng_get_random64(void)
{
    return cvmx_read64_uint64(CVMX_RNG_LOAD_ADDRESS);
}


/**
 * Generates Random number of given length
 * @param len      Size of random number required in Bytes
 * @param output   Output where random number is stored
 * @return       length of random bytes generated on Success
*/
int oct_rand_generate(uint8_t * output, uint64_t len)
{
    uint64_t out;

    /* Check if this core RNG is enabled else call rand_init */
    if (!oct_rng_enabled) {
        cvmx_rng_enable();
# ifdef CRNG_TEST
        oct_prev_hwrng_rand = cvmx_rng_get_random64();
# endif
        oct_rng_enabled = 1;
    }

    while (len >= 8) {
        out = cvmx_rng_get_random64();

# ifdef CRNG_TEST
        if (out == oct_prev_hwrng_rand) {
            return OCT_RAND_CRNG_FAILURE;
        }
        oct_prev_hwrng_rand = out;

#  ifdef CRNG_ERROR_FREQ
        if (oct_is_hwrng_fail && (oct_rng_calls == CRNG_ERROR_FREQ)) {
            oct_rng_calls = 0;
            oct_is_hwrng_fail = 0;
            return OCT_RAND_CRNG_FAILURE;
        }
#  endif
# endif
        *(uint64_t *) output = out;
        len -= 8;
        output += 8;
    }
    /* check remainig Non 8B boundary */
    if (len) {
        out = cvmx_rng_get_random64();
# ifdef CRNG_TEST
        if (out == oct_prev_hwrng_rand) {
            return OCT_RAND_CRNG_FAILURE;
        }
        oct_prev_hwrng_rand = out;
#  ifdef CRNG_ERROR_FREQ
        if (oct_is_hwrng_fail && (oct_rng_calls == CRNG_ERROR_FREQ)) {
            oct_rng_calls = 0;
            oct_is_hwrng_fail = 0;
            return OCT_RAND_CRNG_FAILURE;
        }
#  endif
# endif
# ifdef B_ENDIAN
        *(uint64_t *) output =
            *(uint64_t *) output | (out << (64 - (len << 3)));
# else
    	*(uint64_t*) output = 
            *(uint64_t *) output | (out >> (64 - (len << 3)));

# endif
    }
    return 0;
}

#endif
