/* crypto/rsa/rsa_gen.c */
/* Copyright (C) 1995-1998 Eric Young (eay@cryptsoft.com)
 * All rights reserved.
 *
 * This package is an SSL implementation written
 * by Eric Young (eay@cryptsoft.com).
 * The implementation was written so as to conform with Netscapes SSL.
 *
 * This library is free for commercial and non-commercial use as long as
 * the following conditions are aheared to.  The following conditions
 * apply to all code found in this distribution, be it the RC4, RSA,
 * lhash, DES, etc., code; not just the SSL code.  The SSL documentation
 * included with this distribution is covered by the same copyright terms
 * except that the holder is Tim Hudson (tjh@cryptsoft.com).
 *
 * Copyright remains Eric Young's, and as such any Copyright notices in
 * the code are not to be removed.
 * If this package is used in a product, Eric Young should be given attribution
 * as the author of the parts of the library used.
 * This can be in the form of a textual message at program startup or
 * in documentation (online or textual) provided with the package.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    "This product includes cryptographic software written by
 *     Eric Young (eay@cryptsoft.com)"
 *    The word 'cryptographic' can be left out if the rouines from the library
 *    being used are not cryptographic related :-).
 * 4. If you include any Windows specific code (or a derivative thereof) from
 *    the apps directory (application code) you must include an acknowledgement:
 *    "This product includes software written by Tim Hudson (tjh@cryptsoft.com)"
 *
 * THIS SOFTWARE IS PROVIDED BY ERIC YOUNG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 */

/*
 * NB: these functions have been "upgraded", the deprecated versions (which
 * are compatibility wrappers using these functions) are in rsa_depr.c. -
 * Geoff
 */

#include <stdio.h>
#include <time.h>
#include "cryptlib.h"
#include <openssl/bn.h>
#include <openssl/rsa.h>
#include <openssl/rand.h>
#ifdef OPENSSL_FIPS
# include <openssl/fips.h>
#endif

#define PUBLIC_EXPONENT 65537   /* When generating an RSA key, always use a */
                              /* public exponent of 65537 */

/*
 * They require minimum size factors for p-1, p+1 -- this table gives
 * the size of those "auxilary primes" as a function of the key size (table B.1
 * from 186-3)
 *
 * *N1 will be set to the size of the p1/q1 aux prime (the factor of p-1)
 * *N2 will be set to the size of the p2/q2 aux prime (the factor of p+1)
 */
static void translate_keylen_into_subprime_sizes(unsigned int key_size,
                                                 unsigned int *N1,
                                                 unsigned int *N2)
{
    unsigned int n;

    if (key_size < 1024) {
        /*
         * Not in 186-3 -- a default for keys too small for NIST to recognize
         * Since this is not FIPS approved, we are allowed to notice that
         * there's no particular need for p1 (because our prime generation
         * routine will ensure a large prime factor for (p-1) anyways).  In
         * addition, p1 takes up room that's actually at a premium for primes
         * this small; hence don't bother
         */
        *N1 = 1;                /* This disables p1 */
        *N2 = key_size / 8;
        return;
    } else if (key_size < 2048)
        n = 104;                /* Key sizes 1024-2047 */
    else if (key_size < 3072)
        n = 144;                /* Key sizes 2048-3071 */
    else
        n = 176;                /* Key sizes 3072-4096 */

    *N1 = n;

    /*
     * Note: it doesn't work too well if the same auxiliary prime is picked for
     * both p1 and p2.  While this is quite unlikely, we can avoid that
     * possibility entirely by making the p2/q2 aux primes one bit larger; that
     * way, it can't be the same as p1/q1 (and, yes, 186-3 gives us enough
     * leeway that we can actually do this; it doesn't specify the exact size,
     * but a range of allowable sizes)
     */
    *N2 = n + 1;
}

/*
 * This converts an RSA keylength into a security strength, precisely as
 * endorsed by NIST (800-57).
 */
static unsigned int translate_keylen_into_security_strength(unsigned int
                                                            key_size)
{
    if (key_size < 1024)
        return 64;              /* Not in 800-57 -- a default for keys */
    /* too small for NIST to recognize */
    if (key_size < 2048)
        return 80;              /* Key sizes 1024-2047 */
    if (key_size < 3072)
        return 112;             /* Key sizes 2048-3071 */
    return 128;                 /* Key sizes 3072-4096 (largest we support) */
}

/*
 * This does the actual work in generating the primes, and computing the RSA
 * private exponent
 * This uses algorithm B.3.4 from FIPS 186-3.
 *
 * It's separated out here and made public to make a convienent place to do the
 * FIPS algorithm testing.  Parameters:
 * p, q, n, d -- Where to place the generated RSA key.  Note that this does
 *    not compute the CRT parameters, if the caller needs them, it'll do that
 *    itself.
 * seed, seedlen -- The seed used to generate the primes
 * public_exponent - The public RSA exponent.  This is assumed to be an odd prime.
 * nlen - The RSA keysize that we are generating (in bits).
 */
int rsa_generate_key_internal(BIGNUM **p, BIGNUM **q, BIGNUM **n, BIGNUM **d,
                              void *seed, unsigned int seed_len,
                              unsigned int bitlen1, unsigned int bitlen2,
                              unsigned int bitlen3, unsigned int bitlen4,
                              BIGNUM *e_value, unsigned int nlen,
                              BN_GENCB *cb)
{
    unsigned int security_strength;
    void *working_seed;
    prime_sieve *sieve = NULL;
    BIGNUM *p_minus_1 = NULL, *q_minus_1 = NULL;
    BIGNUM *diff = NULL, *limit = NULL, *half_keysize = NULL;
    BIGNUM *gcd = NULL, *p_1_times_q_1 = NULL, *lcm = NULL;
    unsigned int status = 0;
    BN_CTX *ctx = NULL;
    int ok;

    /*
     * Step 1: check the target keysize (nlen).
     * Limited checking is done here since this
     * could have FIPS mode off.
     * However we do need to make sure nlen is at least 512
     * for the logic to work.
     */
    if (nlen < 512) {
        return status;
    }

    /*
     * Step 2: check the public exponent.  Note that presently we only claim
     * support for 65537, so simply check for that value.
     */

    if (!BN_abs_is_word(e_value, PUBLIC_EXPONENT)) {
        return status;
    }

    /* Step 3: look up the security strength, based on key size */
    security_strength = translate_keylen_into_security_strength(nlen);

    /*
     * Step 4: make sure that the seed len is exactly the length we expect.
     */
    if (8 * seed_len != 2 * security_strength) {
        return status;
    }

    /*
     * Step 5: copy the seed into the working value.
     */
    working_seed = seed;

    /*
     * Allocate sieve as described in FIPS PUB 186-3 C.8
     */
    sieve = BN_allocate_prime_sieve((nlen + 1) / 2);
    if (!sieve) {
        return status;
    }

 retry:
    /*
     * Step 6: generate p
     */
    *p = BN_generate_1863_prime(nlen / 2, bitlen1, bitlen2,
                                working_seed, seed_len, e_value, sieve, cb);

    if (!*p) {
        BN_free_prime_sieve(sieve);
        return status;
    }

    ctx = BN_CTX_new();
    if (ctx == NULL) {
        goto error;
    }
    BN_CTX_start(ctx);

    for (;;) {
        /* Step 7: generate q */
        /* Note: if nlen is odd, we generate a prime 1 bit bigger */
        *q = BN_generate_1863_prime((nlen + 1) / 2, bitlen3, bitlen4,
                                    working_seed, seed_len, e_value, sieve,
                                    cb);
        if (!*q) {
            BN_free(*p);
            *p = NULL;
            BN_free_prime_sieve(sieve);
            BN_CTX_end(ctx);
            BN_CTX_free(ctx);
            return status;
        }

        /* Step 8: if |p-q| < 2**(nlen/2 - 100), generate another q */
        diff = BN_new();
        if (diff == NULL) {
            BN_free(*p);
            *p = NULL;
            BN_free_prime_sieve(sieve);
            BN_CTX_end(ctx);
            BN_CTX_free(ctx);
            return status;
        }
        if (BN_cmp(*p, *q) > 0) {
            BN_sub(diff, *p, *q);
        } else {
            BN_sub(diff, *q, *p);
        }
        limit = BN_2PowerN(nlen / 2 - 100);
        if (!diff || !limit) {
            BN_free(diff);
            BN_free(limit);
            BN_free(*p);
            *p = NULL;
            BN_free(*q);
            *q = NULL;
            BN_free_prime_sieve(sieve);
            BN_CTX_end(ctx);
            BN_CTX_free(ctx);
            return status;
        }
        ok = BN_cmp(diff, limit);
        BN_free(diff);
        BN_free(limit);
        if (ok > 0) {
            break;
        }
        BN_free(*q);
        *q = NULL;              /* Nope: p and q are too close: try again */
    }

    /* Step 9: zeroize the internal seeds */
    OPENSSL_cleanse(working_seed, seed_len);

    /*
     * Step 10: We're formally done with the prime generate steps: do the rest of
     * the RSA key generation logic
     */

    /* Compute the public modulus */
    *n = BN_new();
    if (*n == NULL) {
        goto error;
    }
    BN_mul(*n, *p, *q, ctx);

    /* Compute the private exponent - d such that d*e = 1 mod lcm(p-1, q-1) */
    p_minus_1 = BN_dup(*p);
    if (p_minus_1 == NULL) {
        goto error;
    }
    BN_sub_word(p_minus_1, 1);

    q_minus_1 = BN_dup(*q);
    if (q_minus_1 == NULL) {
        goto error;
    }

    BN_sub_word(q_minus_1, 1);

    gcd = BN_new();
    p_1_times_q_1 = BN_new();
    lcm = BN_new();

    if ((gcd == NULL) || (p_1_times_q_1 == NULL) || (lcm == NULL)) {
        goto error;
    }
    BN_gcd(gcd, p_minus_1, q_minus_1, ctx);
    BN_mul(p_1_times_q_1, p_minus_1, q_minus_1, ctx);

    BN_div(lcm, NULL, p_1_times_q_1, gcd, ctx);

    BN_free(p_1_times_q_1);
    BN_free(gcd);

    *d = BN_mod_inverse(NULL, e_value, lcm, ctx);
    BN_free(lcm);
    if (*d == NULL) {
        goto error;
    }
    /*
     * Note: 186-3 (section B.3.1) insists on a check that d is larger than
     * 2**(keysize/2).
     */
    half_keysize = BN_2PowerN(nlen / 2);
    if (half_keysize == NULL) {
        goto error;
    }
    if (BN_cmp(*d, half_keysize) <= 0) {
        /*
         * 186-3 insists we try again
         */
        BN_free(half_keysize);
        BN_free(*d);
        *d = NULL;
        BN_free(p_minus_1);
        p_minus_1 = NULL;
        BN_free(q_minus_1);
        q_minus_1 = NULL;
        BN_free(*n);
        *n = NULL;
        BN_free(*q);
        *q = NULL;
        BN_free(*p);
        *p = NULL;
        goto retry;
    }
    BN_free(half_keysize);
    status = 1;

 error:
    BN_CTX_end(ctx);
    BN_CTX_free(ctx);
    BN_free_prime_sieve(sieve);
    BN_free(p_minus_1);
    BN_free(q_minus_1);
    return status;
}

static int rsa_builtin_keygen(RSA *rsa, int bits, BIGNUM *e_value,
                              BN_GENCB *cb);

/*
 * NB: this wrapper would normally be placed in rsa_lib.c and the static
 * implementation would probably be in rsa_eay.c. Nonetheless, is kept here
 * so that we don't introduce a new linker dependency. Eg. any application
 * that wasn't previously linking object code related to key-generation won't
 * have to now just because key-generation is part of RSA_METHOD.
 */
int RSA_generate_key_ex(RSA *rsa, int bits, BIGNUM *e_value, BN_GENCB *cb)
{
#ifdef OPENSSL_FIPS
    if (FIPS_mode() && !(rsa->meth->flags & RSA_FLAG_FIPS_METHOD)
        && !(rsa->flags & RSA_FLAG_NON_FIPS_ALLOW)) {
        RSAerr(RSA_F_RSA_GENERATE_KEY_EX, RSA_R_NON_FIPS_RSA_METHOD);
        return 0;
    }
#endif
    if (rsa->meth->rsa_keygen)
        return rsa->meth->rsa_keygen(rsa, bits, e_value, cb);
#ifdef OPENSSL_FIPS
    if (FIPS_mode())
	return FIPS_rsa_generate_key_ex(rsa, bits, e_value, cb);
#endif
    return rsa_builtin_keygen(rsa, bits, e_value, cb);
}

static int rsa_builtin_keygen(RSA *rsa, int bits, BIGNUM *e_value,
                              BN_GENCB *cb)
{
    BIGNUM *r0 = NULL, *r1 = NULL, *r2 = NULL;
    BIGNUM local_r0, local_d, local_p;
    BIGNUM *pr0, *d, *p;
    int ok = -1;
    BN_CTX *ctx = NULL;
    unsigned int seedlen, n1, n2, n3, n4;
    unsigned char *seed;

    /*
     * Generate the random seed that will be used to generate the RSA key.
     */
    seedlen = (2 * translate_keylen_into_security_strength(bits) + 7) / 8;
    seed = OPENSSL_malloc(seedlen);
    if (!seed) {
        goto err;
    }
    if (RAND_bytes(seed, seedlen) <= 0) {
        goto err;
    }

    /* Look up the subprime bitlengths, based on bits */
    translate_keylen_into_subprime_sizes(bits, &n1, &n2);
    n3 = n1;
    n4 = n2;

    ctx = BN_CTX_new();
    if (ctx == NULL)
        goto err;
    BN_CTX_start(ctx);

    if (!rsa_generate_key_internal(&rsa->p, &rsa->q, &rsa->n, &rsa->d,
                                   seed, seedlen,
                                   n1, n2, n3, n4, e_value, bits, cb)) {
        goto err;
    }

    r0 = BN_CTX_get(ctx);
    r1 = BN_CTX_get(ctx);
    r2 = BN_CTX_get(ctx);
    if (r2 == NULL)
        goto err;

    /* We need the RSA components non-NULL */
    if (!rsa->e && ((rsa->e = BN_new()) == NULL))
        goto err;
    if (!rsa->dmp1 && ((rsa->dmp1 = BN_new()) == NULL))
        goto err;
    if (!rsa->dmq1 && ((rsa->dmq1 = BN_new()) == NULL))
        goto err;
    if (!rsa->iqmp && ((rsa->iqmp = BN_new()) == NULL))
        goto err;

    if (BN_copy(rsa->e, e_value) == NULL)
        goto err;

    /* calculate CRT values for efficency later */
    if (!BN_sub(r1, rsa->p, BN_value_one()))
        goto err;               /* p-1 */
    if (!BN_sub(r2, rsa->q, BN_value_one()))
        goto err;               /* q-1 */
    if (!BN_mul(r0, r1, r2, ctx))
        goto err;               /* (p-1)(q-1) */
    if (!(rsa->flags & RSA_FLAG_NO_CONSTTIME)) {
        pr0 = &local_r0;
        BN_with_flags(pr0, r0, BN_FLG_CONSTTIME);
    } else
        pr0 = r0;

    /* set up d for correct BN_FLG_CONSTTIME flag */
    if (!(rsa->flags & RSA_FLAG_NO_CONSTTIME)) {
        d = &local_d;
        BN_with_flags(d, rsa->d, BN_FLG_CONSTTIME);
    } else
        d = rsa->d;

    /* calculate d mod (p-1) */
    if (!BN_mod(rsa->dmp1, d, r1, ctx))
        goto err;

    /* calculate d mod (q-1) */
    if (!BN_mod(rsa->dmq1, d, r2, ctx))
        goto err;

    /* calculate inverse of q mod p */
    if (!(rsa->flags & RSA_FLAG_NO_CONSTTIME)) {
        p = &local_p;
        BN_with_flags(p, rsa->p, BN_FLG_CONSTTIME);
    } else
        p = rsa->p;
    if (!BN_mod_inverse(rsa->iqmp, rsa->q, p, ctx))
        goto err;

    ok = 1;
 err:
    if (seed != NULL) {
        OPENSSL_free(seed);
    }
    if (ok == -1) {
        RSAerr(RSA_F_RSA_BUILTIN_KEYGEN, ERR_LIB_BN);
        ok = 0;
    }
    if (ctx != NULL) {
        BN_CTX_end(ctx);
        BN_CTX_free(ctx);
    }

    return ok;
}
