/* -----------------------------------------------------------------
 * crypto/sha/asm/sha1-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

/* Cavium Standalone Exec doesn't use sockets */
#ifndef OPENSSL_SYS_SE
# define USE_SOCKETS
#endif

#if !defined(OPENSSL_NO_SHA1) && !defined(OPENSSL_NO_SHA)

# undef  SHA_0
# define SHA_1
# include <openssl/opensslv.h>

/* The implementation is in ../md32_common.h */
# include <openssl/crypto.h>
#  ifndef _MIPS_ARCH_OCTEON3
#   include "ciscossl-cvmx.h"
#  else
#   include "cvmx.h"
#  endif


# include "sha.h"
#endif


/*
 * SHA1_Init - initializes a SHA_CTX structure.
 */
int SHA1OCT_Init(SHA_CTX *c)
{
    memset(c, 0, sizeof(*c));
    c->E = 0x67452301EFCDAB89ull;
    c->F = 0x98BADCFE10325476ull;
    c->G = 0xC3D2E1F000000000ull;
    return 1;
}

/*
 * SHA1_Update - can be called repeatedly with chunks of the
 * message to be hashed (len bytes at data).
 */
int SHA1OCT_Update(SHA_CTX *c, const void *data, size_t n)
{
    unsigned long remaining = 0, totlen = 0, copied = 0;
    const uint64_t *ptr = (const uint64_t *)data;

    if (!n) {
        return 1;
    }
    totlen = n;
    if (c->num) {
        totlen += remaining = c->num % 64;
        if (remaining) {
            memcpy((void *)&(c->data) + remaining, data,
                   SHA_CBLOCK - remaining);
            copied = 1;
        }
    }
    if (totlen >= 64) {
        /* Get the running IV loaded */
        CVMX_MT_HSH_IV(c->E, 0);
        CVMX_MT_HSH_IV(c->F, 1);
        CVMX_MT_HSH_IV(c->G, 2);

        /* Iterate through 64 bytes at a time */
        while (totlen >= 64) {
            if (copied)
                copied = 0;
            if (remaining)
                ptr = (uint64_t *) & c->data;

	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 0);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 1);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 2);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 3);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 4);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 5);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 6);
	    CVMX_MT_HSH_STARTSHA (cvmx_htobe64(*ptr++));
            totlen -= 64;
            /* remainig reset set ptr to input data */
            if (remaining) {
                ptr = data + (SHA_CBLOCK - remaining);
                remaining = 0;
            }
        }
        /* Update the IV */
        CVMX_MF_HSH_IV(c->E, 0);
        CVMX_MF_HSH_IV(c->F, 1);
        CVMX_MF_HSH_IV(c->G, 2);
    }                           /* end of if(totlen) */
    c->num += n;
    if (!copied) {
        memcpy(&(c->data), ptr, totlen);
    }
    return 1;
}

/*
 * SHA1_Final - places the message digest in md, which must have space for
 * SHA_DIGEST_LENGTH == 20 bytes of output.
 */
int SHA1OCT_Final(unsigned char *md, SHA_CTX *c)
{
    unsigned long len;
    uint8_t chunk[64];
    const uint64_t *ptr;
    uint64_t bits;

    len = c->num % 64;

    bits = ((uint64_t) c->num * 8); /* SHA1 expects big endian */
    /* The rest of the data will need to be copied into a chunk */
    memset(chunk, 0, 64);
    if (len > 0) {
        memcpy(chunk, c->data, len);
    }
    chunk[len] = 0x80;
    /* Get the running IV loaded */
    CVMX_MT_HSH_IV(c->E, 0);
    CVMX_MT_HSH_IV(c->F, 1);
    CVMX_MT_HSH_IV(c->G, 2);

    ptr = (const uint64_t *)chunk;
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 0);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 1);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 2);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 3);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 4);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 5);
    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 6);

    /* Check to see if there is room for the bit count */
    if (len < 56) {
        CVMX_MT_HSH_STARTSHA(bits);
    } else {
        CVMX_MT_HSH_STARTSHA (cvmx_htobe64(*ptr));
        /* Another block was needed */
        CVMX_MT_HSH_DATZ(0);
        CVMX_MT_HSH_DATZ(1);
        CVMX_MT_HSH_DATZ(2);
        CVMX_MT_HSH_DATZ(3);
        CVMX_MT_HSH_DATZ(4);
        CVMX_MT_HSH_DATZ(5);
        CVMX_MT_HSH_DATZ(6);
        CVMX_MT_HSH_STARTSHA(bits);
    }
    /* Update the IV */
    CVMX_MF_HSH_IV(c->E, 0);
    CVMX_MF_HSH_IV(c->F, 1);
    CVMX_MF_HSH_IV(c->G, 2);

#ifdef L_ENDIAN
    c->E  =  cvmx_betoh64(c->E);
    c->F  =  cvmx_betoh64(c->F);
    c->G  =  cvmx_betoh64(c->G);
#endif

    memcpy(md, (void *)&c->E, 8);
    memcpy(md + 8, (void *)&c->F, 8);
    memcpy(md + 16, (void *)&c->G, 4);
    return 1;
}
