/* -----------------------------------------------------------------
 * crypto/sha/asm/sha512-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <assert.h>
#include <string.h>
#include <openssl/crypto.h>
#ifndef _MIPS_ARCH_OCTEON3
# include "ciscossl-cvmx.h"
#else
# include "cvmx.h"
#endif



#include <openssl/sha.h>

/*
 * SHA384OCT - Perform a single SHA384 digest
 */
unsigned char *SHA384OCT(const unsigned char *d, size_t n, unsigned char *md)
{
    sha_assert(d && md, NULL);
    SHA512_CTX c;
    SHA384OCT_Init(&c);
    SHA384OCT_Update(&c, d, n);
    SHA384OCT_Final(md, &c);
    return md;
}

/*
 * SHA512OCT - Perform a single SHA512 digest
 */
unsigned char *SHA512OCT(const unsigned char *d, size_t n, unsigned char *md)
{
    sha_assert(d && md, NULL);
    SHA512_CTX c;
    SHA512OCT_Init(&c);
    SHA512OCT_Update(&c, d, n);
    SHA512OCT_Final(md, &c);
    return md;
}

/*
 * SHA384OCT_Init - initializes a SHA512_CTX structure.
 */
int SHA384OCT_Init(SHA512_CTX *ctx)
{
    sha_assert(ctx, -1);
    ctx->totalin = 0;
    ctx->done = 0xbabe;
    ctx->iv[0] = 0xcbbb9d5dc1059ed8ull;
    ctx->iv[1] = 0x629a292a367cd507ull;
    ctx->iv[2] = 0x9159015a3070dd17ull;
    ctx->iv[3] = 0x152fecd8f70e5939ull;
    ctx->iv[4] = 0x67332667ffc00b31ull;
    ctx->iv[5] = 0x8eb44a8768581511ull;
    ctx->iv[6] = 0xdb0c2e0d64f98fa7ull;
    ctx->iv[7] = 0x47b5481dbefa4fa4ull;
    ctx->plen = 0;
    ctx->sha512 = 0;
    return 1;
}

/*
 * SHA512OCT_Init - initializes a SHA512_CTX structure.
 */
int SHA512OCT_Init(SHA512_CTX *ctx)
{
    sha_assert(ctx, -1);
    ctx->totalin = 0;
    ctx->done = 0xbabe;
    ctx->iv[0] = 0x6a09e667f3bcc908ull;
    ctx->iv[1] = 0xbb67ae8584caa73bull;
    ctx->iv[2] = 0x3c6ef372fe94f82bull;
    ctx->iv[3] = 0xa54ff53a5f1d36f1ull;
    ctx->iv[4] = 0x510e527fade682d1ull;
    ctx->iv[5] = 0x9b05688c2b3e6c1full;
    ctx->iv[6] = 0x1f83d9abfb41bd6bull;
    ctx->iv[7] = 0x5be0cd19137e2179ull;
    ctx->plen = 0;
    ctx->sha512 = 1;
    return 1;
}

/*
 * SHA384OCT_Update - just calls SHA512_Update
 */
int SHA384OCT_Update(SHA512_CTX *ctx, const void *data, size_t len)
{
    return SHA512OCT_Update(ctx, data, len);
}

/*
 * SHA512OCT_Update - can be called repeatedly with chunks of the
 * message to be hashed (len bytes at data).
 */
int SHA512OCT_Update(SHA512_CTX *ctx, const void *data, size_t len)
{
    sha_assert(ctx, -1);
    if (!data) {
        return -1;
    }
    if (!len) {
        return 1;
    }
    if (ctx->done == 0xdead) {
        return -1;
    }
    sha_assert(ctx->done == 0xbabe, -1);

    // Increment totalin
    ctx->totalin += len;

    if ((ctx->plen + len) >= 128) {
        // restore IV
        CVMX_MT_HSH_IVW(ctx->iv[0], 0);
        CVMX_MT_HSH_IVW(ctx->iv[1], 1);
        CVMX_MT_HSH_IVW(ctx->iv[2], 2);
        CVMX_MT_HSH_IVW(ctx->iv[3], 3);
        CVMX_MT_HSH_IVW(ctx->iv[4], 4);
        CVMX_MT_HSH_IVW(ctx->iv[5], 5);
        CVMX_MT_HSH_IVW(ctx->iv[6], 6);
        CVMX_MT_HSH_IVW(ctx->iv[7], 7);

        if (ctx->plen) {
            memcpy(&ctx->pbuf[ctx->plen], data, 128 - ctx->plen);
            len -= (128 - ctx->plen);
            data = (uint8_t *) data + (128 - ctx->plen);
            ctx->plen = 0;
            uint64_t *ptr = (uint64_t *) ctx->pbuf;
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 0);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 1);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 2);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 3);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 4);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 5);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 6);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 7);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 8);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 9);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 10);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 11);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 12);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 13);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 14);
	    CVMX_MT_HSH_STARTSHA512 (cvmx_htobe64(*ptr++));
        }

        uint64_t *ptr = (uint64_t *) data;
        while (len >= 128) {
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 0);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 1);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 2);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 3);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 4);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 5);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 6);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 7);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 8);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 9);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 10);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 11);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 12);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 13);
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 14);
	    CVMX_MT_HSH_STARTSHA512 (cvmx_htobe64(*ptr++));
            len -= 128;
        }

        if (len) {
            memcpy(ctx->pbuf, ptr, len);
            ctx->plen = len;
        }
        // save IV
        CVMX_MF_HSH_IVW(ctx->iv[0], 0);
        CVMX_MF_HSH_IVW(ctx->iv[1], 1);
        CVMX_MF_HSH_IVW(ctx->iv[2], 2);
        CVMX_MF_HSH_IVW(ctx->iv[3], 3);
        CVMX_MF_HSH_IVW(ctx->iv[4], 4);
        CVMX_MF_HSH_IVW(ctx->iv[5], 5);
        CVMX_MF_HSH_IVW(ctx->iv[6], 6);
        CVMX_MF_HSH_IVW(ctx->iv[7], 7);
    } else {
        memcpy(&ctx->pbuf[ctx->plen], data, len);
        ctx->plen += len;
    }
    return 1;
}

/*
 * SHA384OCT_Final - just call SHA512OCT_Final
 */
int SHA384OCT_Final(unsigned char *md, SHA512_CTX *ctx)
{
    return SHA512OCT_Final(md, ctx);
}

/*
 * SHA512OCT_Final - places the message digest in md
 */
int SHA512OCT_Final(unsigned char *md, SHA512_CTX *ctx)
{
    sha_assert(md && ctx, -1);
    if (ctx->done == 0xdead) {
        return -1;
    }
    sha_assert(ctx->done == 0xbabe, -1);
    if (ctx->done == 0xbabe) {
        ctx->done = 0xdead;

        // Restore IV
        CVMX_MT_HSH_IVW(ctx->iv[0], 0);
        CVMX_MT_HSH_IVW(ctx->iv[1], 1);
        CVMX_MT_HSH_IVW(ctx->iv[2], 2);
        CVMX_MT_HSH_IVW(ctx->iv[3], 3);
        CVMX_MT_HSH_IVW(ctx->iv[4], 4);
        CVMX_MT_HSH_IVW(ctx->iv[5], 5);
        CVMX_MT_HSH_IVW(ctx->iv[6], 6);
        CVMX_MT_HSH_IVW(ctx->iv[7], 7);

        ctx->pbuf[ctx->plen] = 0x80;
        memset(ctx->pbuf + ctx->plen + 1, 0, 128 - ctx->plen - 1);

        uint64_t *ptr = (uint64_t *) ctx->pbuf;
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 0);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 1);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 2);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 3);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 4);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 5);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 6);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 7);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 8);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 9);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 10);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 11);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 12);
	CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 13);

        if (ctx->plen < 112) {
            uint64_t ab[2];
            uint64_t_mul(ab[0], ab[1], ctx->totalin, 8);
            CVMX_MT_HSH_DATW(ab[0], 14);
            CVMX_MT_HSH_STARTSHA512(ab[1]);
        } else {
	    CVMX_MT_HSH_DATW (cvmx_htobe64(*ptr++), 14);
	    CVMX_MT_HSH_STARTSHA512 (cvmx_htobe64(*ptr));
            /* Another block was needed */
            CVMX_MT_HSH_DATWZ(0);
            CVMX_MT_HSH_DATWZ(1);
            CVMX_MT_HSH_DATWZ(2);
            CVMX_MT_HSH_DATWZ(3);
            CVMX_MT_HSH_DATWZ(4);
            CVMX_MT_HSH_DATWZ(5);
            CVMX_MT_HSH_DATWZ(6);
            CVMX_MT_HSH_DATWZ(7);
            CVMX_MT_HSH_DATWZ(8);
            CVMX_MT_HSH_DATWZ(9);
            CVMX_MT_HSH_DATWZ(10);
            CVMX_MT_HSH_DATWZ(11);
            CVMX_MT_HSH_DATWZ(12);
            CVMX_MT_HSH_DATWZ(13);
            uint64_t ab[2];
            uint64_t_mul(ab[0], ab[1], ctx->totalin, 0x8ull);
            CVMX_MT_HSH_DATW(ab[0], 14);
            CVMX_MT_HSH_STARTSHA512(ab[1]);
        }

        CVMX_MF_HSH_IVW(((uint64_t *) md)[0], 0);
        CVMX_MF_HSH_IVW(((uint64_t *) md)[1], 1);
        CVMX_MF_HSH_IVW(((uint64_t *) md)[2], 2);
        CVMX_MF_HSH_IVW(((uint64_t *) md)[3], 3);
        CVMX_MF_HSH_IVW(((uint64_t *) md)[4], 4);
        CVMX_MF_HSH_IVW(((uint64_t *) md)[5], 5);

#ifdef L_ENDIAN
        ((uint64_t *) md)[0]  =  cvmx_betoh64(((uint64_t *) md)[0]);
	((uint64_t *) md)[1]  =  cvmx_betoh64(((uint64_t *) md)[1]);
	((uint64_t *) md)[2]  =  cvmx_betoh64(((uint64_t *) md)[2]);
	((uint64_t *) md)[3]  =  cvmx_betoh64(((uint64_t *) md)[3]);
	((uint64_t *) md)[4]  =  cvmx_betoh64(((uint64_t *) md)[4]);
	((uint64_t *) md)[5]  =  cvmx_betoh64(((uint64_t *) md)[5]);
#endif
        if (ctx->sha512) {
            CVMX_MF_HSH_IVW(((uint64_t *) md)[6], 6);
            CVMX_MF_HSH_IVW(((uint64_t *) md)[7], 7);
#ifdef L_ENDIAN
            ((uint64_t *) md)[6]  =  cvmx_betoh64(((uint64_t *) md)[6]);
	    ((uint64_t *) md)[7]  =  cvmx_betoh64(((uint64_t *) md)[7]);
#endif
        }
    }
    return 1;
}
