/*
 *  ctame.c
 *  Cisco TAM Engine for use with Cisco Trusted Anchor Module Client
 *  OpenSSL engine
 *
 *  Jan 2014
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 *  cloned from cact2.c
 */
#ifdef CTAME_SUPPORT
# include <stdlib.h>
# include <string.h>

# include <openssl/bn.h>
# include <openssl/dso.h>
# include <openssl/objects.h>
# include <openssl/rsa.h>
# include "tam_types.h"
# include "e_ctamc.h"
# include "ctame.h"
# include "e_ctame_err.h"

# ifdef USE_CISCO_SSM
#  include "ssm.h"
# else
#  include "tamc_api.h"
# endif

/**
 * ctame_read_object()
 *
 * Read an object from the CTAME.
 */
CTAMC_STATUS ctame_read_object(const char *ctam_context,
                               const char *object_name,
                               unsigned char **data, unsigned short *length)
{
    tam_status_e status = TAM_ERROR_UNSPECIFIED;
    uint32_t length_read = 0;
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;

    if (!context || !object_name || !data || !length) {
        CTAMEerr(CTAME_F_CTAME_READ_OBJECT, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    status = tamc_so_read(context->session_id,
                          context->app_id,
                          (char *)object_name, data, &length_read);
    if (status != TAM_SUCCESS) {
        CTAMEerr(CTAME_F_CTAME_READ_OBJECT, CTAME_R_OBJ_READ_FAILED);
    }
    *length = length_read;
    return ((status == TAM_SUCCESS) ? CTAMC_SUCCESS : CTAMC_FAILURE);
}

/**
 * ctame_check_object_type()
 *
 * If obj_type matches the value returned from get attribute return CTAMC_SUCCESS
 */
CTAMC_STATUS ctame_check_object_type(const char *ctam_context,
                                     const char *object_name,
                                     unsigned int obj_type)
{
    CTAMC_STATUS rc = CTAMC_FAILURE;
    tam_status_e status = TAM_ERROR_UNSPECIFIED;
    tam_object_attr_t *obj_attr = NULL;
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;

    // Get obj attr
    status = tamc_so_get_attr(context->session_id,
                              context->app_id,
                              (char *)object_name, &obj_attr);
    if ((status != TAM_SUCCESS) || !obj_attr) {
        goto end;
    }
    // Check if obj attribute is a match
    if (obj_attr->object_type == obj_type) {
        rc = CTAMC_SUCCESS;
    }

 end:
    free(obj_attr);
    return (rc);
}

/**
 * ctame_get_object_size()
 *
 * Returns obj size
 * 0 if failed
 */
unsigned int ctame_get_object_size(const char *ctam_context,
                                   const char *object_name)
{
    tam_status_e status = TAM_ERROR_UNSPECIFIED;
    tam_object_attr_t *obj_attr = NULL;
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    unsigned int obj_size = 0;

    // Get obj attr
    status = tamc_so_get_attr(context->session_id,
                              context->app_id,
                              (char *)object_name, &obj_attr);
    if ((status != TAM_SUCCESS) || (!obj_attr)) {
        goto end;
    }

    obj_size = obj_attr->object_size;

 end:
    free(obj_attr);
    return (obj_size);
}

# ifndef OPENSSL_NO_STATIC_ENGINE
int ctame_init(void)
{
    p_ctamc_rsa_private_key_encrypt = ctame_rsa_private_key_encrypt;
    p_ctamc_rsa_public_key_encrypt = ctame_rsa_public_key_encrypt;
    p_ctamc_rsa_private_key_decrypt = ctame_rsa_private_key_decrypt;
    p_ctamc_rsa_public_key_decrypt = ctame_rsa_public_key_decrypt;
    p_ctamc_rsa_gen_keypair = ctame_rsa_gen_keypair;
    p_ctamc_read_object = ctame_read_object;
    p_ctamc_load_rsa_private_key = ctame_load_rsa_private_key;
    p_ctamc_ec_gen_keypair = ctame_ec_gen_keypair;
    p_ctamc_load_ecc_private_key = ctame_load_ecc_private_key;

    p_ctamc_ecdsa_do_sign = ctame_ecdsa_do_sign;
    p_ctamc_ecdsa_do_verify = ctame_ecdsa_do_verify;

    /* Everything's fine. */
    return (1);
}
# endif                         // OPENSSL_NO_STATIC_ENGINE

#endif                          /* CTAME_SUPPORT */
