/*
 *  ctame_ec.c
 *  Cisco TAM Engine for use with Cisco Trusted Anchor Module Client
 *  OpenSSL engine
 *
 *  Jan 2014
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */
#ifdef CTAME_SUPPORT
# include <stdlib.h>
# include <string.h>

# include "../vendor_defns/ctamc.h"
# include "ctame.h"
# include <openssl/x509.h>
# include "e_ctame_err.h"
# include "tam_types.h"
# include "safe_mem_lib.h"
# include "safe_lib_errno.h"
# include "safe_str_lib.h"

# ifdef USE_CISCO_SSM
#  include "ssm.h"
#  include "ssmc/ssmc_ecc_key.c"
#  include "ssmc/ssmc_sign_verify.c"
# else
#  include "tamc_api.h"
# endif

/*
 * convert tamd_ec_curve_e to openssl curve type
 */
uint32_t swtam_tam_curve_type_to_evp(tam_ec_curve_e curve)
{
    uint32_t open_curve = 0;

    switch (curve) {
    case TAM_EC_CURVE_P192:
        open_curve = NID_X9_62_prime192v1;
        break;
    case TAM_EC_CURVE_P256:
        open_curve = NID_X9_62_prime256v1;
        break;
    case TAM_EC_CURVE_P384:
        open_curve = NID_secp384r1;
        break;
    case TAM_EC_CURVE_P521:
        open_curve = NID_secp521r1;
        break;
    case TAM_EC_CURVE_BRAINPOOLP256:
        open_curve = NID_brainpoolP256r1;
        break;
    case TAM_EC_CURVE_BRAINPOOLP384:
        open_curve = NID_brainpoolP384r1;
        break;
    case TAM_EC_CURVE_BRAINPOOLP512:
        open_curve = NID_brainpoolP512r1;
        break;
    default:
        break;
    }
    return ((uint32_t) open_curve);
}

/*
 * ctame_ecc_gen_keypair()
 *
 * Generate an ECC keypair with the specified length and exponent
 * attributes, persistence and name. The key is subsequently
 * referenced by "name."
 */
CTAMC_STATUS ctame_ec_gen_keypair(const char *ctam_context,
                                  ctamc_ecc_key_attr_t * ec_keyparms)
{
    tam_status_e status = TAM_ERROR_UNSPECIFIED;
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    tam_object_attr_t *key_object_attrs = NULL;

    /* Validate input */
    if (!ctam_context || !ec_keyparms) {
        CTAMEerr(CTAME_F_CTAME_EC_GEN_KEYPAIR, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    /*
     * set object store parameter
     */
    key_object_attrs = calloc(1, sizeof(tam_object_attr_t));
    if (!key_object_attrs)
        goto end;
    key_object_attrs->csp_flag = ec_keyparms->zeroize;
    key_object_attrs->nv_flag = ec_keyparms->persistent;
    key_object_attrs->object_type = TAM_SO_TYPE_ECC_KEYPAIR;
    strncpy_s(key_object_attrs->object_name, TAM_OBJECT_NAME_LEN_MAX,
              &ec_keyparms->key_name[0], TAM_OBJECT_NAME_LEN_MAX);

    /*
     * set ecc keygen parameter
     */
    key_object_attrs->key.ecc.ec_curve =
        (tam_ec_curve_e) ec_keyparms->ec_curve;
    key_object_attrs->key.ecc.exportable = ec_keyparms->exportable;
    /*
     * Make key gen request
     */
    status = tamc_ecc_keygen(context->session_id, key_object_attrs);
    if (status != TAM_SUCCESS) {
        CTAMEerr(CTAME_F_CTAME_EC_GEN_KEYPAIR, CTAME_R_KEY_GEN_FAILED);
    }
 end:
    free(key_object_attrs);
    return ((status == TAM_SUCCESS) ? CTAMC_SUCCESS : CTAMC_FAILURE);
}

/*
 * Returns the correlated msg digest algm
 * based on the digest length
 */
int ctame_length_to_md(int md_len)
{
# define CTAME_UNKWN  0
# define CTAME_SHA1   1
# define CTAME_SHA256 2
# define CTAME_SHA384 3
# define CTAME_SHA512 4

# define CTAME_LEN_20 20
# define CTAME_LEN_32 32
# define CTAME_LEN_48 48
# define CTAME_LEN_64 64
    int md_algm;

    switch (md_len) {
    case CTAME_LEN_20:
        md_algm = CTAME_SHA1;
        break;
    case CTAME_LEN_32:
        md_algm = CTAME_SHA256;
        break;
    case CTAME_LEN_48:
        md_algm = CTAME_SHA384;
        break;
    case CTAME_LEN_64:
        md_algm = CTAME_SHA512;
        break;
    default:
        md_algm = CTAME_UNKWN;
        break;
    }
    return (md_algm);
}

CTAMC_STATUS ctame_ecdsa_do_sign(const char *ctamc_context,
                                 const unsigned char *dgst,
                                 int dlen, const char *key_id,
                                 ECDSA_SIG **sig)
{
    tam_status_e status = TAM_ERROR_UNSPECIFIED;
    ctamc_context_t *context = (ctamc_context_t *) ctamc_context;
    uint8_t *ecdsa_sig_buf = NULL;
    ECDSA_SIG *p_ecdsa_sig = NULL;
    uint8_t *p_sig_buf = NULL;
    uint32_t sig_len = 0;
    int md_algm = 0;

    if (!context || !dgst || !sig || !key_id) {
        CTAMEerr(CTAME_F_CTAME_ECDSA_DO_SIGN, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    /*
     * Determine message digest algorithm base on the digest length.
     * The value is ignored by ECDSA but, is needed by the client
     * API.
     */
    md_algm = ctame_length_to_md(dlen);
    if (!md_algm) {
        CTAMEerr(CTAME_F_CTAME_ECDSA_DO_SIGN, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    status = tamc_sign_digest(context->session_id,
                              context->app_id,
                              (char *)key_id,
                              (tam_hash_algo_e) md_algm,
                              (uint8_t *) dgst,
                              dlen, &ecdsa_sig_buf, &sig_len);

    if (status == TAM_SUCCESS) {
        p_sig_buf = ecdsa_sig_buf;
        d2i_ECDSA_SIG(&p_ecdsa_sig, (const unsigned char **)&p_sig_buf,
                      (long)sig_len);
    }

    if (p_ecdsa_sig) {
        *sig = p_ecdsa_sig;     // returned
        free(ecdsa_sig_buf);
        return (CTAMC_SUCCESS);
    } else {
        free(ecdsa_sig_buf);
        return (CTAMC_FAILURE);
    }
}

CTAMC_STATUS ctame_ecdsa_do_verify(const char *ctamc_context,
                                   const unsigned char *dgst,
                                   int dgst_len,
                                   const char *key_id, const ECDSA_SIG *sig)
{
    tam_status_e status = TAM_ERROR_UNSPECIFIED;
    ctamc_context_t *context = (ctamc_context_t *) ctamc_context;
    int md_algm = 0;

    uint8_t *sig_buff = NULL;
    uint8_t *p_sig_buff = NULL;
    int sig_len = 0;

    if (!context || !dgst || !sig || !key_id) {
        CTAMEerr(CTAME_F_CTAME_ECDSA_DO_VERIFY, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    /*
     * Determine message digest algorithm base on the digest length.
     * The value is ignored by ECDSA but, is needed by the client
     * API.
     */
    md_algm = ctame_length_to_md(dgst_len);
    if (!md_algm) {
        CTAMEerr(CTAME_F_CTAME_ECDSA_DO_VERIFY, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }
    // allocate buffer
    sig_len = i2d_ECDSA_SIG(sig, NULL);
    if (sig_len == 0)
        return (CTAMC_FAILURE);

    sig_buff = calloc(1, sig_len);
    if (!sig_buff)
        return (CTAMC_FAILURE);

    p_sig_buff = sig_buff;
    sig_len = i2d_ECDSA_SIG(sig, &p_sig_buff);

    // verify
    status = tamc_verify_digest(context->session_id,
                                context->app_id,
                                (char *)key_id,
                                (tam_hash_algo_e) md_algm,
                                (uint8_t *) dgst,
                                (uint32_t) dgst_len, sig_buff, sig_len);

    free(sig_buff);
    return ((status == TAM_SUCCESS) ? CTAMC_SUCCESS : CTAMC_FAILURE);
}

/*
 * ctame_load_ecc_private_key()
 *
 * Read the specified public key from CTAM
 */
CTAMC_STATUS ctame_load_ecc_private_key(const char *ctam_context,
                                        const char *key_name, EC_KEY **eckey)
{
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    tam_status_e status;
    int pub_key_len = 0;
    unsigned char *pub_key_buf = NULL;

    uint32_t obj_len_read = 0;
    uint8_t *data = NULL;
    EC_KEY *ec_key = NULL;
    tam_object_attr_t *ecc_attr = NULL;
    uint32_t curve_nid;

    if (!context || !eckey || !key_name) {
        CTAMEerr(CTAME_F_CTAME_LOAD_ECC_PRIVATE_KEY, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }
    // Get obj attr
    status = tamc_so_get_attr(context->session_id,
                              context->app_id, (char *)key_name, &ecc_attr);
    if (status != TAM_SUCCESS) {
        CTAMEerr(CTAME_F_CTAME_LOAD_ECC_PRIVATE_KEY,
                 CTAME_R_OBJECT_GET_ATTRIBUTE_FAILED);
        goto error;
    }

    if (ecc_attr->object_type != TAM_SO_TYPE_ECC_KEYPAIR) {
        CTAMEerr(CTAME_F_CTAME_LOAD_ECC_PRIVATE_KEY,
                 CTAME_R_OBJECT_WRONG_ATTRIBUTE_TYPE);
        status = TAM_ERROR_INCORRECT_KEY_TYPE;
        goto error;
    }

    if (!ecc_attr->object_size) {
        CTAMEerr(CTAME_F_CTAME_LOAD_ECC_PRIVATE_KEY,
                 CTAME_R_OBJECT_SIZE_ZERO);
        status = TAM_ERROR_INVALID_PARM;
        goto error;
    }
    // Read ECC Key
    status = tamc_so_read(context->session_id,
                          context->app_id,
                          (char *)key_name, &data, &obj_len_read);
    if (status != TAM_SUCCESS) {
        CTAMEerr(CTAME_F_CTAME_LOAD_ECC_PRIVATE_KEY,
                 CTAME_R_OBJECT_READ_FAILED);
        goto error;
    }
    // Extract public key components
    curve_nid = swtam_tam_curve_type_to_evp(ecc_attr->key.ecc.ec_curve);
    if (curve_nid == 0) {
        CTAMEerr(CTAME_F_CTAME_LOAD_ECC_PRIVATE_KEY,
                 CTAME_R_OBJECT_WRONG_ATTRIBUTE_TYPE);
        status = TAM_ERROR_NO_RESOURCES;
        goto error;
    }

    ec_key = EC_KEY_new_by_curve_name(curve_nid);
    if (!ec_key) {
        CTAMEerr(CTAME_F_CTAME_LOAD_ECC_PRIVATE_KEY, CTAME_R_NO_MEMORY);
        status = TAM_ERROR_NO_RESOURCES;
        goto error;
    }
    pub_key_len = obj_len_read;
    pub_key_buf = data;
    // Build Public key
    if (!d2i_EC_PUBKEY(&ec_key,
                       (const unsigned char **)&pub_key_buf, pub_key_len)) {
        CTAMEerr(CTAME_F_CTAME_LOAD_ECC_PRIVATE_KEY,
                 CTAME_R_ECC_PUBLIC_KEY_FAILED);
        status = TAM_ERROR_UNSPECIFIED;
        EC_KEY_free(ec_key);
        goto error;
    }
    *eckey = ec_key;
 error:
    if (data)
        OPENSSL_free(data);
    data = NULL;
    free(ecc_attr);

    return ((status == TAM_SUCCESS) ? CTAMC_SUCCESS : CTAMC_FAILURE);
}

#endif                          /* CTAME_SUPPORT */
