/*
 *  ctame_rsa.c
 *  Cisco TAM Engine for use with Cisco Trusted Anchor Module Client
 *  OpenSSL engine
 *
 *  Jan 2014
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */
#ifdef CTAME_SUPPORT
# include <stdlib.h>
# include <string.h>

# include <openssl/bn.h>
# include <openssl/dso.h>
# include <openssl/objects.h>
# include <openssl/rsa.h>

# include "../vendor_defns/ctamc.h"
# include "ctame.h"
# include "e_ctame_err.h"
# include "tam_types.h"
# include "safe_mem_lib.h"
# include "safe_lib_errno.h"
# include "safe_str_lib.h"

# ifdef USE_CISCO_SSM
#  include "ssmc/ssmc_rsa_key.c"
# else
#  include "tamc_api.h"
# endif

/*
 * ctame_rsa_private_key_encrypt()
 *
 * Using the specified CTAM, encrypt an object with a private key
 * stored in the CTAM.
 */
CTAMC_STATUS ctame_rsa_private_key_encrypt(const char *ctam_context,
                                           int length,
                                           const unsigned char *input,
                                           unsigned char *output,
                                           unsigned short *output_length,
                                           const char *key_name, int padding)
{
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    tam_status_e status;
    CTAMC_STATUS rc = CTAMC_FAILURE;

    uint8_t *encrypted_data = NULL;
    uint32_t encrypted_data_len = 0;

    if (!context || !input || !output || !key_name || (length == 0)
        || !output_length) {
        CTAMEerr(CTAME_F_CTAME_RSA_PRIVATE_KEY_ENCRYPT, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    status = tamc_rsa_private_encrypt(context->session_id,
                                      context->app_id,
                                      (char *)key_name,
                                      (tam_rsa_padding_e) padding,
                                      (uint8_t *) input,
                                      (uint32_t) length,
                                      &encrypted_data, &encrypted_data_len);
    if (status == TAM_SUCCESS) {
        if (encrypted_data && (encrypted_data_len != 0)) {
            *output_length = encrypted_data_len;
            memcpy_s(output, encrypted_data_len, encrypted_data,
                     encrypted_data_len);
            free(encrypted_data);
            encrypted_data = NULL;
            rc = CTAMC_SUCCESS;
        } else {
            rc = CTAMC_FAILURE;
        }
    }

    return (rc);
}

/*
 * ctame_rsa_public_key_encrypt()
 *
 * Using the specified CTAM, encrypt an object with a public key
 * stored in the CTAM.
 */
CTAMC_STATUS ctame_rsa_public_key_encrypt(const char *ctam_context,
                                          int length,
                                          const unsigned char *input,
                                          unsigned char *output,
                                          unsigned short *output_length,
                                          const char *key_name, int padding)
{
    tam_status_e status;
    CTAMC_STATUS rc = CTAMC_FAILURE;

    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    uint8_t *encrypted_data = NULL;
    uint32_t encrypted_data_len = 0;

    if (!context || !input || !output || !key_name || !output_length
        || (length == 0)) {
        CTAMEerr(CTAME_F_CTAME_RSA_PUBLIC_KEY_ENCRYPT, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    status = tamc_rsa_public_encrypt(context->session_id,
                                     context->app_id,
                                     (char *)key_name,
                                     (tam_rsa_padding_e) padding,
                                     (uint8_t *) input,
                                     (uint32_t) length,
                                     &encrypted_data, &encrypted_data_len);
    if (status == TAM_SUCCESS) {
        if (encrypted_data && (encrypted_data_len != 0)) {
            *output_length = encrypted_data_len;
            memcpy_s(output, encrypted_data_len, encrypted_data,
                     encrypted_data_len);
            free(encrypted_data);
            encrypted_data = NULL;
            rc = CTAMC_SUCCESS;
        } else {
            rc = CTAMC_FAILURE;
        }
    }

    return (rc);
}

/*
 * ctame_rsa_private_key_decrypt()
 *
 * Using the specified CTAM, decrypt an object with a private key
 * stored in the CTAM.
 */
CTAMC_STATUS ctame_rsa_private_key_decrypt(const char *ctam_context,
                                           int length,
                                           const unsigned char *input,
                                           unsigned char *output,
                                           unsigned short *output_length,
                                           const char *key_name, int padding)
{
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    tam_status_e status;
    CTAMC_STATUS rc = CTAMC_FAILURE;

    uint8_t *decrypted_data = NULL;
    uint32_t decrypted_data_len = 0;

    if (!context || !input || !output || !key_name || !output_length
        || (length == 0)) {
        CTAMEerr(CTAME_F_CTAME_RSA_PRIVATE_KEY_DECRYPT, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    status = tamc_rsa_private_decrypt(context->session_id,
                                      context->app_id,
                                      (char *)key_name,
                                      (uint8_t *) input,
                                      (uint32_t) length,
                                      &decrypted_data, &decrypted_data_len);

    if (status == TAM_SUCCESS) {
        if (decrypted_data && (decrypted_data_len != 0)) {
            *output_length = decrypted_data_len;
            memcpy_s(output, decrypted_data_len, decrypted_data,
                     decrypted_data_len);
            free(decrypted_data);
            decrypted_data = NULL;
            rc = CTAMC_SUCCESS;
        } else {
            rc = CTAMC_FAILURE;
        }
    }

    return (rc);
}

/*
 * ctame_rsa_public_key_decrypt()
 *
 * Using the specified CTAM, decrypt an object with a public key
 * stored in the CTAM.
 */
CTAMC_STATUS ctame_rsa_public_key_decrypt(const char *ctam_context,
                                          int length,
                                          const unsigned char *input,
                                          unsigned char *output,
                                          unsigned short *output_length,
                                          const char *key_name, int padding)
{
    tam_status_e status;
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    CTAMC_STATUS rc = CTAMC_FAILURE;

    uint8_t *decrypted_data = NULL;
    uint32_t decrypted_data_len = 0;

    if (!context || !input || !output || !key_name || !output_length
        || (length == 0)) {
        CTAMEerr(CTAME_F_CTAME_RSA_PUBLIC_KEY_DECRYPT, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    status = tamc_rsa_public_decrypt(context->session_id,
                                     context->app_id,
                                     (char *)key_name,
                                     (uint8_t *) input,
                                     (uint32_t) length,
                                     &decrypted_data, &decrypted_data_len);

    if (status == TAM_SUCCESS) {
        if (decrypted_data && (decrypted_data_len != 0)) {
            *output_length = decrypted_data_len;
            memcpy_s(output, decrypted_data_len, decrypted_data,
                     decrypted_data_len);
            free(decrypted_data);
            decrypted_data = NULL;
            rc = CTAMC_SUCCESS;
        } else {
            rc = CTAMC_FAILURE;
        }
    }

    return (rc);
}

/*
 * ctame_load_rsa_private_key()
 *
 * Read the specified public key from CTAM and return its modulus
 * and exponent.
 */
CTAMC_STATUS ctame_load_rsa_private_key(const char *ctam_context,
                                        const char *key_name,
                                        BIGNUM **bn_mod, BIGNUM **bn_exp)
{
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    tam_status_e status;
    int pub_key_len = 0;
    unsigned char *pub_key_buf = NULL;

    uint32_t obj_len_read = 0;
    uint8_t *data = NULL;
    RSA *rsa_key = NULL;
    tam_object_attr_t *rsa_attr = NULL;

    if (!context || !bn_mod || !bn_exp || !key_name) {
        CTAMEerr(CTAME_F_CTAME_LOAD_RSA_PRIVATE_KEY, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }
    // Get obj attr
    status = tamc_so_get_attr(context->session_id,
                              context->app_id, (char *)key_name, &rsa_attr);
    if (status != TAM_SUCCESS) {
        CTAMEerr(CTAME_F_CTAME_LOAD_RSA_PRIVATE_KEY,
                 CTAME_R_OBJECT_GET_ATTRIBUTE_FAILED);
        goto error;
    }
    // Validate Key type. Must be Type RSA
    if (!rsa_attr) {
        CTAMEerr(CTAME_F_CTAME_LOAD_RSA_PRIVATE_KEY,
                 CTAME_R_OBJECT_GET_ATTRIBUTE_FAILED);
        status = TAM_ERROR_INVALID_PARM;
        goto error;
    }

    if (rsa_attr->object_type != TAM_SO_TYPE_RSA_KEYPAIR) {
        CTAMEerr(CTAME_F_CTAME_LOAD_RSA_PRIVATE_KEY,
                 CTAME_R_OBJECT_WRONG_ATTRIBUTE_TYPE);
        status = TAM_ERROR_INCORRECT_KEY_TYPE;
        goto error;
    }

    if (!rsa_attr->object_size) {
        CTAMEerr(CTAME_F_CTAME_LOAD_RSA_PRIVATE_KEY,
                 CTAME_R_OBJECT_SIZE_ZERO);
        status = TAM_ERROR_INVALID_PARM;
        goto error;
    }
    // Read RSA Key
    status = tamc_so_read(context->session_id,
                          context->app_id,
                          (char *)key_name, &data, &obj_len_read);
    if (status != TAM_SUCCESS) {
        CTAMEerr(CTAME_F_CTAME_LOAD_RSA_PRIVATE_KEY,
                 CTAME_R_OBJECT_READ_FAILED);
        goto error;
    }
    // Extract public key components
    rsa_key = RSA_new();
    if (!rsa_key) {
        CTAMEerr(CTAME_F_CTAME_LOAD_RSA_PRIVATE_KEY, CTAME_R_NO_MEMORY);
        status = TAM_ERROR_INVALID_PARM;
        goto error;
    }
    // get public key
    pub_key_len = obj_len_read;
    pub_key_buf = data;

    if (!d2i_RSAPublicKey(&rsa_key,
                          (const unsigned char **)&pub_key_buf,
                          pub_key_len)) {
        CTAMEerr(CTAME_F_CTAME_LOAD_RSA_PRIVATE_KEY,
                 CTAME_R_RSA_PUBLIC_KEY_FAILED);
        status = TAM_ERROR_UNSPECIFIED;
        goto error;
    }

    *bn_exp = BN_dup(rsa_key->e);
    *bn_mod = BN_dup(rsa_key->n);

 error:
    if (data)
        OPENSSL_free(data);
    data = NULL;
    if (rsa_key)
        RSA_free(rsa_key);
    free(rsa_attr);

    return ((status == TAM_SUCCESS) ? CTAMC_SUCCESS : CTAMC_FAILURE);
}

/*
 * ctame_rsa_gen_keypair()
 *
 * Generate an RSA keypair with the specified length and exponent
 * attributes, persistence and name. The key is subsequently
 * referenced by "name."
 */
CTAMC_STATUS ctame_rsa_gen_keypair(const char *ctam_context,
                                   unsigned short key_length,
                                   unsigned int e_value,
                                   unsigned char zeroize,
                                   unsigned char persistent,
                                   unsigned char exportable,
                                   const char *key_name)
{
    tam_status_e status = TAM_ERROR_UNSPECIFIED;
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    tam_object_attr_t *key_object_attrs = NULL;

    /* Validate input */
    if (!ctam_context || !key_name || (key_length == 0) || (e_value == 0)) {
        CTAMEerr(CTAME_F_CTAME_RSA_GEN_KEYPAIR, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }

    /*
     * set object store parameter
     */
    key_object_attrs = calloc(1, sizeof(tam_object_attr_t));
    if (!key_object_attrs)
        goto end;
    key_object_attrs->csp_flag = zeroize;
    key_object_attrs->nv_flag = persistent;
    key_object_attrs->object_type = TAM_SO_TYPE_RSA_KEYPAIR;
    strncpy_s(key_object_attrs->object_name, TAM_OBJECT_NAME_LEN_MAX,
              key_name, TAM_OBJECT_NAME_LEN_MAX);

    /*
     * set rsa keygen parameter
     */
    key_object_attrs->key.rsa.e_value = e_value;
    key_object_attrs->key.rsa.key_len = key_length;
    key_object_attrs->key.rsa.exportable = exportable;
    /*
     * Make key gen request
     */
    status = tamc_rsa_keygen(context->session_id, key_object_attrs);
    if (status != TAM_SUCCESS) {
        CTAMEerr(CTAME_F_CTAME_RSA_GEN_KEYPAIR, CTAME_R_KEY_GEN_FAILED);
    }
 end:
    free(key_object_attrs);
    return ((status == TAM_SUCCESS) ? CTAMC_SUCCESS : CTAMC_FAILURE);
}

#endif                          /* CTAME_SUPPORT */
