/*
 *  e_ctamc.c
 *  Openssl Engine Cisco Trusted Anchor Module Client
 *
 *  cloned from e_ctam.c, January 2014
 *  Max Pritikin, 6/2011
 *  Bill Sulzen, 2/2013
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */

/* ====================================================================
 * Copyright (c) 1999-2001 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    licensing@OpenSSL.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <openssl/crypto.h>
#include <openssl/buffer.h>
#include <openssl/dso.h>
#include <openssl/engine.h>
#ifndef OPENSSL_NO_RSA
# include <openssl/rsa.h>
#endif
#include <openssl/bn.h>
#ifndef OPENSSL_NO_EC
/* This is an unfortunate ugly hack.  The old fecdsa.h use to
 * have private structs in it, which never should have 
 * occurred.  Unfortunately, CTAME was coded to use the private
 * ecdsa_method struct from fecdsa.h.
 */
# include "../../crypto/ecdsa/ecs_locl.h"
#endif

#ifndef OPENSSL_NO_HW
# ifndef OPENSSL_NO_HW_CTAMC

#  include "../vendor_defns/ctamc.h"
#  include "e_ctamc.h"
#  include "ctame.h"
#  include "e_ctamc_err.h"

#  include "tam_types.h"
#  include "safe_mem_lib.h"
#  include "safe_lib_errno.h"
#  include "safe_str_lib.h"

#  ifdef USE_CISCO_SSM
#   include "ssm.h"
#  else
#   include "tamc_api.h"
#  endif

#  define CTAMC_LIB_NAME "ctamc engine"

int ctamc_init(ENGINE *e);
static int ctamc_destroy(ENGINE *e);
static int ctamc_finish(ENGINE *e);
static int ctamc_ctrl(ENGINE *e, int cmd, long i, void *p, void (*f) (void));

/* CTAMC Rand Methods */
#  ifndef USE_CISCO_SSM
static int ctamc_get_true_random_bytes(unsigned char *buf, int trand_size);
static int ctamc_get_pseudo_random_bytes(unsigned char *buf, int rand_size);
static int ctamc_rand_status(void);
#  endif                        // #ifndef USE_CISCO_SSM

#  ifndef OPENSSL_NO_RSA
static int ctamc_rsa_pub_enc(int flen, const unsigned char *from,
                             unsigned char *to, RSA *rsa, int padding);
static int ctamc_rsa_pub_dec(int flen, const unsigned char *from,
                             unsigned char *to, RSA *rsa, int padding);
static int ctamc_rsa_priv_enc(int flen, const unsigned char *from,
                              unsigned char *to, RSA *rsa, int padding);
static int ctamc_rsa_priv_dec(int flen, const unsigned char *from,
                              unsigned char *to, RSA *rsa, int padding);

static EVP_PKEY *ctamc_load_privkey(ENGINE *eng, const char *key_id,
                                    UI_METHOD *ui_method,
                                    void *callback_data);
static void ctamc_ex_free(void *obj, void *item, CRYPTO_EX_DATA *ad, int idx,
                          long argl, void *argp);
#  endif                        // OPENSSL_NO_RSA

#  ifndef OPENSSL_NO_EC
static ECDSA_SIG *ctamc_ecdsa_do_sign(const unsigned char *dgst, int dlen,
                                      const BIGNUM *inv, const BIGNUM *rp,
                                      EC_KEY *eckey);
static int ctamc_ecdsa_do_verify(const unsigned char *dgst, int dgst_len,
                                 const ECDSA_SIG *sig, EC_KEY *eckey);
#  endif                        // OPENSSL_NO_EC

/* The definitions for control commands specific to this engine */
#  define CTAMC_CMD_SO_PATH                  ENGINE_CMD_BASE
#  define CTAMC_CMD_SET_CONTEXT              ENGINE_CMD_BASE+1
#  define CTAMC_CMD_GET_SUDI_CERT            ENGINE_CMD_BASE+2
#  define CTAMC_CMD_GET_SUDI_CERT_CHAIN      ENGINE_CMD_BASE+3
#  define CTAMC_CMD_GEN_RSA_KEYPAIR          ENGINE_CMD_BASE+4
#  define CTAMC_CMD_GEN_EC_KEYPAIR           ENGINE_CMD_BASE+5
#  ifdef USE_CISCO_SSM
#   define CTAMC_CMD_GET_CERT                 ENGINE_CMD_BASE+6
#   define CTAMC_CMD_GET_CERT_CHAIN           ENGINE_CMD_BASE+7
#  else
#   define CTAMC_CMD_GET_RAND                 ENGINE_CMD_BASE+6
#   define CTAMC_CMD_GET_CERT                 ENGINE_CMD_BASE+7
#   define CTAMC_CMD_GET_CERT_CHAIN           ENGINE_CMD_BASE+8
#  endif

static const ENGINE_CMD_DEFN ctamc_cmd_defns[] = {
    {CTAMC_CMD_SO_PATH,
     "CTAMC_SO_PATH",
     "Specifies the path to the 'CTAMC/TAMClient' shared library",
     ENGINE_CMD_FLAG_STRING},
    {CTAMC_CMD_SET_CONTEXT,
     "CTAMC_SET_CONTEXT",
     "Specifies the CTAM module number",
     ENGINE_CMD_FLAG_NUMERIC},
    {CTAMC_CMD_GET_SUDI_CERT,
     "CTAMC_GET_SUDI_CERT",
     "Returns the SUDI certificate",
     ENGINE_CMD_FLAG_INTERNAL},
    {CTAMC_CMD_GET_SUDI_CERT_CHAIN,
     "CTAMC_GET_SUDI_CERT_CHAIN",
     "Returns the SUDI certificate chain",
     ENGINE_CMD_FLAG_INTERNAL},
    {CTAMC_CMD_GEN_RSA_KEYPAIR,
     "CTAMC_GEN_RSA_KEYPAIR",
     "Generate RSA keypair",
     ENGINE_CMD_FLAG_INTERNAL},
    {CTAMC_CMD_GEN_EC_KEYPAIR,
     "CTAMC_GEN_EC_KEYPAIR",
     "Generate EC keypair",
     ENGINE_CMD_FLAG_INTERNAL},
#  ifndef USE_CISCO_SSM
    {CTAMC_CMD_GET_RAND,
     "CTAMC_CMD_GET_RAND",
     "Get Random Seed",
     ENGINE_CMD_FLAG_INTERNAL},
#  endif
    {CTAMC_CMD_GET_CERT,
     "CTAMC_CMD_GET_CERT",
     "Get Stored Certificate",
     ENGINE_CMD_FLAG_INTERNAL},
    {CTAMC_CMD_GET_CERT_CHAIN,
     "CTAMC_CMD_GET_CERT_CHAIN",
     "Get Stored Certificate Chain",
     ENGINE_CMD_FLAG_INTERNAL},
    {0, NULL, NULL, 0}
};

#  ifndef OPENSSL_NO_RSA

/* Our internal RSA_METHOD that we provide pointers to */
// see crypto/rsa/rsa.h for RSA_METHOD definitions
static RSA_METHOD ctamc_rsa = {
    "Cisco TAM RSA method",
    ctamc_rsa_pub_enc,          // rsa_pub_enc
    ctamc_rsa_pub_dec,          // rsa_pub_dec
    ctamc_rsa_priv_enc,         // rsa_priv_enc
    ctamc_rsa_priv_dec,         // rsa_priv_dec
    NULL,                       // rsa_mod_exp
    NULL,                       // bn_mod_exp
    NULL,                       // init, called at new
    NULL,                       // finish, called at free
    0,                          // rsa_sign and rsa_verify vectors
    NULL,                       // reserved for future use (app_data)
    NULL,                       // rsa_sign
    NULL,                       // rsa_verify
    NULL                        // rsa_keygen
};

static int rsa_ctamc_idx = -1;  /* Index for extended key data */
#  endif                        // OPENSSL_NO_RSA

#  ifndef OPENSSL_NO_EC
/* Our internal ECDSA_METHOD that we provide pointers to */
static ECDSA_METHOD ctamc_ecdsa = {
    "Cisco TAM ECDSA method",
    ctamc_ecdsa_do_sign,        // ecdsa_do_sign
    NULL,                       // ecdsa_sign_setup
    ctamc_ecdsa_do_verify,      // ecdsa_do_verify
    0,                          // flags
    NULL,                       // app_data
};

static int ecdsa_ctamc_idx = -1; /* Index for extended key data */
#  endif                        // OPENSSL_NO_EC

#  ifndef USE_CISCO_SSM
static RAND_METHOD ctamc_rand = {
    /* "Cisco CTAMC RAND method", */
    NULL,                       /* seed */
    ctamc_get_true_random_bytes, /* get random bytes */
    NULL,                       /* cleanup */
    NULL,                       /* add */
    ctamc_get_pseudo_random_bytes, /* pseudo drand */
    ctamc_rand_status,          /* status */
};
#  endif                        // #ifndef USE_CISCO_SSM

/*
 * The CTAM module specifies the session associated with the user of the
 * engine instance.
 */
static char *ctamc_session_context = NULL;

/* Constants used when creating the ENGINE */
static const char *engine_ctamc_id = "ctamc";
static const char *engine_ctamc_name = "Cisco TAM client API engine support";
static int ctamc_session_idx = -1; /* Engine Index for TAM Server sessions
                                    * IDs */
/*
 * bind_helper()
 *
 * Load engine function bindings
 */
static int bind_helper(ENGINE *e)
{
#  ifndef OPENSSL_NO_RSA
    const RSA_METHOD *dflt_rsa_meth = RSA_PKCS1_SSLeay();
    ctamc_rsa.rsa_pub_dec = dflt_rsa_meth->rsa_pub_dec;
    ctamc_rsa.rsa_pub_enc = dflt_rsa_meth->rsa_pub_enc;
    ctamc_rsa.bn_mod_exp = dflt_rsa_meth->bn_mod_exp;
    ctamc_rsa.rsa_mod_exp = dflt_rsa_meth->rsa_mod_exp;
#  endif
    const ECDSA_METHOD *dflt_ecc_meth = ECDSA_OpenSSL();
    ctamc_ecdsa.ecdsa_do_verify = dflt_ecc_meth->ecdsa_do_verify;
    ctamc_ecdsa.ecdsa_sign_setup = dflt_ecc_meth->ecdsa_sign_setup;
    if (!ENGINE_set_id(e, engine_ctamc_id) ||
        !ENGINE_set_name(e, engine_ctamc_name) ||
#  ifndef OPENSSL_NO_RSA
        !ENGINE_set_RSA(e, &ctamc_rsa) ||
#  endif                        // OPENSSL_NO_RSA
        !ENGINE_set_ECDSA(e, &ctamc_ecdsa) ||
#  ifndef USE_CISCO_SSM
        !ENGINE_set_RAND(e, &ctamc_rand) ||
#  endif
        !ENGINE_set_destroy_function(e, ctamc_destroy) ||
        !ENGINE_set_init_function(e, ctamc_init) ||
        !ENGINE_set_finish_function(e, ctamc_finish) ||
        !ENGINE_set_ctrl_function(e, ctamc_ctrl) ||
        !ENGINE_set_cmd_defns(e, ctamc_cmd_defns) ||
        !ENGINE_set_load_privkey_function(e, ctamc_load_privkey))
        return (0);

    /* Ensure the ctamc error handling is set up */
    ERR_load_CTAMC_strings();
    return (1);
}

static ENGINE *engine_ctamc(void)
{
    ENGINE *ret = ENGINE_new();
    if (!ret) {
        return (NULL);
    }
    if (!bind_helper(ret)) {
        ENGINE_free(ret);
        return (NULL);
    }
    return (ret);
}

/*
 * ENGINE_load_ctamc()
 *
 * Load the Cisco TAM engine.
 */
void ENGINE_load_ctamc(void)
{
    /* Copied from eng_[openssl|dyn].c */
    ENGINE *toadd;
    toadd = engine_ctamc();
    if (!toadd) {
        return;
    }
    ENGINE_add(toadd);
    ENGINE_free(toadd);
    ERR_clear_error();
}

/*
 * This is a process-global DSO handle used for loading and unloading the
 * Cisco TAM library. NB: This is only set (or unset) during an init() or
 * finish() call (reference counts permitting) and they're operating with
 * global locks, so this should be thread-safe implicitly.
 */
static DSO *ctamc_dso = NULL;

/*
 * These are the function pointers that are (un)set when the library has
 * successfully (un)loaded.
 */

#  ifndef OPENSSL_NO_RSA
t_ctamc_rsa_private_key_encrypt *p_ctamc_rsa_private_key_encrypt = NULL;
t_ctamc_rsa_public_key_encrypt *p_ctamc_rsa_public_key_encrypt = NULL;
t_ctamc_rsa_private_key_decrypt *p_ctamc_rsa_private_key_decrypt = NULL;
t_ctamc_rsa_public_key_decrypt *p_ctamc_rsa_public_key_decrypt = NULL;
t_ctamc_rsa_gen_keypair *p_ctamc_rsa_gen_keypair = NULL;
#  endif                        // OPENSSL_NO_RSA
#  ifndef OPENSSL_NO_EC
t_ctamc_ec_gen_keypair *p_ctamc_ec_gen_keypair = NULL;

t_ctamc_ecdsa_do_sign *p_ctamc_ecdsa_do_sign = NULL;
t_ctamc_ecdsa_do_verify *p_ctamc_ecdsa_do_verify = NULL;
t_ctamc_load_rsa_private_key *p_ctamc_load_rsa_private_key = NULL;
t_ctamc_load_ecc_private_key *p_ctamc_load_ecc_private_key = NULL;
#  endif                        // OPENSSL_NO_EC
t_ctamc_read_object *p_ctamc_read_object = NULL;

/*
 * These are the static string constants for the DSO file name and the
 * function symbol names to bind to
 */
static const char *CTAMC_LIBNAME = NULL;

static const char *get_CTAMC_LIBNAME(void)
{
    if (CTAMC_LIBNAME) {
        return (CTAMC_LIBNAME);
    }
    return ("ctame");
}

static void free_CTAMC_LIBNAME(void)
{
    if (CTAMC_LIBNAME) {
        OPENSSL_free((void *)CTAMC_LIBNAME);
    }
    CTAMC_LIBNAME = NULL;
}

static long set_CTAMC_LIBNAME(const char *name)
{
    free_CTAMC_LIBNAME();
    return (((CTAMC_LIBNAME = BUF_strdup(name)) != NULL) ? 1 : 0);
}

static const char *CTAME_F1 = "ctame_rsa_private_key_encrypt";
static const char *CTAME_F2 = "ctame_rsa_public_key_encrypt";
static const char *CTAME_F3 = "ctame_rsa_private_key_decrypt";
static const char *CTAME_F4 = "ctame_rsa_public_key_decrypt";
static const char *CTAME_F5 = "ctame_rsa_gen_keypair";
static const char *CTAME_F6 = "ctame_read_object";
static const char *CTAME_F7 = "ctame_load_rsa_private_key";
static const char *CTAME_F8 = "ctame_ec_gen_keypair";

static const char *CTAME_F9 = "ctame_ecdsa_do_sign";
static const char *CTAME_F10 = "ctame_ecdsa_do_verify";
static const char *CTAME_F11 = "ctame_load_ecc_private_key";

/* Destructor (complements the "ENGINE_ctamc()" constructor) */
static int ctamc_destroy(ENGINE *e)
{
    free_CTAMC_LIBNAME();
    /*
     * Unload the ctamc error strings so any error state including our functs
     * or reasons won't lead to a segfault (they simply get displayed without
     * corresponding string data because none will be found).
     */
    ERR_unload_CTAMC_strings();
    return (1);
}

/*
 * Sets the session context after the user logins into TAM server
 * 1 = Pass, 0 = Fail
 */
int ctamc_set_session_context(ENGINE *e, ctamc_context_t * ctx)
{
    int rc = 0;
    if (!e)
        goto end;

    if (ctamc_session_idx == -1)
        ctamc_session_idx = ENGINE_get_ex_new_index(0, NULL, NULL, NULL, 0);

    if (ctamc_session_idx == -1)
        goto end;

    rc = ENGINE_set_ex_data(e, ctamc_session_idx, ctx);
    ctamc_session_context = (char *)ctx;
 end:
    return (rc);
}

/* (de)initialisation functions. */
int ctamc_init(ENGINE *e)
{
    t_ctamc_rsa_private_key_encrypt *p1;
    t_ctamc_rsa_public_key_encrypt *p2;
    t_ctamc_rsa_private_key_decrypt *p3;
    t_ctamc_rsa_public_key_decrypt *p4;
    t_ctamc_rsa_gen_keypair *p5;
    t_ctamc_read_object *p6;
    t_ctamc_load_rsa_private_key *p7;
    t_ctamc_ec_gen_keypair *p8;
    t_ctamc_ecdsa_do_sign *p9;
    t_ctamc_ecdsa_do_verify *p10;
    t_ctamc_load_ecc_private_key *p11;

    // Setup session id index
    if (ctamc_session_idx == -1)
        ctamc_session_idx = ENGINE_get_ex_new_index(0, NULL, NULL, NULL, 0);
#  ifndef OPENSSL_NO_RSA
    if (rsa_ctamc_idx == -1)
        rsa_ctamc_idx = RSA_get_ex_new_index(0, "Cisco RSA TAM key data",
                                             NULL, NULL, ctamc_ex_free);
#  endif                        // OPENSSL_NO_RSA
#  ifndef OPENSSL_NO_EC
    if (ecdsa_ctamc_idx == -1)
        ecdsa_ctamc_idx =
            ECDSA_get_ex_new_index(0, "Cisco ECDSA TAM key data", NULL, NULL,
                                   ctamc_ex_free);
#  endif                        // OPENSSL_NO_EC

#  ifndef OPENSSL_NO_STATIC_ENGINE
    return (ctame_init());
#  endif                        // OPENSSL_NO_STATIC_ENGINE

    if (ctamc_dso != NULL) {
        CTAMCerr(CTAMC_F_CTAMC_INIT, CTAMC_R_ALREADY_LOADED);
        goto ctam_init_err;
    }

    ctamc_dso = DSO_load(NULL, get_CTAMC_LIBNAME(), NULL, 0);
    if (ctamc_dso == NULL) {
        CTAMCerr(CTAMC_F_CTAMC_INIT, CTAMC_R_NOT_LOADED);
        goto ctam_init_err;
    }

    /*
     * Vector names for invoking ENGINE_ctrl_cmd().
     */
    if (!(p1 = (t_ctamc_rsa_private_key_encrypt *)
          DSO_bind_func(ctamc_dso, CTAME_F1)) ||
        !(p2 = (t_ctamc_rsa_public_key_encrypt *)
          DSO_bind_func(ctamc_dso, CTAME_F2)) ||
        !(p3 = (t_ctamc_rsa_private_key_decrypt *)
          DSO_bind_func(ctamc_dso, CTAME_F3)) ||
        !(p4 = (t_ctamc_rsa_public_key_decrypt *)
          DSO_bind_func(ctamc_dso, CTAME_F4)) ||
        !(p5 = (t_ctamc_rsa_gen_keypair *)
          DSO_bind_func(ctamc_dso, CTAME_F5)) ||
        !(p6 = (t_ctamc_read_object *)
          DSO_bind_func(ctamc_dso, CTAME_F6)) ||
        !(p7 = (t_ctamc_load_rsa_private_key *)
          DSO_bind_func(ctamc_dso, CTAME_F7)) ||
        !(p8 = (t_ctamc_ec_gen_keypair *)
          DSO_bind_func(ctamc_dso, CTAME_F8)) ||
        !(p9 = (t_ctamc_ecdsa_do_sign *)
          DSO_bind_func(ctamc_dso, CTAME_F9)) ||
        !(p10 = (t_ctamc_ecdsa_do_verify *)
          DSO_bind_func(ctamc_dso, CTAME_F10)) ||
        !(p11 = (t_ctamc_load_ecc_private_key *)
          DSO_bind_func(ctamc_dso, CTAME_F11))) {
        CTAMCerr(CTAMC_F_CTAMC_INIT, CTAMC_R_NOT_LOADED);
        goto ctam_init_err;
    }

    /* Copy the pointers */
#  ifndef OPENSSL_NO_RSA
    p_ctamc_rsa_private_key_encrypt = p1;
    p_ctamc_rsa_public_key_encrypt = p2;
    p_ctamc_rsa_private_key_decrypt = p3;
    p_ctamc_rsa_public_key_decrypt = p4;
    p_ctamc_rsa_gen_keypair = p5;
#  endif
    p_ctamc_read_object = p6;
    p_ctamc_load_rsa_private_key = p7;
#  ifndef OPENSSL_NO_EC
    p_ctamc_ec_gen_keypair = p8;
    p_ctamc_ecdsa_do_sign = p9;
    p_ctamc_ecdsa_do_verify = p10;
    p_ctamc_load_ecc_private_key = p11;
#  endif
    /* Everything's fine. */
    return (1);

 ctam_init_err:
    if (ctamc_dso) {
        DSO_free(ctamc_dso);
        ctamc_dso = NULL;
    }
#  ifndef OPENSSL_NO_RSA
    p_ctamc_rsa_private_key_encrypt = NULL;
    p_ctamc_rsa_public_key_encrypt = NULL;
    p_ctamc_rsa_private_key_decrypt = NULL;
    p_ctamc_rsa_public_key_decrypt = NULL;
    p_ctamc_rsa_gen_keypair = NULL;
#  endif
    p_ctamc_read_object = NULL;
    p_ctamc_load_rsa_private_key = NULL;
#  ifndef OPENSSL_NO_EC
    p_ctamc_ec_gen_keypair = NULL;
    p_ctamc_ecdsa_do_sign = NULL;
    p_ctamc_ecdsa_do_verify = NULL;
    p_ctamc_load_ecc_private_key = NULL;
#  endif
    return (0);
}

static int ctamc_finish(ENGINE *e)
{
    free_CTAMC_LIBNAME();

    p_ctamc_rsa_private_key_encrypt = NULL;
    p_ctamc_rsa_public_key_encrypt = NULL;
    p_ctamc_rsa_private_key_decrypt = NULL;
    p_ctamc_rsa_public_key_decrypt = NULL;
    p_ctamc_rsa_gen_keypair = NULL;
    p_ctamc_read_object = NULL;
    p_ctamc_load_rsa_private_key = NULL;

#  ifndef OPENSSL_NO_EC
    p_ctamc_ec_gen_keypair = NULL;
    p_ctamc_ecdsa_do_sign = NULL;
    p_ctamc_ecdsa_do_verify = NULL;
    p_ctamc_load_ecc_private_key = NULL;
#  endif

    if (ctamc_dso) {
        DSO_free(ctamc_dso);
        ctamc_dso = NULL;
    }

    if (ctamc_session_context) {
        OPENSSL_free(ctamc_session_context);
        ctamc_session_context = NULL;
    }

    return (1);
}

static int ctamc_ctrl(ENGINE *e, int cmd, long i, void *p, void (*f) (void))
{
    ctamc_item_t *item;
    ctamc_rsa_keygen_parms_t *keyparms = NULL;
    ctamc_ecc_key_attr_t *ec_keyparms = NULL;
    ctamc_file_attr_t *file_attr = NULL;

    CTAMC_STATUS rc = CTAMC_SUCCESS;

    switch (cmd) {
    case CTAMC_CMD_SO_PATH:
        if (p == NULL) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_PASSED_NULL_PARAMETER);
            return (0);
        }

        return (set_CTAMC_LIBNAME((const char *)p));
        break;

    case CTAMC_CMD_SET_CONTEXT:
        /*
         * If the given context is null
         * clear the existing one and return.
         */
        if (p == NULL) {
            /* Delete the old context if there is one */
            if (ctamc_session_context) {
                OPENSSL_free(ctamc_session_context);
                ctamc_session_context = NULL;
            }
            return (1);
        }
        /* p is a ctamc_item_t */
        item = (ctamc_item_t *) p;

        /* Delete the old context if there is one */
        if (ctamc_session_context) {
            OPENSSL_free(ctamc_session_context);
            ctamc_session_context = NULL;
        }

        /* Allocate a new one and populate it */
        ctamc_session_context = OPENSSL_malloc(item->len);
        if (!ctamc_session_context) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_INTERNAL_ERROR);
            return (0);
        }

        memcpy(ctamc_session_context, &(item->c[0]), item->len);

        return (1);
        break;

    case CTAMC_CMD_GET_SUDI_CERT:
        if (p == NULL) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_PASSED_NULL_PARAMETER);
            return (0);
        }
        /* p is a ctamc_item_t */
        item = (ctamc_item_t *) p;

        if (!p_ctamc_read_object) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_NOT_LOADED);
            return (0);
        }

        rc = p_ctamc_read_object(ctamc_session_context,
                                 CTAMC_SUDI_CERT_NAME,
                                 &(item->c), &(item->len));

        return ((rc == CTAMC_SUCCESS) ? 1 : 0);
        break;

    case CTAMC_CMD_GET_SUDI_CERT_CHAIN:
        if (p == NULL) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_PASSED_NULL_PARAMETER);
            return (0);
        }
        /* p is a ctamc_item_t */
        item = (ctamc_item_t *) p;
        if (!p_ctamc_read_object) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_NOT_LOADED);
            return (0);
        }

        rc = p_ctamc_read_object(ctamc_session_context,
                                 CTAMC_SUDI_CERT_CHAIN_NAME,
                                 &(item->c), &(item->len));
        return ((rc == CTAMC_SUCCESS) ? 1 : 0);
        break;

    case CTAMC_CMD_GET_CERT:
        if (p == NULL) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_PASSED_NULL_PARAMETER);
            return (0);
        }
        /* p is a ctamc_item_t */
        item = (ctamc_item_t *) p;
        file_attr = (ctamc_file_attr_t *) item->c;

        rc = ctame_check_object_type(ctamc_session_context,
                                     (const char *)file_attr->filename,
                                     TAM_SO_TYPE_CERT);
        if (rc != CTAMC_SUCCESS)
            return (0);

        rc = p_ctamc_read_object(ctamc_session_context,
                                 (const char *)file_attr->filename,
                                 &file_attr->file_buffer,
                                 &(file_attr->file_size));

        return ((rc == CTAMC_SUCCESS) ? 1 : 0);
        break;

    case CTAMC_CMD_GET_CERT_CHAIN:
        if (p == NULL) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_PASSED_NULL_PARAMETER);
            return (0);
        }
        /* p is a ctamc_item_t */
        item = (ctamc_item_t *) p;
        file_attr = (ctamc_file_attr_t *) item->c;

        rc = ctame_check_object_type(ctamc_session_context,
                                     (const char *)file_attr->filename,
                                     TAM_SO_TYPE_CERTCHAIN);
        if (rc != CTAMC_SUCCESS)
            return (0);

        if (!p_ctamc_read_object) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_NOT_LOADED);
            return (0);
        }

        rc = p_ctamc_read_object(ctamc_session_context,
                                 (const char *)file_attr->filename,
                                 &file_attr->file_buffer,
                                 &(file_attr->file_size));

        return ((rc == CTAMC_SUCCESS) ? 1 : 0);
        break;

    case CTAMC_CMD_GEN_RSA_KEYPAIR:
        if (p == NULL) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_PASSED_NULL_PARAMETER);
            return (0);
        }
        /* p is a ctamc_item_t */
        item = (ctamc_item_t *) p;

        keyparms = (ctamc_rsa_keygen_parms_t *) item->c;

        /* Perform some parameter validation */
        if ((keyparms->key_len < 256) ||
            (keyparms->e_value != 65537) ||
            !keyparms->key_name || (strlen(keyparms->key_name) == 0)) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_PARAMETER_ERROR);
            return (0);
        }

        if (!p_ctamc_rsa_gen_keypair) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_NOT_LOADED);
            return (0);
        }

        rc = p_ctamc_rsa_gen_keypair(ctamc_session_context,
                                     keyparms->key_len,
                                     keyparms->e_value,
                                     keyparms->zeroize,
                                     keyparms->persistent,
                                     keyparms->exportable,
                                     keyparms->key_name);

        return ((rc == CTAMC_SUCCESS) ? 1 : 0);
        break;

    case CTAMC_CMD_GEN_EC_KEYPAIR:
        if (p == NULL) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_PASSED_NULL_PARAMETER);
            return (0);
        }
        /* p is a ctamc_item_t */
        item = (ctamc_item_t *) p;

        ec_keyparms = (ctamc_ecc_key_attr_t *) item->c;

        /* Perform some parameter validation */
        if ((ec_keyparms->ec_curve == 0) ||
            (((ec_keyparms->ec_curve >= CTAMC_EC_CURVE_MAX))) ||
            (strlen(&ec_keyparms->key_name[0]) == 0)) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_PARAMETER_ERROR);
            return (0);
        }

        if (!p_ctamc_ec_gen_keypair) {
            CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_NOT_LOADED);
            return (0);
        }

        rc = p_ctamc_ec_gen_keypair(ctamc_session_context, ec_keyparms);

        return ((rc == CTAMC_SUCCESS) ? 1 : 0);
        break;

    default:
        break;
    }

    CTAMCerr(CTAMC_F_CTAMC_CTRL, CTAMC_R_CTRL_COMMAND_NOT_IMPLEMENTED);
    return (0);
}

#  ifndef OPENSSL_NO_RSA
/*
 * ctamc_rsa_pub_enc()
 *
 * Use the CTAM hardware to perform public key encryption.
 *
 * Note that since the public key is cached anyway by ctamc_load_privkey()
 * this vector shouldn't be called. It is here for testing and completeness.
 */
static int ctamc_rsa_pub_enc(int flen, const unsigned char *from,
                             unsigned char *to, RSA *rsa, int padding)
{
    ctame_extra_data_t *extra_data = NULL;
    unsigned short tlen;
    unsigned char *key = NULL;

    if (!p_ctamc_rsa_public_key_encrypt) {
        CTAMCerr(CTAMC_F_CTAMC_RSA_PUB_ENC, CTAMC_R_ENGINE_NOT_INITIALISED);
        return (0);
    }

    /*
     * Extract key_id info from the extended data
     */
    extra_data = RSA_get_ex_data(rsa, rsa_ctamc_idx);
    if (!extra_data) {
        CTAMCerr(CTAMC_F_CTAMC_RSA_PUB_ENC, CTAMC_R_INVALID_KEY);
        return (0);
    }

    key = &extra_data->key_id_str[0];
    if (p_ctamc_rsa_public_key_encrypt
        (ctamc_session_context, flen, from, to, &tlen, (const char *)key,
         padding) == CTAMC_SUCCESS) {
        return (tlen);
    }

    return (0);
}

/*
 * ctamc_rsa_pub_dec()
 *
 * Use the CTAM hardware to perform public key decryption.
 *
 * Note that since the public key is cached anyway by ctamc_load_privkey()
 * this vector shouldn't be called. It is here for testing and completeness.
 */
static int ctamc_rsa_pub_dec(int flen, const unsigned char *from,
                             unsigned char *to, RSA *rsa, int padding)
{
    ctame_extra_data_t *extra_data = NULL;
    unsigned short tlen;
    unsigned char *key = NULL;
    if (!p_ctamc_rsa_public_key_decrypt || !ctamc_session_context) {
        CTAMCerr(CTAMC_F_CTAMC_RSA_PUB_DEC, CTAMC_R_ENGINE_NOT_INITIALISED);
        return (0);
    }

    /*
     * Extract key_id info from the extended data
     */
    extra_data = RSA_get_ex_data(rsa, rsa_ctamc_idx);
    if (!extra_data) {
        CTAMCerr(CTAMC_F_CTAMC_RSA_PUB_DEC, CTAMC_R_INVALID_KEY);
        return (0);
    }

    key = &extra_data->key_id_str[0];
    if (p_ctamc_rsa_public_key_decrypt
        (ctamc_session_context, flen, from, to, &tlen, (const char *)key,
         padding) == CTAMC_SUCCESS) {
        return (tlen);
    }

    return (0);
}

/*
 * ctamc_rsa_priv_enc()
 *
 * Use the CTAM hardware to perform private key encryption.
 */
static int ctamc_rsa_priv_enc(int flen, const unsigned char *from,
                              unsigned char *to, RSA *rsa, int padding)
{
    ctame_extra_data_t *extra_data = NULL;
    unsigned short tlen;
    unsigned char *key = NULL;
    if (!p_ctamc_rsa_private_key_encrypt) {
        CTAMCerr(CTAMC_F_CTAMC_RSA_PRIV_ENC, CTAMC_R_ENGINE_NOT_INITIALISED);
        return (0);
    }

    /*
     * Extract key_id info from the extended data
     */
    extra_data = RSA_get_ex_data(rsa, rsa_ctamc_idx);
    if (!extra_data) {
        CTAMCerr(CTAMC_F_CTAMC_RSA_PRIV_ENC, CTAMC_R_INVALID_KEY);
        return (0);
    }

    key = &extra_data->key_id_str[0];
    if (p_ctamc_rsa_private_key_encrypt
        (ctamc_session_context, flen, from, to, &tlen, (const char *)key,
         padding) == CTAMC_SUCCESS) {
        return (tlen);
    }

    return (0);
}

/*
 * ctamc_rsa_priv_dec()
 *
 * Use the CTAM hardware to perform private key decryption.
 */
static int ctamc_rsa_priv_dec(int flen, const unsigned char *from,
                              unsigned char *to, RSA *rsa, int padding)
{
    ctame_extra_data_t *extra_data = NULL;
    unsigned short tlen;
    unsigned char *key = NULL;
    if (!p_ctamc_rsa_private_key_decrypt) {
        CTAMCerr(CTAMC_F_CTAMC_RSA_PRIV_DEC, CTAMC_R_ENGINE_NOT_INITIALISED);
        return (0);
    }

    /*
     * Extract key_id info from the extended data
     */
    extra_data = RSA_get_ex_data(rsa, rsa_ctamc_idx);
    if (!extra_data) {
        CTAMCerr(CTAMC_F_CTAMC_RSA_PRIV_DEC, CTAMC_R_INVALID_KEY);
        return (0);
    }

    key = &extra_data->key_id_str[0];
    if (p_ctamc_rsa_private_key_decrypt
        (ctamc_session_context, flen, from, to, &tlen, (const char *)key,
         padding) == CTAMC_SUCCESS) {
        return (tlen);
    }

    return (0);
}
#  endif                        /* OPENSSL_NO_RSA */

/*
 * ctamc_load_privkey()
 *
 * Invoked by a call to ENGINE_load_private_key() to load a keypair.
 *
 * Note only RSA keys are currently supported
 */
static EVP_PKEY *ctamc_load_privkey(ENGINE *eng, const char *key_id,
                                    UI_METHOD *ui_method, void *callback_data)
{
    tam_status_e status;
    tam_object_attr_t *obj_attr = NULL;
    ctame_extra_data_t *extra_data = NULL;
    unsigned char *key_id_str = NULL;
    EVP_PKEY *res = NULL;
    ctamc_context_t *context = (ctamc_context_t *) ctamc_session_context;
    tam_object_type_e key_type = TAM_SO_TYPE_INVALID_OBJECT;
#  ifndef OPENSSL_NO_RSA
    RSA *rtmp = NULL;
    BIGNUM *mod = NULL;
    BIGNUM *exp = NULL;
#  endif                        // OPENSSL_NO_RSA

    // declear ecc stuff
    EC_KEY *ec_key = NULL;

    if (!ctamc_session_context) {
        CTAMCerr(CTAMC_F_CTAMC_LOAD_PRIVKEY, CTAMC_R_ENGINE_NOT_INITIALISED);
        goto load_privkey_err;
    }

    if (!key_id) {
        CTAMCerr(CTAMC_F_CTAMC_LOAD_PRIVKEY, CTAMC_R_INVALID_KEY_ID);
        goto load_privkey_err;
    }

    /*
     * Read ctamc attribute
     */
    status = tamc_so_get_attr(context->session_id,
                              context->app_id, (char *)key_id, &obj_attr);
    if (status != TAM_SUCCESS) {
        CTAMCerr(CTAMC_F_CTAMC_LOAD_PRIVKEY, CTAMC_R_KEY_NOT_FOUND);
        goto load_privkey_err;
    }
    key_type = obj_attr->object_type;
    free(obj_attr);
    if ((key_type != TAM_SO_TYPE_RSA_KEYPAIR)
        && (key_type != TAM_SO_TYPE_ECC_KEYPAIR)) {
        CTAMCerr(CTAMC_F_CTAMC_LOAD_PRIVKEY, CTAMC_R_KEY_NOT_FOUND);
        goto load_privkey_err;
    }

    /*
     * Store index of private key for subsequent private key encrypt,
     * decrypt and other operations.
     */
    extra_data = OPENSSL_malloc(sizeof(ctame_extra_data_t));
    if (!extra_data) {
        CTAMCerr(CTAMC_F_CTAMC_LOAD_PRIVKEY, CTAMC_R_INTERNAL_ERROR);
        goto load_privkey_err;
    }
    key_id_str = &extra_data->key_id_str[0];
    memzero_s(extra_data, sizeof(ctame_extra_data_t));
    strncpy((char *)key_id_str, key_id, strlen(key_id));

    /*
     * Retrieve the keypair (which will only return the public key)
     * from the CTAM.
     */
    if (key_type == TAM_SO_TYPE_RSA_KEYPAIR) {
#  ifndef OPENSSL_NO_RSA
        if (p_ctamc_load_rsa_private_key &&
            (p_ctamc_load_rsa_private_key
             (ctamc_session_context, key_id, &mod, &exp)
             == CTAMC_SUCCESS)) {
        } else {
            CTAMCerr(CTAMC_F_CTAMC_LOAD_PRIVKEY, CTAMC_R_KEY_NOT_FOUND);
            goto load_privkey_err;
        }

        rtmp = RSA_new_method(eng);
        RSA_set_ex_data(rtmp, rsa_ctamc_idx, extra_data);

        /*
         * Cache the public key so the host can perform public key
         * encrypt and ecrypt operations.
         */
        rtmp->e = exp;
        rtmp->n = mod;
        /* Call us even when the key isn't fully populated */
        rtmp->flags |= RSA_FLAG_EXT_PKEY;

        /* create an EVP object: engine + RSA key */
        res = EVP_PKEY_new();
        if (!res) {
            CTAMCerr(CTAMC_F_CTAMC_LOAD_PRIVKEY, CTAMC_R_INTERNAL_ERROR);
            goto load_privkey_err;
        }
        EVP_PKEY_assign_RSA(res, rtmp);
#  endif                        // OPENSSL_NO_RSA
    } else {
        // ECC Key type TAM_SO_TYPE_ECC_KEYPAIR

        if (p_ctamc_load_ecc_private_key &&
            (p_ctamc_load_ecc_private_key
             (ctamc_session_context, key_id, &ec_key)
             == CTAMC_SUCCESS)) {
        } else {
            CTAMCerr(CTAMC_F_CTAMC_LOAD_PRIVKEY, CTAMC_R_KEY_NOT_FOUND);
            goto load_privkey_err;
        }

        ECDSA_set_ex_data(ec_key, ecdsa_ctamc_idx, extra_data);
        ECDSA_set_method(ec_key, &ctamc_ecdsa);

        /*
         * Cache the public key so the host can perform public key
         * encrypt and ecrypt operations.
         */

        /* create an EVP object: engine + ECC key */
        res = EVP_PKEY_new();
        if (!res) {
            CTAMCerr(CTAMC_F_CTAMC_LOAD_PRIVKEY, CTAMC_R_INTERNAL_ERROR);
            goto load_privkey_err;
        }
        EVP_PKEY_assign_EC_KEY(res, ec_key);
    }

    return (res);

 load_privkey_err:
    if (extra_data) {
        OPENSSL_free(extra_data);
    }
    return (NULL);
}

#  ifndef OPENSSL_NO_RSA
static void ctamc_ex_free(void *obj, void *item, CRYPTO_EX_DATA *ad,
                          int idx, long argl, void *argp)
{
    if (item) {
        OPENSSL_free(item);
    }
}
#  endif                        // OPENSSL_NO_RSA

#  ifndef OPENSSL_NO_EC
ECDSA_SIG *ctamc_ecdsa_do_sign(const unsigned char *dgst, int dlen,
                               const BIGNUM *inv, const BIGNUM *rp,
                               EC_KEY *eckey)
{
    ECDSA_SIG *sig = NULL;
    ctame_extra_data_t *extra_data = NULL;
    unsigned char *key_id = NULL;

    if (!p_ctamc_ecdsa_do_sign) {
        CTAMCerr(CTAMC_F_CTAMC_ECDSA_DO_SIGN, CTAMC_R_ENGINE_NOT_INITIALISED);
        return (NULL);
    }

    /*
     * Extract key_id info from the extended data
     */
    extra_data = ECDSA_get_ex_data(eckey, ecdsa_ctamc_idx);
    if (!extra_data) {
        CTAMCerr(CTAMC_F_CTAMC_ECDSA_DO_SIGN, CTAMC_R_INVALID_KEY);
        return (NULL);
    }

    key_id = &extra_data->key_id_str[0];

    if (p_ctamc_ecdsa_do_sign(ctamc_session_context, dgst,
                              dlen, (const char *)key_id,
                              (ECDSA_SIG **)&sig) == CTAMC_SUCCESS) {
        return (sig);
    }

    return (NULL);
}

int ctamc_ecdsa_do_verify(const unsigned char *dgst, int dgst_len,
                          const ECDSA_SIG *sig, EC_KEY *eckey)
{
    int rc = 0;
    ctame_extra_data_t *extra_data = NULL;
    unsigned char *key_id = NULL;

    if (!p_ctamc_ecdsa_do_verify) {
        CTAMCerr(CTAMC_F_CTAMC_ECDSA_DO_VERIFY,
                 CTAMC_R_ENGINE_NOT_INITIALISED);
        return (0);
    }

    /*
     * Extract key_id info from the extended data
     */
    extra_data = ECDSA_get_ex_data(eckey, ecdsa_ctamc_idx);
    if (!extra_data) {
        CTAMCerr(CTAMC_F_CTAMC_ECDSA_DO_VERIFY, CTAMC_R_INVALID_KEY);
        return (0);
    }

    key_id = &extra_data->key_id_str[0];

    if (p_ctamc_ecdsa_do_verify(ctamc_session_context, dgst, dgst_len,
                                (const char *)key_id,
                                (ECDSA_SIG *)sig) == CTAMC_SUCCESS) {
        rc = 1;
    }

    return (rc);
}
#  endif                        // OPENSSL_NO_EC

/* CTAMC Rand Methods */
#  ifndef USE_CISCO_SSM
int ctamc_get_true_random_bytes(unsigned char *buf, int trand_size)
{
    CTAMC_STATUS ctamc_rc = CTAMC_FAILURE;
    int rc = 0;

    if (!ctamc_session_context) {
        CTAMCerr(CTAMC_F_CTAMC_GET_TRUE_RANDOM_BYTES,
                 CTAMC_R_ENGINE_NOT_INITIALISED);
        return (rc);
    }

    if (!buf || trand_size == 0) {
        CTAMCerr(CTAMC_F_CTAMC_GET_TRUE_RANDOM_BYTES,
                 CTAMC_R_PARAMETER_ERROR);
        return (rc);
    }

    ctamc_rc =
        ctame_get_true_random_bytes(ctamc_session_context,
                                    (unsigned char *)buf, trand_size);
    if (ctamc_rc == CTAMC_SUCCESS) {
        rc = 1;
    }

    return (rc);
}

int ctamc_get_pseudo_random_bytes(unsigned char *buf, int rand_size)
{
    int rc = 0;
    CTAMC_STATUS ctamc_rc = CTAMC_FAILURE;
    if (!ctamc_session_context) {
        CTAMCerr(CTAMC_F_CTAMC_GET_PSEUDO_RANDOM_BYTES,
                 CTAMC_R_ENGINE_NOT_INITIALISED);
        return (rc);
    }

    if (!buf || rand_size == 0) {
        CTAMCerr(CTAMC_F_CTAMC_GET_PSEUDO_RANDOM_BYTES,
                 CTAMC_R_PARAMETER_ERROR);
        return (rc);
    }

    ctamc_rc =
        ctame_get_pseudo_random_bytes(ctamc_session_context, buf, rand_size);
    if (ctamc_rc == CTAMC_SUCCESS) {
        rc = 1;
    }
    return (rc);
}

int ctamc_rand_status(void)
{
    if (!ctamc_session_context) {
        CTAMCerr(CTAMC_F_CTAMC_RAND_STATUS, CTAMC_R_ENGINE_NOT_INITIALISED);
        return (0);
    }
    return (1);
}
#  endif                        // #ifndef USE_CISCO_SSM

#  ifndef OPENSSL_NO_DYNAMIC_ENGINE
static int bind_fn(ENGINE *e, const char *id)
{
    if (id && (strcmp(id, engine_ctamc_id) != 0)) {
        return (0);
    }
    if (!bind_helper(e)) {
        return (0);
    }
    return (1);
}

IMPLEMENT_DYNAMIC_CHECK_FN();
IMPLEMENT_DYNAMIC_BIND_FN(bind_fn);
#  endif                        /* OPENSSL_NO_DYNAMIC_ENGINE */

# endif                         /* !OPENSSL_NO_HW_CTAMC */
#endif                          /* !OPENSSL_NO_HW */
