/** @file */
/* -----------------------------------------------------------------
 * ssmc.c
 *
 * SSMc General functions
 *
 * Trusted Services interest-ctam@cisco.com
 *
 * Copyright  (c) 2014 by Cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#ifdef USE_CISCO_SSM

# include <stdlib.h>
# include <stdio.h>
# include <stdint.h>
# include <openssl/evp.h>
# include "safe_str_lib.h"
# include "safe_mem_lib.h"
# include "tamc_ssm_api.h"
# include "ssmc.h"

# include <openssl/rsa.h>
# include <openssl/err.h>

# include <openssl/ec.h>
# include <openssl/x509.h>
# include <openssl/ecdsa.h>

static EVP_PKEY *ssmc_build_rsa_key(ssmc_key_buf_t * rsa_key_buff,
                                    uint32_t key_buff_len)
{
    EVP_PKEY *evp_pkey = NULL;
    int32_t openssl_rc;

    uint32_t pub_key_len = key_buff_len;
    uint32_t priv_key_len = 0;
    int32_t openssl_key_type = EVP_PKEY_RSA;

    uint8_t *pub_key_buf = &rsa_key_buff->buffer[0];

    // setup the rsa key we want populated
    RSA *rsa_key = RSA_new();
    if (!rsa_key) {
        goto error;
    }
    // Build Public key
    if (!d2i_RSAPublicKey(&rsa_key,
                          (const unsigned char **)&pub_key_buf,
                          pub_key_len)) {
        tam_debug(TAMD_DBG_DEV_ERR, "Get public key failed. len=%d",
                  pub_key_len);
        goto error;
    }

    priv_key_len = key_buff_len - (pub_key_buf - (uint8_t *) rsa_key_buff);
    if (!d2i_RSAPrivateKey(&rsa_key,
                           (const unsigned char **)&pub_key_buf,
                           priv_key_len)) {
        tam_debug(TAMD_DBG_DEV_ERR, "Get private key failed. len=%d",
                  priv_key_len);
        goto error;
    }
    // Setup EVP Key data
    evp_pkey = EVP_PKEY_new();
    if (!evp_pkey) {
        goto error;
    }
    // Set the Key type
    openssl_rc = EVP_PKEY_set_type(evp_pkey, openssl_key_type);
    if (!openssl_rc) {
        goto error;
    }
    // assign the rsa key to the evp envelope
    openssl_rc = EVP_PKEY_assign_RSA(evp_pkey, rsa_key);
    if (!openssl_rc) {
        tam_debug(TAMD_DBG_DEV_ERR, " Failed to set key in envelope failed");
        goto error;
    }

    return (evp_pkey);

 error:
    if (evp_pkey) {
        EVP_PKEY_free(evp_pkey);
        evp_pkey = NULL;
    }
    if (rsa_key) {
        RSA_free(rsa_key);
        rsa_key = NULL;
    }
    return NULL;
}

/*
 * ssmc_load_key()
 *
 * Load a rsa/ecc key and prepare for use
 * returning an evp key
 */
tam_status_e ssmc_load_key(void *session,
                           char *key_name,
                           char *key_owner, EVP_PKEY **evp_key)
{
    tam_status_e rc = TAM_SUCCESS;
    tam_object_attr_t *attrs = NULL;

    uint8_t *key_buffer = NULL;
    uint32_t key_obj_size = 0;

    /*
     * Check key attribute
     */
    rc = tamc_so_get_attr(session, key_owner, key_name, &attrs);
    if (rc != TAM_SUCCESS) {
        tam_debug(TAMD_DBG_DEV_ERR, "Failed to get key attr.  "
                  " : 0x%x", rc);
        goto end;
    }

    if ((attrs->object_type != TAM_SO_TYPE_RSA_KEYPAIR) &&
        (attrs->object_type != TAM_SO_TYPE_ECC_KEYPAIR)) {
        tam_debug(TAMD_DBG_DEV_ERR,
                  "Requested object isn't a RSA nor ECC key ");
        rc = TAM_ERROR_INCORRECT_KEY_TYPE;
        goto end;
    }

    /*
     * Read key file
     */
    rc = tamc_so_read(session, key_owner, key_name, &key_buffer,
                      &key_obj_size);
    if (rc != TAM_SUCCESS) {
        tam_debug(TAMD_DBG_DEV_ERR, "Failed to read key.  " " : 0x%x", rc);
        goto end;
    }

    if (!key_buffer) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR, "Error retrieving key.  " " : 0x%x", rc);
        goto end;
    }

    rc = ssmc_build_evp_key(attrs, key_buffer, evp_key);
    if (rc != TAM_SUCCESS) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR, "Error building key.  " " : 0x%x", rc);
        goto end;
    }
 end:
    if (attrs)
        free(attrs);
    if (key_buffer)
        free(key_buffer);
    return (rc);
}

/*
 * ssmc_build_evp_key()
 *
 * Builds and returns an EVP_PKEY read from DER encoded buffer
 */
tam_status_e ssmc_build_evp_key(tam_object_attr_t * attrs,
                                uint8_t * key_buffer, EVP_PKEY **key)
{
    tam_status_e rc = TAM_SUCCESS;
    uint8_t key_type;
    ssmc_key_buf_t *key_buff = NULL;

    if (!attrs || !key_buffer || !key) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    key_buff = (ssmc_key_buf_t *) key_buffer;
    key_type = attrs->object_type;
    switch (key_type) {
    case TAM_SO_TYPE_RSA_KEYPAIR:
        *key = ssmc_build_rsa_key(key_buff, attrs->object_size);
        if (*key == NULL)
            rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        break;

    case TAM_SO_TYPE_ECC_KEYPAIR:
        *key =
            ssmc_build_ec_key(key_buff, attrs->key.ecc.ec_curve,
                              attrs->object_size);
        if (*key == NULL)
            rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        break;

    default:
        rc = TAM_ERROR_INCORRECT_KEY_TYPE;
        break;
    }
 end:
    return (rc);
}

/*
 * convert tamd_ec_curve_e to openssl curve type
 */
uint32_t ssmc_tam_curve_type_to_evp(tam_ec_curve_e curve)
{
    uint32_t open_curve = 0;

    switch (curve) {
    case TAM_EC_CURVE_P192:
        open_curve = NID_X9_62_prime192v1;
        break;
    case TAM_EC_CURVE_P256:
        open_curve = NID_X9_62_prime256v1;
        break;
    case TAM_EC_CURVE_P384:
        open_curve = NID_secp384r1;
        break;
    case TAM_EC_CURVE_P521:
        open_curve = NID_secp521r1;
        break;
    case TAM_EC_CURVE_BRAINPOOLP256:
        open_curve = NID_brainpoolP256r1;
        break;
    case TAM_EC_CURVE_BRAINPOOLP384:
        open_curve = NID_brainpoolP384r1;
        break;
    case TAM_EC_CURVE_BRAINPOOLP512:
        open_curve = NID_brainpoolP512r1;
        break;

    case TAM_EC_CURVE_MAX:
    default:
        break;
    }
    return ((uint32_t) open_curve);
}

/*
 * convert tamd_hash_algo_e to  openssl md type
 */
const EVP_MD *ssmc_tam_hash_type_to_evp(tam_hash_algo_e algo)
{
    const EVP_MD *type = NULL;
    switch (algo) {
    case TAM_HASH_ALGO_SHA1:
        type = EVP_sha1();
        break;
    case TAM_HASH_ALGO_SHA256:
        type = EVP_sha256();
        break;
    case TAM_HASH_ALGO_SHA384:
        type = EVP_sha384();
        break;
    case TAM_HASH_ALGO_SHA512:
        type = EVP_sha512();
        break;
    default:
        break;
    }
    return (type);
}

/*
 * Build ECC EVP_PKEY from a DER encoded buffer
 */
EVP_PKEY *ssmc_build_ec_key(ssmc_key_buf_t * ecc_key_buff,
                            tam_ec_curve_e ec_curve, uint32_t key_buff_len)
{
    EVP_PKEY *evp_pkey = NULL;
    int32_t openssl_rc;
    uint32_t pub_key_len = 0;
    uint32_t priv_key_len = 0;
    int32_t openssl_key_type = EVP_PKEY_EC;
    uint8_t *pub_key_buf = NULL;
    EC_KEY *ec_key = NULL;
    uint32_t curve_nid;

    if (!ecc_key_buff) {
        goto error;
    }

    pub_key_len = key_buff_len;
    priv_key_len = 0;
    pub_key_buf = &ecc_key_buff->buffer[0];

    // setup the ec key we want populated
    curve_nid = ssmc_tam_curve_type_to_evp(ec_curve);
    ec_key = EC_KEY_new_by_curve_name(curve_nid);

    if (!ec_key) {
        goto error;
    }
    // Build Public key
    if (!d2i_EC_PUBKEY(&ec_key,
                       (const unsigned char **)&pub_key_buf, pub_key_len)) {
        tam_debug(TAMD_DBG_DEV_ERR, "Get public key failed.");
        goto error;
    }

    priv_key_len = key_buff_len - (pub_key_buf - (uint8_t *) ecc_key_buff);
    if (!d2i_ECPrivateKey(&ec_key,
                          (const unsigned char **)&pub_key_buf,
                          priv_key_len)) {
        tam_debug(TAMD_DBG_DEV_ERR, " Get private key failed.");
        goto error;
    }
    // Setup EVP Key data
    evp_pkey = EVP_PKEY_new();
    if (!evp_pkey) {
        goto error;
    }
    // Set the Key type
    openssl_rc = EVP_PKEY_set_type(evp_pkey, openssl_key_type);
    if (!openssl_rc) {
        goto error;
    }
    // assign the ec key to the evp envelope
    openssl_rc = EVP_PKEY_set1_EC_KEY(evp_pkey, ec_key);
    if (!openssl_rc) {
        tam_debug(TAMD_DBG_DEV_ERR, " Failed to set key in envelope failed");
        goto error;
    }

    /*
     * if we are here then we have an ec key
     * Decrement ref count
     * since the block is done with the key
     */
    EC_KEY_free(ec_key);
    return (evp_pkey);

 error:
    if (evp_pkey) {
        EVP_PKEY_free(evp_pkey);
        evp_pkey = NULL;
    }
    if (ec_key) {
        EC_KEY_free(ec_key);
        ec_key = NULL;
    }
    return NULL;
}

const ECDSA_METHOD *ssmc_get_ecdsa_method(void)
{
# ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
        return FIPS_ecdsa_openssl();
    } else
# endif
    {
        return ECDSA_OpenSSL();
    }
}

const RSA_METHOD *ssmc_get_rsa_method(void)
{
# ifdef OPENSSL_FIPS
    if (FIPS_mode())
        return (RSA_METHOD *)FIPS_rsa_pkcs1_ssleay();
    else
        return (RSA_METHOD *)RSA_PKCS1_SSLeay();
# else
    return (RSA_METHOD *)RSA_PKCS1_SSLeay();
# endif
}

#endif
