/** @file */
/* -----------------------------------------------------------------
 * ssmc_ecc_key.c
 *
 * SSMc ECC Related functionality
 *
 * Trusted Services interest-ctam@cisco.com
 *
 * Copyright  (c) 2014 by Cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include "safe_str_lib.h"
#include "safe_mem_lib.h"
#include "ssmc.h"

/*
 * ECC Keypair Generation
 */
/** @brief tamc_ecc_keygen - Generate ECC Key pair with the specified parameters.
 * Blocking call to generate a ECC key pair.
 * @param *vsession - client/server session used to request services.
 * @param *ecc_key_attr - ECC Key parameter/curve name
 * @return TAM_SUCCESS or error code.
 */
static tam_status_e tamc_ecc_keygen(void *session,
                                    tam_object_attr_t * ecc_key_attr)
{
    tam_status_e rc = TAM_SUCCESS;
    EC_KEY *ec_key = NULL;
    EC_GROUP *group = NULL;

    uint32_t curve_nid = 0;

    ssmc_key_buf_t *ecc_key_buff = NULL;
    unsigned char *pub_buf_ptr = NULL;
    unsigned char *pri_buf_ptr = NULL;

    int32_t pub_key_buff_size = 0, pub_size_read = 0;
    int32_t pri_key_buff_size = 0, pri_size_read = 0;
    uint32_t buffer_length = 0;

    int openssl_rc;

    /*
     * Parameter check
     */
    if (!session || !ecc_key_attr) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }
    // Verify ECC Params
    if (ecc_key_attr->key.ecc.ec_curve > TAM_EC_CURVE_MAX) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    /*
     * Get EC Curve NID and ec key
     */

    curve_nid = ssmc_tam_curve_type_to_evp(ecc_key_attr->key.ecc.ec_curve);
    if (curve_nid == 0) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    /*
     * allocate group based on curve value
     */
    group = EC_GROUP_new_by_curve_name(curve_nid);
    if (!group) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV, "Error creating EC group. 0x%x", rc);
        goto end;
    }

    /*
     * Set group parameters
     */
    EC_GROUP_set_asn1_flag(group, OPENSSL_EC_NAMED_CURVE);
    EC_GROUP_set_point_conversion_form(group, POINT_CONVERSION_UNCOMPRESSED);

    /*
     * allocate a new key struct based on the curve
     */
    ec_key = EC_KEY_new();
    if (!ec_key) {
        rc = TAM_ERROR_NO_RESOURCES;
        tam_debug(TAMD_DBG_DEV, "Error creating EC KEY: 0x%x", rc);
        goto end;
    }

    if (EC_KEY_set_group(ec_key, group) == 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV, "Error setting EC group in key. 0x%x", rc);
        goto end;
    }

    /* Generate Keys */
    openssl_rc = EC_KEY_generate_key(ec_key);
    if (openssl_rc != 1) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV, "Error Check EC Key. 0x%x", rc);
        goto end;
    }

    /*
     * Read with null buffer to get the space needed
     */
    pub_key_buff_size = i2d_EC_PUBKEY(ec_key, NULL);
    if (pub_key_buff_size <= 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV,
                  "Error converting Pub EC Key to file format. 0x%x", rc);
        goto end;
    }

    /*
     * Read with null buffer to get the space needed
     */
    pri_key_buff_size = i2d_ECPrivateKey(ec_key, NULL);
    if (pri_key_buff_size <= 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV,
                  "Error converting Priv EC Key to file format. 0x%x", rc);
        goto end;
    }
    // buffer needed
    buffer_length =
        sizeof(ssmc_key_buf_t) + (pub_key_buff_size +
                                  pri_key_buff_size) * sizeof(uint8_t);
    ecc_key_buff = calloc(1, buffer_length);
    if (!ecc_key_buff) {
        rc = TAM_ERROR_NO_RESOURCES;
        goto end;
    }

    /*
     * Set write buffer pointers
     */
    pub_buf_ptr = &ecc_key_buff->buffer[0];
    pri_buf_ptr = &ecc_key_buff->buffer[pub_key_buff_size];

    /*
     * Write data to the buffer
     */
    pub_size_read = i2d_EC_PUBKEY(ec_key, &pub_buf_ptr);
    if (pub_size_read != pub_key_buff_size) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_STATE_ERR, "Read/Write mismatch Pub EC Key.");
        goto end;
    }

    /*
     * Write data to the buffer
     */
    pri_size_read = i2d_ECPrivateKey(ec_key, &pri_buf_ptr);
    if (pri_size_read != pri_key_buff_size) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_STATE_ERR, "Read/Write mismatch Priv EC Key.");
        goto end;
    }

    /* setup object store key */
    ecc_key_attr->generated = 1;
    ecc_key_attr->object_type = TAM_SO_TYPE_ECC_KEYPAIR;
    ecc_key_attr->object_size = buffer_length;

    /* create object */
    rc = tamc_so_create(session, ecc_key_attr);
    if (rc != TAM_SUCCESS) {
        tam_debug(TAMD_DBG_STATE_ERR,
                  "Error creating stored object: 0x%x", rc);
        goto end;
    }

    /* write the object */
    rc = tamc_so_write(session, ecc_key_attr->object_owner,
                       ecc_key_attr->object_name, (uint8_t *) ecc_key_buff,
                       buffer_length);
    if (rc != TAM_SUCCESS) {
        tam_debug(TAMD_DBG_STATE_ERR,
                  "Error writing stored object: 0x%x", rc);
        goto end;
    }

 end:
    if (ecc_key_buff) {
        memzero_s(ecc_key_buff, buffer_length);
        free(ecc_key_buff);
    }
    if (ec_key)
        EC_KEY_free(ec_key);
    if (group)
        EC_GROUP_free(group);

    return (rc);
}
