/** @file */
/* -----------------------------------------------------------------
 * ssmc_rsa_key.c
 *
 * SSMc RSA related functionality
 *
 * Trusted Services interest-ctam@cisco.com
 *
 * Copyright  (c) 2014 by Cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <openssl/evp.h>
#include "safe_str_lib.h"
#include "safe_mem_lib.h"
#include "tamc_ssm_api.h"
#include "ssmc.h"

#include <openssl/rsa.h>
#include <openssl/err.h>

#define PKCS1_PAD_LEN 11

/*
 * Crypto Functions
 */
/*
 * RSA Keypair Generation
 */
/** @brief tamc_rsa_keygen - Generate RSA Key pair with the
 * specified parameters
 * @param *vsession - client/server session used to request services.
 * @param *key_object_attr - attributes for requested key
 * @return TAM_SUCCESS or error code.
 */
static tam_status_e tamc_rsa_keygen(void *session,
                                    tam_object_attr_t * rsa_key_attr)
{
    tam_status_e rc = TAM_SUCCESS;
    BIGNUM *e = NULL;
    RSA *rsa_key = NULL;
    int32_t pub_key_buff_size = 0, pub_size_read = 0;
    int32_t pri_key_buff_size = 0, pri_size_read = 0;
    uint32_t buffer_length = 0;
    unsigned char *pub_buf_ptr = NULL;
    unsigned char *pri_buf_ptr = NULL;
    ssmc_key_buf_t *rsa_key_buff = NULL;

    /*
     * Validate params
     */
    if (!session || !rsa_key_attr
        || rsa_key_attr->object_type != TAM_SO_TYPE_RSA_KEYPAIR
        || (rsa_key_attr->key.rsa.e_value != 0x10001)
        || strnlen_s(rsa_key_attr->object_name,
                     TAM_OBJECT_NAME_LEN_MAX) > TAM_OBJECT_NAME_LEN_MAX) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    /* Generate Keys */
    rsa_key = RSA_new();
    e = BN_new();
    if (!rsa_key || !e) {
        rc = TAM_ERROR_NO_RESOURCES;
        tam_debug(TAMD_DBG_DEV_ERR,
                  "Error getting resources for RSA keygen: 0x%x", rc);
        goto end;
    }
    BN_set_word(e, rsa_key_attr->key.rsa.e_value);

    if (RSA_generate_key_ex(rsa_key, rsa_key_attr->key.rsa.key_len, e, NULL)
        == 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR, "Error generating RSA Key:%lx. 0x%x",
                  ERR_peek_last_error(), rc);
        goto end;
    }

    pri_key_buff_size = i2d_RSAPrivateKey(rsa_key, NULL);
    if (pri_key_buff_size < 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR,
                  "Error converting Priv RSA Key to file format. 0x%x", rc);
        goto end;
    }

    pub_key_buff_size = i2d_RSAPublicKey(rsa_key, NULL);
    if (pub_key_buff_size < 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR,
                  "Error converting Pub RSA Key to file format. 0x%x", rc);
        goto end;
    }
    buffer_length =
        sizeof(ssmc_key_buf_t) + pub_key_buff_size + pri_key_buff_size;
    rsa_key_buff = calloc(1, buffer_length);
    if (!rsa_key_buff) {
        rc = TAM_ERROR_NO_RESOURCES;
        goto end;
    }

    /*
     * Set write buffer pointers
     */
    pub_buf_ptr = &rsa_key_buff->buffer[0];
    pri_buf_ptr = &rsa_key_buff->buffer[pub_key_buff_size];

    pub_size_read = i2d_RSAPublicKey(rsa_key, &pub_buf_ptr);
    if (pub_size_read != pub_key_buff_size) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_STATE_ERR, "Read/Write mismatch Pub RSA Key.");
        goto end;
    }

    pri_size_read = i2d_RSAPrivateKey(rsa_key, &pri_buf_ptr);
    if (pri_size_read != pri_key_buff_size) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_STATE_ERR, "Read/Write mismatch Priv RSA Key.");
        goto end;
    }

    /* setup object store key */
    rsa_key_attr->generated = 1;
    rsa_key_attr->object_type = TAM_SO_TYPE_RSA_KEYPAIR;
    rsa_key_attr->object_size = buffer_length;

    /* create object */
    rc = tamc_so_create(session, rsa_key_attr);
    if (rc != TAM_SUCCESS) {
        tam_debug(TAMD_DBG_STATE_ERR,
                  "Error creating stored object: 0x%x", rc);
        goto end;
    }

    /* write the object */
    rc = tamc_so_write(session, rsa_key_attr->object_owner,
                       rsa_key_attr->object_name, (uint8_t *) rsa_key_buff,
                       buffer_length);
    if (rc != TAM_SUCCESS) {
        tam_debug(TAMD_DBG_STATE_ERR,
                  "Error writing stored object: 0x%x", rc);
        goto end;
    }

 end:
    if (e)
        BN_free(e);
    if (rsa_key_buff) {
        memzero_s(rsa_key_buff, buffer_length);
        free(rsa_key_buff);
    }
    if (rsa_key)
        RSA_free(rsa_key);
    return (rc);
}

/** @brief tamc_rsa_private_encrypt - Encrypt with RSA private key
 * and generate cipher text.
 * @param *vsession - client/server session used to request services.
 * @param *key_owner - application/user name of the owner of the RSA key
 * @param *key_name - name of the RSA key to be used
 * @param padding - Type of padding to be used when encrypting
 * @param *cleartext - buffer with data to be encrypted
 * @param cleartext_len - amount of data to be encrypted, in bytes
 * @param **ciphertext - returned buffer with encrypted data
 * @param ciphertext_len - returned size of the ciphertext buffer, in bytes.
 * May be different from cleartext_len due to padding.
 * @return TAM_SUCCESS or error code.
 */
static tam_status_e tamc_rsa_private_encrypt(void *session,
                                             char *key_owner,
                                             char *key_name,
                                             tam_rsa_padding_e padding,
                                             uint8_t * cleartext,
                                             uint32_t cleartext_len,
                                             uint8_t ** ciphertext,
                                             uint32_t * ciphertext_len)
{
    tam_status_e rc = TAM_SUCCESS;
    EVP_PKEY *key = NULL;
    RSA_METHOD *tmp_rsa_meth = NULL;

    int outlen = 0;

    /*
     * Validate input
     */
    if (!session || !key_name || !cleartext || (cleartext_len == 0)
        || !ciphertext || !ciphertext_len) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    /*
     * Read and Process RSA Key
     */
    rc = ssmc_load_key(session, key_name, key_owner, &key);
    if (rc != TAM_SUCCESS || !key) {
        tam_debug(TAMD_DBG_DEV_ERR, "Couldn't load rsa key  " " : 0x%x", rc);
        goto end;
    }

    if (padding != TAM_RSA_PADDING_PKCS1_V15) {
        rc = TAM_ERROR_INVALID_PARM;
        tam_debug(TAMD_DBG_DEV_ERR, "Invalid rsa padding.");
        goto end;
    }

    if ((int32_t) cleartext_len > RSA_size(key->pkey.rsa) - PKCS1_PAD_LEN) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR,
                  "Encrypt request (%d) too large for rsa key size (%d).",
                  cleartext_len, RSA_size(key->pkey.rsa));
        goto end;
    }

    *ciphertext = calloc(1, RSA_size(key->pkey.rsa));
    if (!*ciphertext) {
        rc = TAM_ERROR_NO_RESOURCES;
        goto end;
    }

    /*
     * Temporarily set the the RSA method to the default meth
     * otherwise we run recursively through the engine.
     */
    tmp_rsa_meth = (RSA_METHOD *)key->pkey.rsa->meth;
    key->pkey.rsa->meth = ssmc_get_rsa_method();
    /* encrypt buffer */
    outlen = RSA_private_encrypt(cleartext_len, cleartext, *ciphertext,
                                 key->pkey.rsa, RSA_PKCS1_PADDING);
    /*
     * restore the original method
     */
    key->pkey.rsa->meth = tmp_rsa_meth;
    if (outlen <= 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR, "Error during RSA private key encrypt.");
        goto end;
    }

    *ciphertext_len = outlen;
 end:
    if (rc != TAM_SUCCESS) {
        if (ciphertext && *ciphertext) {
            free(*ciphertext);
            *ciphertext = NULL;
        }
        if (ciphertext_len)
            *ciphertext_len = 0;
    }
    if (key)
        EVP_PKEY_free(key);
    return rc;
}

/** @brief tamc_rsa_public_encrypt - Encrypt with RSA public key
 * and generate cipher text.
 * @param *vsession - client/server session used to request services.
 * @param *key_owner - application/user name of the owner of the RSA key
 * @param *key_name - name of the RSA key to be used
 * @param padding - Type of padding to be used when encrypting
 * @param *cleartext - buffer with data to be encrypted
 * @param cleartext_len - amount of data to be encrypted, in bytes
 * @param **ciphertext - returned buffer with encrypted data
 * @param ciphertext_len - returned size of the ciphertext buffer, in bytes.
 * May be different from cleartext_len due to padding.
 * @return TAM_SUCCESS or error code.
 */
static tam_status_e tamc_rsa_public_encrypt(void *session,
                                            char *key_owner,
                                            char *key_name,
                                            tam_rsa_padding_e padding,
                                            uint8_t * cleartext,
                                            uint32_t cleartext_len,
                                            uint8_t ** ciphertext,
                                            uint32_t * ciphertext_len)
{
    tam_status_e rc = TAM_SUCCESS;
    EVP_PKEY *key = NULL;
    RSA_METHOD *tmp_rsa_meth = NULL;
    int outlen = 0;

    /*
     * Validate input
     */
    if (!session || !key_name || !cleartext || (cleartext_len == 0)
        || !ciphertext || !ciphertext_len) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    /*
     * Read and Process RSA Key
     */
    rc = ssmc_load_key(session, key_name, key_owner, &key);
    if (rc != TAM_SUCCESS || !key) {
        tam_debug(TAMD_DBG_DEV_ERR, "Couldn't load rsa key  " " : 0x%x", rc);
        goto end;
    }

    if (padding != TAM_RSA_PADDING_PKCS1_V15) {
        rc = TAM_ERROR_INVALID_PARM;
        tam_debug(TAMD_DBG_DEV_ERR, "Invalid rsa padding.");
        goto end;
    }

    if ((int32_t) cleartext_len > RSA_size(key->pkey.rsa) - PKCS1_PAD_LEN) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR,
                  "Encrypt request (%d) too large for rsa key size (%d).",
                  cleartext_len, RSA_size(key->pkey.rsa));
        goto end;
    }

    *ciphertext = calloc(1, RSA_size(key->pkey.rsa));
    if (!*ciphertext) {
        rc = TAM_ERROR_NO_RESOURCES;
        goto end;
    }

    /*
     * Temporarily set the the RSA method to the default meth
     * otherwise we run recursively through the engine.
     */
    tmp_rsa_meth = (RSA_METHOD *)key->pkey.rsa->meth;
    key->pkey.rsa->meth = ssmc_get_rsa_method();

    /* encrypt buffer */
    outlen = RSA_public_encrypt(cleartext_len, cleartext, *ciphertext,
                                key->pkey.rsa, RSA_PKCS1_PADDING);
    /*
     * restore the original method
     */
    key->pkey.rsa->meth = tmp_rsa_meth;
    if (outlen <= 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR, "Error during RSA public key encrypt.");
        goto end;
    }

    *ciphertext_len = outlen;
 end:
    if (rc != TAM_SUCCESS) {
        if (ciphertext && *ciphertext) {
            free(*ciphertext);
            *ciphertext = NULL;
        }
        if (ciphertext_len)
            *ciphertext_len = 0;
    }
    if (key)
        EVP_PKEY_free(key);
    return rc;
}

/*
 * RSA Private Key Decryption
 */
/** @brief tamc_rsa_private_decrypt - Decrypt with RSA private key
 * and generate clear text.
 * @param *vsession - client/server session used to request services.
 * @param *key_owner - application/user name of the owner of the RSA key
 * @param *key_name - name of the RSA key to be used
 * @param *ciphertext - buffer with data to be decrypted
 * @param ciphertext_len - amount of data to be decrypted, in bytes
 * @param **cleartext - returned buffer with decrypted data
 * @param cleartext_len - returned size of the cleartext buffer, in bytes.
 * May be different from ciphertext_len due to padding.
 * @return TAM_SUCCESS or error code.
 */
static tam_status_e tamc_rsa_private_decrypt(void *session,
                                             char *key_owner,
                                             char *key_name,
                                             uint8_t * ciphertext,
                                             uint32_t ciphertext_len,
                                             uint8_t ** cleartext,
                                             uint32_t * cleartext_len)
{
    tam_status_e rc = TAM_SUCCESS;
    EVP_PKEY *key = NULL;
    RSA_METHOD *tmp_rsa_meth = NULL;

    int outlen = 0;

    /*
     * Validate input
     */
    if (!session || !key_name || !cleartext || !cleartext_len || !ciphertext
        || (ciphertext_len == 0)) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    /*
     * Read and Process RSA Key
     */
    rc = ssmc_load_key(session, key_name, key_owner, &key);
    if (rc != TAM_SUCCESS || !key) {
        tam_debug(TAMD_DBG_DEV_ERR, "Couldn't load rsa key  " " : 0x%x", rc);
        goto end;
    }

    if (ciphertext_len != RSA_size(key->pkey.rsa)) {
        rc = TAM_ERROR_INVALID_PARM;
        tam_debug(TAMD_DBG_DEV_ERR,
                  "Decrypt request doesn't match rsa key size.");
        goto end;
    }

    *cleartext = calloc(1, ciphertext_len);
    if (!*cleartext) {
        rc = TAM_ERROR_NO_RESOURCES;
        goto end;
    }

    /*
     * Temporarily set the the RSA method to the default meth
     * otherwise we run recursively through the engine.
     */
    tmp_rsa_meth = (RSA_METHOD *)key->pkey.rsa->meth;
    key->pkey.rsa->meth = ssmc_get_rsa_method();

    /* Decrypt buffer */
    outlen = RSA_private_decrypt(ciphertext_len, ciphertext, *cleartext,
                                 key->pkey.rsa, RSA_PKCS1_PADDING);
    /*
     * restore the original method
     */
    key->pkey.rsa->meth = tmp_rsa_meth;
    if (outlen <= 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR, "Error during RSA private key decrypt.");
        goto end;
    }

    *cleartext_len = outlen;
 end:
    if (rc != TAM_SUCCESS) {
        if (cleartext && *cleartext) {
            free(*cleartext);
            *cleartext = NULL;
        }
        if (cleartext_len)
            *cleartext_len = 0;
    }

    if (key)
        EVP_PKEY_free(key);
    return rc;
}

/** @brief tamc_rsa_public_decrypt - Decrypt with RSA public key
 * and generate clear text.
 * @param *vsession - client/server session used to request services.
 * @param *key_owner - application/user name of the owner of the RSA key
 * @param *key_name - name of the RSA key to be used
 * @param *ciphertext - buffer with data to be decrypted
 * @param ciphertext_len - amount of data to be decrypted, in bytes
 * @param **cleartext - returned buffer with decrypted data
 * @param cleartext_len - returned size of the cleartext buffer, in bytes.
 * May be different from ciphertext_len due to padding.
 * @return TAM_SUCCESS or error code.
 */
static tam_status_e tamc_rsa_public_decrypt(void *session,
                                            char *key_owner,
                                            char *key_name,
                                            uint8_t * ciphertext,
                                            uint32_t ciphertext_len,
                                            uint8_t ** cleartext,
                                            uint32_t * cleartext_len)
{
    tam_status_e rc = TAM_SUCCESS;
    EVP_PKEY *key = NULL;
    RSA_METHOD *tmp_rsa_meth = NULL;
    int outlen = 0;

    /*
     * Validate input
     */
    if (!session || !key_name || !ciphertext ||
        (ciphertext_len == 0) || !cleartext || !cleartext_len) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    /*
     * Read and Process RSA Key
     */
    rc = ssmc_load_key(session, key_name, key_owner, &key);
    if (rc != TAM_SUCCESS || !key) {
        tam_debug(TAMD_DBG_DEV_ERR, "Couldn't load rsa key  " " : 0x%x", rc);
        goto end;
    }

    if (ciphertext_len != RSA_size(key->pkey.rsa)) {
        rc = TAM_ERROR_INVALID_PARM;
        tam_debug(TAMD_DBG_DEV_ERR,
                  "Decrypt request doesn't match rsa key size.");
        goto end;
    }

    *cleartext = calloc(1, ciphertext_len);
    if (!*cleartext) {
        rc = TAM_ERROR_NO_RESOURCES;
        goto end;
    }

    /*
     * Temporarily set the the RSA method to the default meth
     * otherwise we run recursively through the engine.
     */
    tmp_rsa_meth = (RSA_METHOD *)key->pkey.rsa->meth;
    key->pkey.rsa->meth = ssmc_get_rsa_method();
    /* Decrypt buffer */
    outlen = RSA_public_decrypt(ciphertext_len, ciphertext, *cleartext,
                                key->pkey.rsa, RSA_PKCS1_PADDING);
    /*
     * restore the original method
     */
    key->pkey.rsa->meth = tmp_rsa_meth;
    if (outlen <= 0) {
        rc = TAM_ERROR_DEVICE_OPERATION_FAIL;
        tam_debug(TAMD_DBG_DEV_ERR, "Error during RSA public key decrypt.");
        goto end;
    }

    *cleartext_len = outlen;
 end:
    if (rc != TAM_SUCCESS) {
        if (cleartext && *cleartext) {
            free(*cleartext);
            *cleartext = NULL;
        }
        if (cleartext_len)
            *cleartext_len = 0;
    }
    if (key)
        EVP_PKEY_free(key);
    return rc;
}
