/** @file */
/* -----------------------------------------------------------------
 * ssmc_sign_verify.c
 *
 * SSMc sign/verify digest
 *
 * Trusted Services interest-ctam@cisco.com
 *
 * Copyright  (c) 2014 by Cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <openssl/evp.h>
#include "safe_str_lib.h"
#include "safe_mem_lib.h"
#include "tamc_ssm_api.h"
#include "ssmc.h"

#include <openssl/rsa.h>
#include <openssl/err.h>
#include <openssl/engine.h>
#include <openssl/x509.h>
//#include "../../../crypto/fec/fec_lcl.h"
#include <openssl/ec.h>
#include <openssl/ecdsa.h>

/** @brief tamc_verify_digest - Verify that the signature is valid for the given hash. Blocking.
 * @param *vsession  - client/server session used to request services.
 * @param *key_owner - Application that created/owns the key
 * @param *key_name  - Filename containing the public key
 * @param  algo      - Hash Algorithm used to generate digest
 * @param *digest    - buffer containing the digset
 * @param  digest_len - Length of the digest
 * @param *signature  - buffer to signature to be verified
 * @param  signature_len - length of signature
 * @return TAM_SUCCESS or error code.
 */
static tam_status_e tamc_verify_digest(void *session,
                                       char *key_owner,
                                       char *key_name,
                                       tam_hash_algo_e algo,
                                       uint8_t * digest,
                                       uint32_t digest_len,
                                       uint8_t * signature,
                                       uint32_t signature_len)
{
    tam_status_e rc = TAM_SUCCESS;
    uint8_t *sig_p = NULL;
    EVP_PKEY *verifier_key = NULL;
    tam_object_attr_t *attrs = NULL;
    int verify_rc = 0;          // failed
    EVP_MD *md = NULL;
    int md_type = 0;
    int ecdsa_not_used = 0;
    RSA_METHOD *tmp_rsa_meth = NULL;
    const ECDSA_METHOD *method_data = NULL;
    const ECDSA_METHOD *tmp_method_data = NULL;
    ENGINE *e = NULL;
    /*
     * Parameter check
     */
    if (!session || !key_name || !signature
        || !digest || (signature_len == 0)) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    sig_p = signature;

    rc = tamc_so_get_attr(session, key_owner, key_name, &attrs);
    if (rc != TAM_SUCCESS) {
        goto end;
    }
    if ((attrs->object_type != TAM_SO_TYPE_RSA_KEYPAIR) &&
        (attrs->object_type != TAM_SO_TYPE_ECC_KEYPAIR)) {
        tam_debug(TAMD_DBG_DEV_ERR, "Unsupported Key type.  "
                  " : %d", attrs->object_type);
        rc = TAM_ERROR_INCORRECT_KEY_TYPE;
        goto end;
    }

    /*
     * Read and Load key
     */
    rc = ssmc_load_key(session, key_name, key_owner, &verifier_key);
    if ((rc != TAM_SUCCESS) || !verifier_key) {
        tam_debug(TAMD_DBG_DEV, "Error build verifier key.  " " : 0x%x", rc);
        goto end;
    }
    // Verify Signature
    if (attrs->object_type == TAM_SO_TYPE_RSA_KEYPAIR) {
        // if RSA key, process rsa verify/
        md = (EVP_MD *)ssmc_tam_hash_type_to_evp(algo);
        if (!md) {
            rc = TAM_ERROR_UNSUPPORTED_HASH_ALGO;
            goto end;
        }
        md_type = EVP_MD_type(md);

        /*
         * Temporarily set the the RSA method to the default meth
         * otherwise we run recursively through the engine.
         */
        tmp_rsa_meth = (RSA_METHOD *)verifier_key->pkey.rsa->meth;
        verifier_key->pkey.rsa->meth = ssmc_get_rsa_method();

        /* Verify buffer */
        verify_rc =
            RSA_verify(md_type, (const unsigned char *)digest, digest_len,
                       sig_p, signature_len, verifier_key->pkey.rsa);
        /*
         * Restore method
         */
        verifier_key->pkey.rsa->meth = tmp_rsa_meth;
    } else if (attrs->object_type == TAM_SO_TYPE_ECC_KEYPAIR) {
        /*
         * Temporarily set the the ECDSA method to the default meth
         * otherwise we run recursively through the engine.
         */
        /*
         * Set ECDSA method temporarily back to openssl methods.
         */

        e = ENGINE_get_default_ECDSA();
        tmp_method_data = ECDSA_OpenSSL();
        method_data = ENGINE_get_ECDSA(e);
        ENGINE_set_ECDSA(e, tmp_method_data);

        verify_rc = ECDSA_verify(ecdsa_not_used, digest, digest_len,
                                 sig_p, signature_len, verifier_key->pkey.ec);
        /*
         * Reset to engine method
         */
        ENGINE_set_ECDSA(e, method_data);
    }

    if (verify_rc == 1) {
        rc = TAM_SUCCESS;
    } else {
        rc = TAM_ERROR_SIGN_VERIFY_FAIL;
    }

 end:
    free(attrs);
    if (verifier_key)
        EVP_PKEY_free(verifier_key);
    return (rc);
}

/*
 * Sign Digest
 */
/** @brief tamc_sign_digest - Blocking call to create a signature for the given hash.
 * @param *vsession  - client/server session used to request services.
 * @param *key_owner - Owing application which created the key
 * @param *key_name  - Filename containing the private key
 * @param  algo      - Hash Algorithm used to generate digest
 * @param *digest    - buffer containing the digest
 * @param  digest_len    - Length of the digest
 * @param **signature    - Pointer to buffer which will hold signature.
 * @param  signature_len - length of signature
 * @return TAM_SUCCESS or error code.
 */
static tam_status_e tamc_sign_digest(void *vsession,
                                     char *key_owner,
                                     char *key_name,
                                     tam_hash_algo_e algo,
                                     uint8_t * digest,
                                     uint32_t digest_len,
                                     uint8_t ** signature,
                                     uint32_t * signature_len)
{
    tam_status_e rc = TAM_SUCCESS;
    uint8_t *sig_buff = NULL;
    EVP_PKEY *signer_key = NULL;
    const ECDSA_METHOD *method_data = NULL;
    const ECDSA_METHOD *tmp_method_data = NULL;
    ENGINE *e = NULL;
    tam_object_attr_t *attrs = NULL;
    uint8_t *ptr_sig_buf = NULL;
    uint32_t sig_len = 0;
    int sign_rc = 0;
    int ecdsa_not_used = 0;
    EVP_MD *md = NULL;
    int md_type = 0;
    /*
     * Parameter check
     */
    if (!vsession || !key_name || !digest || (digest_len == 0) ||
        !signature || !signature_len) {
        rc = TAM_ERROR_INVALID_PARM;
        goto end;
    }

    rc = tamc_so_get_attr(vsession, key_owner, key_name, &attrs);
    if (rc != TAM_SUCCESS) {
        goto end;
    }

    if ((attrs->object_type != TAM_SO_TYPE_RSA_KEYPAIR) &&
        (attrs->object_type != TAM_SO_TYPE_ECC_KEYPAIR)) {
        tam_debug(TAMD_DBG_DEV_ERR, "Unsupported Key type.  "
                  " : %d", attrs->object_type);
        rc = TAM_ERROR_INCORRECT_KEY_TYPE;
        goto end;
    }

    /*
     * Read and Load key
     */
    rc = ssmc_load_key(vsession, key_name, key_owner, &signer_key);
    if ((rc != TAM_SUCCESS) || !signer_key) {
        tam_debug(TAMD_DBG_DEV, "Error getting signer key.  " " : 0x%x", rc);
        goto end;
    }
    // Calculate Signature
    if (attrs->object_type == TAM_SO_TYPE_RSA_KEYPAIR) {
        // if RSA key, process rsa sign
        md = (EVP_MD *)ssmc_tam_hash_type_to_evp(algo);
        if (!md) {
            rc = TAM_ERROR_UNSUPPORTED_HASH_ALGO;
            goto end;
        }
        md_type = EVP_MD_type(md);

        sig_buff = calloc(1, RSA_size(signer_key->pkey.rsa));
        if (!sig_buff) {
            rc = TAM_ERROR_NO_RESOURCES;
            goto end;
        }
        ptr_sig_buf = sig_buff;

        sign_rc = RSA_sign(md_type, digest, digest_len,
                           ptr_sig_buf, &sig_len, signer_key->pkey.rsa);

        if (sign_rc == 1) {
            rc = TAM_SUCCESS;
            goto end;
        } else {
            rc = TAM_ERROR_SIGN_VERIFY_FAIL;
        }
    } else if (attrs->object_type == TAM_SO_TYPE_ECC_KEYPAIR) {
        /*
         * if we have an ECC key, then ecdsa sign with it.
         */
        sig_len = ECDSA_size(signer_key->pkey.ec);
        if (sig_len == 0) {
            rc = TAM_ERROR_SIGN_VERIFY_FAIL;
            tam_debug(TAMD_DBG_DEV_ERR,
                      "Error during ECDSA sign. Signature length = 0.");
            goto end;
        }

        /*
         * allocate ecdsa sig buffer
         */
        sig_buff = OPENSSL_malloc(sig_len);
        if (!signature) {
            rc = TAM_ERROR_NO_RESOURCES;
            goto end;
        }
        ptr_sig_buf = sig_buff;

        /*
         * Set ECDSA method temporarily back to openssl methods.
         */

        e = ENGINE_get_default_ECDSA();
        if (e) {
            tmp_method_data = ECDSA_OpenSSL();
            method_data = ENGINE_get_ECDSA(e);
            ENGINE_set_ECDSA(e, tmp_method_data);
        }
        sign_rc = ECDSA_sign(ecdsa_not_used, digest, digest_len,
                             ptr_sig_buf, &sig_len,
                             (EC_KEY *)signer_key->pkey.ec);
        if (e) {
            /*
             * Reset to engine method
             */
            ENGINE_set_ECDSA(e, method_data);
        }
        if (sign_rc != 1) {
            rc = TAM_ERROR_SIGN_VERIFY_FAIL;
            tam_debug(TAMD_DBG_DEV_ERR, "Error during ECDSA sign.");
            goto end;
        }
    }
    // Set results
    *signature = sig_buff;
    *signature_len = sig_len;

 end:
    if (signer_key)
        EVP_PKEY_free(signer_key);
    free(attrs);

    return (rc);
}
