/*
 *  ctamc_tam_config.c
 *  Openssl Engine Cisco Trusted Anchor Module client Device Tests
 *
 *  January, 2014
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ctamc_tam_config.h"

#ifndef USE_CISCO_SSM

/* Client Interface APIs */
# include "tamc_api.h"
# include "tam_types.h"

/* Management Interface APIs */
# include "tamm_app_mgmt_api.h"
# include "tamm_dev_mgmt_api.h"
# include "safe_str_lib.h"
# include "safe_mem_lib.h"
# include "tam_common.h"
# include "tam_linux_ipc.h"
# include "tam_linux_mem.h"

static const char *ctamc_usage[] = {
    "usage: ctamc args\n",
    "Configure SW or HW TAM device.\n",
    "Each application will create one session. TAM Sever process needs to be already running.\n",
    "\n",
    " -d  <0,1>           - Debugs 0-OFF, 1- ON.\n",
    " -hw arg1 arg2 arg3  - hw device name, app name, app password.\n",
    " -sw arg1 arg2 arg3  - sw device name, app name, app password.\n",
    " -deprov arg         - deprovison device device name.\n",
    NULL
};

tam_status_e ct_unprovision_all_apps(const char *device_name);
tam_status_e config_devices(int dev_type, int num_of_dev,
                            int num_apps_per_device, char *dev_name,
                            char *app_name, char *app_passwd);
void deprov_device(char *device_name);

void *my_mem_alloc(size_t size)
{
    return (calloc(1, size));
}

void my_free(void *ptr)
{
    free(ptr);
}

tam_status_e config_devices(int dev_type, int num_of_dev,
                            int num_apps_per_device, char *dev_name,
                            char *app_name, char *app_passwd)
{

    tam_credentials_t credentials;
    tam_device_attr_t dev_attrib;
    tam_status_e rc = TAM_SUCCESS;
    uint32_t dev_context = 0;

    if (!dev_name || !app_name || !app_passwd)
        goto error;

    dev_attrib.perm1 =
        TAM_DEVICE_AUTH | TAM_SUDI_USE | TAM_SO_READ | TAM_SO_WRITE |
        TAM_CRYPTO_ENG | TAM_RNG_USE | TAM_EEPROM_USE;
    dev_attrib.perm2 = 0;
    dev_attrib.object_quota = 0;

    // register devs
    rc = tamm_register_device(dev_name, dev_type, dev_context);
    if (rc != TAM_SUCCESS)
        goto error;

    strncpy_s(credentials.type.userpass.appid, TAM_APPID_LEN_MAX, app_name,
              TAM_APPID_LEN_MAX);
    strncpy_s(credentials.type.userpass.pw, TAM_PW_LEN_MAX, app_passwd,
              TAM_PW_LEN_MAX);
    rc = tamm_register_app(TAM_CREDENTIAL_PW, &credentials, dev_name, 1, 20);
    if (rc != TAM_SUCCESS)
        goto error;

    rc = tamm_app_set_attr(app_name, dev_name, &dev_attrib);
    if (rc != TAM_SUCCESS)
        goto error;

    return (rc);

 error:
    // clean up tam devices
    deprov_device(dev_name);
    return (rc);
}

/*
 * unprovision_all_apps()
 */
tam_status_e ct_unprovision_all_apps(const char *device_name)
{
    tam_status_e rc = TAM_SUCCESS;
    void *dev_app_context = NULL;
    uint32_t num_apps;
    uint32_t i;

    /*
     * Enumerate apps with permissions on device and delete them
     */
    rc = tamm_get_device_app_context(device_name, &num_apps,
                                     &dev_app_context);

    if (rc != TAM_SUCCESS)
        goto error;

    for (i = 0; i < num_apps; i++) {
        rc = tamm_unregister_app(tamm_enumerate_device_app
                                 (dev_app_context, i), device_name);
    }

 error:
    if (dev_app_context)
        free(dev_app_context);

    return (rc);
}

void deprov_device(char *device_name)
{
    if (!device_name)
        return;

    ct_unprovision_all_apps(device_name);
    tamm_unregister_device(device_name);
}

void print_help()
{
    const char **pp2;

    for (pp2 = ctamc_usage; (*pp2 != NULL); pp2++)
        printf("%s", *pp2);
}

#endif

int main(int argc, char **argv)
{
#ifndef USE_CISCO_SSM

    tam_status_e status = TAM_SUCCESS;
    int badops = 0;
    char dev_name[TAM_DEVICE_NAME_MAX_LEN];
    char app_name[TAM_APPID_LEN_MAX];
    char app_passwd[TAM_PW_LEN_MAX];

    memzero_s(dev_name, TAM_DEVICE_NAME_MAX_LEN);
    memzero_s(app_name, TAM_APPID_LEN_MAX);
    memzero_s(app_passwd, TAM_PW_LEN_MAX);

    argc--;
    argv++;
    if (argc < 1)
        print_help();

    /*
     * Register PD linux IPC and memory mgmt vectors
     */
    tam_register_ipc_cb(tam_linux_ipc_init,
                        tam_linux_ipc_cleanup,
                        tam_linux_ipc_open,
                        tam_linux_ipc_close,
                        tam_linux_ipc_send, tam_linux_ipc_receive);

    tam_register_mem_cb(my_mem_alloc, my_free);

    while (argc >= 1) {
        if (strcmp(*argv, "-d") == 0) {
            if (--argc < 1)
                goto bad;
            if (atoi(*(++argv)) == 0) {
                tam_set_debug_level(0); // /Debug off
            } else {
                tam_set_debug_level(TAMC_DBG_ERR_ALL); // /Debug On
            }
        } else if (strcmp(*argv, "-deprov") == 0) {
            if (--argc < 1)
                goto bad;
            deprov_device(*(++argv));
        } else if (strcmp(*argv, "-hw") == 0) {
            if (--argc < 3)
                goto bad;
            strncpy_s(&dev_name[0], TAM_DEVICE_NAME_MAX_LEN, *(++argv),
                      TAM_DEVICE_NAME_MAX_LEN);
            argc--;
            strncpy_s(&app_name[0], TAM_APPID_LEN_MAX, *(++argv),
                      TAM_APPID_LEN_MAX);
            argc--;
            strncpy_s(&app_passwd[0], TAM_PW_LEN_MAX, *(++argv),
                      TAM_PW_LEN_MAX);
            status =
                config_devices(DEVICE_TYPE_ACT2, 1, 1, dev_name, app_name,
                               app_passwd);
            printf("Device configuration status = %s\n",
                   tam_status_string(status));
        } else if (strcmp(*argv, "-sw") == 0) {
            if (--argc < 3)
                goto bad;
            strncpy_s(&dev_name[0], TAM_DEVICE_NAME_MAX_LEN, *(++argv),
                      TAM_DEVICE_NAME_MAX_LEN);
            argc--;
            strncpy_s(&app_name[0], TAM_APPID_LEN_MAX, *(++argv),
                      TAM_APPID_LEN_MAX);
            argc--;
            strncpy_s(&app_passwd[0], TAM_PW_LEN_MAX, *(++argv),
                      TAM_PW_LEN_MAX);
            status =
                config_devices(DEVICE_TYPE_SWTAM, 1, 1, dev_name, app_name,
                               app_passwd);
            printf("Device configuration status = %s\n",
                   tam_status_string(status));
        } else {
 bad:
            printf("unknown option %s\n", *argv);
            badops = 1;
            break;
        }
        argc--;
        argv++;
    }                           // while loop

    if (badops) {
        print_help();
        goto err;
    }

 err:
#endif
    exit(0);
}
