/*
 *  ctamc_test_certs.c
 *  Openssl Engine Cisco Trusted Anchor CLient Module
 *  Test storing and getting certificates and certificate chain
 *
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <openssl/engine.h>
#include <openssl/rand.h>
#include "../../vendor_defns/ctamc.h"
#include "../ctame.h"
#include "tam_types.h"
#ifdef USE_CISCO_SSM
# include "ssm.h"
#else
# include "tamc_api.h"
#endif
#include "safe_mem_lib.h"
#include "safe_str_lib.h"
#include "ctamc_utility.c"
#include <sys/stat.h>

#define CERT_FILE 1             // Certificate file type
#define CHAIN_CERT_FILE 2       // Cert Chain file
static BIO *bio_out = NULL;
int test_ctamc_store_cert(void *session, char *id, unsigned char *filename,
                          int chain_type);
int test_ctamc_get_cert(ENGINE *e, unsigned char *filename, int chain_type);
int read_platform_file(unsigned char *filename,
                       unsigned char **cert_raw_buffer,
                       uint32_t * cert_file_size);

void print_help();

int main(int argc, char *argv[])
{
    ENGINE *e_ctamc = NULL;
    char *app_id = NULL;
    char *app_password = NULL;
    void *ctx = NULL;
    void *session_id = NULL;
    int rc = 0;
    char *in_cert_name = NULL;
    char *in_chain_name = NULL;
    char *out_cert_name = NULL;
    char *out_chain_name = NULL;
    ctamc_context_t *ctamc_session_ctx;

    bio_out = BIO_new_fp(stdout, BIO_NOCLOSE);

    BIO_printf(bio_out, "Certificate Test entered\n");

    argc--;
    argv++;
    if (argc < 4) {
        print_help();
        goto end;
    }

    while (argc >= 1) {
        if (strcmp(*argv, "-app_id") == 0) {
            if (--argc < 2)
                goto bad;

            app_id = *(++argv);
            argc--;
            app_password = *(++argv);
        } else if (strcmp(*argv, "-in_cert") == 0) {
            if (--argc < 1)
                goto bad;
            if (!in_cert_name) {
                in_cert_name = *(++argv);
            } else {
                BIO_printf(bio_out, "Cert name already set.\n");
                goto end;
            }
        } else if (strcmp(*argv, "-in_chain") == 0) {
            if (--argc < 1)
                goto bad;
            if (!in_chain_name) {
                in_chain_name = *(++argv);
            } else {
                BIO_printf(bio_out, "Chain cert name already set.\n");
                goto end;
            }
        } else if (strcmp(*argv, "-out_cert") == 0) {
            if (--argc < 1)
                goto bad;
            if (!out_cert_name) {
                out_cert_name = *(++argv);
            } else {
                BIO_printf(bio_out, "Cert name already set.\n");
                goto end;
            }
        } else if (strcmp(*argv, "-out_chain") == 0) {
            if (--argc < 1)
                goto bad;
            if (!out_chain_name) {
                out_chain_name = *(++argv);
            } else {
                BIO_printf(bio_out, "Chain Cert name already set.\n");
                goto end;
            }
        } else {
 bad:
            print_help();
            goto end;
        }
        argc--;
        argv++;
    }                           // while

    load_ctamc_engine(&e_ctamc);
    if (!e_ctamc) {
        BIO_printf(bio_out, "load_ctamc_engine failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }

    ctamc_login(app_id, app_password, &ctx, &session_id);
    if (session_id)
        BIO_printf(bio_out, "Session context 0x%x.\n", *(int *)session_id);
    if (!ctx) {
        BIO_printf(bio_out, "login failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }

    ctamc_session_ctx = calloc(1, sizeof(ctamc_context_t));
    ctamc_session_ctx->session_id = session_id;
    strncpy_s(&ctamc_session_ctx->app_id[0], TAM_APPID_LEN_MAX, app_id,
              TAM_APPID_LEN_MAX);
    ctamc_item_t ctamc_item;
    ctamc_item.c = (unsigned char *)ctamc_session_ctx;
    ctamc_item.len = sizeof(ctamc_context_t);

    rc = ENGINE_ctrl_cmd(e_ctamc, "CTAMC_SET_CONTEXT", 0, &ctamc_item, 0, 0);

    if (rc == 1) {
        BIO_printf(bio_out, "Passed set session context\n");
    } else {
        BIO_printf(bio_out, "Failed set session context\n");
        rc = 1;
        goto end;
    }

    // store cert
    if (in_cert_name) {
        rc = test_ctamc_store_cert(session_id, app_id,
                                   (unsigned char *)in_cert_name,
                                   TAM_SO_TYPE_CERT);
        if (!rc) {
            BIO_printf(bio_out,
                       "Failed to store certificate in TAM device %s\n",
                       in_cert_name);
            goto end;
        }
    }
    // store chain
    if (in_chain_name) {
        rc = test_ctamc_store_cert(session_id, app_id,
                                   (unsigned char *)in_chain_name,
                                   TAM_SO_TYPE_CERTCHAIN);
        if (!rc) {
            BIO_printf(bio_out,
                       "Failed to store certificate chain in TAM device %s\n",
                       in_chain_name);
            goto end;
        }
    }
    // get cert
    if (out_cert_name) {
        rc = test_ctamc_get_cert(e_ctamc, (unsigned char *)out_cert_name,
                                 TAM_SO_TYPE_CERT);
        if (!rc) {
            BIO_printf(bio_out,
                       "Failed to read certificate stored in TAM device %s\n",
                       out_cert_name);
            goto end;
        }
    }
    // get chain
    if (out_chain_name) {
        rc = test_ctamc_get_cert(e_ctamc, (unsigned char *)out_chain_name,
                                 TAM_SO_TYPE_CERTCHAIN);
        if (!rc) {
            BIO_printf(bio_out,
                       "Failed to read certificate chain stored in TAM device %s\n",
                       out_chain_name);
            goto end;
        }
    }

    BIO_printf(bio_out, "Test complete!\n");

 end:
    if (ctx) {
        ctamc_logout(&session_id);
        tamc_delete_context(&ctx);
    }
    if (rc == 1)
        rc = 0;                 // exit pass
    else
        rc = 1;                 // exit fail
    exit(rc);
}

int test_ctamc_store_cert(void *session, char *app_id,
                          unsigned char *filename, int chain_type)
{
    int rc = 0;
    tam_status_e status = TAM_ERROR_UNSPECIFIED;
    unsigned char *cert_raw_buffer = NULL;
    uint32_t cert_file_size = 0;
    tam_object_attr_t *attr = NULL;

    // read file from disk
    read_platform_file(filename, &cert_raw_buffer, &cert_file_size);

    // prep to write data to TAM device
    if (cert_raw_buffer) {
        // setup stored obj attr
        attr = calloc(1, sizeof(tam_object_attr_t));
        if (!attr) {
            free(cert_raw_buffer);
            cert_raw_buffer = NULL;
            return 0;
        }
        attr->nv_flag = TAM_RAM_OBJECT;
        attr->csp_flag = TAM_CSP_ZEROIZE;
        attr->encryption_flag = 0;
        attr->object_size = cert_file_size;
        attr->object_type = chain_type;
        strncpy_s(attr->object_name, TAM_OBJECT_NAME_LEN_MAX,
                  (const char *)filename, TAM_OBJECT_NAME_LEN_MAX);
        strncpy_s(attr->object_owner, TAM_APPID_LEN_MAX, app_id,
                  TAM_APPID_LEN_MAX);

        status = tamc_so_create(session, attr);
        if (status == TAM_SUCCESS) {
            status =
                tamc_so_write(session, app_id, (char *)filename,
                              cert_raw_buffer, cert_file_size);
        }
    }

    if (status == TAM_SUCCESS) {
        rc = 1;                 // pass
    }
    free(attr);
    free(cert_raw_buffer);
    return (rc);
}

int test_ctamc_get_cert(ENGINE *e, unsigned char *filename, int chain_type)
{
    int rc = 0;
    int ctame_rc = 0;
    ctamc_item_t ctame_dcert;

    X509 *ctame_x509_cert = NULL;

    long ctame_cert_len;

    long ctame_bytes_to_read = 0;
    int ctame_chain_index = 1;

    const unsigned char *ctame_cert_ptr = NULL;

    ctamc_file_attr_t ctamc_cert_file_attr;

    memzero_s(&ctamc_cert_file_attr, sizeof(ctamc_file_attr_t));
    ctame_dcert.c = (unsigned char *)&ctamc_cert_file_attr;
    /*
     * Retrieve certificate chain stored in TAM.
     */

    memcpy(ctamc_cert_file_attr.filename, filename,
           strlen((const char *)filename));
    if (chain_type == TAM_SO_TYPE_CERT) {
        ctame_rc =
            ENGINE_ctrl_cmd(e, "CTAMC_CMD_GET_CERT", 0, &ctame_dcert, 0, 0);
        if (!ctame_rc) {
            BIO_printf(bio_out, "ERROR Failed to read CTAME cert.\n");
            return rc;
        }
    } else if (chain_type == TAM_SO_TYPE_CERTCHAIN) {
        ctame_rc =
            ENGINE_ctrl_cmd(e, "CTAMC_CMD_GET_CERT_CHAIN", 0, &ctame_dcert, 0,
                            0);
        if (!ctame_rc) {
            BIO_printf(bio_out, "ERROR Failed to read CTAME cert chain\n");
            return rc;
        }
    } else {
        // error out
        BIO_printf(bio_out, "Unknown file type. Not cert nor chain.\n");
        OPENSSL_free(ctamc_cert_file_attr.file_buffer);
        return rc;
    }

    ctame_bytes_to_read = ctamc_cert_file_attr.file_size;
    ctame_cert_ptr = ctamc_cert_file_attr.file_buffer;
    while (ctame_bytes_to_read) {
        /*
         * DER Formatted files begins with a tag followed by a length field
         * length of the certificate
         */
        ctame_cert_len = 0;
        ctame_cert_len = ctame_cert_ptr[02] << 8;
        ctame_cert_len += ctame_cert_ptr[03];
        ctame_cert_len += 4;

        ctame_x509_cert =
            d2i_X509(NULL, (const unsigned char **)&ctame_cert_ptr,
                     ctame_cert_len);

        if (ctame_x509_cert == NULL) {
            BIO_printf(bio_out,
                       "Error retrieving Cert chain index %d Certificate\n",
                       ctame_chain_index);
            OPENSSL_free(ctamc_cert_file_attr.file_buffer);
            ctamc_cert_file_attr.file_buffer = NULL;
            return rc;
        }

        BIO_printf(bio_out, "Read Chain index #%d Certificate\n\n",
                   ctame_chain_index);
        X509_print(bio_out, ctame_x509_cert);

        ctame_bytes_to_read =
            ctamc_cert_file_attr.file_size - (ctame_cert_ptr -
                                              ctamc_cert_file_attr.file_buffer);
        ctame_chain_index++;
        if (ctame_x509_cert)
            X509_free(ctame_x509_cert);
        ctame_x509_cert = NULL;
    }

    OPENSSL_free(ctamc_cert_file_attr.file_buffer);
    rc = 1;
    return (rc);
}

void print_help()
{
    BIO_printf(bio_out, "Stores DER formatted certificates and chains.\n");
    BIO_printf(bio_out,
               "./ctamc_test_certs -app_id user_id password option\n");
    BIO_printf(bio_out,
               "-out_cert obj_name: read certificate file in obj_name.\n");
    BIO_printf(bio_out,
               "-in_cert obj_name: store certificate in TAM Device as obj_name.\n");
    BIO_printf(bio_out, "-out_chain chain_filename.\n");
    BIO_printf(bio_out, "-in_chain chain_filename.\n");
    BIO_printf(bio_out,
               "./ctamc_test_certs -app_id app1 passwd1 -in_cert cert-name -in_chain chain-name ");
    BIO_printf(bio_out, "-out_cert cert-name -out_chain chain-name\n");
}

int read_platform_file(unsigned char *filename,
                       unsigned char **cert_raw_buffer,
                       uint32_t * cert_file_size)
{
    FILE *fdr;
    int file_size;
    unsigned char *file_contents = NULL;
    struct stat s_cert_stat;
    int rc = 0;
    size_t cnt;

    fdr = fopen((const char *)filename, "r");
    if (fdr == NULL) {
        return (0);
    }

    /* Get space in ram for encrypted file */
    if (fstat(fileno(fdr), &s_cert_stat) == -1) {
        perror(strerror(errno));
        fclose(fdr);
        *cert_raw_buffer = 0;
        *cert_file_size = 0;
        return (0);
    }
    file_size = s_cert_stat.st_size;
    file_contents = calloc(1, file_size);
    if (!file_contents) {
        file_size = 0;
        rc = 0;
    } else {

        /* Read in encrypted file */
        cnt = fread(file_contents, file_size, 1, fdr);
        if (cnt != 1) {
            free(file_contents);
            file_contents = 0;
            file_size = 0;
            rc = 0;
        }
    }
    fclose(fdr);

    *cert_raw_buffer = file_contents;
    *cert_file_size = file_size;

    return (rc);
}
