/*
 *  ctamc_test_ec_crypto.c
 *  Openssl Engine Cisco Trusted Anchor CLient Module EC Tests
 *
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <openssl/engine.h>
#include "../../vendor_defns/ctamc.h"
#include "../ctame.h"
#include "tam_types.h"
#ifdef USE_CISCO_SSM
# include "ssm.h"
#else
# include "tamc_api.h"
#endif
#include "safe_str_lib.h"
#include "safe_mem_lib.h"
#include "ctamc_utility.c"

/* hard coded for now */
static int list_max = 0;
static BIO *bio_out = NULL;
int test_ctamc_ec_keypairgen_test(ENGINE *e_ctamc);
int test_ec_signature(ENGINE *e, char *key_id, int eng_as_deflt);
int test_ctamc_evp_digest_sign_verify(ENGINE *e);
int test_ctamc_evp_sign_verify(ENGINE *e);
void print_help();

#define the_message "Rock Chalk Jayhawk!! Go KU!!!!"

ctamc_ecc_key_attr_t ec_key_parms_list[] = {
    {TAM_EC_CURVE_P256, 1, 1, 0, "ecc-P256"},
    {TAM_EC_CURVE_P384, 1, 1, 0, "ecc-P384"},
    {TAM_EC_CURVE_P521, 1, 1, 0, "ecc-P521"},
    {TAM_EC_CURVE_P192, 1, 1, 0, "ecc-P192"},
    {TAM_EC_CURVE_BRAINPOOLP256, 1, 1, 0, "ecc-BRAINPOOLP256"},
    {TAM_EC_CURVE_BRAINPOOLP384, 1, 1, 0, "ecc-BRAINPOOLP384"},
    {TAM_EC_CURVE_BRAINPOOLP512, 1, 1, 0, "ecc-BRAINPOOLP512"},
};
#define FLIST_MAX 3 // Subset of fips curves at beginning of struct
#define KLIST_MAX sizeof(ec_key_parms_list)/sizeof(ctamc_ecc_key_attr_t) // All curves

int test_load_private_key(ENGINE *e, char *key_id)
{
    int rc = 0;
    EVP_PKEY *evp_key = NULL;
    // Load generate keys and do some work with them
    evp_key = ENGINE_load_private_key(e, key_id, NULL, NULL);

    if (!evp_key) {
        BIO_printf(bio_out, "Error retrieving %s key pair\n", key_id);
        rc = 0;
    } else {
        rc = 1;
        BIO_printf(bio_out, "Loaded %s private key.\n", key_id);
    }
    if (evp_key)
        EVP_PKEY_free(evp_key);
    evp_key = NULL;
    return (rc);
}

int main(int argc, char *argv[])
{
    ENGINE *e_ctamc = NULL;
    char app_id[TAM_APPID_LEN_MAX] = "engine_9";
    char app_password[TAM_PW_LEN_MAX] = "engine_9_passwd";
    void *ctx = NULL;
    void *session_id = NULL;
    int engine_as_default = 1;  // set engine as default
    int rc = 0;
    int env_fips_val = 0;
    ctamc_context_t *ctamc_session_ctx;

    list_max = KLIST_MAX; // all curves
    if (getenv("OPENSSL_FIPS")) {
#ifdef OPENSSL_FIPS
        env_fips_val = atoi(getenv("OPENSSL_FIPS"));
        if (env_fips_val == 0) {
            env_fips_val = 1;
        }
        if (!FIPS_mode_set(env_fips_val)) {
            ERR_load_crypto_strings();
            ERR_print_errors(BIO_new_fp(stderr, BIO_NOCLOSE));
            exit(1);
        }
        list_max = FLIST_MAX; // use fips subset of curves
	printf("Running in FIPS mode...\n");
#else
        fprintf(stderr, "FIPS mode not supported.\n");
        exit(1);
#endif
    }

    bio_out = BIO_new_fp(stdout, BIO_NOCLOSE);

    BIO_printf(bio_out, "ECC Test entered\n");
    argc--;
    argv++;
    if (argc == 3) {
        if (strcmp(*argv, "-app_id") == 0) {
            if (--argc < 2)
                goto bad;

            strncpy_s(&app_id[0], TAM_APPID_LEN_MAX, *(++argv),
                      TAM_APPID_LEN_MAX);
            argc--;
            strncpy_s(&app_password[0], TAM_PW_LEN_MAX, *(++argv),
                      TAM_PW_LEN_MAX);

        } else {
 bad:
            printf("unknown option %s. Using default user.\n", *argv);
        }
    }

    load_ctamc_engine(&e_ctamc);
    if (!e_ctamc) {
        BIO_printf(bio_out, "load_ctamc_engine failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }

    ctamc_login(app_id, app_password, &ctx, &session_id);
    if (session_id)
        BIO_printf(bio_out, "Session context 0x%x.\n", *(int *)session_id);
    if (!ctx) {
        BIO_printf(bio_out, "login failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }

    ctamc_session_ctx = calloc(1, sizeof(ctamc_context_t));
    ctamc_session_ctx->session_id = session_id;
    strncpy_s(&ctamc_session_ctx->app_id[0], TAM_APPID_LEN_MAX, app_id,
              TAM_APPID_LEN_MAX);
    ctamc_item_t ctamc_item;
    ctamc_item.c = (unsigned char *)ctamc_session_ctx;
    ctamc_item.len = sizeof(ctamc_context_t);

    rc = ENGINE_ctrl_cmd(e_ctamc, "CTAMC_SET_CONTEXT", 0, &ctamc_item, 0, 0);

    if (rc == 1) {
        BIO_printf(bio_out, "Passed set session context\n");
    } else {
        BIO_printf(bio_out, "Failed set session context\n");
        rc = 1;
        goto end;
    }
    // EC Key Gen
    rc = test_ctamc_ec_keypairgen_test(e_ctamc);
    if (!rc) {
        BIO_printf(bio_out, "EC Key Gen failed.\n");
        goto end;
    }

    BIO_printf(bio_out, "EC Key Gen completed.\n");

    // Test sign
    rc = test_ec_signature(e_ctamc, "ecc-P256", 0);
    if (!rc) {
        BIO_printf(bio_out, "EC signature test failed.\n");
        goto end;
    }
    
    if (!env_fips_val) {
        rc = test_ec_signature(e_ctamc, "ecc-BRAINPOOLP256", 0);
        if (!rc) {
            BIO_printf(bio_out, "EC signature test failed.\n");
            goto end;
        }
        rc = test_ec_signature(e_ctamc, "ecc-BRAINPOOLP384", 0);
        if (!rc) {
            BIO_printf(bio_out, "EC signature test failed.\n");
            goto end;
        }
        rc = test_ec_signature(e_ctamc, "ecc-BRAINPOOLP512", 0);
        if (!rc) {
            BIO_printf(bio_out, "EC signature test failed.\n");
            goto end;
        }
    }
    BIO_printf(bio_out, "EC signature test Passed.\n");

    rc = test_ctamc_evp_digest_sign_verify(e_ctamc);
    if (!rc) {
        BIO_printf(bio_out, "EC test_ctamc_evp_digest_sign_verify failed.\n");
        goto end;
    }
    BIO_printf(bio_out, "EC test_ctamc_evp_digest_sign_verify Passed.\n");

    rc = test_ctamc_evp_sign_verify(e_ctamc);
    if (!rc) {
        BIO_printf(bio_out, "EC test_ctamc_evp_sign_verify failed.\n");
        goto end;
    }
    BIO_printf(bio_out, "\nECtest_ctamc_evp_sign_verify Passed.\n");

    // Repeat Test with the engine as the default provider.
    BIO_printf(bio_out, "\nRepeating Tests with ENGINE_set_default_ECDSA.\n");
    rc = test_ec_signature(e_ctamc, "ecc-P256", engine_as_default);
    if (!rc) {
        BIO_printf(bio_out, "EC signature test failed.\n");
        goto end;
    }
    if (!env_fips_val) {
        rc = test_ec_signature(e_ctamc, "ecc-BRAINPOOLP256", engine_as_default);
        if (!rc) {
            BIO_printf(bio_out, "EC signature test failed.\n");
            goto end;
        }
        rc = test_ec_signature(e_ctamc, "ecc-BRAINPOOLP384", engine_as_default);
        if (!rc) {
            BIO_printf(bio_out, "EC signature test failed.\n");
            goto end;
        }
        rc = test_ec_signature(e_ctamc, "ecc-BRAINPOOLP512", engine_as_default);
        if (!rc) {
            BIO_printf(bio_out, "EC signature test failed.\n");
            goto end;
        }
    }
    BIO_printf(bio_out, "EC signature test Passed.\n");

    rc = test_ctamc_evp_digest_sign_verify(e_ctamc);
    if (!rc) {
        BIO_printf(bio_out, "EC test_ctamc_evp_digest_sign_verify failed.\n");
        goto end;
    }
    BIO_printf(bio_out, "EC test_ctamc_evp_digest_sign_verify Passed.\n");

    rc = test_ctamc_evp_sign_verify(e_ctamc);
    if (!rc) {
        BIO_printf(bio_out, "EC test_ctamc_evp_sign_verify failed.\n");
        goto end;
    }
    BIO_printf(bio_out, "\nECtest_ctamc_evp_sign_verify Passed.\n");

    BIO_printf(bio_out, "EC Engine Test PASSED!.\n");
 end:

    BIO_printf(bio_out, "Test complete!\n");

    if (ctx) {
        /* cleanup stored files */
        int loop_idx;
        for (loop_idx = 0; loop_idx < list_max; loop_idx++) {
            tamc_so_delete(session_id, app_id,
                           ec_key_parms_list[loop_idx].key_name);
        }
        ctamc_logout(&session_id);
        tamc_delete_context(&ctx);
    }
    if (rc == 1)
        rc = 0;                 // exit pass
    else {
	ERR_print_errors_fp(stderr);
        rc = 1;                 // exit fail
    }
    exit(rc);
}

/*
 * int test_ctamc_keypairgen_test description: This function performs ctamc
 * ECC key pair generation API. in param Return int : 1 for PASS 0 for FAIL
 */
int test_ctamc_ec_keypairgen_test(ENGINE *e_ctamc)
{
    ctamc_item_t ctamc_item;
    int ret = 1;
    int loop_idx;

    for (loop_idx = 0; loop_idx < list_max; loop_idx++) {
        // Generate a key pair with key name and key size
        ctamc_item.c = (unsigned char *)&ec_key_parms_list[loop_idx];
        ctamc_item.len = sizeof(ctamc_ecc_key_attr_t);

        if (ENGINE_ctrl_cmd(e_ctamc, "CTAMC_GEN_EC_KEYPAIR", 0, &ctamc_item,
                            0, 0)) {
            BIO_printf(bio_out, "EC Key pair %s generated.\n",
                       ec_key_parms_list[loop_idx].key_name);
            ret = 1;
        } else {
            BIO_printf(bio_out, "EC KEY PAIR Gen FAILED!!! %s\n",
                       ec_key_parms_list[loop_idx].key_name);
            ret = 0;
        }
    }
    return ret;
}

int test_ec_signature(ENGINE *e, char *key_id, int eng_as_deflt)
{
    int rc = 0;
    unsigned int md_len, i, m;
    EVP_PKEY *evp_key = NULL;
    EVP_MD_CTX *mdctx;
    EVP_MD *md[4];
    unsigned char md_value[EVP_MAX_MD_SIZE];

    if (eng_as_deflt)
        ENGINE_set_default_ECDSA(e);
    // Load generate keys and do some work with them
    evp_key = ENGINE_load_private_key(e, key_id, NULL, NULL);

    if (!evp_key) {
        BIO_printf(bio_out, "Error retrieving %s key pair\n", key_id);
        rc = 0;
        return (rc);
    } else {
        rc = 1;
        BIO_printf(bio_out, "Loaded %s private key.\n", key_id);
    }

    OpenSSL_add_all_digests();
    md[0] = (EVP_MD *)EVP_sha1();
    md[1] = (EVP_MD *)EVP_sha256();
    md[2] = (EVP_MD *)EVP_sha384();
    md[3] = (EVP_MD *)EVP_sha512();

    for (m = 0; m < 4; m++) {
        mdctx = EVP_MD_CTX_create();

        EVP_DigestInit_ex(mdctx, md[m], NULL);
        EVP_DigestUpdate(mdctx, the_message, strlen(the_message));
        EVP_DigestFinal_ex(mdctx, md_value, &md_len);
        EVP_MD_CTX_destroy(mdctx);
        BIO_printf(bio_out, "Digest type %d ", md[m]->type);
        BIO_printf(bio_out, "Digest is: ");
        for (i = 0; i < md_len; i++)
            BIO_printf(bio_out, "%02x", md_value[i]);
        BIO_printf(bio_out, "\n");

        // Sign hash
        ECDSA_SIG *signature =
            ECDSA_do_sign(md_value, md_len, evp_key->pkey.ec);
        if (!signature) {
            BIO_printf(bio_out, "Failed to generate EC Signature\n");
            rc = 0;
            break;
        } else {
            BIO_printf(bio_out, "EC Signature generated\n");
            rc = 1;
        }

        // Verify signature
        if (ECDSA_do_verify(md_value, md_len, signature, evp_key->pkey.ec) !=
            1) {
            BIO_printf(bio_out, "Failed to Verify EC Signature\n");
            rc = 0;
        } else {
            BIO_printf(bio_out, "EC Signature Verified\n");
            rc = 1;
        }
        if (signature)
            ECDSA_SIG_free(signature);
    }
    if (evp_key)
        EVP_PKEY_free(evp_key);
    evp_key = NULL;
    return (rc);
}

int test_ctamc_evp_digest_sign_verify(ENGINE *e)
{
    EVP_MD_CTX *mdctx = NULL;
    int ret = 0;
    int key_i = 0;
    int md_j = 0;
    EVP_PKEY *key = NULL;
    unsigned char *sig = NULL;
    size_t slen = 0;
    EVP_MD *md[4];

    OpenSSL_add_all_digests();
    md[0] = (EVP_MD *)EVP_sha1();
    md[1] = (EVP_MD *)EVP_sha256();
    md[2] = (EVP_MD *)EVP_sha384();
    md[3] = (EVP_MD *)EVP_sha512();

    for (key_i = 0; key_i < list_max; key_i++) {
        key =
            ENGINE_load_private_key(e, ec_key_parms_list[key_i].key_name,
                                    NULL, NULL);

        for (md_j = 0; md_j < 4; md_j++) {
            /* Create & init the Message Digest Context */
            if (!(mdctx = EVP_MD_CTX_create()))
                goto err;

#ifdef OPENSSL_FIPS
            if (FIPS_mode() && (md[md_j] == EVP_sha1())) {
                EVP_MD_CTX_set_flags(mdctx, EVP_MD_CTX_FLAG_NON_FIPS_ALLOW);
            }
#endif

            /* Init DigestSign */
            ret = EVP_DigestSignInit(mdctx, NULL, md[md_j], NULL, key);
            if (ret != 1)
                goto err;

            /* Update message */
            ret =
                EVP_DigestSignUpdate(mdctx, the_message, strlen(the_message));
            if (ret != 1)
                goto err;

            /* Finalize the DigestSign operation */
            /* get sig len */
            ret = EVP_DigestSignFinal(mdctx, NULL, &slen);
            if (ret != 1)
                goto err;

            /* Allocate sig */
            sig = OPENSSL_malloc(sizeof(unsigned char) * slen);
            if (!sig)
                goto err;

            /* Obtain the signature */
            ret = EVP_DigestSignFinal(mdctx, sig, &slen);
            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: Digest sign failed.\n");
                goto err;
            }

            if (mdctx)
                EVP_MD_CTX_destroy(mdctx);
            mdctx = NULL;

            // //Verify
            /* Create the Message Digest Context */
            if (!(mdctx = EVP_MD_CTX_create()))
                goto err;

            /* Initialize Verify operation  */
            ret = EVP_DigestVerifyInit(mdctx, NULL, md[md_j], NULL, key);
            if (ret != 1)
                goto err;

            /* Call update with the message */
            ret =
                EVP_DigestVerifyUpdate(mdctx, the_message,
                                       strlen(the_message));
            if (ret != 1)
                goto err;

            /* Finalize */
            ret = EVP_DigestVerifyFinal(mdctx, sig, slen);
            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: Digest verify failed.\n");
                goto err;
            }

 err:
            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: Digest sign/verify failed.\n");
            }

            /* Clean up */
            if (sig)
                OPENSSL_free(sig);
            sig = NULL;
            slen = 0;
            if (mdctx)
                EVP_MD_CTX_destroy(mdctx);
            mdctx = NULL;
        }                       // md_j
        if (key)
            EVP_PKEY_free(key);
        key = NULL;
    }                           // key_i
    return (ret);
}

int test_ctamc_evp_sign_verify(ENGINE *e)
{
    EVP_MD_CTX *mdctx = NULL;
    int ret = 0;
    int key_i = 0;
    int md_j = 0;
    EVP_PKEY *key = NULL;
    unsigned char *sig = NULL;
    unsigned int slen = 0;
    EVP_MD *md[4];

    OpenSSL_add_all_digests();
    md[0] = (EVP_MD *)EVP_sha1();
    md[1] = (EVP_MD *)EVP_sha256();
    md[2] = (EVP_MD *)EVP_sha384();
    md[3] = (EVP_MD *)EVP_sha512();

    for (key_i = 0; key_i < list_max; key_i++) {
        key =
            ENGINE_load_private_key(e, ec_key_parms_list[key_i].key_name,
                                    NULL, NULL);

        for (md_j = 0; md_j < 4; md_j++) {
            /* Create & init the Message Digest Context */
            if (!(mdctx = EVP_MD_CTX_create()))
                goto err;

#ifdef OPENSSL_FIPS
            if (FIPS_mode() && (md[md_j] == EVP_sha1())) {
                EVP_MD_CTX_set_flags(mdctx, EVP_MD_CTX_FLAG_NON_FIPS_ALLOW);
            }
#endif

            /* Init the DigestSign */
            ret = EVP_SignInit_ex(mdctx, md[md_j], NULL);
            if (ret != 1)
                goto err;

            /* Update with the message */
            ret = EVP_SignUpdate(mdctx, the_message, strlen(the_message));
            if (ret != 1)
                goto err;

            /* Finalize the DigestSign operation */
            /* get Sig len */
            ret = EVP_SignFinal(mdctx, NULL, &slen, key);
            if (ret != 1)
                goto err;

            /* Allocate sig memory  */
            sig = OPENSSL_malloc(sizeof(unsigned char) * slen);
            if (!sig)
                goto err;

            /* Calc signature */
            ret = EVP_SignFinal(mdctx, sig, &slen, key);

            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: ECC sign failed.\n");
                goto err;
            }

            if (mdctx)
                EVP_MD_CTX_destroy(mdctx);
            mdctx = NULL;

            // //Verify
            /* Create the Message Digest Context */
            if (!(mdctx = EVP_MD_CTX_create()))
                goto err;

            /* Init Verify operation  */
            ret = EVP_VerifyInit_ex(mdctx, md[md_j], NULL);
            if (ret != 1)
                goto err;

            /* Update message */
            ret = EVP_VerifyUpdate(mdctx, the_message, strlen(the_message));
            if (ret != 1)
                goto err;

            /* Finalize */
            ret = EVP_VerifyFinal(mdctx, sig, slen, key);
            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: verify failed.\n");
                goto err;
            }

 err:
            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: sign/verify failed.\n");
            }

            /* Clean up */
            if (sig)
                OPENSSL_free(sig);
            sig = NULL;
            slen = 0;
            if (mdctx)
                EVP_MD_CTX_destroy(mdctx);
            mdctx = NULL;
        }                       // md_j
        if (key)
            EVP_PKEY_free(key);
        key = NULL;
    }                           // key_i
    return (ret);
}

void print_help()
{
    const char *ec_usage[] = {
        "usage: ctamc args\n",
        "\n",
        " -app_id  <app_id app_password>\n",
        NULL
    };
    const char **pp2;

    for (pp2 = ec_usage; (*pp2 != NULL); pp2++)
        printf("%s", *pp2);
}
