/*
 *  ctamc_test_rando.c
 *  Openssl Engine Cisco Trusted Anchor CLient Module Random Num Gen Tests
 *
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <openssl/engine.h>
#include <openssl/rand.h>
#include "../../vendor_defns/ctamc.h"
#include "../ctame.h"
#include "tam_types.h"
#ifdef USE_CISCO_SSM
# include "ssm.h"
#else
# include "tamc_api.h"
#endif
#include "safe_mem_lib.h"
#include "safe_str_lib.h"
#include "ctamc_utility.c"

/* hard coded for now */
static BIO *bio_out = NULL;
int test_ctamc_rand_seed(ENGINE *e, unsigned char *buf, int rand_size);
int test_ctamc_rand_bytes(ENGINE *e, unsigned char *buf, int rand_size);
void print_help();

int main(int argc, char *argv[])
{
    ENGINE *e_ctamc = NULL;
    char app_id[TAM_APPID_LEN_MAX] = "engine_9";
    char app_password[TAM_PW_LEN_MAX] = "engine_9_passwd";
    void *ctx = NULL;
    void *session_id = NULL;
    int rc = 0;
    int drand_size = 32;        // /default size
    int trand_size = 8;         // /default size
    ENGINE *test_e = NULL;

    ctamc_context_t *ctamc_session_ctx;
    uint8_t *trand = NULL;
    uint8_t *drand = NULL;

    bio_out = BIO_new_fp(stdout, BIO_NOCLOSE);

    BIO_printf(bio_out, "Random Number Gen Test entered\n");

    argc--;
    argv++;
    printf("argc = %d\n", argc);
    if (argc == 3) {
        if (strcmp(*argv, "-app_id") == 0) {
            if (--argc < 2)
                goto bad;

            strncpy_s(&app_id[0], TAM_APPID_LEN_MAX, *(++argv),
                      TAM_APPID_LEN_MAX);
            argc--;
            strncpy_s(&app_password[0], TAM_PW_LEN_MAX, *(++argv),
                      TAM_PW_LEN_MAX);

        } else {
 bad:
            printf("unknown option %s. Using default user.\n", *argv);
        }
    }

    load_ctamc_engine(&e_ctamc);
    if (!e_ctamc) {
        BIO_printf(bio_out, "load_ctamc_engine failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }

    ctamc_login(app_id, app_password, &ctx, &session_id);
    if (session_id)
        BIO_printf(bio_out, "Session context 0x%x.\n", *(int *)session_id);
    if (!ctx) {
        BIO_printf(bio_out, "login failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }

    ctamc_session_ctx = calloc(1, sizeof(ctamc_context_t));
    ctamc_session_ctx->session_id = session_id;
    strncpy_s(&ctamc_session_ctx->app_id[0], TAM_APPID_LEN_MAX, app_id,
              TAM_APPID_LEN_MAX);
    ctamc_item_t ctamc_item;
    ctamc_item.c = (unsigned char *)ctamc_session_ctx;
    ctamc_item.len = sizeof(ctamc_context_t);

    rc = ENGINE_ctrl_cmd(e_ctamc, "CTAMC_SET_CONTEXT", 0, &ctamc_item, 0, 0);

    if (rc == 1) {
        BIO_printf(bio_out, "Passed set session context\n");
    } else {
        BIO_printf(bio_out, "Failed set session context\n");
        rc = 1;
        goto end;
    }

    trand = calloc(1, trand_size);
    drand = calloc(1, drand_size);
    if (!trand || !drand)
        goto end;

    rc = ENGINE_set_default_RAND(e_ctamc);
    RAND_set_rand_engine(e_ctamc);
    if (!rc) {
        BIO_printf(bio_out, "ENGINE_set_default_RAND failed.");
        goto end;
    }

    test_e = ENGINE_get_default_RAND();
    if (test_e)
        BIO_printf(bio_out, "Default engine name %s.\n",
                   ENGINE_get_id(test_e));

    rc = test_ctamc_rand_seed(e_ctamc, (unsigned char *)trand, trand_size);
    if (rc) {
        int i;
        BIO_printf(bio_out, "Seed Data: 0x");
        for (i = 0; i < trand_size; i++) {
            BIO_printf(bio_out, "%x", trand[i]);
        }
        BIO_printf(bio_out, "\n");
    }
    rc = test_ctamc_rand_bytes(e_ctamc, (unsigned char *)drand, drand_size);
    if (rc) {
        int i;
        BIO_printf(bio_out, "Pseudo Random Data: 0x");
        for (i = 0; i < drand_size; i++) {
            BIO_printf(bio_out, "%x", drand[i]);
        }
        BIO_printf(bio_out, "\n");
    }
    BIO_printf(bio_out, "Test complete!\n");

 end:
    if (ctx) {
        ctamc_logout(&session_id);
        tamc_delete_context(&ctx);
    }
    if (rc == 1)
        rc = 0;                 // exit pass
    else {
	ERR_print_errors_fp(stderr);
        rc = 1;                 // exit fail
    }

    free(trand);
    free(drand);
    exit(rc);
}

int test_ctamc_rand_seed(ENGINE *e, unsigned char *buf, int trand_size)
{
    int rc = 1;
    RAND_bytes(buf, trand_size);
    RAND_seed(buf, trand_size);
    return (rc);
}

int test_ctamc_rand_bytes(ENGINE *e, unsigned char *buf, int drand_size)
{
    int rc = 0;
    rc = RAND_pseudo_bytes(buf, drand_size);
    return (rc);
}

void print_help()
{
    const char *rsa_usage[] = {
        "usage: ctamc args\n",
        "\n",
        " -app_id  <app_id app_password>\n",
        NULL
    };
    const char **pp2;

    for (pp2 = rsa_usage; (*pp2 != NULL); pp2++)
        printf("%s", *pp2);
}
