/*
 *  ctamc_test_rsa_crypto.c
 *  Openssl Engine Cisco Trusted Anchor CLient Module RSA Tests
 *
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <openssl/engine.h>
#include <openssl/err.h>
#include "../../vendor_defns/ctamc.h"
#include "../ctame.h"
#include "tam_types.h"
#ifdef USE_CISCO_SSM
# include "ssm.h"
#else
# include "tamc_api.h"
#endif
#include "safe_str_lib.h"
#include "safe_mem_lib.h"
#include "ctamc_utility.c"

/* hard coded for now */
static BIO *bio_out = NULL;
static int debug = 1;

int test_ctamc_read_object(ctamc_context_t * ctame_context, char *obj_name);
int test_ctamc_keypairgen_test(ENGINE *e_ctamc);
static int test_rsa_encrypt_decrypt(ENGINE *e);
int test_ctamc_priv_encr_pub_decry(RSA *rsa);
int test_ctamc_pub_encr_priv_decry(RSA *rsa);
int test_ctamc_rsa_evp_sign_verify(ENGINE *e);
int test_ctamc_rsa_evp_digest_sign_verify(ENGINE *e);
void print_help();

#define the_message "Rock Chalk Jayhawk!! Go KU!!!!"

#define KLIST_MAX 2
ctamc_rsa_keygen_parms_t key_parms_list[KLIST_MAX] = {
    {1024, 65537, 1, 1, 0, "rsa-1024"},
    {2048, 65537, 1, 1, 0, "rsa-2048"},
};

int main(int argc, char *argv[])
{
    ENGINE *e_ctamc = NULL;
    char app_id[TAM_APPID_LEN_MAX] = "engine_9";
    char app_password[TAM_PW_LEN_MAX] = "engine_9_passwd";
    void *ctx = NULL;
    void *session_id = NULL;
    int rc = 0;
    ctamc_context_t *ctamc_session_ctx;
    ENGINE *test_e = NULL;

    bio_out = BIO_new_fp(stdout, BIO_NOCLOSE);

    BIO_printf(bio_out, "Test entered\n");

    argc--;
    argv++;
    printf("argc = %d\n", argc);
    if (argc == 3) {
        if (strcmp(*argv, "-app_id") == 0) {
            if (--argc < 2)
                goto bad;

            strncpy_s(&app_id[0], TAM_APPID_LEN_MAX, *(++argv),
                      TAM_APPID_LEN_MAX);
            argc--;
            strncpy_s(&app_password[0], TAM_PW_LEN_MAX, *(++argv),
                      TAM_PW_LEN_MAX);

        } else {
 bad:
            printf("unknown option %s. Using default user.\n", *argv);
        }
    }

    load_ctamc_engine(&e_ctamc);
    if (!e_ctamc) {
        BIO_printf(bio_out, "load_ctamc_engine failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }

    ctamc_login(app_id, app_password, &ctx, &session_id);
    if (session_id)
        BIO_printf(bio_out, "Session context 0x%x.\n", *(int *)session_id);
    if (!ctx) {
        BIO_printf(bio_out, "login failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }
#ifndef USE_CISCO_SSM
    tam_set_debug_level(TAMC_DBG_ERR_ALL);
#endif

    ctamc_session_ctx = calloc(1, sizeof(ctamc_context_t));
    ctamc_session_ctx->session_id = session_id;
    strncpy_s(&ctamc_session_ctx->app_id[0], TAM_APPID_LEN_MAX, app_id,
              TAM_APPID_LEN_MAX);
    ctamc_item_t ctamc_item;
    ctamc_item.c = (unsigned char *)ctamc_session_ctx;
    ctamc_item.len = sizeof(ctamc_context_t);

    rc = ENGINE_ctrl_cmd(e_ctamc, "CTAMC_SET_CONTEXT", 0, &ctamc_item, 0, 0);

    if (rc == 1) {
        BIO_printf(bio_out, "Passed set session context\n");
    } else {
        BIO_printf(bio_out, "Failed set session context\n");
        rc = 1;
        goto end;
    }
    // RSA Key Gen
    rc = test_ctamc_keypairgen_test(e_ctamc);
    if (!rc) {
        BIO_printf(bio_out, "RSA Key Gen failed.\n");
        goto end;
    }

    test_e = ENGINE_get_default_RSA();
    if (test_e)
        BIO_printf(bio_out, "Default engine name %s.\n",
                   ENGINE_get_id(test_e));

    rc = test_rsa_encrypt_decrypt(e_ctamc);
    if (rc != 1) {
        BIO_printf(bio_out, "\ntest_rsa_encrypt_decrypt FAILED!!!\n");
        goto end;
    }
    BIO_printf(bio_out, "\ntest_rsa_encrypt_decrypt PASSED!\n");

    rc = test_ctamc_rsa_evp_digest_sign_verify(e_ctamc);
    if (rc != 1) {
        BIO_printf(bio_out,
                   "\ntest_ctamc_rsa_evp_digest_sign_verify FAILED!!!\n");
        goto end;
    }
    BIO_printf(bio_out, "\ntest_ctamc_rsa_evp_digest_sign_verify PASSED!\n");

    rc = test_ctamc_rsa_evp_sign_verify(e_ctamc);
    if (rc != 1) {
        BIO_printf(bio_out, "\ntest_ctamc_rsa_evp_sign_verify FAILED!!!\n");
        goto end;
    }
    BIO_printf(bio_out, "\ntest_ctamc_rsa_evp_sign_verify PASSED!\n");

    // Repeat tests with ctamc engine as default engine

    BIO_printf(bio_out, "\nRepeating Tests with ENGINE_set_default_RSA.\n");
    rc = ENGINE_set_default_RSA(e_ctamc);
    if (rc != 1) {
        BIO_printf(bio_out, "ENGINE_set_default_RSA FAILED\n");
        goto end;
    }
    test_e = ENGINE_get_default_RSA();
    if (test_e)
        BIO_printf(bio_out, "Default engine name %s.\n",
                   ENGINE_get_id(test_e));

    rc = test_rsa_encrypt_decrypt(e_ctamc);
    if (rc != 1) {
        BIO_printf(bio_out, "\ntest_rsa_encrypt_decrypt FAILED!!!\n");
        goto end;
    }
    BIO_printf(bio_out, "\ntest_rsa_encrypt_decrypt PASSED!\n");

    rc = test_ctamc_rsa_evp_digest_sign_verify(e_ctamc);
    if (rc != 1) {
        BIO_printf(bio_out,
                   "\ntest_ctamc_rsa_evp_digest_sign_verify FAILED!!!\n");
        goto end;
    }
    BIO_printf(bio_out, "\ntest_ctamc_rsa_evp_digest_sign_verify PASSED!\n");

    rc = test_ctamc_rsa_evp_sign_verify(e_ctamc);
    if (rc != 1) {
        BIO_printf(bio_out, "\ntest_ctamc_rsa_evp_sign_verify FAILED!!!\n");
        goto end;
    }
    BIO_printf(bio_out, "\ntest_ctamc_rsa_evp_sign_verify PASSED!\n");

 end:

    BIO_printf(bio_out, "Test complete!\n");

    if (ctx) {
        /* cleanup stored files */
        int loop_idx;
        for (loop_idx = 0; loop_idx < KLIST_MAX; loop_idx++) {
            tamc_so_delete(session_id, app_id,
                           key_parms_list[loop_idx].key_name);
        }
        ctamc_logout(&session_id);
        tamc_delete_context(&ctx);
    }
    if (rc == 1) {
        rc = 0;
    } else {
	ERR_print_errors_fp(stderr);
        rc = 1;
    }
    exit(rc);
}

/*
 * int test_ctamc_keypairgen_test(RSA *rsa, RSA *pub_key, int key_size)
 * description: This function performs ctamc RSA key pair generation API. in
 * param rsa : RSA key mapped to CTAM Device Key pair pub_key : Only Public
 * key part of the ctamc dev key pair key_size : size of the key pair
 * (128/256) Return int : 1 for PASS 0 for FAIL
 */
int test_ctamc_keypairgen_test(ENGINE *e_ctamc)
{
    ctamc_item_t ctamc_item;
    int ret = 1;
    int loop_idx;

    for (loop_idx = 0; loop_idx < KLIST_MAX; loop_idx++) {
        // Generate a key pair with key name and key size
        ctamc_item.c = (unsigned char *)&key_parms_list[loop_idx];
        ctamc_item.len = sizeof(ctamc_rsa_keygen_parms_t);

        if (ENGINE_ctrl_cmd(e_ctamc, "CTAMC_GEN_RSA_KEYPAIR", 0, &ctamc_item,
                            0, 0)) {
            BIO_printf(bio_out, "RSA Key pair %s generated.\n",
                       key_parms_list[loop_idx].key_name);
            ret = 1;
        } else {
            BIO_printf(bio_out, "RSA KEY PAIR Gen FAILED!!! %s\n",
                       key_parms_list[loop_idx].key_name);
            ret = 0;
        }
    }
    return ret;
}

int test_ctamc_priv_encr_pub_decry(RSA *rsa)
{
    unsigned char *cipher_buffer = NULL;
    unsigned char *decrypted_buffer = NULL;

    int pt_data_len = 0;
    int ct_data_len = 0;
    int dt_data_len = 0;
    int memcmp_cnt = 0;
    int rsa_key_size = RSA_size(rsa);
    int ret = 0;

    pt_data_len = strlen(the_message);

    // Allocate buffer storage
    cipher_buffer = calloc(1, rsa_key_size);
    decrypted_buffer = calloc(1, pt_data_len);

    if (!cipher_buffer || !decrypted_buffer) {
        BIO_printf(bio_out, "ERROR out of memory!!! keysize = %d\n",
                   rsa_key_size);
        ret = 0;
        goto end;
    }

    BIO_printf(bio_out, "RSA key size = %d bytes.\n", rsa_key_size);

    ct_data_len =
        RSA_private_encrypt(pt_data_len, (unsigned char *)the_message,
                            cipher_buffer, rsa, RSA_PKCS1_PADDING);
    if ((ct_data_len <= 0) || (ct_data_len > rsa_key_size)) {
        BIO_printf(bio_out, "ERROR Failed to do private key "
                   "encryption: plain text length=%d, ct len = %d\n",
                   pt_data_len, ct_data_len);
        ret = 0;
    } else {
        BIO_printf(bio_out, "Private key "
                   " encryption passed.: cipher text length=%d\n",
                   ct_data_len);
        dt_data_len =
            RSA_public_decrypt(ct_data_len, cipher_buffer, decrypted_buffer,
                               rsa, RSA_PKCS1_PADDING);
        if (dt_data_len <= 0) {
            BIO_printf(bio_out, "ERROR TAMC private key encrypted"
                       " data cannot be decrypted with openssl function. \n\n");
            ret = 0;
        } else {
            BIO_printf(bio_out, "Decrypted with public key.\n");
            ret = 1;
        }
    }

    if (ret == 1) {
        if (memcmp_s(the_message, pt_data_len, decrypted_buffer,
                     dt_data_len, &memcmp_cnt) != 0) {
            BIO_printf(bio_out, "ERROR TAMC private key encrypted"
                       "data does not match with openssl decrytped data.\n\n");
            ret = 0;
            goto end;
        }
    }
 end:
    free(decrypted_buffer);
    free(cipher_buffer);

    return (ret);
}

int test_ctamc_pub_encr_priv_decry(RSA *rsa)
{
    unsigned char *cipher_buffer = NULL;
    unsigned char *decrypted_buffer = NULL;

    int pt_data_len = 0;
    int ct_data_len = 0;
    int dt_data_len = 0;
    int memcmp_cnt = 0;
    int rsa_key_size = RSA_size(rsa);
    int ret = 0;

    pt_data_len = strlen(the_message);
    // Allocate buffer storage
    cipher_buffer = calloc(1, rsa_key_size);
    decrypted_buffer = calloc(1, pt_data_len);

    if (!cipher_buffer || !decrypted_buffer) {
        BIO_printf(bio_out, "ERROR out of memory!!! keysize = %d\n",
                   rsa_key_size);
        ret = 0;
        goto end;
    }

    BIO_printf(bio_out, "RSA key size = %d bytes.\n", rsa_key_size);
    pt_data_len = strlen(the_message);
    ct_data_len =
        RSA_public_encrypt(pt_data_len, (unsigned char *)the_message,
                           cipher_buffer, rsa, RSA_PKCS1_PADDING);
    if ((ct_data_len <= 0) || (ct_data_len > rsa_key_size)) {
        BIO_printf(bio_out, "ERROR Failed to do public key "
                   "encryption: plain text length=%d, ct len = %d\n",
                   pt_data_len, ct_data_len);
        ret = 0;
    } else {
        BIO_printf(bio_out, "Public key "
                   " encryption passed.: cipher text length=%d\n",
                   ct_data_len);
        dt_data_len =
            RSA_private_decrypt(ct_data_len, cipher_buffer, decrypted_buffer,
                                rsa, RSA_PKCS1_PADDING);
        if (dt_data_len <= 0) {
            BIO_printf(bio_out, "ERROR TAMC public key encrypted"
                       " data cannot be decrypted with openssl function. \n\n");
            ret = 0;
        } else {
            BIO_printf(bio_out, "Decrypted with private key.\n");
            ret = 1;
        }
    }

    if (ret == 1) {
        if (memcmp_s(the_message, pt_data_len, decrypted_buffer,
                     dt_data_len, &memcmp_cnt) != 0) {
            BIO_printf(bio_out, "ERROR TAMC public key encrypted"
                       "data does not match with openssl decrytped data.\n\n");
            ret = 0;
            goto end;
        }
    }
 end:
    free(decrypted_buffer);
    free(cipher_buffer);

    return (ret);
}

static int test_rsa_encrypt_decrypt(ENGINE *e)
{
    int rc = 0;
    int i;
    EVP_PKEY *key = NULL;
    RSA *rsa = NULL;

    for (i = 0; i < KLIST_MAX; i++) {

        // Load generate keys and do some work with them
        key =
            ENGINE_load_private_key(e, key_parms_list[i].key_name, NULL,
                                    NULL);

        if (!key) {
            BIO_printf(bio_out, "Error retrieving %s key pair\n",
                       key_parms_list[i].key_name);
            goto end;
        }

        /* Retrieve rsa struct from the key */
        rsa = EVP_PKEY_get1_RSA(key);
        if (!rsa) {
            BIO_printf(bio_out, "%s RSA Key not loaded\n",
                       key_parms_list[i].key_name);
            continue;
        }

        if (debug) {
            BIO_printf(bio_out, "%s RSA Keys loaded\n",
                       key_parms_list[i].key_name);
        }

        rc = test_ctamc_priv_encr_pub_decry(rsa);
        rc = test_ctamc_pub_encr_priv_decry(rsa);

        EVP_PKEY_free(key);
        key = NULL;
        if (rsa)
            RSA_free(rsa);
        rsa = NULL;
    }
 end:
    return (rc);
}

/*
 * int test_ctamc_read_object(RSA *rsa, int key_size) description: This
 * function read the specified object name for the tam device in param rsa :
 * RSA key mapped to CTAM Device Key pair key_size : size of the key pair
 * (128/256) Return int : 1 for PASS 0 for FAIL
 */
int test_ctamc_read_object(ctamc_context_t * ctame_context, char *object_name)
{

    int ret = 1;
    ctamc_item_t mydata;
    unsigned short length;
    const char *context = (const char *)ctame_context;
    mydata.c = NULL;
    mydata.len = 0;

    mydata.len = 2048;
    mydata.c = OPENSSL_malloc(2048);

    if (CTAMC_SUCCESS != ctame_read_object(context, (const char *)object_name,
                                           &(mydata.c), &length)) {
        BIO_printf(bio_out, "ERROR ctame_read_object");
        ret = 0;
    }
    free(mydata.c);
    mydata.c = NULL;

    printf("length =%d", length);
    /* specify object name null */
    if (CTAMC_SUCCESS == ctame_read_object(context, NULL,
                                           &(mydata.c), &length)) {
        BIO_printf(bio_out, "ERROR ctame_read_object\n");
        ret = 0;
    }
    free(mydata.c);
    return ret;
}

int test_ctamc_rsa_evp_digest_sign_verify(ENGINE *e)
{
    EVP_MD_CTX *mdctx = NULL;
    int ret = 0;
    int key_i = 0;
    int md_j = 0;
    EVP_PKEY *key = NULL;
    unsigned char *sig = NULL;
    size_t slen = 0;
    EVP_MD *md[4];

    OpenSSL_add_all_digests();
    md[0] = (EVP_MD *)EVP_sha1();
    md[1] = (EVP_MD *)EVP_sha256();
    md[2] = (EVP_MD *)EVP_sha384();
    md[3] = (EVP_MD *)EVP_sha512();

    for (key_i = 0; key_i < KLIST_MAX; key_i++) {
        key =
            ENGINE_load_private_key(e, key_parms_list[key_i].key_name, NULL,
                                    NULL);

        for (md_j = 0; md_j < 4; md_j++) {
            /* Create & init the Message Digest Context */
            if (!(mdctx = EVP_MD_CTX_create()))
                goto err;

#ifdef OPENSSL_FIPS
            if (FIPS_mode() && (md[md_j] == EVP_sha1())) {
                EVP_MD_CTX_set_flags(mdctx, EVP_MD_CTX_FLAG_NON_FIPS_ALLOW);
            }
#endif

            /* Init DigestSign */
            ret = EVP_DigestSignInit(mdctx, NULL, md[md_j], NULL, key);
            if (ret != 1)
                goto err;

            /* Update message */
            ret =
                EVP_DigestSignUpdate(mdctx, the_message, strlen(the_message));
            if (ret != 1)
                goto err;

            /* Finalize the DigestSign operation */
            /* get sig len */
            ret = EVP_DigestSignFinal(mdctx, NULL, &slen);
            if (ret != 1)
                goto err;

            /* Allocate sig */
            sig = OPENSSL_malloc(sizeof(unsigned char) * slen);
            if (!sig)
                goto err;

            /* Obtain the signature */
            ret = EVP_DigestSignFinal(mdctx, sig, &slen);
            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: RSA Digest sign failed.\n");
                goto err;
            }

            if (mdctx)
                EVP_MD_CTX_destroy(mdctx);
            mdctx = NULL;

            // //Verify
            /* Create the Message Digest Context */
            if (!(mdctx = EVP_MD_CTX_create()))
                goto err;

            /* Initialize Verify operation  */
            ret = EVP_DigestVerifyInit(mdctx, NULL, md[md_j], NULL, key);
            if (ret != 1)
                goto err;

            /* Call update with the message */
            ret =
                EVP_DigestVerifyUpdate(mdctx, the_message,
                                       strlen(the_message));
            if (ret != 1)
                goto err;

            /* Finalize */
            ret = EVP_DigestVerifyFinal(mdctx, sig, slen);
            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: RSA Digest verify failed.\n");
                goto err;
            }

 err:
            if (ret != 1) {
                BIO_printf(bio_out,
                           "ERROR: RSA Digest sign/verify failed.\n");
            }

            /* Clean up */
            if (sig)
                OPENSSL_free(sig);
            sig = NULL;
            slen = 0;
            if (mdctx)
                EVP_MD_CTX_destroy(mdctx);
            mdctx = NULL;
        }                       // md_j
        if (key)
            EVP_PKEY_free(key);
        key = NULL;
    }                           // key_i
    return (ret);
}

int test_ctamc_rsa_evp_sign_verify(ENGINE *e)
{
    EVP_MD_CTX *mdctx = NULL;
    int ret = 0;
    int key_i = 0;
    int md_j = 0;
    EVP_PKEY *key = NULL;
    unsigned char *sig = NULL;
    unsigned int slen = 0;
    EVP_MD *md[4];

    OpenSSL_add_all_digests();
    md[0] = (EVP_MD *)EVP_sha1();
    md[1] = (EVP_MD *)EVP_sha256();
    md[2] = (EVP_MD *)EVP_sha384();
    md[3] = (EVP_MD *)EVP_sha512();

    for (key_i = 0; key_i < KLIST_MAX; key_i++) {
        key =
            ENGINE_load_private_key(e, key_parms_list[key_i].key_name, NULL,
                                    NULL);

        for (md_j = 0; md_j < 4; md_j++) {
            /* Create & init the Message Digest Context */
            if (!(mdctx = EVP_MD_CTX_create()))
                goto err;

#ifdef OPENSSL_FIPS
            if (FIPS_mode() && (md[md_j] == EVP_sha1())) {
                EVP_MD_CTX_set_flags(mdctx, EVP_MD_CTX_FLAG_NON_FIPS_ALLOW);
            }
#endif

            /* Init the DigestSign */
            ret = EVP_SignInit_ex(mdctx, md[md_j], NULL);
            if (ret != 1)
                goto err;

            /* Update with the message */
            ret = EVP_SignUpdate(mdctx, the_message, strlen(the_message));
            if (ret != 1)
                goto err;

            /* Finalize the DigestSign operation */
            /* get Sig len */
            ret = EVP_SignFinal(mdctx, NULL, &slen, key);
            if (ret != 1)
                goto err;

            /* Allocate sig memory  */
            sig = OPENSSL_malloc(sizeof(unsigned char) * slen);
            if (!sig)
                goto err;

            /* Calc signature */
            ret = EVP_SignFinal(mdctx, sig, &slen, key);

            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: RSA sign failed.\n");
                goto err;
            }

            if (mdctx)
                EVP_MD_CTX_destroy(mdctx);
            mdctx = NULL;

            // //Verify
            /* Create the Message Digest Context */
            if (!(mdctx = EVP_MD_CTX_create()))
                goto err;

            /* Init Verify operation  */
            ret = EVP_VerifyInit_ex(mdctx, md[md_j], NULL);
            if (ret != 1)
                goto err;

            /* Update message */
            ret = EVP_VerifyUpdate(mdctx, the_message, strlen(the_message));
            if (ret != 1)
                goto err;

            /* Finalize */
            ret = EVP_VerifyFinal(mdctx, sig, slen, key);
            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: RSA verify failed.\n");
                goto err;
            }

 err:
            if (ret != 1) {
                BIO_printf(bio_out, "ERROR: RSA sign/verify failed.\n");
            }

            /* Clean up */
            if (sig)
                OPENSSL_free(sig);
            sig = NULL;
            slen = 0;
            if (mdctx)
                EVP_MD_CTX_destroy(mdctx);
            mdctx = NULL;
        }                       // md_j
        if (key)
            EVP_PKEY_free(key);
        key = NULL;
    }                           // key_i
    return (ret);
}

void print_help()
{
    const char *rsa_usage[] = {
        "usage: ctamc args\n",
        "\n",
        " -app_id  <app_id app_password>\n",
        NULL
    };
    const char **pp2;

    for (pp2 = rsa_usage; (*pp2 != NULL); pp2++)
        printf("%s", *pp2);
}
