/*
 *  ctamc_test_tls.c
 *  Openssl Engine Cisco Trusted Anchor CLient Module RSA/EC Tests
 *
 *  Copyright 2015 Cisco Systems, Inc. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/engine.h>
#include <openssl/err.h>
#include <openssl/x509v3.h>
#include <openssl/evp.h>
#include <openssl/ssl.h>
#include "../../vendor_defns/ctamc.h"
#include "../ctame.h"
#include "tam_types.h"
#ifdef USE_CISCO_SSM
# include "ssm.h"
#else
# include "tamc_api.h"
#endif
#include "ctamc_utility.c"

/* hard coded for now */
static BIO *bio_out = NULL;

#define KEY_NAME_RSA "rsa-1024"
#define KEY_NAME_EC "ec-256"

#define KEY_TYPE_RSA    1
#define KEY_TYPE_EC     2

static void test_ctamc_keypair_create(ENGINE *e_ctamc, int key_type);
static void test_ctamc_keypair_delete(void *session_id, char *app_id, char *key_name);
static int test_ctamc_cert_create(X509 **new_cert, EVP_PKEY *keypair);
static int test_ctamc_tls_session(X509 *s_cert, EVP_PKEY *s_key, X509 *c_cert, EVP_PKEY *c_key);
static int doit(SSL *s_ssl, SSL *c_ssl, long count);

/*
 * This is the entry point into the TLS test.  This routine will:
 *
 * 1. create a RSA keypair and a certificate
 * 2. create an EC keypair and a certificate
 * 3. Use these certs/keys to setup a TLS session
 *
 */
int main (int argc, char *argv[])
{
    ENGINE *e_ctamc = NULL;
    char app_id[TAM_APPID_LEN_MAX] = "engine_9";
    char app_password[TAM_PW_LEN_MAX] = "engine_9_passwd";
    void *ctx = NULL;
    void *session_id = NULL;
    int rc = 0;
    ctamc_context_t *ctamc_session_ctx;
    X509 *s_cert = NULL;
    X509 *c_cert = NULL;
    EVP_PKEY *s_keypair;
    EVP_PKEY *c_keypair;

    bio_out = BIO_new_fp(stdout, BIO_NOCLOSE);

    BIO_printf(bio_out, "Test entered\n");

    load_ctamc_engine(&e_ctamc);
    if (!e_ctamc) {
        BIO_printf(bio_out, "load_ctamc_engine failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }

    ctamc_login(app_id, app_password, &ctx, &session_id);
    if (session_id) {
        BIO_printf(bio_out, "Session context 0x%x.\n", *(int*)session_id);
    }
    if (!ctx) {
        BIO_printf(bio_out, "login failed.\n");
        BIO_printf(bio_out, "\n");
        goto end;
    }
#ifndef USE_CISCO_SSM
    tam_set_debug_level(TAMC_DBG_ERR_ALL);
#endif

    ctamc_session_ctx = calloc(1, sizeof(ctamc_context_t));
    ctamc_session_ctx->session_id = session_id;
    strncpy_s(&ctamc_session_ctx->app_id[0], TAM_APPID_LEN_MAX, app_id,
              TAM_APPID_LEN_MAX);
    ctamc_item_t ctamc_item;
    ctamc_item.c = (unsigned char*)ctamc_session_ctx;
    ctamc_item.len = sizeof(ctamc_context_t);

    rc = ENGINE_ctrl_cmd(e_ctamc, "CTAMC_SET_CONTEXT", 0, &ctamc_item, 0, 0);

    if (rc == 1) {
        BIO_printf(bio_out, "Passed set session context\n");
    } else {
        BIO_printf(bio_out, "Failed set session context\n");
        rc = 1;
        goto end;
    }

    /* Generate a new RSA keypair in the engine */
    test_ctamc_keypair_create(e_ctamc, KEY_TYPE_RSA);
    /* Generate a new EC keypair in the engine */
    test_ctamc_keypair_create(e_ctamc, KEY_TYPE_EC);

    /* Get a reference to the new keypair to use with the TLS server */
    s_keypair = ENGINE_load_private_key(e_ctamc, KEY_NAME_RSA, NULL, NULL);
    if (!s_keypair) {
        BIO_printf(bio_out, "Failed to load new server keypair into EVP_PKEY\n");
        rc = 1;
        goto end;
    }

    /* Get a reference to the new keypair to use with the TLS client */
    c_keypair = ENGINE_load_private_key(e_ctamc, KEY_NAME_EC, NULL, NULL);
    if (!c_keypair) {
        BIO_printf(bio_out, "Failed to load new client keypair into EVP_PKEY\n");
        rc = 1;
        goto end;
    }

    /* Create a new cert using the keypair to use with the TLS server */
    if (test_ctamc_cert_create(&s_cert, s_keypair)) {
        BIO_printf(bio_out, "Failed to create server certificate\n");
        rc = 1;
        goto end;
    }
    X509_print_fp(stdout, s_cert);

    /* Create a new cert using the keypair to use with the TLS client */
    if (test_ctamc_cert_create(&c_cert, c_keypair)) {
        BIO_printf(bio_out, "Failed to create client certificate\n");
        rc = 1;
        goto end;
    }
    X509_print_fp(stdout, c_cert);

    /*
     * Now that we have a valid cert and keypair, use it to test
     * the TLS stack against the engine.
     */
    rc = test_ctamc_tls_session(s_cert, s_keypair, c_cert, c_keypair);
    if (rc) {
        BIO_printf(bio_out, "TLS test failed!!!\n");
        rc = 1;
        goto end;
    }

end:

    BIO_printf(bio_out, "Test complete!\n");

    test_ctamc_keypair_delete(session_id, app_id, KEY_NAME_RSA);
    test_ctamc_keypair_delete(session_id, app_id, KEY_NAME_EC);

    if (s_cert) {
        X509_free(s_cert);
    }
    if (c_cert) {
        X509_free(c_cert);
    }

    if (ctx) {
        /* cleanup */
        ctamc_logout(&session_id);
        tamc_delete_context(&ctx);
    }

    if (rc) {
        ERR_print_errors_fp(stderr);
    }
    exit(rc);
}

/*
 * int test_ctamc_keypairgen_test(ENGINE *e_ctamc, int key_type)
 * description: This function performs ctamc key pair generation API. 
 * in  *e_ctamc : key parameters
 *     key_type : KEY_TYPE_RSA/KEY_TYPE_EC
 * Return void
 */
static void test_ctamc_keypair_create (ENGINE *e_ctamc, int key_type)
{
    ctamc_item_t ctamc_item;
    ctamc_rsa_keygen_parms_t rsa_key_parms_list = { 1024, 65537, 1, 1, 0, KEY_NAME_RSA };
    ctamc_ecc_key_attr_t ec_key_parms_list = { TAM_EC_CURVE_P256, 1, 1, 0, KEY_NAME_EC };

    switch (key_type) {
    case KEY_TYPE_RSA:
        // Generate a key pair with key name and key size
        ctamc_item.c = (unsigned char*)&rsa_key_parms_list;
        ctamc_item.len = sizeof(ctamc_rsa_keygen_parms_t);

        if (ENGINE_ctrl_cmd(e_ctamc, "CTAMC_GEN_RSA_KEYPAIR", 0, &ctamc_item, 0, 0)) {
            BIO_printf(bio_out, "RSA Key pair %s generated.\n", rsa_key_parms_list.key_name);
        } else {
            BIO_printf(bio_out, "Warning: failed to generate RSA key %s.  The key may already exist.\n",
                       rsa_key_parms_list.key_name);
        }
        break;
    case KEY_TYPE_EC:
        // Generate a key pair with key name and key size
        ctamc_item.c = (unsigned char*)&ec_key_parms_list;
        ctamc_item.len = sizeof(ctamc_ecc_key_attr_t);

        if (ENGINE_ctrl_cmd(e_ctamc, "CTAMC_GEN_EC_KEYPAIR", 0, &ctamc_item, 0, 0)) {
            BIO_printf(bio_out, "EC Key pair %s generated.\n", ec_key_parms_list.key_name);
        } else {
            BIO_printf(bio_out, "Warning: failed to generate EC key %s.  The key may already exist.\n",
                       ec_key_parms_list.key_name);
        }
        break;
    default:
        BIO_printf(bio_out, "Warning: invalid key type requested.\n");
        break;
    }
}

/*
 * This is used to remove the key from the SSM store.
 */
static void test_ctamc_keypair_delete (void *session_id, char *app_id, char *key_name)
{
    tamc_so_delete(session_id, app_id, key_name);
}


/*
 * This routine will create a self-signed certificate that
 * can be used with TLS.
 *
 * Returns zero on success
 */
static int test_ctamc_cert_create (X509 **new_cert, EVP_PKEY *keypair)
{
    ASN1_INTEGER *serial = NULL;
    X509_NAME *subj;
    EVP_MD_CTX mctx;
    EVP_PKEY_CTX *pkctx = NULL;
    const EVP_MD *md = EVP_sha256();

    if ((*new_cert = X509_new()) == NULL) {
        BIO_printf(bio_out, "Failed to alloc X509.\n");
        return 1;
    }

    /* Set version to V3 */
    if (!X509_set_version(*new_cert, 2)) {
        BIO_printf(bio_out, "Failed to set X509 version.\n");
        return 1;
    }

    /*
     * Set the serial#
     */
    serial = s2i_ASN1_INTEGER(NULL, "12345678");
    if (!X509_set_serialNumber(*new_cert, serial)) {
        BIO_printf(bio_out, "Failed to setX509_print_fp X509 serial#.\n");
        return 1;
    }

    /*
     * Add Common Name entry
     */
    subj = X509_get_subject_name(*new_cert);
    if (!X509_NAME_add_entry_by_txt(subj, "CN", MBSTRING_ASC,
                                    (unsigned char*)"localhost", -1, -1, 0)) {
        BIO_printf(bio_out, "Failed to set X509 subject name.\n");
        return 1;
    }
    if (!X509_set_issuer_name(*new_cert, subj)) {
        BIO_printf(bio_out, "Failed to set Issuer.\n");
        return 1;
    }

    if (!X509_gmtime_adj(X509_get_notBefore(*new_cert), 0)) {
        BIO_printf(bio_out, "Failed to set notBefore.\n");
        return 1;
    }

    if (!X509_time_adj_ex(X509_get_notAfter(*new_cert), 1, 0, NULL)) {
        BIO_printf(bio_out, "Failed to set notAfter.\n");
        return 1;
    }

    if (!X509_set_pubkey(*new_cert, keypair)) {
        BIO_printf(bio_out, "Failed to public key.\n");
        return 1;
    }

    /*
     * Add a signature to the cert
     */
    EVP_MD_CTX_init(&mctx);
    EVP_DigestSignInit(&mctx, &pkctx, md, NULL, keypair);
    X509_sign_ctx(*new_cert, &mctx);
    EVP_MD_CTX_cleanup(&mctx);

    ASN1_INTEGER_free(serial);
    return 0;
}

/*
 * This function is a callback used by OpenSSL's verify_cert function.
 * It's called at the end of a cert verification to allow an opportunity to
 * gather more information regarding a failing cert verification, and to
 * possibly change the result of the verfication.
 */
static int test_ctamc_verify_cb (int ok, X509_STORE_CTX *ctx)
{
    int cert_error = X509_STORE_CTX_get_error(ctx);
    X509 *current_cert = X509_STORE_CTX_get_current_cert(ctx);

    BIO_printf(bio_out, "enter function: ok=%d cert_error=%d\n", ok, cert_error);

    if (!ok) {
        if (current_cert) {
            X509_NAME_print_ex_fp(stdout,
                                  X509_get_subject_name(current_cert),
                                  0, XN_FLAG_ONELINE);
            printf("\n");
        }
        BIO_printf(bio_out, "%s warning %d at %d depth lookup: %s\n",
                   X509_STORE_CTX_get0_parent_ctx(ctx) ? "[CRL path]" : "",
                   cert_error,
                   X509_STORE_CTX_get_error_depth(ctx),
                   X509_verify_cert_error_string(cert_error));
    }
    /*
     * Normally not Kosher, but we'll always allow verification to pass
     * since we know the client is using a self-signed cert.
     */
    return (1);
}

/*
 * This routine establishes a TLS server using the cert/key provided
 * by the caller.  It then creates a TLS client to perform a TLS
 * handshake with the server.
 *
 * Returns 0 if the handshake is established, non-zero for a failed handshake.
 */
static int test_ctamc_tls_session (X509 *s_cert, EVP_PKEY *s_key, X509 *c_cert, EVP_PKEY *c_key)
{
    SSL_CTX *c_ctx, *s_ctx;
    SSL *c_ssl, *s_ssl;
    int rc;

    SSL_library_init();
    SSL_load_error_strings();

    c_ctx = SSL_CTX_new(TLSv1_2_method());
    s_ctx = SSL_CTX_new(TLSv1_2_method());
    if ((c_ctx == NULL) || (s_ctx == NULL)) {
        BIO_printf(bio_out, "SSL_CTX_new failed.\n");
        return 1;
    }

    /*
     * Setup server-side context
     */
    SSL_CTX_use_certificate(s_ctx, s_cert);
    SSL_CTX_use_PrivateKey(s_ctx, s_key);
    SSL_CTX_set_verify(s_ctx, SSL_VERIFY_PEER, test_ctamc_verify_cb);

    /*
     * Setup client-side context
     */
    SSL_CTX_use_certificate(c_ctx, c_cert);
    SSL_CTX_use_PrivateKey(c_ctx, c_key);

    /*
     * Create the client and server sessions
     */
    c_ssl = SSL_new(c_ctx);
    s_ssl = SSL_new(s_ctx);
    if ((c_ssl == NULL) || (s_ssl == NULL)) {
        BIO_printf(bio_out, "SSL_new failed.\n");
        return 1;
    }

    /*
     * Establish the session between the client/server
     */
    rc = doit(s_ssl, c_ssl, 256);

    SSL_free(s_ssl);
    SSL_free(c_ssl);
    SSL_CTX_free(c_ctx);
    SSL_CTX_free(s_ctx);

    return rc;
}



/*
 * *********************************************************************************
 * *********************************************************************************
 *
 * Routines below this point were shamelessly taken from ssl/ssltest.c
 *
 * *********************************************************************************
 * *********************************************************************************
 */
static void print_details (SSL *c_ssl, const char *prefix)
{
    const SSL_CIPHER *ciph;
    X509 *cert;

    ciph = SSL_get_current_cipher(c_ssl);
    BIO_printf(bio_out, "%s%s, cipher %s %s",
               prefix,
               SSL_get_version(c_ssl),
               SSL_CIPHER_get_version(ciph), SSL_CIPHER_get_name(ciph));
    cert = SSL_get_peer_certificate(c_ssl);
    if (cert != NULL) {
        EVP_PKEY *pkey = X509_get_pubkey(cert);
        if (pkey != NULL) {
            if (0) {
                ;
            }
#ifndef OPENSSL_NO_RSA
            else if (pkey->type == EVP_PKEY_RSA && pkey->pkey.rsa != NULL
                     && pkey->pkey.rsa->n != NULL) {
                BIO_printf(bio_out, ", %d bit RSA",
                           BN_num_bits(pkey->pkey.rsa->n));
            }
#endif
#ifndef OPENSSL_NO_DSA
            else if (pkey->type == EVP_PKEY_DSA && pkey->pkey.dsa != NULL
                     && pkey->pkey.dsa->p != NULL) {
                BIO_printf(bio_out, ", %d bit DSA",
                           BN_num_bits(pkey->pkey.dsa->p));
            }
#endif
            EVP_PKEY_free(pkey);
        }
        X509_free(cert);
    }
    /*
     * The SSL API does not allow us to look at temporary RSA/DH keys,
     * otherwise we should print their lengths too
     */
    BIO_printf(bio_out, "\n");
}


#define W_READ  1
#define W_WRITE 2
#define C_DONE  1
#define S_DONE  2
static int doit (SSL *s_ssl, SSL *c_ssl, long count)
{
    char *cbuf = NULL, *sbuf = NULL;
    long bufsiz;
    long cw_num = count, cr_num = count;
    long sw_num = count, sr_num = count;
    int ret = 1;
    BIO *c_to_s = NULL;
    BIO *s_to_c = NULL;
    BIO *c_bio = NULL;
    BIO *s_bio = NULL;
    int c_r, c_w, s_r, s_w;
    int i, j;
    int done = 0;
    int c_write, s_write;
    int do_server = 0, do_client = 0;
    int max_frag = 5 * 1024;

    bufsiz = count > 40 * 1024 ? 40 * 1024 : count;

    if ((cbuf = OPENSSL_malloc(bufsiz)) == NULL) {
        goto err;
    }
    if ((sbuf = OPENSSL_malloc(bufsiz)) == NULL) {
        goto err;
    }

    memset(cbuf, 0, bufsiz);
    memset(sbuf, 0, bufsiz);

    c_to_s = BIO_new(BIO_s_mem());
    s_to_c = BIO_new(BIO_s_mem());
    if ((s_to_c == NULL) || (c_to_s == NULL)) {
        ERR_print_errors_fp(stderr);
        goto err;
    }

    c_bio = BIO_new(BIO_f_ssl());
    s_bio = BIO_new(BIO_f_ssl());
    if ((c_bio == NULL) || (s_bio == NULL)) {
        ERR_print_errors_fp(stderr);
        goto err;
    }

    SSL_set_connect_state(c_ssl);
    SSL_set_bio(c_ssl, s_to_c, c_to_s);
    SSL_set_max_send_fragment(c_ssl, max_frag);
    BIO_set_ssl(c_bio, c_ssl, BIO_NOCLOSE);

    SSL_set_accept_state(s_ssl);
    SSL_set_bio(s_ssl, c_to_s, s_to_c);
    SSL_set_max_send_fragment(s_ssl, max_frag);
    BIO_set_ssl(s_bio, s_ssl, BIO_NOCLOSE);

    c_r = 0;
    s_r = 1;
    c_w = 1;
    s_w = 0;
    c_write = 1, s_write = 0;

    /* We can always do writes */
    for (;; ) {
        do_server = 0;
        do_client = 0;

        i = (int)BIO_pending(s_bio);
        if ((i && s_r) || s_w) {
            do_server = 1;
        }

        i = (int)BIO_pending(c_bio);
        if ((i && c_r) || c_w) {
            do_client = 1;
        }

        if (do_server) {
            if (SSL_in_init(s_ssl)) {
                printf("server waiting in SSL_accept - %s\n",
                       SSL_state_string_long(s_ssl));
            }
        }

        if (do_client) {
            if (SSL_in_init(c_ssl)) {
                printf("client waiting in SSL_connect - %s\n",
                       SSL_state_string_long(c_ssl));
            }
        }

        if (!do_client && !do_server) {
            fprintf(stdout, "ERROR IN STARTUP\n");
            ERR_print_errors_fp(stderr);
            break;
        }
        if (do_client && !(done & C_DONE)) {
            if (c_write) {
                j = (cw_num > bufsiz) ? (int)bufsiz : (int)cw_num;
                i = BIO_write(c_bio, cbuf, j);
                if (i < 0) {
                    c_r = 0;
                    c_w = 0;
                    if (BIO_should_retry(c_bio)) {
                        if (BIO_should_read(c_bio)) {
                            c_r = 1;
                        }
                        if (BIO_should_write(c_bio)) {
                            c_w = 1;
                        }
                    } else {
                        fprintf(stderr, "ERROR in CLIENT\n");
                        ERR_print_errors_fp(stderr);
                        goto err;
                    }
                } else if (i == 0) {
                    fprintf(stderr, "SSL CLIENT STARTUP FAILED\n");
                    goto err;
                } else {
                    printf("client wrote %d\n", i);
                    /* ok */
                    s_r = 1;
                    c_write = 0;
                    cw_num -= i;
                    if (max_frag > 1029) {
                        SSL_set_max_send_fragment(c_ssl, max_frag -= 5);
                    }
                }
            } else {
                i = BIO_read(c_bio, cbuf, bufsiz);
                if (i < 0) {
                    c_r = 0;
                    c_w = 0;
                    if (BIO_should_retry(c_bio)) {
                        if (BIO_should_read(c_bio)) {
                            c_r = 1;
                        }
                        if (BIO_should_write(c_bio)) {
                            c_w = 1;
                        }
                    } else {
                        fprintf(stderr, "ERROR in CLIENT\n");
                        ERR_print_errors_fp(stderr);
                        goto err;
                    }
                } else if (i == 0) {
                    fprintf(stderr, "SSL CLIENT STARTUP FAILED\n");
                    goto err;
                } else {
                    printf("client read %d\n", i);
                    cr_num -= i;
                    if (sw_num > 0) {
                        s_write = 1;
                        s_w = 1;
                    }
                    if (cr_num <= 0) {
                        s_write = 1;
                        s_w = 1;
                        done = S_DONE | C_DONE;
                    }
                }
            }
        }

        if (do_server && !(done & S_DONE)) {
            if (!s_write) {
                i = BIO_read(s_bio, sbuf, bufsiz);
                if (i < 0) {
                    s_r = 0;
                    s_w = 0;
                    if (BIO_should_retry(s_bio)) {
                        if (BIO_should_read(s_bio)) {
                            s_r = 1;
                        }
                        if (BIO_should_write(s_bio)) {
                            s_w = 1;
                        }
                    } else {
                        fprintf(stderr, "ERROR in SERVER\n");
                        ERR_print_errors_fp(stderr);
                        goto err;
                    }
                } else if (i == 0) {
                    ERR_print_errors_fp(stderr);
                    fprintf(stderr,
                            "SSL SERVER STARTUP FAILED in SSL_read\n");
                    goto err;
                } else {
                    printf("server read %d\n", i);
                    sr_num -= i;
                    if (cw_num > 0) {
                        c_write = 1;
                        c_w = 1;
                    }
                    if (sr_num <= 0) {
                        s_write = 1;
                        s_w = 1;
                        c_write = 0;
                    }
                }
            } else {
                j = (sw_num > bufsiz) ? (int)bufsiz : (int)sw_num;
                i = BIO_write(s_bio, sbuf, j);
                if (i < 0) {
                    s_r = 0;
                    s_w = 0;
                    if (BIO_should_retry(s_bio)) {
                        if (BIO_should_read(s_bio)) {
                            s_r = 1;
                        }
                        if (BIO_should_write(s_bio)) {
                            s_w = 1;
                        }
                    } else {
                        fprintf(stderr, "ERROR in SERVER\n");
                        ERR_print_errors_fp(stderr);
                        goto err;
                    }
                } else if (i == 0) {
                    ERR_print_errors_fp(stderr);
                    fprintf(stderr,
                            "SSL SERVER STARTUP FAILED in SSL_write\n");
                    goto err;
                } else {
                    printf("server wrote %d\n", i);
                    sw_num -= i;
                    s_write = 0;
                    c_r = 1;
                    if (sw_num <= 0) {
                        done |= S_DONE;
                    }
                    if (max_frag > 1029) {
                        SSL_set_max_send_fragment(s_ssl, max_frag -= 5);
                    }
                }
            }
        }

        if ((done & S_DONE) && (done & C_DONE)) {
            break;
        }
    }
    SSL_SESSION_print(bio_out, SSL_get_session(c_ssl));
    print_details(c_ssl, "DONE: ");
    ret = 0;
err:
    /*
     * We have to set the BIO's to NULL otherwise they will be
     * OPENSSL_free()ed twice.  Once when th s_ssl is SSL_free()ed and again
     * when c_ssl is SSL_free()ed. This is a hack required because s_ssl and
     * c_ssl are sharing the same BIO structure and SSL_set_bio() and
     * SSL_free() automatically BIO_free non NULL entries. You should not
     * normally do this or be required to do this
     */
    if (s_ssl != NULL) {
        s_ssl->rbio = NULL;
        s_ssl->wbio = NULL;
    }
    if (c_ssl != NULL) {
        c_ssl->rbio = NULL;
        c_ssl->wbio = NULL;
    }

    if (c_to_s != NULL) {
        BIO_free(c_to_s);
    }
    if (s_to_c != NULL) {
        BIO_free(s_to_c);
    }
    if (c_bio != NULL) {
        BIO_free_all(c_bio);
    }
    if (s_bio != NULL) {
        BIO_free_all(s_bio);
    }

    if (cbuf) {
        OPENSSL_free(cbuf);
    }
    if (sbuf) {
        OPENSSL_free(sbuf);
    }

    return (ret);
}

