#define _BSD_SOURCE 1           /* Or gethostname won't be declared properly
                                 * on Linux and GNU platforms. */
#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>

#ifdef  OPENSSL_SYSNAME_WIN32
# include <windows.h>
# include <Lmcons.h>
#else
# include <sys/resource.h>
# include <pwd.h>
#endif

#define USE_SOCKETS
#include "e_os.h"

#define _XOPEN_SOURCE 500       /* Or isascii won't be declared properly on
                                 * VMS (at least with DECompHP C).  */
#include <ctype.h>

#include <openssl/bio.h>
#include <openssl/crypto.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/ssl.h>
#ifndef OPENSSL_NO_ENGINE
# include <openssl/engine.h>
#endif
#include <openssl/err.h>
#include <openssl/rand.h>
#ifndef OPENSSL_NO_RSA
# include <openssl/rsa.h>
#endif
#ifndef OPENSSL_NO_DSA
# include <openssl/dsa.h>
#endif
#ifndef OPENSSL_NO_DH
# include <openssl/dh.h>
#endif
#include <openssl/bn.h>
#include "../ssl/tls1.h"

#define _XOPEN_SOURCE_EXTENDED  1 /* Or gethostname won't be declared
                                   * properly on Compaq platforms (at least
                                   * with DEC C). Do not try to put it
                                   * earlier, or IPv6 includes get screwed... */
#ifdef OPENSSL_SYSNAME_WIN32
# include <winsock.h>
#else
# include OPENSSL_UNISTD
#endif

#define ALGO_RSA        1
#define ALGO_DSA        2
#define ALGO_ECDSA      3

/**
 * Arbitrary max length for paths (needed to allow for user-specified directories)
 */
#define SSL_PATHLEN 500
/**
 * Arbitrary max length for command strings (to system call) (derived from previous revision).
 */
#define SSL_CMDLEN 256

#define HASH_MD5        1
#define HASH_SHA1       2
#define HASH_SHA224     3
#define HASH_SHA256     4
#define HASH_SHA384     5
#define HASH_SHA512     6

/*
 * Note: the following values correspond to the entries
 *       in the certs[] table below.  If you add an entry to
 *       the table, you'll need a new #define here.  Be sure
 *       to keep them in order!!!
 *       Yes, this is weak, but this is just UT code.
 */
#define CERT_RSA_SHA1_IDX   0
#define CERT_DSA_SHA1_IDX   1
#define CERT_ECDSA_SHA1_IDX   2
#define CERT_RSA_MD5_IDX      3
#define CERT_RSA_SHA224_IDX   4
#define CERT_RSA_SHA256_IDX   5
#define CERT_RSA_SHA384_IDX   6
#define CERT_RSA_SHA512_IDX   7
#define CERT_ECDSA_SHA224_IDX   8
#define CERT_ECDSA_SHA256_IDX   9
#define CERT_ECDSA_SHA384_IDX   10
#define CERT_ECDSA_SHA512_IDX   11

typedef struct cert_tc {
    char cert[SSL_PATHLEN + 1];
    char key[SSL_PATHLEN + 1];
    const char *cert_file;
    const char *key_file;
    const int algo;
    const int hash;
} CERT_TC;

#define NUM_CERTS            12
CERT_TC certs[] = {
    /* CERT_RSA_SHA1_IDX */
    {
     "",
     "",
     "s_cert_rsa_sha1",
     "s_key_rsa_sha1",
     ALGO_RSA,
     HASH_SHA1},
    /* CERT_DSA_SHA1_IDX */
    {
     "",
     "",
     "s_cert_dsa_sha1",
     "s_key_dsa_sha1",
     ALGO_DSA,
     HASH_SHA1},
    /* CERT_ECDSA_SHA1_IDX */
    {
     "",
     "",
     "s_cert_ecdsa_sha1",
     "s_key_ecdsa_sha1",
     ALGO_ECDSA,
     HASH_SHA1},
    /* CERT_RSA_MD5_IDX */
    {
     "",
     "",
     "s_cert_rsa_md5",
     "s_key_rsa_md5",
     ALGO_RSA,
     HASH_MD5},
    /* CERT_RSA_SHA224_IDX */
    {
     "",
     "",
     "s_cert_rsa_sha224",
     "s_key_rsa_sha224",
     ALGO_RSA,
     HASH_SHA224},
    /* CERT_RSA_SHA256_IDX */
    {
     "",
     "",
     "s_cert_rsa_sha256",
     "s_key_rsa_sha256",
     ALGO_RSA,
     HASH_SHA256},
    /* CERT_RSA_SHA384_IDX */
    {
     "",
     "",
     "s_cert_rsa_sha384",
     "s_key_rsa_sha384",
     ALGO_RSA,
     HASH_SHA384},
    /* CERT_RSA_SHA512_IDX */
    {
     "",
     "",
     "s_cert_rsa_sha512",
     "s_key_rsa_sha512",
     ALGO_RSA,
     HASH_SHA512},
    /* CERT_ECDSA_SHA224_IDX */
    {
     "",
     "",
     "s_cert_ecdsa_sha224",
     "s_key_ecdsa_sha224",
     ALGO_ECDSA,
     HASH_SHA224},
    /* CERT_ECDSA_SHA256_IDX */
    {
     "",
     "",
     "s_cert_ecdsa_sha256",
     "s_key_ecdsa_sha256",
     ALGO_ECDSA,
     HASH_SHA256},
    /* CERT_ECDSA_SHA384_IDX */
    {
     "",
     "",
     "s_cert_ecdsa_sha384",
     "s_key_ecdsa_sha384",
     ALGO_ECDSA,
     HASH_SHA384},
    /* CERT_ECDSA_SHA512_IDX */
    {
     "",
     "",
     "s_cert_ecdsa_sha512",
     "s_key_ecdsa_sha512",
     ALGO_ECDSA,
     HASH_SHA512},
};

typedef struct cipher_suite_tc {
    const char *cipher_suite;
    const char *desc;
    const int server_cert_idx;
    const int expect_failure;   /* for negative test cases */
    void (*prep_function) (void); /* pre-test specific logic */
    void (*post_function) (void); /* post-test specific logic */
    const int secg_valid;       /* is this test case valid for tls.secg.org
                                 * interop? */
    const int fips_valid;       /* can this test case run in fips mode? */
} CIPHER_SUITE_TC;

#define NUM_TEST_CASES  (sizeof(testcases)/sizeof(CIPHER_SUITE_TC))

CIPHER_SUITE_TC testcases[] = {
    /* rfc 5246 */
#if !defined(OPENSSL_NO_RSA)
    {
     TLS1_TXT_RSA_WITH_NULL_SHA256,
     "RSA NULL SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     0,
     },
    {
     TLS1_TXT_RSA_WITH_AES_128_SHA256,
     "RSA AES-128 CBC SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_RSA_WITH_AES_256_SHA256,
     "RSA AES-256 CBC SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
#endif
#if !defined(OPENSSL_NO_DH)
# if !defined(OPENSSL_NO_RSA)
    {
     TLS1_TXT_DHE_RSA_WITH_AES_128_SHA256,
     "DHE RSA AES-128 CBC SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_DHE_RSA_WITH_AES_256_SHA256,
     "DHE RSA AES-256 CBC SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
# endif
# if !defined(OPENSSL_NO_DSA)
    {
     TLS1_TXT_DHE_DSS_WITH_AES_128_SHA256,
     "DHE DSS AES-128 CBC SHA256 with DSA-SHA1 server cert",
     CERT_DSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_DHE_DSS_WITH_AES_256_SHA256,
     "DHE DSS AES-256 CBC SHA256 with DSA-SHA1 server cert",
     CERT_DSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
# endif
#endif
    /* rfc 5288 */
#if !defined(OPENSSL_NO_RSA)
    {
     TLS1_TXT_RSA_WITH_AES_128_GCM_SHA256,
     "RSA AES128 GCM SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_RSA_WITH_AES_256_GCM_SHA384,
     "RSA AES256 GCM SHA384 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
#endif
#if !defined(OPENSSL_NO_DH)
# if !defined(OPENSSL_NO_RSA)
    {
     TLS1_TXT_DHE_RSA_WITH_AES_128_GCM_SHA256,
     "DHE RSA AES128 GCM SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_DHE_RSA_WITH_AES_256_GCM_SHA384,
     "DHE RSA AES256 GCM SHA384 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
# endif
# if !defined(OPENSSL_NO_DSA)
    {
     TLS1_TXT_DHE_DSS_WITH_AES_128_GCM_SHA256,
     "DHE DSS AES128 GCM SHA256 with DSA-SHA1 server cert",
     CERT_DSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_DHE_DSS_WITH_AES_256_GCM_SHA384,
     "DHE DSS AES256 GCM SHA384 with DSA-SHA1 server cert",
     CERT_DSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
# endif
    {
     TLS1_TXT_ADH_WITH_AES_128_GCM_SHA256,
     "Anonymous DH AES128 GCM SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ADH_WITH_AES_256_GCM_SHA384,
     "Anonymous DH AES256 GCM SHA384 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
#endif

    /* rfc 5289 */
    {
     TLS1_TXT_ECDHE_ECDSA_WITH_AES_128_SHA256,
     "ECDHE AES128 CBC SHA256 with ECDSA-SHA1 server cert",
     CERT_ECDSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ECDHE_ECDSA_WITH_AES_256_SHA384,
     "ECDHE AES256 CBC SHA384 with ECDSA-SHA1 server cert",
     CERT_ECDSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ECDH_ECDSA_WITH_AES_128_SHA256,
     "ECDH AES128 CBC SHA256 with ECDSA-SHA1 server cert",
     CERT_ECDSA_SHA1_IDX,
     1,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ECDH_ECDSA_WITH_AES_256_SHA384,
     "ECDH AES256 CBC SHA384 with ECDSA-SHA1 server cert",
     CERT_ECDSA_SHA1_IDX,
     1,
     NULL,
     NULL,
     1,
     1,
     },
#if !defined(OPENSSL_NO_RSA)
    {
     TLS1_TXT_ECDHE_RSA_WITH_AES_128_SHA256,
     "ECDHE AES128 CBC SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     0,                         /* tls.secg.org doesn't appear to have RSA
                                 * signed certs */
     1,
     },
    {
     TLS1_TXT_ECDHE_RSA_WITH_AES_256_SHA384,
     "ECDHE AES256 CBC SHA384 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     0,                         /* tls.secg.org doesn't appear to have RSA
                                 * signed certs */
     1,
     },
#endif
    /*
     * These are currently not working, can't get any ECDH-RSA cipher suites
     * to work.  Does not work on stock 0.9.8l as well.  There is demo code
     * to test this in demos/ssltest-ecc.  But there appears to be no
     * support for ECDH_RSA (defined in RFC 4492).  Looking at the other
     * legacy DH-RSA cipher suites in s3_lib.c, they are disabled. OpenSSL
     * does not implement non-ephemeral DH. Hence, these will not work.
     * Taking a look at ssl_cert_type(), there isn't even a handler to
     * return SSL_PKEY_DH_DSA as a cert type.
     *
     * Note: openssl 1.0.0d does support these.  They added a new capability
     * bit called SSL_kECDHr
     *
     * { TLS1_TXT_ECDH_RSA_WITH_AES_128_CBC_SHA256, "ECDH AES128 CBC SHA256
     * with RSA-SHA1 server cert", CERT_RSA_SHA1_IDX }, {
     * TLS1_TXT_ECDH_RSA_WITH_AES_256_CBC_SHA384, "ECDH AES256 CBC SHA384
     * with RSA-SHA1 server cert", CERT_RSA_SHA1_IDX },
     */
    {
     TLS1_TXT_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
     "ECDHE AES128 GCM SHA256 with ECDSA-SHA1 server cert",
     CERT_ECDSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
     "ECDHE AES256 GCM SHA384 with ECDSA-SHA1 server cert",
     CERT_ECDSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ECDH_ECDSA_WITH_AES_128_GCM_SHA256,
     "ECDH AES128 GCM SHA256 with ECDSA-SHA1 server cert",
     CERT_ECDSA_SHA1_IDX,
     1,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ECDH_ECDSA_WITH_AES_256_GCM_SHA384,
     "ECDH AES256 GCM SHA384 with ECDSA-SHA1 server cert",
     CERT_ECDSA_SHA1_IDX,
     1,
     NULL,
     NULL,
     1,
     1,
     },
#if !defined(OPENSSL_NO_RSA)
    {
     TLS1_TXT_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
     "ECDHE AES128 GCM SHA256 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     0,                         /* tls.secg.org doesn't appear to have RSA
                                 * signed certs */
     1,
     },
    {
     TLS1_TXT_ECDHE_RSA_WITH_AES_256_GCM_SHA384,
     "ECDHE AES256 GCM SHA384 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     0,                         /* tls.secg.org doesn't appear to have RSA
                                 * signed certs */
     1,
     },
#endif
    /*
     * ECDH-RSA is not implemented in OpenSSL {
     * TLS1_TXT_ECDH_RSA_WITH_AES_128_GCM_SHA256, "ECDH AES128 GCM SHA256
     * with RSA-SHA1 server cert", CERT_RSA_SHA1_IDX }, {
     * TLS1_TXT_ECDH_RSA_WITH_AES_256_GCM_SHA384, "ECDH AES256 GCM SHA384
     * with RSA-SHA1 server cert", CERT_RSA_SHA1_IDX },
     */

    /*
     * spot check a few EC cipher suites using SHA2 hash algorithms
     */
    {
     TLS1_TXT_ECDH_ECDSA_WITH_AES_128_GCM_SHA256,
     "ECDH AES128 GCM SHA256 with ECDSA-SHA224 server cert",
     CERT_ECDSA_SHA224_IDX,
     1,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ECDH_ECDSA_WITH_AES_128_GCM_SHA256,
     "ECDH AES128 GCM SHA256 with ECDSA-SHA256 server cert",
     CERT_ECDSA_SHA256_IDX,
     1,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ECDH_ECDSA_WITH_AES_128_GCM_SHA256,
     "ECDH AES128 GCM SHA256 with ECDSA-SHA384 server cert",
     CERT_ECDSA_SHA384_IDX,
     1,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_ECDH_ECDSA_WITH_AES_128_GCM_SHA256,
     "ECDH AES128 GCM SHA256 with ECDSA-SHA512 server cert",
     CERT_ECDSA_SHA512_IDX,
     1,
     NULL,
     NULL,
     1,
     1,
     },

    /*
     * spot check a few RSA cipher suites using SHA2 hash algorithms
     */
#if !defined(OPENSSL_NO_RSA)
    {
     TLS1_TXT_RSA_WITH_AES_256_GCM_SHA384,
     "RSA AES256 GCM SHA384 with RSA-MD5 server cert",
     CERT_RSA_MD5_IDX,
     0,
     NULL,
     NULL,
     1,
     0,
     },
    {
     TLS1_TXT_RSA_WITH_AES_256_GCM_SHA384,
     "RSA AES256 GCM SHA384 with RSA-SHA224 server cert",
     CERT_RSA_SHA224_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_RSA_WITH_AES_256_GCM_SHA384,
     "RSA AES256 GCM SHA384 with RSA-SHA256 server cert",
     CERT_RSA_SHA256_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_RSA_WITH_AES_256_GCM_SHA384,
     "RSA AES256 GCM SHA384 with RSA-SHA384 server cert",
     CERT_RSA_SHA384_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
    {
     TLS1_TXT_RSA_WITH_AES_256_GCM_SHA384,
     "RSA AES256 GCM SHA384 with RSA-SHA512 server cert",
     CERT_RSA_SHA512_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },

    /* legacy */
    {
     TLS1_TXT_RSA_WITH_AES_128_SHA,
     "RSA AES128 CBC SHA1 with RSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
#endif
#if !defined(OPENSSL_NO_DH)
# if !defined(OPENSSL_NO_DSA)
    {
     TLS1_TXT_DHE_DSS_WITH_AES_128_SHA,
     "DHE DSS AES128 CBC SHA1 with DSA-SHA1 server cert",
     CERT_DSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     },
# endif
# if !defined(OPENSSL_NO_RSA)
    {
     TLS1_TXT_DHE_RSA_WITH_AES_128_SHA,
     "DHE RSA AES128 CBC SHA1 with DSA-SHA1 server cert",
     CERT_RSA_SHA1_IDX,
     0,
     NULL,
     NULL,
     1,
     1,
     }
# endif
#endif
};

static int passwd_cb(char *buf, int size, int rwflag, void *userdata);
#ifndef OPENSSL_NO_RSA
static RSA MS_CALLBACK *tmp_rsa_cb(SSL *s, int is_export, int keylength);
static void free_tmp_rsa(void);
#endif
#define APP_CALLBACK_STRING "Test Callback Argument"
struct app_verify_arg {
    char *string;
    int app_verify;
    int allow_proxy_certs;
    char *proxy_auth;
    char *proxy_cond;
};

static BIO *bio_err = NULL;
static BIO *bio_stdout = NULL;

static char *cipher = NULL;
static int verbose = 0;
static int debug = 0;
static int fips_m = 0;

static char *CApath = NULL, *CAfile = NULL;
static char *curve_name = NULL;
static char workpath[SSL_PATHLEN + 1];

/* Utility functions for performance timing */
#define START   0
#define STOP    1
#define TM_START 0

#ifndef OPENSSL_SYSNAME_WIN32
double app_tminterval(int stop)
{
    double ret = 0;
    struct rusage rus;
    struct rusage rus_children;
    struct timeval now;
    struct timeval now_children;
    static struct timeval tmstart;
    static struct timeval tmstart_children;

    getrusage(RUSAGE_SELF, &rus);
    getrusage(RUSAGE_CHILDREN, &rus_children);
    now = rus.ru_utime;
    now_children = rus_children.ru_utime;

    if (stop == TM_START) {
        tmstart = now;
        tmstart_children = now_children;
    } else {
        ret = ((now.tv_sec + now.tv_usec * 1e-6)
               - (tmstart.tv_sec + tmstart.tv_usec * 1e-6));
        ret += ((now_children.tv_sec + now_children.tv_usec * 1e-6)
                - (tmstart_children.tv_sec +
                   tmstart_children.tv_usec * 1e-6));

        printf("\nTiming stats\n");
        printf("-----------------------------\n");
        printf("Parent ctx switches %lu %lu\n", rus.ru_nvcsw, rus.ru_nivcsw);
        printf("Parent messages %lu %lu\n", rus.ru_msgsnd, rus.ru_msgrcv);
        printf("Children ctx switches %lu %lu\n",
               rus_children.ru_nvcsw, rus_children.ru_nivcsw);
        printf("Children messages %lu %lu\n",
               rus_children.ru_msgsnd, rus_children.ru_msgrcv);
    }

    return ret;
}

static double Time_F(int s)
{
    return app_tminterval(s);
}
#endif

static const char rnd_seed[] =
    "string to make the random number generator think it has entropy";

// int doit_biopair(SSL *s_ssl,SSL *c_ssl,long bytes,clock_t *s_time,clock_t
// *c_time);
int doit(SSL *s_ssl, SSL *c_ssl, long bytes);
int doit_secg(SSL *c_ssl, long bytes);

static void print_details(SSL *c_ssl, const char *prefix)
{
    const SSL_CIPHER *ciph;
    X509 *cert;

    ciph = SSL_get_current_cipher(c_ssl);
    BIO_printf(bio_stdout, "%s%s, cipher %s %s",
               prefix,
               SSL_get_version(c_ssl),
               SSL_CIPHER_get_version(ciph), SSL_CIPHER_get_name(ciph));
    cert = SSL_get_peer_certificate(c_ssl);
    if (cert != NULL) {
        EVP_PKEY *pkey = X509_get_pubkey(cert);
        if (pkey != NULL) {
            if (0) ;
#ifndef OPENSSL_NO_RSA
            else if (pkey->type == EVP_PKEY_RSA && pkey->pkey.rsa != NULL
                     && pkey->pkey.rsa->n != NULL) {
                BIO_printf(bio_stdout, ", %d bit RSA",
                           BN_num_bits(pkey->pkey.rsa->n));
            }
#endif
#ifndef OPENSSL_NO_DSA
            else if (pkey->type == EVP_PKEY_DSA && pkey->pkey.dsa != NULL
                     && pkey->pkey.dsa->p != NULL) {
                BIO_printf(bio_stdout, ", %d bit DSA",
                           BN_num_bits(pkey->pkey.dsa->p));
            }
#endif
            EVP_PKEY_free(pkey);
        }
        X509_free(cert);
    }
    /*
     * The SSL API does not allow us to look at temporary RSA/DH keys,
     * otherwise we should print their lengths too
     */
    BIO_printf(bio_stdout, "\n");
}

static void lock_dbg_cb(int mode, int type, const char *file, int line)
{
    static int modes[CRYPTO_NUM_LOCKS]; /* = {0, 0, ... } */
    const char *errstr = NULL;
    int rw;

    rw = mode & (CRYPTO_READ | CRYPTO_WRITE);
    if (!((rw == CRYPTO_READ) || (rw == CRYPTO_WRITE))) {
        errstr = "invalid mode";
        goto err;
    }
    if (type < 0 || type >= CRYPTO_NUM_LOCKS) {
        errstr = "type out of bounds";
        goto err;
    }

    if (mode & CRYPTO_LOCK) {
        if (modes[type]) {
            errstr = "already locked";
            /*
             * must not happen in a single-threaded program (would deadlock)
             */
            goto err;
        }

        modes[type] = rw;
    } else if (mode & CRYPTO_UNLOCK) {
        if (!modes[type]) {
            errstr = "not locked";
            goto err;
        }

        if (modes[type] != rw) {
            errstr = (rw == CRYPTO_READ) ?
                "CRYPTO_r_unlock on write lock" :
                "CRYPTO_w_unlock on read lock";
        }

        modes[type] = 0;
    } else {
        errstr = "invalid mode";
        goto err;
    }

 err:
    if (errstr) {
        /* we cannot use bio_err here */
        fprintf(stderr,
                "openssl (lock_dbg_cb): %s (mode=%d, type=%d) at %s:%d\n",
                errstr, mode, type, file, line);
    }
}

void create_certs(void)
{
    int i, rc;
    int runcmd = 1;
    char cmd[SSL_CMDLEN + 1] = "";
    char *hash_cmd;

    for (i = 0; i < NUM_CERTS; i++) {

        /*
         * MD5 doesn't work in FIPS mode
         */
        if (fips_m) {
            if (certs[i].hash == HASH_MD5) {
                continue;
            }
        }

        /* set the hash algorithm to use on the certificate */
        switch (certs[i].hash) {
        case HASH_MD5:
            hash_cmd = "-md5";
            break;
        case HASH_SHA1:
            hash_cmd = "-sha1";
            break;
        case HASH_SHA224:
            hash_cmd = "-sha224";
            break;
        case HASH_SHA256:
            hash_cmd = "-sha256";
            break;
        case HASH_SHA384:
            hash_cmd = "-sha384";
            break;
        case HASH_SHA512:
            hash_cmd = "-sha512";
            break;
        default:
            printf("\nBogus Hash algorithm type specified in certs array\n");
            exit(-1);
            break;
        }

        /*
         * create the key and cert request
         */
        switch (certs[i].algo) {
        case ALGO_RSA:
            BIO_snprintf(cmd, sizeof(cmd) - 1,
                         "../apps/openssl req -newkey rsa:1024 %s -keyout %s -keyform PEM -passout pass:hello -out %s/req -outform PEM -batch\n",
                         hash_cmd, certs[i].key, workpath);
            break;
        case ALGO_DSA:
            BIO_snprintf(cmd, sizeof(cmd) - 1,
                         "../apps/openssl dsaparam 1024 -out %s/dsaparms\n",
                         workpath);
            rc = system(cmd);
            if (rc != 0) {
                printf("\nUnable to create dsaparms (rc=%d)\n", rc);
                exit(-1);
            }
            BIO_snprintf(cmd, sizeof(cmd) - 1,
                         "../apps/openssl req -newkey dsa:%s/dsaparms %s -keyout %s -keyform PEM -passout pass:hello -out %s/req -outform PEM -batch\n",
                         workpath, hash_cmd, certs[i].key, workpath);
            break;
        case ALGO_ECDSA:
            BIO_snprintf(cmd, sizeof(cmd) - 1,
                         "../apps/openssl ecparam -name %s -out %s/ecparms\n",
                         (curve_name != NULL ? curve_name : "prime192v1"),
                         workpath);
            rc = system(cmd);
            if (rc != 0) {
                printf("\nUnable to create ecparms (rc=%d)\n", rc);
                exit(-1);
            }
            BIO_snprintf(cmd, sizeof(cmd) - 1,
                         "../apps/openssl req -newkey ec:%s/ecparms %s -keyout %s -keyform PEM -passout pass:hello -out %s/req -outform PEM -batch\n",
                         workpath, hash_cmd, certs[i].key, workpath);
            break;
        default:
            runcmd = 0;
            break;
        }

        if (runcmd) {
            if (debug) {
                printf("\nRunning command: %s\n", cmd);
            }
            rc = system(cmd);
            if (rc != 0) {
                printf("\nUnable to create key (rc=%d [0x%x])\n", rc, rc);
                exit(-1);
            }
        }

        /*
         * create the cert using the appropriate hash algorithm
         */
        BIO_snprintf(cmd, sizeof(cmd) - 1,
                     "../apps/openssl x509 %s -req -days 10 -in %s/req -signkey %s -out %s -passin pass:hello\n",
                     hash_cmd, workpath, certs[i].key, certs[i].cert);
        if (debug) {
            printf("\nRunning command: %s\n", cmd);
        }
        rc = system(cmd);
        if (rc != 0) {
            printf("\nUnable to create cert (rc=%d [0x%x])\n", rc, rc);
            exit(-1);
        }
    }

}

#if !defined(OPENSSL_NO_DH)
static DH *get_dh1024dsa()
{
    static unsigned char dh1024_p[] = {
        0xC8, 0x00, 0xF7, 0x08, 0x07, 0x89, 0x4D, 0x90, 0x53, 0xF3, 0xD5,
        0x00,
        0x21, 0x1B, 0xF7, 0x31, 0xA6, 0xA2, 0xDA, 0x23, 0x9A, 0xC7, 0x87,
        0x19,
        0x3B, 0x47, 0xB6, 0x8C, 0x04, 0x6F, 0xFF, 0xC6, 0x9B, 0xB8, 0x65,
        0xD2,
        0xC2, 0x5F, 0x31, 0x83, 0x4A, 0xA7, 0x5F, 0x2F, 0x88, 0x38, 0xB6,
        0x55,
        0xCF, 0xD9, 0x87, 0x6D, 0x6F, 0x9F, 0xDA, 0xAC, 0xA6, 0x48, 0xAF,
        0xFC,
        0x33, 0x84, 0x37, 0x5B, 0x82, 0x4A, 0x31, 0x5D, 0xE7, 0xBD, 0x52,
        0x97,
        0xA1, 0x77, 0xBF, 0x10, 0x9E, 0x37, 0xEA, 0x64, 0xFA, 0xCA, 0x28,
        0x8D,
        0x9D, 0x3B, 0xD2, 0x6E, 0x09, 0x5C, 0x68, 0xC7, 0x45, 0x90, 0xFD,
        0xBB,
        0x70, 0xC9, 0x3A, 0xBB, 0xDF, 0xD4, 0x21, 0x0F, 0xC4, 0x6A, 0x3C,
        0xF6,
        0x61, 0xCF, 0x3F, 0xD6, 0x13, 0xF1, 0x5F, 0xBC, 0xCF, 0xBC, 0x26,
        0x9E,
        0xBC, 0x0B, 0xBD, 0xAB, 0x5D, 0xC9, 0x54, 0x39,
    };
    static unsigned char dh1024_g[] = {
        0x3B, 0x40, 0x86, 0xE7, 0xF3, 0x6C, 0xDE, 0x67, 0x1C, 0xCC, 0x80,
        0x05,
        0x5A, 0xDF, 0xFE, 0xBD, 0x20, 0x27, 0x74, 0x6C, 0x24, 0xC9, 0x03,
        0xF3,
        0xE1, 0x8D, 0xC3, 0x7D, 0x98, 0x27, 0x40, 0x08, 0xB8, 0x8C, 0x6A,
        0xE9,
        0xBB, 0x1A, 0x3A, 0xD6, 0x86, 0x83, 0x5E, 0x72, 0x41, 0xCE, 0x85,
        0x3C,
        0xD2, 0xB3, 0xFC, 0x13, 0xCE, 0x37, 0x81, 0x9E, 0x4C, 0x1C, 0x7B,
        0x65,
        0xD3, 0xE6, 0xA6, 0x00, 0xF5, 0x5A, 0x95, 0x43, 0x5E, 0x81, 0xCF,
        0x60,
        0xA2, 0x23, 0xFC, 0x36, 0xA7, 0x5D, 0x7A, 0x4C, 0x06, 0x91, 0x6E,
        0xF6,
        0x57, 0xEE, 0x36, 0xCB, 0x06, 0xEA, 0xF5, 0x3D, 0x95, 0x49, 0xCB,
        0xA7,
        0xDD, 0x81, 0xDF, 0x80, 0x09, 0x4A, 0x97, 0x4D, 0xA8, 0x22, 0x72,
        0xA1,
        0x7F, 0xC4, 0x70, 0x56, 0x70, 0xE8, 0x20, 0x10, 0x18, 0x8F, 0x2E,
        0x60,
        0x07, 0xE7, 0x68, 0x1A, 0x82, 0x5D, 0x32, 0xA2,
    };
    DH *dh;

    if ((dh = DH_new()) == NULL)
        return (NULL);
    dh->p = BN_bin2bn(dh1024_p, sizeof(dh1024_p), NULL);
    dh->g = BN_bin2bn(dh1024_g, sizeof(dh1024_g), NULL);
    if ((dh->p == NULL) || (dh->g == NULL)) {
        DH_free(dh);
        return (NULL);
    }
    dh->length = 160;
    return (dh);
}
#endif

int run_test_case_secg(CIPHER_SUITE_TC * tc)
{
    SSL *c_ssl = NULL;
    SSL_CTX *c_ctx = NULL;
    BIO *conn;
    int ret = 1;
    long bytes = 256L;

    printf("\nUsing:  cert=%s\n", certs[tc->server_cert_idx].cert);
    printf("         key=%s\n", certs[tc->server_cert_idx].key);
    printf("      cipher=%s\n", tc->cipher_suite);

#ifdef OPENSSL_FIPS
    if (fips_m) {
        if (!FIPS_mode_set(1)) {
            ERR_load_crypto_strings();
            ERR_print_errors(BIO_new_fp(stderr, BIO_NOCLOSE));
            EXIT(1);
        } else {
            printf("*** IN FIPS MODE ***\n");
        }
    }
#endif

    SSL_library_init();
    SSL_load_error_strings();

    c_ctx = SSL_CTX_new(TLSv1_2_method());
    if (c_ctx == NULL) {
        ERR_print_errors(bio_err);
        goto end;
    }
    if (SSL_CTX_set_cipher_list(c_ctx, tc->cipher_suite) != 1) {
        printf("\nWARNING: no ciphers set");
        goto end;
    }
    c_ssl = SSL_new(c_ctx);
    SSL_set_session(c_ssl, NULL);

    conn = BIO_new_connect("tls.secg.org:443");
    if (BIO_do_connect(conn) <= 0) {
        printf("\nUnable to connect to tls.secg.org");
        goto end;
    }
    SSL_set_bio(c_ssl, conn, conn);

    ret = doit_secg(c_ssl, bytes);
    if (!verbose) {
        print_details(c_ssl, "");
    }
    BIO_printf(bio_stdout, "1 handshakes of %ld bytes done\n", bytes);

 end:
    if (c_ssl != NULL)
        SSL_free(c_ssl);
    if (c_ctx != NULL)
        SSL_CTX_free(c_ctx);
    return ret;
}

int run_test_case(CIPHER_SUITE_TC * tc)
{
    SSL_CTX *s_ctx = NULL;
    SSL_CTX *c_ctx = NULL;
    const SSL_METHOD *meth = NULL;
    SSL *c_ssl, *s_ssl;
    int ret = 1;
    long bytes = 256L;
#if !defined(OPENSSL_NO_DH)
    DH *dh;
#endif
#ifndef OPENSSL_NO_EC
    int nid;
    EC_KEY *ecdh;
#endif

#ifdef OPENSSL_FIPS
    if (fips_m) {
        if (!FIPS_mode_set(1)) {
            ERR_load_crypto_strings();
            ERR_print_errors(BIO_new_fp(stderr, BIO_NOCLOSE));
            EXIT(1);
        } else {
            printf("*** IN FIPS MODE ***\n");
        }
    }
#endif

    SSL_library_init();
    SSL_load_error_strings();

    printf("\nUsing:  cert=%s\n", certs[tc->server_cert_idx].cert);
    printf("         key=%s\n", certs[tc->server_cert_idx].key);
    printf("      cipher=%s\n", tc->cipher_suite);

    /* only TLS 1.2 supported for now */
    meth = TLSv1_2_method();
    // meth=TLSv1_method();

    c_ctx = SSL_CTX_new(meth);
    s_ctx = SSL_CTX_new(meth);
    if ((c_ctx == NULL) || (s_ctx == NULL)) {
        ERR_print_errors(bio_err);
        goto end;
    }

    SSL_CTX_set_default_passwd_cb(s_ctx, passwd_cb);
#if 0
    SSL_CTX_set_default_passwd_cb(c_ctx, passwd_cb);
#endif

    if (!SSL_CTX_set_cipher_list(c_ctx, tc->cipher_suite)) {
        printf("\nWARNING: no ciphers set");
    }
    if (!SSL_CTX_set_cipher_list(s_ctx, tc->cipher_suite)) {
        printf("\nWARNING: no ciphers set");
    }
#if !defined(OPENSSL_NO_DH)
    /* use SSL_OP_SINGLE_DH_USE to avoid small subgroup attacks */
    SSL_CTX_set_options(s_ctx, SSL_OP_SINGLE_DH_USE);
    dh = get_dh1024dsa();
    SSL_CTX_set_tmp_dh(s_ctx, dh);
    DH_free(dh);
#endif

#ifndef OPENSSL_NO_EC
    /* setup ECDH curve - only p192 is supported for now */
    if (curve_name != NULL) {
        nid = OBJ_sn2nid(curve_name);
    } else {
        nid = NID_X9_62_prime192v1;
    }
    ecdh = EC_KEY_new_by_curve_name(nid);
    if (ecdh == NULL) {
        printf("\nERROR: unable to create curve\n");
        exit(-1);
    }
    SSL_CTX_set_tmp_ecdh(s_ctx, ecdh);
    SSL_CTX_set_options(s_ctx, SSL_OP_SINGLE_ECDH_USE);
    EC_KEY_free(ecdh);
#endif

#ifndef OPENSSL_NO_RSA
    SSL_CTX_set_tmp_rsa_callback(s_ctx, tmp_rsa_cb);
#endif

    if (!SSL_CTX_use_certificate_file(s_ctx, certs[tc->server_cert_idx].cert,
                                      SSL_FILETYPE_PEM)) {
        ERR_print_errors(bio_err);
    } else
        if (!SSL_CTX_use_PrivateKey_file
            (s_ctx, certs[tc->server_cert_idx].key, SSL_FILETYPE_PEM)) {
        ERR_print_errors(bio_err);
        goto end;
    }

    if ((!SSL_CTX_load_verify_locations(s_ctx, CAfile, CApath)) ||
        (!SSL_CTX_set_default_verify_paths(s_ctx)) ||
        (!SSL_CTX_load_verify_locations(c_ctx, CAfile, CApath)) ||
        (!SSL_CTX_set_default_verify_paths(c_ctx))) {
        /* fprintf(stderr,"SSL_load_verify_locations\n"); */
        ERR_print_errors(bio_err);
        /* goto end; */
    }

    {
        int session_id_context = 0;
        SSL_CTX_set_session_id_context(s_ctx, (void *)&session_id_context,
                                       sizeof session_id_context);
    }

    c_ssl = SSL_new(c_ctx);
    s_ssl = SSL_new(s_ctx);

#ifndef OPENSSL_NO_KRB5
    if (c_ssl && c_ssl->kssl_ctx) {
        char localhost[MAXHOSTNAMELEN + 2];

        if (gethostname(localhost, sizeof localhost - 1) == 0) {
            localhost[sizeof localhost - 1] = '\0';
            if (strlen(localhost) == sizeof localhost - 1) {
                BIO_printf(bio_err, "localhost name too long\n");
                goto end;
            }
            kssl_ctx_setstring(c_ssl->kssl_ctx, KSSL_SERVER, localhost);
        }
    }
#endif                          /* OPENSSL_NO_KRB5 */

    SSL_set_session(c_ssl, NULL);

    /*
     * Is there a test-case specific function to be called
     * before we start the TLS handshake?
     */
    if (tc->prep_function) {
        (tc->prep_function) ();
    }

    ret = doit(s_ssl, c_ssl, bytes);

    if (!verbose) {
        print_details(c_ssl, "");
    }
    BIO_printf(bio_stdout, "1 handshakes of %ld bytes done\n", bytes);

    SSL_free(s_ssl);
    SSL_free(c_ssl);

    /*
     * Is there a test-case specific function to be called
     * after we finish the TLS handshake?
     */
    if (tc->post_function) {
        (tc->post_function) ();
    }

 end:
    if (s_ctx != NULL)
        SSL_CTX_free(s_ctx);
    if (c_ctx != NULL)
        SSL_CTX_free(c_ctx);

#ifndef OPENSSL_NO_RSA
    free_tmp_rsa();
#endif
    return (ret);

}

static void print_test_array(void)
{
    int i;

    printf("\n\nTest cases...\n\n");
    for (i = 0; i < NUM_TEST_CASES; i++) {
        printf("ID: %d\n", i);
        printf("Cipher suite: %s\n", testcases[i].cipher_suite);
        printf("Test description: %s\n", testcases[i].desc);
        printf("Expect failure (negative test): %d\n",
               testcases[i].expect_failure);
        printf("Valid for SECG test: %d\n", testcases[i].secg_valid);
        printf("Valid for FIPS: %d\n", testcases[i].fips_valid);
        printf("\n");
    }
    exit(0);
}

static void print_usage(void)
{
    printf("\n\nusage: ssltest2 [args ...]\n\n");
    printf
        ("-c           - Generate certs required for the test (certs will be in /tmp/<processid>)\n");
    printf("-CApath      - PEM format directory of CA's\n");
    printf("-CAfile      - PEM format file of CA's\n");
    printf("-curve arg   - Named curve to use for Elliptic Curve crypto\n");
    printf("-d           - Enable debug output\n");
    printf("-dir         - working directory for storing files\n");
#ifdef OPENSSL_FIPS
    printf("-fips        - Enable FIPS mode while running the test\n");
#endif
    printf
        ("-i arg       - Iterate test multiple times, arg=iteration count\n");
    printf("-list        - List the test case matrix\n");
    printf("-secg        - Run interop test with tls.secg.org test server\n");
    printf("-tc arg      - Run an individual test case\n");
    printf("-v           - Enable verbose output\n");
}

int main(int argc, char *argv[])
{
    int i;
    int make_certs = 0;
    int single_tc = 0;
    int secg = 0;
    int single_tc_num = 0;
    int ret;
    int rc = 1;                 /* assume failure */
    int iterations = 1;
    char cmd[SSL_CMDLEN + 1];
#ifdef OPENSSL_SYSNAME_WIN32
    TCHAR username[UNLEN + 1];
    DWORD size = UNLEN + 1;
#else
    double d = 0;
    struct passwd *pwd;
    char  *username = "";

    pwd = getpwuid(getuid());
    if (pwd) {
        username = pwd->pw_name;
    }
#endif
    verbose = 0;
    cipher = 0;

    bio_err = BIO_new_fp(stderr, BIO_NOCLOSE);

    CRYPTO_set_locking_callback(lock_dbg_cb);

    /* enable memory leak checking unless explicitly disabled */
    if (!((getenv("OPENSSL_DEBUG_MEMORY") != NULL)
          && (0 == strcmp(getenv("OPENSSL_DEBUG_MEMORY"), "off")))) {
        CRYPTO_malloc_debug_init();
        CRYPTO_set_mem_debug_options(V_CRYPTO_MDEBUG_ALL);
    } else {
        /* OPENSSL_DEBUG_MEMORY=off */
        CRYPTO_set_mem_debug_functions(0, 0, 0, 0, 0);
    }
    CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);

    RAND_seed(rnd_seed, sizeof rnd_seed);

    bio_stdout = BIO_new_fp(stdout, BIO_NOCLOSE);

    argc--;
    argv++;

    while (argc >= 1) {
        if (strcmp(*argv, "-v") == 0)
            verbose = 1;
        else if (strcmp(*argv, "-d") == 0)
            debug = 1;
        else if (strcmp(*argv, "-c") == 0)
            make_certs = 1;
        else if (strcmp(*argv, "-secg") == 0)
            secg = 1;
#ifdef OPENSSL_FIPS
        else if (strcmp(*argv, "-fips") == 0)
            fips_m = 1;
#endif
        else if (strcmp(*argv, "-list") == 0)
            print_test_array();
        else if (strcmp(*argv, "-CApath") == 0) {
            if (--argc < 1)
                exit(-1);
            CApath = *(++argv);
        } else if (strcmp(*argv, "-CAfile") == 0) {
            if (--argc < 1)
                exit(-1);
            CAfile = *(++argv);
        } else if (strcmp(*argv, "-dir") == 0) {
            if (--argc < 1)
                exit(-1);
            BIO_snprintf(workpath, sizeof(workpath) - 1, "%s", *(++argv));
        } else if (strcmp(*argv, "-tc") == 0) {
            single_tc = 1;
            if (--argc < 1)
                exit(-1);
            single_tc_num = atoi(*(++argv));
        } else if (strcmp(*argv, "-i") == 0) {
            if (--argc < 1)
                exit(-1);
            iterations = atoi(*(++argv));
        } else if (strcmp(*argv, "-curve") == 0) {
            if (--argc < 1)
                exit(-1);
            curve_name = *(++argv);
        } else {
            printf("Unknown option %s\n", *argv);
            print_usage();
            exit(1);
            break;
        }
        argc--;
        argv++;
    }

    if (fips_m && !single_tc) {
        printf
            ("\nFIPS mode is only supported when running a single test case.  Use -tc option with -fips. \n");
        exit(1);
    }

    /**
     * Set up the cert/key directory paths based on CLI and/or directory derived
     * from the user id subdirectory of the hard-coded root (/tmp). Note that path
     * lengths are maxed at PATHLEN chars.
     */
    if (strlen(workpath) == 0) {
#ifdef OPENSSL_SYSNAME_WIN32
        GetUserName(username, &size);
        BIO_snprintf(workpath, sizeof(workpath) - 1, "/tmp/%s", username);
#else
	/*
	 * Build temp directory name that uses the username and PID
	 */
        int mypid = 0;
        mypid = getpid();
        BIO_snprintf(workpath, sizeof(workpath) - 1, "/tmp/%s/%d", username, mypid);
#endif
    }

    /**
     * Create (potentially user-defined or user-specific) directory
     * if it does not exist for storing certs and keys.
     */
    BIO_snprintf(cmd, sizeof(cmd) - 1, "mkdir -p \"%s\"\n", workpath);
    if (debug) {
        printf("\nRunning command: %s\n", cmd);
    }
    rc = system(cmd);
    if (rc != 0) {
        printf
            ("\nUnable to find/create key directory (rc=%d [0x%x]). Will not exit (maybe directory already exists).\n",
             rc, rc);
        perror("reason");
        // exit(-1);
    }
    /*
     * rc = mkdir(workpath, S_IRWXU | S_IRWXG); if(rc != 0 && errno !=
     * EEXIST) { printf("\nUnable to find/create key directory (rc=%d
     * [0x%x]). Will not exit (maybe directory already exists).\n", rc, rc);
     * perror("reason"); exit(-1); }
     */

    for (i = 0; i < NUM_CERTS; i++) {
        if (strlen(certs[i].cert) == 0) {
            BIO_snprintf(certs[i].cert, sizeof(certs[i].cert) - 1, "%s/%s",
                         workpath, certs[i].cert_file);
        }

        if (strlen(certs[i].key) == 0) {
            BIO_snprintf(certs[i].key, sizeof(certs[i].key) - 1, "%s/%s",
                         workpath, certs[i].key_file);
        }
    }

    /*
     * Create the keys and certs used for the test
     */
    if (make_certs) {
        create_certs();
    }
#ifndef OPENSSL_SYSNAME_WIN32
    Time_F(START);
#endif
    while (iterations-- > 0) {
        if (single_tc) {
            if (single_tc_num >= NUM_TEST_CASES) {
                printf
                    ("\nInvalid test case number provide.  Value range is 0 to %d\n",
                     (int)NUM_TEST_CASES - 1);
                goto bail;
            }
            if (fips_m && !testcases[single_tc_num].fips_valid) {
                printf("\nTest case %d is not valid in FIPS mode\n",
                       single_tc_num);
                goto bail;
            }
            if (secg && !testcases[single_tc_num].secg_valid) {
                printf
                    ("\nTest case %d is not valid for tls.secg.org interop testing\n",
                     single_tc_num);
                goto bail;
            }

            printf("\nRunning single test case# %d.", single_tc_num);
            if (secg) {
                ret = run_test_case_secg(&testcases[single_tc_num]);
            } else {
                ret = run_test_case(&testcases[single_tc_num]);
            }
            if ((ret != 0 && !testcases[single_tc_num].expect_failure) ||
                (ret == 0 && testcases[single_tc_num].expect_failure)) {
                printf("\nTest case %d failed!!!\n", single_tc_num);
                goto bail;
            } else {
                printf("Test case %d passed.\n", single_tc_num);
            }
        } else {
            for (i = 0; i < NUM_TEST_CASES; i++) {
                if (fips_m && !testcases[i].fips_valid) {
                    continue;
                }
                if (secg && !testcases[i].secg_valid) {
                    continue;
                }
                printf("\n\nExecute TLS test case: %s", testcases[i].desc);
                if (secg) {
                    ret = run_test_case_secg(&testcases[i]);
                } else {
                    ret = run_test_case(&testcases[i]);
                }
                printf("\nret=%d (0x%x)\n", ret, ret);
                if ((ret != 0 && !testcases[i].expect_failure) ||
                    (ret == 0 && testcases[i].expect_failure)) {
                    printf("Test case %d failed!!!\n", i);
                    goto bail;
                } else {
                    printf("Test case %d passed.\n", i);
                }
            }
        }
    }
#ifndef OPENSSL_SYSNAME_WIN32
    d = Time_F(STOP);
    printf("\nTotal CPU seconds: %f\n", d);
#endif

    rc = 0;
 bail:
    if (bio_stdout != NULL)
        BIO_free(bio_stdout);
#ifndef OPENSSL_NO_ENGINE
    ENGINE_cleanup();
#endif
    CRYPTO_cleanup_all_ex_data();
    ERR_free_strings();
    ERR_remove_state(0);
    EVP_cleanup();
    CRYPTO_mem_leaks(bio_err);
    if (bio_err != NULL)
        BIO_free(bio_err);
    return rc;
}

#define W_READ  1
#define W_WRITE 2
#define C_DONE  1
#define S_DONE  2

static unsigned char secg_get[] = {
    0x47, 0x45, 0x54, 0x20, 0x2f, 0x3f, 0x61, 0x63, 0x74, 0x69, 0x6f, 0x6e,
    0x3d, 0x69, 0x6e, 0x74,
    0x72, 0x6f, 0x20, 0x48, 0x54, 0x54, 0x50, 0x2f, 0x31, 0x2e, 0x31, 0x0d,
    0x0a, 0x48, 0x6f, 0x73,
    0x74, 0x3a, 0x20, 0x74, 0x6c, 0x73, 0x2e, 0x73, 0x65, 0x63, 0x67, 0x2e,
    0x6f, 0x72, 0x67, 0x0d,
    0x0a, 0x55, 0x73, 0x65, 0x72, 0x2d, 0x41, 0x67, 0x65, 0x6e, 0x74, 0x3a,
    0x20, 0x4d, 0x6f, 0x7a,
    0x69, 0x6c, 0x6c, 0x61, 0x2f, 0x35, 0x2e, 0x30, 0x20, 0x28, 0x58, 0x31,
    0x31, 0x3b, 0x20, 0x55,
    0x3b, 0x20, 0x4c, 0x69, 0x6e, 0x75, 0x78, 0x20, 0x69, 0x36, 0x38, 0x36,
    0x3b, 0x20, 0x65, 0x6e,
    0x2d, 0x55, 0x53, 0x3b, 0x20, 0x72, 0x76, 0x3a, 0x31, 0x2e, 0x39, 0x2e,
    0x32, 0x2e, 0x31, 0x38,
    0x29, 0x20, 0x47, 0x65, 0x63, 0x6b, 0x6f, 0x2f, 0x32, 0x30, 0x31, 0x31,
    0x30, 0x36, 0x32, 0x38,
    0x20, 0x55, 0x62, 0x75, 0x6e, 0x74, 0x75, 0x2f, 0x31, 0x30, 0x2e, 0x30,
    0x34, 0x20, 0x28, 0x6c,
    0x75, 0x63, 0x69, 0x64, 0x29, 0x20, 0x46, 0x69, 0x72, 0x65, 0x66, 0x6f,
    0x78, 0x2f, 0x33, 0x2e,
    0x36, 0x2e, 0x31, 0x38, 0x0d, 0x0a, 0x41, 0x63, 0x63, 0x65, 0x70, 0x74,
    0x3a, 0x20, 0x74, 0x65,
    0x78, 0x74, 0x2f, 0x68, 0x74, 0x6d, 0x6c, 0x2c, 0x61, 0x70, 0x70, 0x6c,
    0x69, 0x63, 0x61, 0x74,
    0x69, 0x6f, 0x6e, 0x2f, 0x78, 0x68, 0x74, 0x6d, 0x6c, 0x2b, 0x78, 0x6d,
    0x6c, 0x2c, 0x61, 0x70,
    0x70, 0x6c, 0x69, 0x63, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x2f, 0x78, 0x6d,
    0x6c, 0x3b, 0x71, 0x3d,
    0x30, 0x2e, 0x39, 0x2c, 0x2a, 0x2f, 0x2a, 0x3b, 0x71, 0x3d, 0x30, 0x2e,
    0x38, 0x0d, 0x0a, 0x41,
    0x63, 0x63, 0x65, 0x70, 0x74, 0x2d, 0x4c, 0x61, 0x6e, 0x67, 0x75, 0x61,
    0x67, 0x65, 0x3a, 0x20,
    0x65, 0x6e, 0x2d, 0x75, 0x73, 0x2c, 0x65, 0x6e, 0x3b, 0x71, 0x3d, 0x30,
    0x2e, 0x35, 0x0d, 0x0a,
    0x41, 0x63, 0x63, 0x65, 0x70, 0x74, 0x2d, 0x45, 0x6e, 0x63, 0x6f, 0x64,
    0x69, 0x6e, 0x67, 0x3a,
    0x20, 0x67, 0x7a, 0x69, 0x70, 0x2c, 0x64, 0x65, 0x66, 0x6c, 0x61, 0x74,
    0x65, 0x0d, 0x0a, 0x41,
    0x63, 0x63, 0x65, 0x70, 0x74, 0x2d, 0x43, 0x68, 0x61, 0x72, 0x73, 0x65,
    0x74, 0x3a, 0x20, 0x49,
    0x53, 0x4f, 0x2d, 0x38, 0x38, 0x35, 0x39, 0x2d, 0x31, 0x2c, 0x75, 0x74,
    0x66, 0x2d, 0x38, 0x3b,
    0x71, 0x3d, 0x30, 0x2e, 0x37, 0x2c, 0x2a, 0x3b, 0x71, 0x3d, 0x30, 0x2e,
    0x37, 0x0d, 0x0a, 0x4b,
    0x65, 0x65, 0x70, 0x2d, 0x41, 0x6c, 0x69, 0x76, 0x65, 0x3a, 0x20, 0x31,
    0x31, 0x35, 0x0d, 0x0a,
    0x43, 0x6f, 0x6e, 0x6e, 0x65, 0x63, 0x74, 0x69, 0x6f, 0x6e, 0x3a, 0x20,
    0x6b, 0x65, 0x65, 0x70,
    0x2d, 0x61, 0x6c, 0x69, 0x76, 0x65, 0x0d, 0x0a, 0x0d, 0x0a
};

static int secg_len = 394;

int doit_secg(SSL *c_ssl, long count)
{
    int ret = 1;
    int err;
    unsigned char buf[2048];

    printf("\nstub");
    if (SSL_connect(c_ssl) <= 0) {
        printf("\nError connecting SSL object");
        goto err;
    }
    SSL_write(c_ssl, secg_get, secg_len);
    err = SSL_read(c_ssl, buf, sizeof(buf));
    if (err <= 0) {
        printf("\nSSL read error");
        goto err;
    }
    printf("\n%s\n", buf);
    SSL_shutdown(c_ssl);
    ret = 0;
 err:
    return ret;
}

int doit(SSL *s_ssl, SSL *c_ssl, long count)
{
    MS_STATIC char cbuf[1024 * 8], sbuf[1024 * 8];
    long cw_num = count, cr_num = count;
    long sw_num = count, sr_num = count;
    int ret = 1;
    BIO *c_to_s = NULL;
    BIO *s_to_c = NULL;
    BIO *c_bio = NULL;
    BIO *s_bio = NULL;
    int c_r, c_w, s_r, s_w;
    int i, j;
    int done = 0;
    int c_write, s_write;
    int do_server = 0, do_client = 0;

    memset(cbuf, 0, sizeof(cbuf));
    memset(sbuf, 0, sizeof(sbuf));

    c_to_s = BIO_new(BIO_s_mem());
    s_to_c = BIO_new(BIO_s_mem());
    if ((s_to_c == NULL) || (c_to_s == NULL)) {
        ERR_print_errors(bio_err);
        goto err;
    }

    c_bio = BIO_new(BIO_f_ssl());
    s_bio = BIO_new(BIO_f_ssl());
    if ((c_bio == NULL) || (s_bio == NULL)) {
        ERR_print_errors(bio_err);
        goto err;
    }

    SSL_set_connect_state(c_ssl);
    SSL_set_bio(c_ssl, s_to_c, c_to_s);
    BIO_set_ssl(c_bio, c_ssl, BIO_NOCLOSE);

    SSL_set_accept_state(s_ssl);
    SSL_set_bio(s_ssl, c_to_s, s_to_c);
    BIO_set_ssl(s_bio, s_ssl, BIO_NOCLOSE);

    c_r = 0;
    s_r = 1;
    c_w = 1;
    s_w = 0;
    c_write = 1, s_write = 0;

    /* We can always do writes */
    for (;;) {
        do_server = 0;
        do_client = 0;

        i = (int)BIO_pending(s_bio);
        if ((i && s_r) || s_w)
            do_server = 1;

        i = (int)BIO_pending(c_bio);
        if ((i && c_r) || c_w)
            do_client = 1;

        if (do_server && debug) {
            if (SSL_in_init(s_ssl))
                printf("server waiting in SSL_accept - %s\n",
                       SSL_state_string_long(s_ssl));
            /*
             * else if (s_write) printf("server:SSL_write()\n"); else
             * printf("server:SSL_read()\n");
             */
        }

        if (do_client && debug) {
            if (SSL_in_init(c_ssl))
                printf("client waiting in SSL_connect - %s\n",
                       SSL_state_string_long(c_ssl));
            /*
             * else if (c_write) printf("client:SSL_write()\n"); else
             * printf("client:SSL_read()\n");
             */
        }

        if (!do_client && !do_server) {
            fprintf(stdout, "ERROR IN STARTUP\n");
            ERR_print_errors(bio_err);
            break;
        }
        if (do_client && !(done & C_DONE)) {
            if (c_write) {
                j = (cw_num > (long)sizeof(cbuf)) ?
                    (int)sizeof(cbuf) : (int)cw_num;
                i = BIO_write(c_bio, cbuf, j);
                if (i < 0) {
                    c_r = 0;
                    c_w = 0;
                    if (BIO_should_retry(c_bio)) {
                        if (BIO_should_read(c_bio))
                            c_r = 1;
                        if (BIO_should_write(c_bio))
                            c_w = 1;
                    } else {
                        fprintf(stderr, "ERROR in CLIENT\n");
                        ERR_print_errors(bio_err);
                        goto err;
                    }
                } else if (i == 0) {
                    fprintf(stderr, "SSL CLIENT STARTUP FAILED\n");
                    goto err;
                } else {
                    if (debug)
                        printf("client wrote %d\n", i);
                    /* ok */
                    s_r = 1;
                    c_write = 0;
                    cw_num -= i;
                }
            } else {
                i = BIO_read(c_bio, cbuf, sizeof(cbuf));
                if (i < 0) {
                    c_r = 0;
                    c_w = 0;
                    if (BIO_should_retry(c_bio)) {
                        if (BIO_should_read(c_bio))
                            c_r = 1;
                        if (BIO_should_write(c_bio))
                            c_w = 1;
                    } else {
                        fprintf(stderr, "ERROR in CLIENT\n");
                        ERR_print_errors(bio_err);
                        goto err;
                    }
                } else if (i == 0) {
                    fprintf(stderr, "SSL CLIENT STARTUP FAILED\n");
                    goto err;
                } else {
                    if (debug)
                        printf("client read %d\n", i);
                    cr_num -= i;
                    if (sw_num > 0) {
                        s_write = 1;
                        s_w = 1;
                    }
                    if (cr_num <= 0) {
                        s_write = 1;
                        s_w = 1;
                        done = S_DONE | C_DONE;
                    }
                }
            }
        }

        if (do_server && !(done & S_DONE)) {
            if (!s_write) {
                i = BIO_read(s_bio, sbuf, sizeof(cbuf));
                if (i < 0) {
                    s_r = 0;
                    s_w = 0;
                    if (BIO_should_retry(s_bio)) {
                        if (BIO_should_read(s_bio))
                            s_r = 1;
                        if (BIO_should_write(s_bio))
                            s_w = 1;
                    } else {
                        fprintf(stderr, "ERROR in SERVER\n");
                        ERR_print_errors(bio_err);
                        goto err;
                    }
                } else if (i == 0) {
                    ERR_print_errors(bio_err);
                    fprintf(stderr,
                            "SSL SERVER STARTUP FAILED in SSL_read\n");
                    goto err;
                } else {
                    if (debug)
                        printf("server read %d\n", i);
                    sr_num -= i;
                    if (cw_num > 0) {
                        c_write = 1;
                        c_w = 1;
                    }
                    if (sr_num <= 0) {
                        s_write = 1;
                        s_w = 1;
                        c_write = 0;
                    }
                }
            } else {
                j = (sw_num > (long)sizeof(sbuf)) ?
                    (int)sizeof(sbuf) : (int)sw_num;
                i = BIO_write(s_bio, sbuf, j);
                if (i < 0) {
                    s_r = 0;
                    s_w = 0;
                    if (BIO_should_retry(s_bio)) {
                        if (BIO_should_read(s_bio))
                            s_r = 1;
                        if (BIO_should_write(s_bio))
                            s_w = 1;
                    } else {
                        fprintf(stderr, "ERROR in SERVER\n");
                        ERR_print_errors(bio_err);
                        goto err;
                    }
                } else if (i == 0) {
                    ERR_print_errors(bio_err);
                    fprintf(stderr,
                            "SSL SERVER STARTUP FAILED in SSL_write\n");
                    goto err;
                } else {
                    if (debug)
                        printf("server wrote %d\n", i);
                    sw_num -= i;
                    s_write = 0;
                    c_r = 1;
                    if (sw_num <= 0)
                        done |= S_DONE;
                }
            }
        }

        if ((done & S_DONE) && (done & C_DONE))
            break;
    }

    if (verbose)
        print_details(c_ssl, "DONE: ");
    ret = 0;
 err:
    /*
     * We have to set the BIO's to NULL otherwise they will be
     * OPENSSL_free()ed twice.  Once when th s_ssl is SSL_free()ed and again
     * when c_ssl is SSL_free()ed. This is a hack required because s_ssl and
     * c_ssl are sharing the same BIO structure and SSL_set_bio() and
     * SSL_free() automatically BIO_free non NULL entries. You should not
     * normally do this or be required to do this
     */
    if (s_ssl != NULL) {
        s_ssl->rbio = NULL;
        s_ssl->wbio = NULL;
    }
    if (c_ssl != NULL) {
        c_ssl->rbio = NULL;
        c_ssl->wbio = NULL;
    }

    if (c_to_s != NULL)
        BIO_free(c_to_s);
    if (s_to_c != NULL)
        BIO_free(s_to_c);
    if (c_bio != NULL)
        BIO_free_all(c_bio);
    if (s_bio != NULL)
        BIO_free_all(s_bio);
    return (ret);
}

#ifndef OPENSSL_NO_RSA
static RSA *rsa_tmp = NULL;

static RSA MS_CALLBACK *tmp_rsa_cb(SSL *s, int is_export, int keylength)
{
    BIGNUM *bn = NULL;
    if (rsa_tmp == NULL) {
        bn = BN_new();
        rsa_tmp = RSA_new();
        if (!bn || !rsa_tmp || !BN_set_word(bn, RSA_F4)) {
            BIO_printf(bio_err, "Memory error...");
            goto end;
        }
        BIO_printf(bio_err, "Generating temp (%d bit) RSA key...", keylength);
        (void)BIO_flush(bio_err);
        if (!RSA_generate_key_ex(rsa_tmp, keylength, bn, NULL)) {
            BIO_printf(bio_err, "Error generating key.");
            RSA_free(rsa_tmp);
            rsa_tmp = NULL;
        }
 end:
        BIO_printf(bio_err, "\n");
        (void)BIO_flush(bio_err);
    }
    if (bn)
        BN_free(bn);
    return (rsa_tmp);
}

static void free_tmp_rsa(void)
{
    if (rsa_tmp != NULL) {
        RSA_free(rsa_tmp);
        rsa_tmp = NULL;
    }
}
#endif

static int passwd_cb(char *buf, int size, int rwflag, void *userdata)
{
    BIO_snprintf(buf, strlen("hello") + 1, "hello");
    return (strlen(buf));
}
