/*
 *  ssltest_cc.c
 *  Unit test for TLS Common Criteria mode
 *
 *  Copyright 2015 Cisco Systems, Inc. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/err.h>
#include <openssl/rsa.h>
#include <openssl/x509v3.h>
#include <openssl/evp.h>
#include <openssl/ssl.h>

#ifndef OPENSSL_FIPS
int main (int argc, char *argv[])
{
    printf("Common Criteria TLS tests will be skipped (not a FIPS build)...\n");
    return (0);
}
#else

#define HOSTNAME    "localhost"
#define KEY_TYPE_RSA    1
#define KEY_TYPE_EC     2
#define TLSv1_2     1
#define TLSv1_1     2
#define TLSv1_0     3
#define TLSvALL     4
#define CC_ON       1
#define CC_OFF      0
#define DH_1024     0
#define DH_2048     1
#define SERVER_VERIFY 0
#define MUTUAL_VERIFY 1
#define NO_EXTKEYUSAGE 0
#define BAD_VERSION 1
#define ADD_BOTH_EXTKEYUSAGE 2
#define ADD_CLIENT_EXTKEYUSAGE 3

typedef struct cc_test {
    /*
     * Server options
     */
    int s_version;
    char *s_cipher_suite;
    int s_cert_type;
    int s_cert_size;
    int s_ecdhe_nid;
    int s_dhe_size;
    /*
     * Client options
     */
    int c_version;
    char *c_cipher_suite;
    /*
     * Common options
     */
    int cert_options;
    int verify_type;
    int cc_enabled;
    int expected_result; /* 0=handshake completes, 1=handshake fails */
} CC_TEST;

/*
 * This matrix contains the unit test cases.
 * Note: since all tests are performed with FIPS enabled, we're unable to
 *       test key sizes not allowed by FIPS.  This includes the NIST prime 192 bit curve
 *       and RSA 1024 bit keys.
 *
 */
static CC_TEST test_matrix[] = {
    /* Server Auth only */
    /* TLS version tests */
    {TLSv1_2, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},
    {TLSv1_2, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", NO_EXTKEYUSAGE, SERVER_VERIFY, CC_OFF, 0},
    {TLSvALL, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},
    {TLSvALL, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_1, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},
    {TLSvALL, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_0, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 1},
    {TLSvALL, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_0, "", NO_EXTKEYUSAGE, SERVER_VERIFY, CC_OFF, 0},
    {TLSv1_0, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSvALL, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 1},
    /* TLS cipher suite tests */
    {TLSvALL, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSvALL, "AES128-SHA", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},
    {TLSvALL, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSvALL, "DES-CBC3-SHA", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 1},
    {TLSvALL, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSvALL, "DES-CBC3-SHA", NO_EXTKEYUSAGE, SERVER_VERIFY, CC_OFF, 0},
    {TLSvALL, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSvALL, "ECDHE-RSA-AES256-GCM-SHA384", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},

    {TLSvALL, "DHE-RSA-AES128-GCM-SHA256", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSvALL, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 1},
    {TLSvALL, "DHE-RSA-AES128-GCM-SHA256", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSvALL, "", NO_EXTKEYUSAGE, SERVER_VERIFY, CC_OFF, 0},
    {TLSvALL, "DHE-RSA-AES256-SHA256",     KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSvALL, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},

    /* TLS ephemeral DH/ECDH tests */
    {TLSvALL, "", KEY_TYPE_EC,  NID_X9_62_prime256v1, NID_X9_62_prime256v1, DH_2048,      TLSvALL, "ECDHE-ECDSA-AES256-GCM-SHA384", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},
    {TLSvALL, "", KEY_TYPE_EC,  NID_X9_62_prime256v1, NID_secp384r1,        DH_2048,      TLSvALL, "ECDHE-ECDSA-AES256-GCM-SHA384", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},
    {TLSvALL, "", KEY_TYPE_EC,  NID_X9_62_prime256v1, NID_secp521r1,        DH_2048,      TLSvALL, "ECDHE-ECDSA-AES256-GCM-SHA384", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},
    {TLSvALL, "", KEY_TYPE_EC,  NID_X9_62_prime256v1, NID_secp521r1,        DH_2048,      TLSvALL, "ECDHE-ECDSA-AES256-GCM-SHA384", NO_EXTKEYUSAGE, SERVER_VERIFY, CC_OFF, 0},
    {TLSvALL, "", KEY_TYPE_RSA, 2048,                 NID_X9_62_prime256v1, DH_1024,      TLSvALL, "DHE-RSA-AES256-SHA256", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 1},
    {TLSvALL, "", KEY_TYPE_RSA, 2048,                 NID_X9_62_prime256v1, DH_2048,      TLSvALL, "DHE-RSA-AES256-SHA256", NO_EXTKEYUSAGE, SERVER_VERIFY, CC_OFF, 0},

    /* TLS public key size tests */
    {TLSv1_2, "", KEY_TYPE_RSA, 4096,          NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},
    {TLSv1_2, "", KEY_TYPE_EC,  NID_secp384r1, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},
    {TLSv1_2, "", KEY_TYPE_EC,  NID_secp521r1, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", ADD_BOTH_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 0},


    /* Mutual Auth */
    {TLSv1_2, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", ADD_BOTH_EXTKEYUSAGE, MUTUAL_VERIFY, CC_ON, 0},

    /* Server Auth only negative tests */
    {TLSv1_2, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", NO_EXTKEYUSAGE, SERVER_VERIFY, CC_ON, 1},

    /* Mutual Auth negative tests */
    {TLSv1_2, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", ADD_CLIENT_EXTKEYUSAGE, MUTUAL_VERIFY, CC_ON, 1},

    /* Server Version negative tests */
    {TLSv1_2, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", BAD_VERSION, SERVER_VERIFY, CC_ON, 1},

    /* Client Version negative tests */
    {TLSv1_2, "", KEY_TYPE_RSA, 2048, NID_X9_62_prime256v1, DH_2048,      TLSv1_2, "", BAD_VERSION, MUTUAL_VERIFY, CC_ON, 1},

};


/* Some forward prototypes */
static EVP_PKEY* ssltest_keypair_create(int key_type, int key_size);
static int ssltest_cert_create(X509 **new_cert, EVP_PKEY *keypair, int server, int extkeyusag);
static int ssltest_run_test_case(int idx);
static DH *ssltest_get_dh1024(void);
static DH *ssltest_get_dh2048(void);
static int doit(SSL *s_ssl, SSL *c_ssl, long count);

/*
 * This is the entry point into the TLS test.
 */
int main (int argc, char *argv[])
{
    int rc = 0;
    int i, test_cnt;
    BIO *bio_err;

    bio_err = BIO_new_fp(stderr, BIO_NOCLOSE);

    /*
     * Enable memory leak checking unless explicitly disabled
     */
    if (!((getenv("OPENSSL_DEBUG_MEMORY") != NULL)
          && (0 == strcmp(getenv("OPENSSL_DEBUG_MEMORY"), "off")))) {
        CRYPTO_malloc_debug_init();
        CRYPTO_set_mem_debug_options(V_CRYPTO_MDEBUG_ALL);
    } else {
        /* OPENSSL_DEBUG_MEMORY=off */
        CRYPTO_set_mem_debug_functions(0, 0, 0, 0, 0);
    }
    CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);

    /*
     * All the tests assume FIPS is enabled
     */
    if (!FIPS_mode_set(1)) {
        ERR_load_crypto_strings();
        ERR_print_errors(BIO_new_fp(stderr, BIO_NOCLOSE));
        exit(1);
    }

    SSL_library_init();
    SSL_load_error_strings();

    /*
     * Check if the user only wants to run a single test case.
     */
    if (argc > 1) {
        i = atoi(argv[1]);
        printf("Running test case %d\n", i);
        rc = ssltest_run_test_case(i);
        if (rc != test_matrix[i].expected_result) {
            printf("TLS test failed (rc=%d)!!!\n", rc);
            rc = 1;
            goto err;
        } else {
            printf("Test case %d passed.\n", i);
        }
        goto end;
    }

    /*
     * Iterate through each test case in the matrix and run it
     */
    test_cnt = sizeof(test_matrix)/sizeof(test_matrix[0]);
    for (i = 0; i < test_cnt; i++) {
        printf("Starting test %d:\n", i);
        rc = ssltest_run_test_case(i);
        if (rc != test_matrix[i].expected_result) {
            printf("TLS test failed!!!\n");
            rc = 1;
            goto err;
        }
        printf("Finshed test %d:\n\n", i);
    }

    printf("All tests passed.\n");
    rc = 0;
    goto end;
err:
    ERR_print_errors_fp(stderr);
end:
    CRYPTO_cleanup_all_ex_data();
    ERR_free_strings();
    ERR_remove_state(0);
    EVP_cleanup();
    CRYPTO_mem_leaks(bio_err);
    if (bio_err != NULL) BIO_free(bio_err);
    exit(rc);
}

/*
 * Simple function that returns one of the TLS methods
 */
static const SSL_METHOD * ssltest_get_method(int m)
{
    switch (m) {
    case TLSv1_2:
        return TLSv1_2_method();
        break;
    case TLSv1_1:
        return TLSv1_1_method();
        break;
    case TLSv1_0:
        return TLSv1_method();
        break;
    default:
        return SSLv23_method();
    }
}

/*
 * This function does the work for an individual test case.  The test
 * case parameters are defined in test_matrix[].  The idx argument is
 * the entry in the matrix that will be executed.
 *
 * Each test case entry specifies how the server and client side
 * SSL contexts will be created, along with an expected outcome
 * of the handshake.  This function will create a keypair and certificate
 * to be used for the TLS server, setup both the client and
 * server side contexts, and attempt a handshake between the
 * two contexts.
 */
static int ssltest_run_test_case (int idx)
{
    SSL_CTX *c_ctx, *s_ctx;
    SSL *c_ssl, *s_ssl;
    int rc;
    X509 *s_cert = NULL;
    X509 *c_cert = NULL;
    EVP_PKEY *s_key = NULL;
    EVP_PKEY *c_key = NULL;
    X509_STORE *s_vfy = NULL;
    X509_STORE *c_vfy = NULL;
    EC_KEY *ecdh = NULL;
#ifndef OPENSSL_NO_DH
    DH *dh = NULL;
#endif

    /*
     * Create a key pair
     */
    s_key = ssltest_keypair_create (test_matrix[idx].s_cert_type, test_matrix[idx].s_cert_size);
    if (!s_key) {
        printf("Failed to create key: %d %d\n", test_matrix[idx].s_cert_type, test_matrix[idx].s_cert_size);
        exit(1);
    }

    if (test_matrix[idx].verify_type) {
        c_key = ssltest_keypair_create (test_matrix[idx].s_cert_type, test_matrix[idx].s_cert_size);
        if (!c_key) {
            printf("Failed to create key: %d %d\n", test_matrix[idx].s_cert_type, test_matrix[idx].s_cert_size);
	    exit(1);
        }
    }
    /*
     * Create a certificate to be used by the server
     */
    if (ssltest_cert_create (&s_cert, s_key, 1, test_matrix[idx].cert_options)) {
        printf("Failed to create cert\n");
        exit(1);
    }

    /*
     * Create a certificate to be used by the client
     */
    if (test_matrix[idx].verify_type) {
        if (ssltest_cert_create (&c_cert, c_key, 0, test_matrix[idx].cert_options)) {
            printf("Failed to create cert\n");
            exit(1);
        }
    }
    /*
     * Enable common criteria if this test case requires it
     */
    if (!FIPS_cc_mode() && test_matrix[idx].cc_enabled) {
        if (!FIPS_cc_set(1)) {
            ERR_load_crypto_strings();
            ERR_print_errors(BIO_new_fp(stderr, BIO_NOCLOSE));
            exit(1);
        }
    } else if (FIPS_cc_mode() && !test_matrix[idx].cc_enabled) {
        if (!FIPS_cc_set(0)) {
            ERR_load_crypto_strings();
            ERR_print_errors(BIO_new_fp(stderr, BIO_NOCLOSE));
            exit(1);
        }
    }

    /*
     * Create client & server contexts
     */
    c_ctx = SSL_CTX_new(ssltest_get_method(test_matrix[idx].c_version));
    s_ctx = SSL_CTX_new(ssltest_get_method(test_matrix[idx].s_version));
    if ((c_ctx == NULL) || (s_ctx == NULL)) {
        printf("SSL_CTX_new failed.\n");
        return 1;
    }

    /*
     * Setup server context
     *   1. Set the cert/key
     *   2. Set the cipher suite
     *   3. Set temp DH
     *   4. Set temp ECDH
     */
    if (test_matrix[idx].verify_type) {
        s_vfy = X509_STORE_new();
	X509_STORE_set_flags(s_vfy, 0);
	X509_STORE_add_cert(s_vfy, c_cert);
	SSL_CTX_set1_verify_cert_store(s_ctx, s_vfy);
	SSL_CTX_set_verify(s_ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);
    }
    SSL_CTX_use_certificate(s_ctx, s_cert);
    SSL_CTX_use_PrivateKey(s_ctx, s_key);
    if (strlen(test_matrix[idx].s_cipher_suite)) {
        SSL_CTX_set_cipher_list(s_ctx, test_matrix[idx].s_cipher_suite);
    }
    ecdh = EC_KEY_new_by_curve_name(test_matrix[idx].s_ecdhe_nid);
    if (ecdh == NULL) {
        printf("unable to create curve to set tmp ECDH\n");
        exit(1);
    }
    SSL_CTX_set_tmp_ecdh(s_ctx, ecdh);
    SSL_CTX_set_options(s_ctx, SSL_OP_SINGLE_ECDH_USE);
    EC_KEY_free(ecdh);
#ifndef OPENSSL_NO_DH
    if (test_matrix[idx].s_dhe_size == DH_1024) {
        dh = ssltest_get_dh1024();
    } else {
        dh = ssltest_get_dh2048();
    }
    SSL_CTX_set_tmp_dh(s_ctx, dh);
    DH_free(dh);
#endif


    /*
     * Setup client-side context
     *   1. Add server cert to client trust anchor store
     *   2. Set the cipher suite
     */
    c_vfy = X509_STORE_new();
    X509_STORE_set_flags(c_vfy, 0);
    X509_STORE_add_cert(c_vfy, s_cert);
    SSL_CTX_set1_verify_cert_store(c_ctx, c_vfy);
    if (strlen(test_matrix[idx].c_cipher_suite)) {
        SSL_CTX_set_cipher_list(c_ctx, test_matrix[idx].c_cipher_suite);
    }

    if (test_matrix[idx].verify_type) {
        SSL_CTX_use_certificate(c_ctx, c_cert);
	SSL_CTX_use_PrivateKey(c_ctx, c_key);
	SSL_CTX_set_verify(s_ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);
	ecdh = EC_KEY_new_by_curve_name(test_matrix[idx].s_ecdhe_nid);
	if (ecdh == NULL) {
            printf("unable to create curve to set tmp ECDH\n");
	    exit(1);
        }
	SSL_CTX_set_tmp_ecdh(c_ctx, ecdh);
	SSL_CTX_set_options(c_ctx, SSL_OP_SINGLE_ECDH_USE);
	EC_KEY_free(ecdh);
#ifndef OPENSSL_NO_DH
	if (test_matrix[idx].s_dhe_size == DH_1024) {
            dh = ssltest_get_dh1024();
        } else {
            dh = ssltest_get_dh2048();
        }
	SSL_CTX_set_tmp_dh(c_ctx, dh);
	DH_free(dh);
#endif
    }

    /*
     * Create the client and server sessions
     */
    c_ssl = SSL_new(c_ctx);
    s_ssl = SSL_new(s_ctx);
    if ((c_ssl == NULL) || (s_ssl == NULL)) {
        printf("SSL_new failed.\n");
        exit(1);
    }

    /*
     * Establish TLS session using the two contexts
     */
    rc = doit(s_ssl, c_ssl, 256);

    /*
     * Clean up contexts, certs, and keys
     */
    if (s_cert) X509_free(s_cert);
    if (c_cert) X509_free(c_cert);
    if (s_key) EVP_PKEY_free(s_key);
    if (c_key) EVP_PKEY_free(c_key);
    SSL_free(s_ssl);
    SSL_free(c_ssl);
    SSL_CTX_free(c_ctx);
    SSL_CTX_free(s_ctx);
    if (s_vfy) X509_STORE_free(s_vfy);
    if (c_vfy) X509_STORE_free(c_vfy);

    return rc;
}


/*
 * description: This function performs ctamc key pair generation API.
 * in  *e_ctamc : key parameters
 *     key_type : KEY_TYPE_RSA/KEY_TYPE_EC
 * Return void
 */
static EVP_PKEY* ssltest_keypair_create (int key_type, int key_size)
{
    EVP_PKEY *key = NULL;
    RSA *rsa = NULL;
    BIGNUM *bn = NULL;
    EC_GROUP *group = NULL;
    EC_KEY *eckey = NULL;

    switch (key_type) {
    case KEY_TYPE_RSA:
        rsa = RSA_new();
        bn = BN_new();
        BN_set_word(bn, 0x10001);
        RSA_generate_key_ex(rsa, key_size, bn, NULL);
        key = EVP_PKEY_new();
        EVP_PKEY_set1_RSA(key, rsa);
        RSA_free(rsa);
        BN_free(bn);
        break;
    case KEY_TYPE_EC:
        /* We overload the key_size argument to hold the NID */
        group = EC_GROUP_new_by_curve_name(key_size);
        EC_GROUP_set_asn1_flag(group, OPENSSL_EC_NAMED_CURVE);
        EC_GROUP_set_point_conversion_form(group, POINT_CONVERSION_UNCOMPRESSED);
        eckey = EC_KEY_new();
        EC_KEY_set_group(eckey, group);
        if (!EC_KEY_generate_key(eckey)) {
            printf("Failed to generate EC key\n");
            exit(1);
        }
        key = EVP_PKEY_new();
        EVP_PKEY_set1_EC_KEY(key, eckey);
        EC_KEY_free(eckey);
        EC_GROUP_free(group);
        break;
    default:
        printf("Warning: invalid key type requested.\n");
        break;
    }

    return key;
}


/*
 * This routine will create a self-signed certificate that
 * can be used with TLS.
 *
 * Returns zero on success
 */
static int ssltest_cert_create (X509 **new_cert, EVP_PKEY *keypair, int server, int cert_options)
{
    ASN1_INTEGER *serial = NULL;
    X509_NAME *subj;
    EVP_MD_CTX mctx;
    EVP_PKEY_CTX *pkctx = NULL;
    const EVP_MD *md = EVP_sha256();
    EXTENDED_KEY_USAGE *extusage = NULL;
    ASN1_OBJECT *usage = NULL;

    if ((*new_cert = X509_new()) == NULL) {
        printf("Failed to alloc X509.\n");
        return 1;
    }

    if (cert_options == BAD_VERSION) {
        /* Set version to V1 */
        if (!X509_set_version(*new_cert, 0)) {
            printf("Failed to set X509 version.\n");
            return 1;
        }
    } else {
        /* Set version to V3 */
        if (!X509_set_version(*new_cert, 2)) {
            printf("Failed to set X509 version.\n");
            return 1;
        }
    }
 
    /*
     * Set the serial#
     */
    serial = s2i_ASN1_INTEGER(NULL, "12345678");
    if (!X509_set_serialNumber(*new_cert, serial)) {
        printf("Failed to setX509_print_fp X509 serial#.\n");
        return 1;
    }

    switch (cert_options) {

        case BAD_VERSION:
	case ADD_BOTH_EXTKEYUSAGE:
	    /* BAD and both always set extendedKeyUsage */
	    if (server) {
                usage = OBJ_nid2obj(NID_server_auth);
	    } else {
                usage = OBJ_nid2obj(NID_client_auth);
            }

	    if (!usage) {
                printf("Failed to get usage object.\n");
	        return 1;
            }

	    extusage = sk_ASN1_OBJECT_new_null();
	    if (!extusage) {
                printf("Failed to get extended key usage object.\n");
		return 1;
            }

	    sk_ASN1_OBJECT_push(extusage, usage);

	    if (!X509_add1_ext_i2d(*new_cert, NID_ext_key_usage, extusage, 
                                   0, X509V3_ADD_DEFAULT)) {
                printf("Failed to set X509 extended key usage.\n");
		return 1;
            }
	    break;

	case ADD_CLIENT_EXTKEYUSAGE:
	    /* only set extendedKeyUsage for client */
	    if (!server) {
                usage = OBJ_nid2obj(NID_client_auth);

		if (!usage) {
		    printf("Failed to get usage object.\n");
		    return 1;
                }

		extusage = sk_ASN1_OBJECT_new_null();
		if (!extusage) {
		    printf("Failed to get extended key usage object.\n");
		    return 1;
                }

		sk_ASN1_OBJECT_push(extusage, usage);

		if (!X509_add1_ext_i2d(*new_cert, NID_ext_key_usage, extusage, 
                               0, X509V3_ADD_DEFAULT)) {
	            printf("Failed to set X509 extended key usage.\n");
		    return 1;
                }
	    }
	    break;

        case NO_EXTKEYUSAGE:
	default:
	    break;

    }
    /*
     * Add Common Name entry
     */
    subj = X509_get_subject_name(*new_cert);
    if (!X509_NAME_add_entry_by_txt(subj, "CN", MBSTRING_ASC,
                                    (unsigned char*)HOSTNAME, -1, -1, 0)) {
        printf("Failed to set X509 subject name.\n");
        return 1;
    }
    if (!X509_set_issuer_name(*new_cert, subj)) {
        printf("Failed to set Issuer.\n");
        return 1;
    }

    if (!X509_gmtime_adj(X509_get_notBefore(*new_cert), 0)) {
        printf("Failed to set notBefore.\n");
        return 1;
    }

    if (!X509_time_adj_ex(X509_get_notAfter(*new_cert), 1, 0, NULL)) {
        printf("Failed to set notAfter.\n");
        return 1;
    }

    if (!X509_set_pubkey(*new_cert, keypair)) {
        printf("Failed to public key.\n");
        return 1;
    }

    /*
     * Add a signature to the cert
     */
    EVP_MD_CTX_init(&mctx);
    EVP_DigestSignInit(&mctx, &pkctx, md, NULL, keypair);
    X509_sign_ctx(*new_cert, &mctx);
    EVP_MD_CTX_cleanup(&mctx);

    usage = sk_ASN1_OBJECT_pop(extusage);
    ASN1_OBJECT_free(usage);
    sk_ASN1_OBJECT_free(extusage);
    ASN1_INTEGER_free(serial);
    X509_print_fp(stdout, *new_cert);
    return 0;
}

/*
 * Verifies the subj alt name in the TLS peer
 * certificate.  Returns zero if the hostname
 * is OK, non-zero if it doesn't match.
 * Common Criteria requires the client to verify the server's
 * certificate against the hostname used to open the connection
 * to the server.  CiscoSSL provides the X509_check_host()
 * utility to achieve this.
 */
static int ssltest_check_hostname(SSL* ssl)
{
    X509 *cert;
    int ret;
    int rv=1;

    /*
     * First get the certificate we authenticated from the TLS peer
     */
    cert = SSL_get_peer_certificate(ssl);
    if (cert) {
	/*
	 * Now compare the hostname we used to open the connection
	 * against the FQDN in the certificate.  In this case we've
	 * hardcoded the hostname to "localhost".
	 */
	ret = X509_check_host(cert, HOSTNAME, strlen(HOSTNAME), 0, NULL);
	if (ret < 0) {
	    printf("Error in X509_check_host\n");
	} else if (ret == 1) {
	    rv = 0;
	}
    } else {
	printf("Peer cert is not available, unable to check FQDN in cert\n");
    }
    X509_free(cert);
    return rv;
}

/*
 * *********************************************************************************
 * *********************************************************************************
 *
 * Routines below this point were shamelessly taken from ssl/ssltest.c
 *
 * *********************************************************************************
 * *********************************************************************************
 */
#ifndef OPENSSL_NO_DH
static DH *ssltest_get_dh1024()
{
    static unsigned char dh1024_p[] = {
        0xF8, 0x81, 0x89, 0x7D, 0x14, 0x24, 0xC5, 0xD1, 0xE6, 0xF7, 0xBF,
        0x3A,
        0xE4, 0x90, 0xF4, 0xFC, 0x73, 0xFB, 0x34, 0xB5, 0xFA, 0x4C, 0x56,
        0xA2,
        0xEA, 0xA7, 0xE9, 0xC0, 0xC0, 0xCE, 0x89, 0xE1, 0xFA, 0x63, 0x3F,
        0xB0,
        0x6B, 0x32, 0x66, 0xF1, 0xD1, 0x7B, 0xB0, 0x00, 0x8F, 0xCA, 0x87,
        0xC2,
        0xAE, 0x98, 0x89, 0x26, 0x17, 0xC2, 0x05, 0xD2, 0xEC, 0x08, 0xD0,
        0x8C,
        0xFF, 0x17, 0x52, 0x8C, 0xC5, 0x07, 0x93, 0x03, 0xB1, 0xF6, 0x2F,
        0xB8,
        0x1C, 0x52, 0x47, 0x27, 0x1B, 0xDB, 0xD1, 0x8D, 0x9D, 0x69, 0x1D,
        0x52,
        0x4B, 0x32, 0x81, 0xAA, 0x7F, 0x00, 0xC8, 0xDC, 0xE6, 0xD9, 0xCC,
        0xC1,
        0x11, 0x2D, 0x37, 0x34, 0x6C, 0xEA, 0x02, 0x97, 0x4B, 0x0E, 0xBB,
        0xB1,
        0x71, 0x33, 0x09, 0x15, 0xFD, 0xDD, 0x23, 0x87, 0x07, 0x5E, 0x89,
        0xAB,
        0x6B, 0x7C, 0x5F, 0xEC, 0xA6, 0x24, 0xDC, 0x53,
    };
    static unsigned char dh1024_g[] = {
        0x02,
    };
    DH *dh;

    if ((dh = DH_new()) == NULL)
        return (NULL);
    dh->p = BN_bin2bn(dh1024_p, sizeof(dh1024_p), NULL);
    dh->g = BN_bin2bn(dh1024_g, sizeof(dh1024_g), NULL);
    if ((dh->p == NULL) || (dh->g == NULL)) {
        DH_free(dh);
        return (NULL);
    }
    /*
     * The FOM no longer allows 1024 bit DH.  We enable the override
     * flag here so we can test that CC mode is preventing 1024 bit
     */
    dh->flags |= DH_FLAG_NON_FIPS_ALLOW; 
    return (dh);
}
static unsigned char dh2048_p[] = {
    0xF6,0x42,0x57,0xB7,0x08,0x7F,0x08,0x17,0x72,0xA2,0xBA,0xD6,
    0xA9,0x42,0xF3,0x05,0xE8,0xF9,0x53,0x11,0x39,0x4F,0xB6,0xF1,
    0x6E,0xB9,0x4B,0x38,0x20,0xDA,0x01,0xA7,0x56,0xA3,0x14,0xE9,
    0x8F,0x40,0x55,0xF3,0xD0,0x07,0xC6,0xCB,0x43,0xA9,0x94,0xAD,
    0xF7,0x4C,0x64,0x86,0x49,0xF8,0x0C,0x83,0xBD,0x65,0xE9,0x17,
    0xD4,0xA1,0xD3,0x50,0xF8,0xF5,0x59,0x5F,0xDC,0x76,0x52,0x4F,
    0x3D,0x3D,0x8D,0xDB,0xCE,0x99,0xE1,0x57,0x92,0x59,0xCD,0xFD,
    0xB8,0xAE,0x74,0x4F,0xC5,0xFC,0x76,0xBC,0x83,0xC5,0x47,0x30,
    0x61,0xCE,0x7C,0xC9,0x66,0xFF,0x15,0xF9,0xBB,0xFD,0x91,0x5E,
    0xC7,0x01,0xAA,0xD3,0x5B,0x9E,0x8D,0xA0,0xA5,0x72,0x3A,0xD4,
    0x1A,0xF0,0xBF,0x46,0x00,0x58,0x2B,0xE5,0xF4,0x88,0xFD,0x58,
    0x4E,0x49,0xDB,0xCD,0x20,0xB4,0x9D,0xE4,0x91,0x07,0x36,0x6B,
    0x33,0x6C,0x38,0x0D,0x45,0x1D,0x0F,0x7C,0x88,0xB3,0x1C,0x7C,
    0x5B,0x2D,0x8E,0xF6,0xF3,0xC9,0x23,0xC0,0x43,0xF0,0xA5,0x5B,
    0x18,0x8D,0x8E,0xBB,0x55,0x8C,0xB8,0x5D,0x38,0xD3,0x34,0xFD,
    0x7C,0x17,0x57,0x43,0xA3,0x1D,0x18,0x6C,0xDE,0x33,0x21,0x2C,
    0xB5,0x2A,0xFF,0x3C,0xE1,0xB1,0x29,0x40,0x18,0x11,0x8D,0x7C,
    0x84,0xA7,0x0A,0x72,0xD6,0x86,0xC4,0x03,0x19,0xC8,0x07,0x29,
    0x7A,0xCA,0x95,0x0C,0xD9,0x96,0x9F,0xAB,0xD0,0x0A,0x50,0x9B,
    0x02,0x46,0xD3,0x08,0x3D,0x66,0xA4,0x5D,0x41,0x9F,0x9C,0x7C,
    0xBD,0x89,0x4B,0x22,0x19,0x26,0xBA,0xAB,0xA2,0x5E,0xC3,0x55,
    0xE9,0x32,0x0B,0x3B,
};

static unsigned char dh2048_g[] = {
    0x02,
};

static DH *ssltest_get_dh2048()
{
    DH *dh;

    if ((dh = DH_new()) == NULL)
        return NULL;
    dh->p = BN_bin2bn(dh2048_p, sizeof(dh2048_p), NULL);
    dh->g = BN_bin2bn(dh2048_g, sizeof(dh2048_g), NULL);
    if (dh->p == NULL || dh->g == NULL) {
        DH_free(dh);
        return NULL;
    }
    return dh;
}
#endif

static void print_details (SSL *c_ssl, const char *prefix)
{
    const SSL_CIPHER *ciph;
    X509 *cert;

    ciph = SSL_get_current_cipher(c_ssl);
    printf("%s%s, cipher %s %s",
           prefix,
           SSL_get_version(c_ssl),
           SSL_CIPHER_get_version(ciph), SSL_CIPHER_get_name(ciph));
    cert = SSL_get_peer_certificate(c_ssl);
    if (cert != NULL) {
        EVP_PKEY *pkey = X509_get_pubkey(cert);
        if (pkey != NULL) {
            if (0) {
                ;
            }
#ifndef OPENSSL_NO_RSA
            else if (pkey->type == EVP_PKEY_RSA && pkey->pkey.rsa != NULL
                     && pkey->pkey.rsa->n != NULL) {
                printf(", %d bit RSA",
                       BN_num_bits(pkey->pkey.rsa->n));
            }
#endif
#ifndef OPENSSL_NO_DSA
            else if (pkey->type == EVP_PKEY_DSA && pkey->pkey.dsa != NULL
                     && pkey->pkey.dsa->p != NULL) {
                printf(", %d bit DSA",
                       BN_num_bits(pkey->pkey.dsa->p));
            }
#endif
            EVP_PKEY_free(pkey);
        }
        X509_free(cert);
    }
    /*
     * The SSL API does not allow us to look at temporary RSA/DH keys,
     * otherwise we should print their lengths too
     */
    printf("\n");
}


#define W_READ  1
#define W_WRITE 2
#define C_DONE  1
#define S_DONE  2
static int doit (SSL *s_ssl, SSL *c_ssl, long count)
{
    char *cbuf = NULL, *sbuf = NULL;
    long bufsiz;
    long cw_num = count, cr_num = count;
    long sw_num = count, sr_num = count;
    int ret = 1;
    BIO *c_to_s = NULL;
    BIO *s_to_c = NULL;
    BIO *c_bio = NULL;
    BIO *s_bio = NULL;
    int c_r, c_w, s_r, s_w;
    int i, j;
    int done = 0;
    int c_write, s_write;
    int do_server = 0, do_client = 0;
    int max_frag = 5 * 1024;

    bufsiz = count > 40 * 1024 ? 40 * 1024 : count;

    if ((cbuf = OPENSSL_malloc(bufsiz)) == NULL) {
        goto err;
    }
    if ((sbuf = OPENSSL_malloc(bufsiz)) == NULL) {
        goto err;
    }

    memset(cbuf, 0, bufsiz);
    memset(sbuf, 0, bufsiz);

    c_to_s = BIO_new(BIO_s_mem());
    s_to_c = BIO_new(BIO_s_mem());
    if ((s_to_c == NULL) || (c_to_s == NULL)) {
        ERR_print_errors_fp(stderr);
        goto err;
    }

    c_bio = BIO_new(BIO_f_ssl());
    s_bio = BIO_new(BIO_f_ssl());
    if ((c_bio == NULL) || (s_bio == NULL)) {
        ERR_print_errors_fp(stderr);
        goto err;
    }

    SSL_set_connect_state(c_ssl);
    SSL_set_bio(c_ssl, s_to_c, c_to_s);
    SSL_set_max_send_fragment(c_ssl, max_frag);
    BIO_set_ssl(c_bio, c_ssl, BIO_NOCLOSE);

    SSL_set_accept_state(s_ssl);
    SSL_set_bio(s_ssl, c_to_s, s_to_c);
    SSL_set_max_send_fragment(s_ssl, max_frag);
    BIO_set_ssl(s_bio, s_ssl, BIO_NOCLOSE);

    c_r = 0;
    s_r = 1;
    c_w = 1;
    s_w = 0;
    c_write = 1, s_write = 0;

    /* We can always do writes */
    for (;; ) {
        do_server = 0;
        do_client = 0;

        i = (int)BIO_pending(s_bio);
        if ((i && s_r) || s_w) {
            do_server = 1;
        }

        i = (int)BIO_pending(c_bio);
        if ((i && c_r) || c_w) {
            do_client = 1;
        }

        if (do_server) {
            if (SSL_in_init(s_ssl)) {
                printf("server waiting in SSL_accept - %s\n",
                       SSL_state_string_long(s_ssl));
            }
        }

        if (do_client) {
            if (SSL_in_init(c_ssl)) {
                printf("client waiting in SSL_connect - %s\n",
                       SSL_state_string_long(c_ssl));
            }
        }

        if (!do_client && !do_server) {
            fprintf(stdout, "ERROR IN STARTUP\n");
            ERR_print_errors_fp(stderr);
            break;
        }

        if (do_client && !(done & C_DONE)) {
            if (c_write) {
                j = (cw_num > bufsiz) ? (int)bufsiz : (int)cw_num;
                i = BIO_write(c_bio, cbuf, j);
                if (i < 0) {
                    c_r = 0;
                    c_w = 0;
                    if (BIO_should_retry(c_bio)) {
                        if (BIO_should_read(c_bio)) {
                            c_r = 1;
                        }
                        if (BIO_should_write(c_bio)) {
                            c_w = 1;
                        }
                    } else {
                        fprintf(stderr, "ERROR in CLIENT\n");
                        ERR_print_errors_fp(stderr);
                        goto err;
                    }
                } else if (i == 0) {
                    fprintf(stderr, "SSL CLIENT STARTUP FAILED\n");
                    goto err;
                } else {
		    if (ssltest_check_hostname(c_ssl)) {
			fprintf(stderr, "FQDN check failed\n");
			goto err;
		    }
                    printf("client wrote %d\n", i);
                    /* ok */
                    s_r = 1;
                    c_write = 0;
                    cw_num -= i;
                    if (max_frag > 1029) {
                        SSL_set_max_send_fragment(c_ssl, max_frag -= 5);
                    }
                }
            } else {
                i = BIO_read(c_bio, cbuf, bufsiz);
                if (i < 0) {
                    c_r = 0;
                    c_w = 0;
                    if (BIO_should_retry(c_bio)) {
                        if (BIO_should_read(c_bio)) {
                            c_r = 1;
                        }
                        if (BIO_should_write(c_bio)) {
                            c_w = 1;
                        }
                    } else {
                        fprintf(stderr, "ERROR in CLIENT\n");
                        ERR_print_errors_fp(stderr);
                        goto err;
                    }
                } else if (i == 0) {
                    fprintf(stderr, "SSL CLIENT STARTUP FAILED\n");
                    goto err;
                } else {
                    printf("client read %d\n", i);
                    cr_num -= i;
                    if (sw_num > 0) {
                        s_write = 1;
                        s_w = 1;
                    }
                    if (cr_num <= 0) {
                        s_write = 1;
                        s_w = 1;
                        done = S_DONE | C_DONE;
                    }
                }
            }
        }

        if (do_server && !(done & S_DONE)) {
            if (!s_write) {
                i = BIO_read(s_bio, sbuf, bufsiz);
                if (i < 0) {
                    s_r = 0;
                    s_w = 0;
                    if (BIO_should_retry(s_bio)) {
                        if (BIO_should_read(s_bio)) {
                            s_r = 1;
                        }
                        if (BIO_should_write(s_bio)) {
                            s_w = 1;
                        }
                    } else {
                        fprintf(stderr, "ERROR in SERVER\n");
                        ERR_print_errors_fp(stderr);
                        goto err;
                    }
                } else if (i == 0) {
                    ERR_print_errors_fp(stderr);
                    fprintf(stderr,
                            "SSL SERVER STARTUP FAILED in SSL_read\n");
                    goto err;
                } else {
                    printf("server read %d\n", i);
                    sr_num -= i;
                    if (cw_num > 0) {
                        c_write = 1;
                        c_w = 1;
                    }
                    if (sr_num <= 0) {
                        s_write = 1;
                        s_w = 1;
                        c_write = 0;
                    }
                }
            } else {
                j = (sw_num > bufsiz) ? (int)bufsiz : (int)sw_num;
                i = BIO_write(s_bio, sbuf, j);
                if (i < 0) {
                    s_r = 0;
                    s_w = 0;
                    if (BIO_should_retry(s_bio)) {
                        if (BIO_should_read(s_bio)) {
                            s_r = 1;
                        }
                        if (BIO_should_write(s_bio)) {
                            s_w = 1;
                        }
                    } else {
                        fprintf(stderr, "ERROR in SERVER\n");
                        ERR_print_errors_fp(stderr);
                        goto err;
                    }
                } else if (i == 0) {
                    ERR_print_errors_fp(stderr);
                    fprintf(stderr,
                            "SSL SERVER STARTUP FAILED in SSL_write\n");
                    goto err;
                } else {
                    printf("server wrote %d\n", i);
                    sw_num -= i;
                    s_write = 0;
                    c_r = 1;
                    if (sw_num <= 0) {
                        done |= S_DONE;
                    }
                    if (max_frag > 1029) {
                        SSL_set_max_send_fragment(s_ssl, max_frag -= 5);
                    }
                }
            }
        }

        if ((done & S_DONE) && (done & C_DONE)) {
            break;
        }
    }
    //SSL_SESSION_print(bio_out, SSL_get_session(c_ssl));
    print_details(c_ssl, "DONE: ");
    ret = 0;
err:
    /*
     * We have to set the BIO's to NULL otherwise they will be
     * OPENSSL_free()ed twice.  Once when th s_ssl is SSL_free()ed and again
     * when c_ssl is SSL_free()ed. This is a hack required because s_ssl and
     * c_ssl are sharing the same BIO structure and SSL_set_bio() and
     * SSL_free() automatically BIO_free non NULL entries. You should not
     * normally do this or be required to do this
     */
    if (s_ssl != NULL) {
        s_ssl->rbio = NULL;
        s_ssl->wbio = NULL;
    }
    if (c_ssl != NULL) {
        c_ssl->rbio = NULL;
        c_ssl->wbio = NULL;
    }

    if (c_to_s != NULL) {
        BIO_free(c_to_s);
    }
    if (s_to_c != NULL) {
        BIO_free(s_to_c);
    }
    if (c_bio != NULL) {
        BIO_free_all(c_bio);
    }
    if (s_bio != NULL) {
        BIO_free_all(s_bio);
    }

    if (cbuf) {
        OPENSSL_free(cbuf);
    }
    if (sbuf) {
        OPENSSL_free(sbuf);
    }

    return (ret);
}
#endif /* OPENSSL_FIPS */
