/* -----------------------------------------------------------------
 * crypto/aes/asm/aes_xts-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#include <openssl/crypto.h>
#include <openssl/aes.h>
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif
#include "modes_lcl.h"
#include "cvmx-aes.h"

/*
 * octeon_xts_set_key() - setup key and calculate tweak
 */
void octeon_xts_set_key(EVP_AES_OCT_XTS_CTX * xctx)
{

    CVMX_MT_AES_KEY(xctx->key2[0], 0);
    CVMX_MT_AES_KEY(xctx->key2[1], 1);
    CVMX_MT_AES_KEY(xctx->key2[2], 2);
    CVMX_MT_AES_KEY(xctx->key2[3], 3);
    CVMX_MT_AES_KEYLENGTH(xctx->keylen);

    CVMX_MT_AES_ENC0(xctx->tweak[0]);
    CVMX_MT_AES_ENC1(xctx->tweak[1]);
    CVMX_MF_AES_RESULT(xctx->tweak[0], 0);
    CVMX_MF_AES_RESULT(xctx->tweak[1], 1);
}

/*
 * octeon_xts_decrypt() - perform the decrypt
 */
int octeon_xts_decrypt(const uint8_t * ct_in, uint32_t len, uint8_t * pt,
                       EVP_AES_OCT_XTS_CTX * ctx)
{
    uint64_t *plain;
    uint64_t *cipher;
    uint64_t *ctxtweak;
    uint64_t oldtweak[2];
    uint64_t x[2];
    uint64_t rx[2];
    uint8_t *ct = NULL;
    int32_t i = 0, j;

    if (!pt || !ct_in || !ctx) {
        return XTS_AES_NULL_POINTER_ARGUMENT;
    }
    if (len < 16) {
        return XTS_AES_INVALID_DATALENGTH;
    }

    /* special case in == out and len not 16 byte divisible */
    if ((ct_in == pt) && (len % 16)) {
        ct = OPENSSL_malloc(len);
	memcpy(ct, ct_in, len);
    } else {
        ct = (uint8_t *)ct_in;
    }

    CVMX_PREFETCH0(ct);

    plain = (uint64_t *) pt;
    cipher = (uint64_t *) ct;
    ctxtweak = (uint64_t *) ctx->tweak;

    octeon_xts_set_key(ctx);

    CVMX_MT_AES_KEY(ctx->key1[0], 0);
    CVMX_MT_AES_KEY(ctx->key1[1], 1);
    CVMX_MT_AES_KEY(ctx->key1[2], 2);
    CVMX_MT_AES_KEY(ctx->key1[3], 3);
    CVMX_MT_AES_KEYLENGTH(ctx->keylen);

    uint64_t all80 = 0x8080808080808080ull;
    uint64_t all7f = ~all80;

    uint64_t lenby8;
    uint64_t upperbits, lsb, cout;
    /* uint64_t start,end; */

    lenby8 = len / 16 * 2;
    lenby8 -= 2;

    /* start  = cvmx_get_cycle(); */

    if (len >= 48) {
        x[0] = cipher[0] ^ ctxtweak[0];
        x[1] = cipher[1] ^ ctxtweak[1];

        CVMX_MT_AES_DEC0(x[0]);
        CVMX_MT_AES_DEC1(x[1]);

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        cipher += 2;
        x[0] = cipher[0] ^ ctxtweak[0];
        x[1] = cipher[1] ^ ctxtweak[1];

        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        for (i = 2; i < lenby8; i += 2) {
            CVMX_MT_AES_DEC0(x[0]);
            CVMX_MT_AES_DEC1(x[1]);

            CVMX_PREFETCH0(cipher + 16);

            plain[0] = rx[0] ^ oldtweak[0];
            plain[1] = rx[1] ^ oldtweak[1];
            plain += 2;

            oldtweak[0] = ctxtweak[0];
            oldtweak[1] = ctxtweak[1];

            upperbits = ctxtweak[0] & all80;
            ctxtweak[0] &= all7f;
            lsb = upperbits & 0xffull;
            ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

            upperbits = ctxtweak[1] & all80;
            ctxtweak[1] &= all7f;
            cout = upperbits & 0xffull;
            upperbits = (lsb << 56) | (upperbits >> 8);
            ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

            ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

            cipher += 2;
            x[0] = cipher[0] ^ ctxtweak[0];
            x[1] = cipher[1] ^ ctxtweak[1];

            CVMX_MF_AES_RESULT(rx[0], 0);
            CVMX_MF_AES_RESULT(rx[1], 1);
        }

        CVMX_MT_AES_DEC0(x[0]);
        CVMX_MT_AES_DEC1(x[1]);

        plain[0] = rx[0] ^ oldtweak[0];
        plain[1] = rx[1] ^ oldtweak[1];
        plain += 2;

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        i += 2;
        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        plain[0] = rx[0] ^ oldtweak[0];
        plain[1] = rx[1] ^ oldtweak[1];

    } else if (len >= 32) {

        x[0] = cipher[0] ^ ctxtweak[0];
        x[1] = cipher[1] ^ ctxtweak[1];

        CVMX_MT_AES_DEC0(x[0]);
        CVMX_MT_AES_DEC1(x[1]);

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        cipher += 2;
        x[0] = cipher[0] ^ ctxtweak[0];
        x[1] = cipher[1] ^ ctxtweak[1];

        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        CVMX_MT_AES_DEC0(x[0]);
        CVMX_MT_AES_DEC1(x[1]);

        plain[0] = rx[0] ^ oldtweak[0];
        plain[1] = rx[1] ^ oldtweak[1];
        plain += 2;

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        i = 4;
        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        plain[0] = rx[0] ^ oldtweak[0];
        plain[1] = rx[1] ^ oldtweak[1];

    } else if (len >= 16) {
        x[0] = cipher[0] ^ ctxtweak[0];
        x[1] = cipher[1] ^ ctxtweak[1];

        CVMX_MT_AES_DEC0(x[0]);
        CVMX_MT_AES_DEC1(x[1]);

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        i = 2;
        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        plain[0] = rx[0] ^ oldtweak[0];
        plain[1] = rx[1] ^ oldtweak[1];
    }

    if ((len % 16)) {
        uint8_t y[16];
        uint64_t *Y = (uint64_t *) y;
        uint8_t *T = (uint8_t *) oldtweak;

        ctx->state = 0xdead;
        x[0] = cipher[0] ^ ctx->tweak[0];
        x[1] = cipher[1] ^ ctx->tweak[1];

        CVMX_MT_AES_DEC0(x[0]);
        CVMX_MT_AES_DEC1(x[1]);

        CVMX_MF_AES_RESULT(x[0], 0);
        CVMX_MF_AES_RESULT(x[1], 1);

        plain[0] = x[0] ^ ctx->tweak[0];
        plain[1] = x[1] ^ ctx->tweak[1];

        i <<= 3;
        for (j = 0; (i + j) < len; j++) {
            pt[i + j] = pt[i + j - 16];
            y[j] = ct[i + j] ^ T[j];
        }

        for (; j < 16; j++) {
            y[j] = pt[i + j - 16] ^ T[j];
        }

        CVMX_MT_AES_DEC0(Y[0]);
        CVMX_MT_AES_DEC1(Y[1]);
        CVMX_MF_AES_RESULT(Y[0], 0);
        CVMX_MF_AES_RESULT(Y[1], 1);

        plain = (uint64_t *) & pt[i - 16];
        plain[0] = Y[0] ^ oldtweak[0];
        plain[1] = Y[1] ^ oldtweak[1];
    }

    if ((ct_in == pt) && (len % 16)) {
        OPENSSL_free(ct);
    }
    return 1;
}

/*
 * octeon_xts_encrypt() - perform the encrypt
 */
int octeon_xts_encrypt(const uint8_t * pt, uint32_t len, uint8_t * ct,
                       EVP_AES_OCT_XTS_CTX * ctx)
{
    uint64_t *plain;
    uint64_t *cipher;
    uint64_t *ctxtweak;
    uint64_t oldtweak[2];
    uint64_t x[2];
    uint64_t rx[2];
    int32_t i = 0, j;

    if (!pt || !ct || !ctx) {
        return XTS_AES_NULL_POINTER_ARGUMENT;
    }
    if (len < 16) {
        return XTS_AES_INVALID_DATALENGTH;
    }

    CVMX_PREFETCH0(pt);

    plain = (uint64_t *) pt;
    cipher = (uint64_t *) ct;
    ctxtweak = (uint64_t *) ctx->tweak;

    octeon_xts_set_key(ctx);

    CVMX_MT_AES_KEY(ctx->key1[0], 0);
    CVMX_MT_AES_KEY(ctx->key1[1], 1);
    CVMX_MT_AES_KEY(ctx->key1[2], 2);
    CVMX_MT_AES_KEY(ctx->key1[3], 3);
    CVMX_MT_AES_KEYLENGTH(ctx->keylen);

    uint64_t all80 = 0x8080808080808080ull;
    uint64_t all7f = ~all80;

    uint64_t lenby8;
    uint64_t upperbits, lsb, cout;
    /* uint64_t start,end; */

    lenby8 = len / 16 * 2;
    lenby8 -= 2;

    /* start  = cvmx_get_cycle(); */

    if (len >= 48) {
        x[0] = plain[0] ^ ctxtweak[0];
        x[1] = plain[1] ^ ctxtweak[1];

        CVMX_MT_AES_ENC0(x[0]);
        CVMX_MT_AES_ENC1(x[1]);

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        plain += 2;
        x[0] = plain[0] ^ ctxtweak[0];
        x[1] = plain[1] ^ ctxtweak[1];

        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        for (i = 2; i < lenby8; i += 2) {
            CVMX_MT_AES_ENC0(x[0]);
            CVMX_MT_AES_ENC1(x[1]);

            CVMX_PREFETCH0(plain + 16);

            cipher[0] = rx[0] ^ oldtweak[0];
            cipher[1] = rx[1] ^ oldtweak[1];
            cipher += 2;

            oldtweak[0] = ctxtweak[0];
            oldtweak[1] = ctxtweak[1];

            upperbits = ctxtweak[0] & all80;
            ctxtweak[0] &= all7f;
            lsb = upperbits & 0xffull;
            ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

            upperbits = ctxtweak[1] & all80;
            ctxtweak[1] &= all7f;
            cout = upperbits & 0xffull;
            upperbits = (lsb << 56) | (upperbits >> 8);
            ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

            ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

            plain += 2;
            x[0] = plain[0] ^ ctxtweak[0];
            x[1] = plain[1] ^ ctxtweak[1];

            CVMX_MF_AES_RESULT(rx[0], 0);
            CVMX_MF_AES_RESULT(rx[1], 1);
        }

        CVMX_MT_AES_ENC0(x[0]);
        CVMX_MT_AES_ENC1(x[1]);

        cipher[0] = rx[0] ^ oldtweak[0];
        cipher[1] = rx[1] ^ oldtweak[1];
        cipher += 2;

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        i += 2;
        i <<= 3;
        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        cipher[0] = rx[0] ^ oldtweak[0];
        cipher[1] = rx[1] ^ oldtweak[1];

    } else if (len >= 32) {

        x[0] = plain[0] ^ ctxtweak[0];
        x[1] = plain[1] ^ ctxtweak[1];

        CVMX_MT_AES_ENC0(x[0]);
        CVMX_MT_AES_ENC1(x[1]);

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        plain += 2;
        x[0] = plain[0] ^ ctxtweak[0];
        x[1] = plain[1] ^ ctxtweak[1];

        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        CVMX_MT_AES_ENC0(x[0]);
        CVMX_MT_AES_ENC1(x[1]);

        cipher[0] = rx[0] ^ oldtweak[0];
        cipher[1] = rx[1] ^ oldtweak[1];
        cipher += 2;

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        i = 32;
        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        cipher[0] = rx[0] ^ oldtweak[0];
        cipher[1] = rx[1] ^ oldtweak[1];

    } else if (len >= 16) {
        x[0] = plain[0] ^ ctxtweak[0];
        x[1] = plain[1] ^ ctxtweak[1];

        CVMX_MT_AES_ENC0(x[0]);
        CVMX_MT_AES_ENC1(x[1]);

        oldtweak[0] = ctxtweak[0];
        oldtweak[1] = ctxtweak[1];

        upperbits = ctxtweak[0] & all80;
        ctxtweak[0] &= all7f;
        lsb = upperbits & 0xffull;
        ctxtweak[0] = (ctxtweak[0] << 1) + (upperbits >> 15);

        upperbits = ctxtweak[1] & all80;
        ctxtweak[1] &= all7f;
        cout = upperbits & 0xffull;
        upperbits = (lsb << 56) | (upperbits >> 8);
        ctxtweak[1] = (ctxtweak[1] << 1) + (upperbits >> 7);

        ctxtweak[0] ^= (cout ? 0x87ull << 56 : 0);

        i = 16;
        CVMX_MF_AES_RESULT(rx[0], 0);
        CVMX_MF_AES_RESULT(rx[1], 1);

        cipher[0] = rx[0] ^ oldtweak[0];
        cipher[1] = rx[1] ^ oldtweak[1];
    }

    /* end = cvmx_get_cycle(); */
    /* printf("cycles = %ld\n",end-start); */

    if ((len % 16)) {
        uint8_t x[16];
        uint64_t *CT;
        uint64_t *X;
        ctx->state = 0xdead;

        for (j = 0; (i + j) < len; j++) {
            x[j] = pt[i + j];
            ct[i + j] = ct[i + j - 16];
        }

        for (; j < 16; j++) {
            x[j] = ct[i + j - 16];
        }

        X = (uint64_t *) x;
        X[0] = X[0] ^ ctx->tweak[0];
        X[1] = X[1] ^ ctx->tweak[1];
        CVMX_MT_AES_ENC0(X[0]);
        CVMX_MT_AES_ENC1(X[1]);
        CVMX_MF_AES_RESULT(X[0], 0);
        CVMX_MF_AES_RESULT(X[1], 1);

        CT = (uint64_t *) & ct[i - 16];

        CT[0] = X[0] ^ ctx->tweak[0];
        CT[1] = X[1] ^ ctx->tweak[1];
    }

    return 1;
}
