/* ====================================================================
 * Copyright (c) 2011-2013 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    licensing@OpenSSL.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#include <openssl/opensslconf.h>

#include <stdio.h>
#include <string.h>

#if !defined(OPENSSL_NO_AES) && !defined(OPENSSL_NO_SHA1)

# include <openssl/evp.h>
# include <openssl/objects.h>
# include <openssl/aes.h>
# include <openssl/sha.h>
# include <openssl/rand.h>
# include <openssl/crypto.h>
# include "modes_lcl.h"
# ifdef OCTEON_SINGLE_PASS
# include "octeon_papi.h"
# endif

# define SSL3_RT_MAX_PLAIN_LENGTH         16384
# ifndef EVP_CIPH_FLAG_AEAD_CIPHER
#  define EVP_CIPH_FLAG_AEAD_CIPHER       0x200000
#  define EVP_CTRL_AEAD_TLS1_AAD          0x16
#  define EVP_CTRL_AEAD_SET_MAC_KEY       0x17
# endif

# if !defined(EVP_CIPH_FLAG_DEFAULT_ASN1)
#  define EVP_CIPH_FLAG_DEFAULT_ASN1 0
# endif

# define SSL3_VERSION   0x0300
# define TLS1_VERSION   0x0301
# define TLS1_1_VERSION 0x0302
# define TLS1_2_VERSION 0x0303

# define NO_PAYLOAD_LENGTH       ((size_t)-1)

#if defined(OCTEON_SINGLE_PASS)

#  define data(ctx) ((EVP_AESOCT_HMAC_SHA1 *)(ctx)->cipher_data)

static
int aesoct_cbc_hmac_sha1_init_key (EVP_CIPHER_CTX *ctx, const unsigned char * inkey,
                                    const unsigned char *iv, int enc)
{
    EVP_AESOCT_HMAC_SHA1 *key = data(ctx);
    
    key->ex_mode = EX_MODE_DEFAULT;
    switch(M_EVP_CIPHER_key_length(ctx->cipher)) {
        case 32:
            key->cipher_type = OCT_AES_256;
            break;
        case 16:
            key->cipher_type = OCT_AES_128;
            break;
        default:
            return 0;
    }
    if(inkey)
        memcpy((unsigned char *) key->cipher_key,inkey,M_EVP_CIPHER_key_length(ctx->cipher));
    if(iv)
        memcpy((unsigned char *)key->cipher_iv,iv,M_EVP_CIPHER_iv_length(ctx->cipher));

    key->payload_length = NO_PAYLOAD_LENGTH;

    return  1;
}

static
int aesoct_cbc_hmac_sha1_ctrl(EVP_CIPHER_CTX *ctx, int type, int arg,
                                    void *ptr)
{
    EVP_AESOCT_HMAC_SHA1 *key = data(ctx);

    switch (type) {
    case EVP_CTRL_AEAD_SET_MAC_KEY:
        {
            /* Store MAC SECRET to cipher data structure*/
            unsigned char hmac_key[64];

            memset(hmac_key, 0, sizeof(hmac_key));

            if (arg > (int)sizeof(hmac_key)) {
                SHA_CTX key_ex;
                SHA1OCT_Init(&key_ex);
                SHA1OCT_Update(&key_ex, ptr, arg);
                SHA1OCT_Final(hmac_key, &key_ex);
            } else {
                memcpy(hmac_key, ptr, arg);
            }
            memcpy(key->mac_secret, hmac_key,64);
            OPENSSL_cleanse(hmac_key, sizeof(hmac_key));

            return 1;
        }
    case EVP_CTRL_AEAD_TLS1_AAD:
        {
            /* Store TLS AAD in cipher data structure, 
             * Encryption : Return pad length from payload length and HMAC size
             * Decryption : Return HMAC size
             */
            unsigned char *p = ptr;
            unsigned int len;

            if (arg != EVP_AEAD_TLS1_AAD_LEN)
                return -1;
            if(key->ex_mode == EX_MODE_DEFAULT) {
 
                len = p[arg - 2] << 8 | p[arg - 1];

                 if (ctx->encrypt) {
                    key->payload_length = len;
                    if ((p[arg - 4] << 8 | p[arg - 3]) >= TLS1_1_VERSION) {
                        len -= AES_BLOCK_SIZE;
                        p[arg - 2] = len >> 8;
                        p[arg - 1] = len;
                    }
                    memcpy(key->aux.tls_aad_8, ptr, arg);
                    return (int)(((len + SHA_DIGEST_LENGTH +
                                   AES_BLOCK_SIZE) & -AES_BLOCK_SIZE)
                                 - len);
                }
                 else {
                    memcpy(key->aux.tls_aad_8, ptr, arg);
                    key->payload_length = arg;
                    return SHA_DIGEST_LENGTH;
                }
            } else {
                memcpy(key->aux.tls_aad_8, ptr, arg);
                return 1;
            }
        }
    case EVP_CTRL_CIPH_SET_MODE :
        {
            /* Set execution mode, 
             * Support for NIST vector testing */
            switch(arg) {
                case EX_MODE_ENCRYPT:
                    key->ex_mode = EX_MODE_ENCRYPT;
                    return 1;
                case EX_MODE_DIGEST:
                    key->ex_mode = EX_MODE_DIGEST;
                    return 1;
                case EX_MODE_HMAC:
                    key->ex_mode = EX_MODE_HMAC;
                    return 1;
                case EX_MODE_DEFAULT:
                    key->ex_mode = EX_MODE_DEFAULT;
                    return 1;
                default:
                    return 0;
            }
        } 
    default:
        return 0;
    }
}

static
int aesoct_cbc_hmac_sha1_cipher(EVP_CIPHER_CTX *ctx, unsigned char *out, const unsigned char *in, size_t in_len)
{
    EVP_AESOCT_HMAC_SHA1 *key = data(ctx);
    struct iovec_cav inp_vec[1],outp_vec[1];
    uint16_t inp_cnt = 1, outp_cnt = 1;
    uint8_t *ext_iv_ptr = NULL;
    int ret = 0;
    struct tls_rec_hdr *rec_header = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];
    if(key->ex_mode == EX_MODE_DEFAULT) {
        /* Set explicit IV pointer and move input buffer to IV length bytes for 
         * protocol > TLS 1.0 in case of encryption[API requirement].
         * Leave IV length bytes space in output buffer for explicit IV in case of 
         * decryption.
         */
        if(cvmx_htobe16(rec_header->u.version)  >= TLS1_1_VERSION && ctx->encrypt){
            ext_iv_ptr = (uint8_t *)in;
            inp_vec[0].ptr = (uint8_t *)in+M_EVP_CIPHER_iv_length(ctx->cipher);
            inp_vec[0].size = key->payload_length - M_EVP_CIPHER_iv_length(ctx->cipher);
            outp_vec[0].ptr = (uint8_t *)out;
        } else {
            inp_vec[0].ptr = (uint8_t *)in;
            inp_vec[0].size = (ctx->encrypt)?key->payload_length:in_len;
            outp_vec[0].ptr = (uint8_t *)out;
            if(cvmx_htobe16(rec_header->u.version) >= TLS1_1_VERSION)
                outp_vec[0].ptr += M_EVP_CIPHER_iv_length(ctx->cipher);
        }
    } else {
        inp_vec[0].ptr = (uint8_t *)in;
        inp_vec[0].size = in_len;
        outp_vec[0].ptr = out;
    }
    outp_vec[0].size = SSL3_RT_MAX_PLAIN_LENGTH;

    switch(cvmx_htobe16(rec_header->u.version)) {
        case TLS1_VERSION:
            if(ctx->encrypt)
                ret = oct_aes_sha1_enc_tls10 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,(uint16_t *)&in_len);
            else
                ret = oct_aes_sha1_dec_tls10 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,(uint16_t *)&in_len);
            break;
        case TLS1_1_VERSION:
        case TLS1_2_VERSION:
            if(ctx->encrypt)
                ret = oct_aes_sha1_enc_tls11 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,ext_iv_ptr, (uint16_t *)&in_len);
            else
                ret = oct_aes_sha1_dec_tls11 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,(uint16_t *)&in_len);
            break;
        default:
            return ret;
    }
    return ret;
}

static
EVP_CIPHER aesoct_256_cbc_hmac_sha1_cipher = {
#  ifdef NID_aes_256_cbc_hmac_sha1
    NID_aes_256_cbc_hmac_sha1,
#  else
    NID_undef,
#  endif
    16, 32, 16,
    EVP_CIPH_CBC_MODE | EVP_CIPH_FLAG_DEFAULT_ASN1 |
        EVP_CIPH_FLAG_AEAD_CIPHER | EVP_CIPH_FLAG_FIPS,
    aesoct_cbc_hmac_sha1_init_key,
    aesoct_cbc_hmac_sha1_cipher,
    NULL,
    sizeof(EVP_AESOCT_HMAC_SHA1),
    EVP_CIPH_FLAG_DEFAULT_ASN1 ? NULL : EVP_CIPHER_set_asn1_iv,
    EVP_CIPH_FLAG_DEFAULT_ASN1 ? NULL : EVP_CIPHER_get_asn1_iv,
    aesoct_cbc_hmac_sha1_ctrl,
    NULL
};


static
EVP_CIPHER aesoct_128_cbc_hmac_sha1_cipher = {
#  ifdef NID_aes_128_cbc_hmac_sha1
    NID_aes_128_cbc_hmac_sha1,
#  else
    NID_undef,
#  endif
    16, 16, 16,
    EVP_CIPH_CBC_MODE | EVP_CIPH_FLAG_DEFAULT_ASN1 |
        EVP_CIPH_FLAG_AEAD_CIPHER | EVP_CIPH_FLAG_FIPS,
    aesoct_cbc_hmac_sha1_init_key,
    aesoct_cbc_hmac_sha1_cipher,
    NULL,
    sizeof(EVP_AESOCT_HMAC_SHA1),
    EVP_CIPH_FLAG_DEFAULT_ASN1 ? NULL : EVP_CIPHER_set_asn1_iv,
    EVP_CIPH_FLAG_DEFAULT_ASN1 ? NULL : EVP_CIPHER_get_asn1_iv,
    aesoct_cbc_hmac_sha1_ctrl,
    NULL
};
const EVP_CIPHER *EVP_aes_128_cbc_hmac_sha1(void)
{
    return &aesoct_128_cbc_hmac_sha1_cipher;
}

const EVP_CIPHER *EVP_aes_256_cbc_hmac_sha1(void)
{
    return &aesoct_256_cbc_hmac_sha1_cipher;
}
# else
const EVP_CIPHER *EVP_aes_128_cbc_hmac_sha1(void)
{
    return NULL;
}

const EVP_CIPHER *EVP_aes_256_cbc_hmac_sha1(void)
{
    return NULL;
}
# endif
#endif
