/*------------------------------------------------------------------
 * kdf/kdf_snmp.c - Key Derivation Function for SNMP
 *
 * February 2015
 *
 * Copyright (c) 2015 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/kdf.h>
#include <openssl/sha.h>

/*
 * kdf_snmp - in compliance with SP800-135 and RFC2574, calculate
 *            a master key using the engine and password.
 *
 *    e_id - engine ID
 *    e_len - engineID length
 *    password - password
 *    pw_len - password length
 *    digest - key output, always length of 20 bytes.
 */
int kdf_snmp(unsigned char *e_id, int e_len, const char *password, 
             int pw_len, unsigned char *digest)
{
    EVP_MD_CTX md;
    int len;
    unsigned int md_len = -1;

    if (!e_id || !e_len || !password || !pw_len || !digest) {
        KDFerr(KDF_F_KDF_SNMP, KDF_R_INPUT_PARAMETER_ERROR);
        return -1;
    }

    EVP_MD_CTX_init(&md);
    if (!EVP_DigestInit_ex(&md, EVP_sha1(), NULL)) {
	goto err;
    }
    for (len = 0; len < PASSWORD_HASH_AMOUNT - pw_len; len += pw_len ) {
        if (!EVP_DigestUpdate(&md, password, pw_len)) {
	    goto err;
        }
    }
    if (!EVP_DigestUpdate(&md, password, PASSWORD_HASH_AMOUNT - len)) {
	goto err;
    }
    if (!EVP_DigestFinal_ex(&md, digest, &md_len)) {
        goto err;
    }

    if (!EVP_DigestInit_ex(&md, EVP_sha1(), NULL)) {
        goto err;
    }
    if (!EVP_DigestUpdate(&md, digest, SHA_DIGEST_LENGTH)) {
        goto err;
    }
    if (!EVP_DigestUpdate(&md, e_id, e_len)) {
        goto err;
    }
    if (!EVP_DigestUpdate(&md, digest, SHA_DIGEST_LENGTH)) {
        goto err;
    }
    if (!EVP_DigestFinal_ex(&md, digest, &md_len)) {
        goto err;
    }

err:
    EVP_MD_CTX_cleanup(&md);
    return md_len;
}
