/*------------------------------------------------------------------
 * kdf/kdf_ssh.c - Key Derivation Function for SSH
 *
 * March 2015
 *
 * Copyright (c) 2015 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cryptlib.h"
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/kdf.h>
#include <openssl/sha.h>
#ifdef OPENSSL_FIPS
# include <openssl/fips.h>
#endif

/*
 * kdf_ssh - in compliance with SP800-135 and RFC 4253, calculate
 *            IVs and keys.
 *
 */
int kdf_ssh(const EVP_MD *evp_md, int id, unsigned int need, char *shared_secret, 
            int ss_len, char *session_id, int session_id_len, char *hash, 
	    int hash_len, unsigned char *digest)
{

    char c = id;
    EVP_MD_CTX md;
    unsigned int mdsz, have;

    if (!evp_md || !id || !need || !shared_secret || !ss_len || !session_id ||
        !session_id_len || !hash || !hash_len || !digest) {
        KDFerr(KDF_F_KDF_SSH, KDF_R_INPUT_PARAMETER_ERROR);
        return -1;
    }

    mdsz = evp_md->md_size;
    switch(evp_md->type) {
        case NID_sha1:
        case NID_sha224:
        case NID_sha256:
        case NID_sha384:
        case NID_sha512:
	    EVP_MD_CTX_init(&md);
	    /* K1 = HASH(K || H || "A" || session_id) */
	    if (!EVP_DigestInit_ex(&md, evp_md, NULL)) {
	        goto err;
	    }
	    if (!EVP_DigestUpdate(&md, shared_secret, ss_len)) {
	        goto err;
	    }
	    if (!EVP_DigestUpdate(&md, hash, hash_len)) {
	        goto err;
	    }
	    if (!EVP_DigestUpdate(&md, &c, 1)) {
	        goto err;
	    }
	    if (!EVP_DigestUpdate(&md, session_id, session_id_len)) {
	        goto err;
	    }
	    if (!EVP_DigestFinal_ex(&md, digest, NULL)) {
	        goto err;
	    }

	    /*
	     * expand key:
	     * Kn = HASH(K || H || K1 || K2 || ... || Kn-1)
	     * Key = K1 || K2 || ... || Kn
	     */
	    for (have = mdsz; need > have; have += mdsz) {
	        if (!EVP_DigestInit_ex(&md, evp_md, NULL)) {
	            goto err;
                 }
		 if (!EVP_DigestUpdate(&md, shared_secret, ss_len)) {
 		     goto err;
                 }
		 if (!EVP_DigestUpdate(&md, hash, hash_len)) {
 		     goto err;
                 }
		 if (!EVP_DigestUpdate(&md, digest, have)) {
 		     goto err;
                 }
		 if (!EVP_DigestFinal_ex(&md, digest + have, NULL)) {
 		     goto err;
                 }
	    }
	    break;
        default:
            KDFerr(KDF_F_KDF_SSH, KDF_R_BAD_DIGEST);
	    return -1;
	    break;
    }    
err:
    EVP_MD_CTX_cleanup(&md);
    return 0;
}
