/* -----------------------------------------------------------------
 * crypto/sha/asm/sha256-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <assert.h>
#include <string.h>
#include <openssl/crypto.h>
#ifndef _MIPS_ARCH_OCTEON3
# include "ciscossl-cvmx.h"
#else
# include "cvmx.h"
#endif



#include <openssl/sha.h>

/*
 * SHA224OCT - Perform a single SHA224 digest
 */
unsigned char *SHA224OCT(const unsigned char *d, size_t n, unsigned char *md)
{
    sha_assert((d && md), NULL);
    SHA256_CTX c;
    SHA224OCT_Init(&c);
    SHA256OCT_Update(&c, d, n);
    SHA256OCT_Final(md, &c);
    return md;
}

/*
 * SHA256 - Perform a single SHA256 digest
 */
unsigned char *SHA256OCT(const unsigned char *d, size_t n, unsigned char *md)
{
    sha_assert((d && md), NULL);
    SHA256_CTX c;
    SHA256OCT_Init(&c);
    SHA256OCT_Update(&c, d, n);
    SHA256OCT_Final(md, &c);
    return md;
}

/*
 * SHA224OCT_Init - initializes a SHA256_CTX structure.
 */
int SHA224OCT_Init(SHA256_CTX *ctx)
{
    sha_assert(ctx, -1);
    ctx->iv[0] = 0xc1059ed8367cd507ull;
    ctx->iv[1] = 0x3070dd17f70e5939ull;
    ctx->iv[2] = 0xffc00b3168581511ull;
    ctx->iv[3] = 0x64f98fa7befa4fa4ull;
    ctx->totalin = 0;
    ctx->done = 0xbabe;
    ctx->plen = 0;
    ctx->sha256 = 0;
    return 1;
}

/*
 * SHA256OCT_Init - initializes a SHA256_CTX structure.
 */
int SHA256OCT_Init(SHA256_CTX *ctx)
{
    sha_assert(ctx, -1);
    ctx->iv[0] = 0x6a09e667bb67ae85ull;
    ctx->iv[1] = 0x3c6ef372a54ff53aull;
    ctx->iv[2] = 0x510e527f9b05688cull;
    ctx->iv[3] = 0x1f83d9ab5be0cd19ull;
    ctx->totalin = 0;
    ctx->done = 0xbabe;
    ctx->plen = 0;
    ctx->sha256 = 1;
    return 1;
}

/*
 * SHA256OCT_Update - can be called repeatedly with chunks of the
 * message to be hashed (len bytes at data).
 */
int SHA256OCT_Update(SHA256_CTX *ctx, const void *data, size_t len)
{
    sha_assert(ctx, -1);
    if (!data) {
        return -1;
    }
    if (!len) {
        return 1;
    }
    if (ctx->done == 0xdead) {
        return -1;
    }
    sha_assert(ctx->done == 0xbabe, -1);

    // Increment totalin
    ctx->totalin += len;

    if ((ctx->plen + len) >= 64) {
        // restore IV
        CVMX_MT_HSH_IV(ctx->iv[0], 0);
        CVMX_MT_HSH_IV(ctx->iv[1], 1);
        CVMX_MT_HSH_IV(ctx->iv[2], 2);
        CVMX_MT_HSH_IV(ctx->iv[3], 3);

        if (ctx->plen) {
            memcpy(&ctx->pbuf[ctx->plen], data, 64 - ctx->plen);
            len -= (64 - ctx->plen);
            data = (uint8_t *) data + (64 - ctx->plen);
            ctx->plen = 0;
            uint64_t *ptr = (uint64_t *) ctx->pbuf;
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 0);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 1);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 2);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 3);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 4);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 5);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 6);
	    CVMX_MT_HSH_STARTSHA256 (cvmx_htobe64(*ptr++));
        }

        uint64_t *ptr = (uint64_t *) data;
        while (len >= 64) {
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 0);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 1);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 2);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 3);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 4);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 5);
	    CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 6);
	    CVMX_MT_HSH_STARTSHA256 (cvmx_htobe64(*ptr++));
            len -= 64;
        }

        if (len) {
            memcpy(ctx->pbuf, ptr, len);
            ctx->plen = len;
        }
        // save IV
        CVMX_MF_HSH_IV(ctx->iv[0], 0);
        CVMX_MF_HSH_IV(ctx->iv[1], 1);
        CVMX_MF_HSH_IV(ctx->iv[2], 2);
        CVMX_MF_HSH_IV(ctx->iv[3], 3);
    } else {
        memcpy(&ctx->pbuf[ctx->plen], data, len);
        ctx->plen += len;
    }
    return 1;
}

/*
 * SHA256OCT_Final - places the message digest in md
 */
int SHA256OCT_Final(unsigned char *md, SHA256_CTX *ctx)
{
    sha_assert(md && ctx, -1);
    if (ctx->done == 0xdead) {
        return -1;
    }
    sha_assert(ctx->done == 0xbabe, -1);
    if (ctx->done == 0xbabe) {
        ctx->done = 0xdead;
        // RESTORE IV
        CVMX_MT_HSH_IV(ctx->iv[0], 0);
        CVMX_MT_HSH_IV(ctx->iv[1], 1);
        CVMX_MT_HSH_IV(ctx->iv[2], 2);
        CVMX_MT_HSH_IV(ctx->iv[3], 3);

        ctx->pbuf[ctx->plen] = 0x80;
        memset(ctx->pbuf + ctx->plen + 1, 0, 64 - ctx->plen - 1);

        uint64_t *ptr = (uint64_t *) ctx->pbuf;
	CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 0);
	CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 1);
	CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 2);
	CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 3);
	CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 4);
	CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 5);
	CVMX_MT_HSH_DAT (cvmx_htobe64(*ptr++), 6);
        if (ctx->plen < 56) {
            CVMX_MT_HSH_STARTSHA256(ctx->totalin * 8);
        } else {
            CVMX_MT_HSH_STARTSHA256 (cvmx_htobe64(*ptr));
            /* Another block was needed */
            CVMX_MT_HSH_DATZ(0);
            CVMX_MT_HSH_DATZ(1);
            CVMX_MT_HSH_DATZ(2);
            CVMX_MT_HSH_DATZ(3);
            CVMX_MT_HSH_DATZ(4);
            CVMX_MT_HSH_DATZ(5);
            CVMX_MT_HSH_DATZ(6);
            CVMX_MT_HSH_STARTSHA256(ctx->totalin * 8);
        }

        // WRITE IV
        CVMX_MF_HSH_IV(((uint64_t *) md)[0], 0);
        CVMX_MF_HSH_IV(((uint64_t *) md)[1], 1);
        CVMX_MF_HSH_IV(((uint64_t *) md)[2], 2);
#ifdef L_ENDIAN
       ((uint64_t *) md)[0]  =  cvmx_betoh64(((uint64_t *) md)[0]);
       ((uint64_t *) md)[1]  =  cvmx_betoh64(((uint64_t *) md)[1]);
       ((uint64_t *) md)[2]  =  cvmx_betoh64(((uint64_t *) md)[2]);
#endif
        if (ctx->sha256) {
            CVMX_MF_HSH_IV(((uint64_t *) md)[3], 3);
#ifdef L_ENDIAN
            ((uint64_t *) md)[3]  =  cvmx_betoh64(((uint64_t *) md)[3]);
#endif
        } else {
            CVMX_MF_HSH_IV(ctx->iv[3], 3);
#ifdef L_ENDIAN
	    ctx->iv[3]  =  cvmx_betoh64(ctx->iv[3]);
#endif
            memcpy(md + 24, &ctx->iv[3], 4);
        }
    }
    return 1;
}
