/*------------------------------------------------------------------
 * fips_test_ig95.c - Unit test for IG 9.5 support
 *
 * April, 2015
 *
 * Copyright (c) 2015 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
/* ====================================================================
 * Copyright (c) 2003 The OpenSSL Project.  All rights reserved.
 *
 *
 * This command is intended as a test driver for the FIPS-140 testing
 * lab performing FIPS-140 validation.  It demonstrates the use of the
 * OpenSSL library ito perform a variety of common cryptographic
 * functions.  A power-up self test is demonstrated by deliberately
 * pointing to an invalid executable hash
 *
 * Contributed by Steve Marquess.
 *
 */

#define OPENSSL_FIPSAPI

#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>
#include <openssl/ec.h>
#include <openssl/ecdsa.h>
#include <openssl/ecdh.h>
#include <openssl/cmac.h>
#include <openssl/sha.h>
#include <openssl/err.h>
#include <openssl/rsa.h>
#include <openssl/dsa.h>
#include <openssl/dh.h>
#include <openssl/bn.h>
#include <openssl/rand.h>

#ifndef OPENSSL_FIPS
int main(int argc, char *argv[])
{
    printf("No FIPS support\n");
    return(0);
}
#else

#define ERR_clear_error() while (0)


#include <openssl/fips.h>
#include <openssl/fips_rand.h>
#include "fips_utl.h"

/* AES: encrypt and decrypt known plaintext, verify result matches original plaintext */
static int FIPS_aes_test(void)
{
    int ret = 0;
    unsigned char pltmp[16];
    unsigned char citmp[16];
    unsigned char key[16] = {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16};
    unsigned char plaintext[16] = "etaonrishdlcu";
    EVP_CIPHER_CTX ctx;
    FIPS_cipher_ctx_init(&ctx);
    if (FIPS_cipherinit(&ctx, EVP_aes_128_ecb(), key, NULL, 1) <= 0) goto err;
    FIPS_cipher(&ctx, citmp, plaintext, 16);
    if (FIPS_cipherinit(&ctx, EVP_aes_128_ecb(), key, NULL, 0) <= 0) goto err;
    FIPS_cipher(&ctx, pltmp, citmp, 16);
    if (memcmp(pltmp, plaintext, 16)) goto err;
    ret = 1;
err:
    FIPS_cipher_ctx_cleanup(&ctx);
    return ret;
}

static int FIPS_aes_gcm_test(void)
{
    int ret = 0;
    unsigned char pltmp[16];
    unsigned char citmp[16];
    unsigned char tagtmp[16];
    unsigned char key[16] = {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16};
    unsigned char iv[16] = {21,22,23,24,25,26,27,28,29,30,31,32};
    unsigned char aad[] = "Some text AAD";
    unsigned char plaintext[16] = "etaonrishdlcu";
    EVP_CIPHER_CTX ctx;
    FIPS_cipher_ctx_init(&ctx);
    if (FIPS_cipherinit(&ctx, EVP_aes_128_gcm(), key, iv, 1) <= 0) goto err;
    FIPS_cipher(&ctx, NULL, aad, sizeof(aad));
    FIPS_cipher(&ctx, citmp, plaintext, 16);
    FIPS_cipher(&ctx, NULL, NULL, 0);
    if (!FIPS_cipher_ctx_ctrl(&ctx, EVP_CTRL_GCM_GET_TAG, 16, tagtmp)) goto err;

    if (FIPS_cipherinit(&ctx, EVP_aes_128_gcm(), key, iv, 0) <= 0) goto err;
    if (!FIPS_cipher_ctx_ctrl(&ctx, EVP_CTRL_GCM_SET_TAG, 16, tagtmp)) goto err;

    FIPS_cipher(&ctx, NULL, aad, sizeof(aad));

    FIPS_cipher(&ctx, pltmp, citmp, 16);

    if (FIPS_cipher(&ctx, NULL, NULL, 0) < 0) goto err;

    if (memcmp(pltmp, plaintext, 16)) goto err;

    ret = 1;
err:
    FIPS_cipher_ctx_cleanup(&ctx);
    return ret;
}

static int FIPS_des3_test(void)
{
    int ret = 0;
    unsigned char pltmp[8];
    unsigned char citmp[8];
    unsigned char key[] = {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,
                           19,20,21,22,23,24};
    unsigned char plaintext[] = {'e', 't', 'a', 'o', 'n', 'r', 'i', 's'};
    EVP_CIPHER_CTX ctx;
    FIPS_cipher_ctx_init(&ctx);
    if (FIPS_cipherinit(&ctx, EVP_des_ede3_ecb(), key, NULL, 1) <= 0) goto err;
    FIPS_cipher(&ctx, citmp, plaintext, 8);
    if (FIPS_cipherinit(&ctx, EVP_des_ede3_ecb(), key, NULL, 0) <= 0) goto err;
    FIPS_cipher(&ctx, pltmp, citmp, 8);
    if (memcmp(pltmp, plaintext, 8)) goto err;
    ret = 1;
err:
    FIPS_cipher_ctx_cleanup(&ctx);
    return ret;
}

/*
 * DSA: generate keys and sign, verify input plaintext.
 */
static int FIPS_dsa_test()
{
    DSA *dsa = NULL;
    unsigned char dgst[] = "etaonrishdlc";
    int r = 0;
    DSA_SIG *sig = NULL;

    ERR_clear_error();
    dsa = FIPS_dsa_new();
    if (!dsa) goto end;
    if (!DSA_generate_parameters_ex(dsa, 2048,NULL,0,NULL,NULL,NULL)) goto end;
    if (!DSA_generate_key(dsa)) goto end;

    sig = FIPS_dsa_sign(dsa, dgst, sizeof(dgst) -1, EVP_sha256());
    if (!sig) goto end;

    r = FIPS_dsa_verify(dsa, dgst, sizeof(dgst) -1, EVP_sha256(), sig);
end:
    if (sig) FIPS_dsa_sig_free(sig);
    if (dsa) FIPS_dsa_free(dsa);
    if (r != 1) return 0;
    return 1;
}

/*
 * RSA: generate keys and sign, verify input plaintext.
 */
static int FIPS_rsa_test()
{
    RSA *key;
    unsigned char input_ptext[] = "etaonrishdlc";
    unsigned char buf[256];
    unsigned int slen;
    BIGNUM *bn;
    int r = 0;

    ERR_clear_error();
    key = FIPS_rsa_new();
    bn = BN_new();
    if (!key || !bn) return 0;
    BN_set_word(bn, 65537);
    if (!RSA_generate_key_ex(key, 2048,bn,NULL)) {
	r = 0;
    }
    BN_free(bn);

    if (!FIPS_rsa_sign(key, input_ptext, sizeof(input_ptext) - 1, EVP_sha256(),
                       RSA_PKCS1_PADDING, 0, NULL, buf, &slen)) {
	r = 0;
    }

    r = FIPS_rsa_verify(key, input_ptext, sizeof(input_ptext) - 1, EVP_sha256(),
                        RSA_PKCS1_PADDING, 0, NULL, buf, slen);
    if (key) FIPS_rsa_free(key);
    if (r != 1) return 0;
    return 1;
}

/*
 * This is used for ECDH computation
 */
static void *KDF1_SHA1(const void *in, size_t inlen, void *out, size_t *outlen)
{
    SHA_CTX c;

    if (out == NULL) return NULL;
    if (*outlen < SHA_DIGEST_LENGTH) return NULL;
    else *outlen = SHA_DIGEST_LENGTH;

    if (!SHA1_Init(&c)) return NULL;
    SHA1_Update(&c, in, inlen);
    SHA1_Final(out, &c);
    return out;
}

/*
 * ECDSA: generate keys and sign, verify input plaintext.
 */
static int FIPS_ecdsa_test()
{
    int rv = 1;
    EC_GROUP *curve = NULL;
    EC_KEY *K = NULL;
    unsigned char hash[32];
    ECDSA_SIG *sig = NULL;
    int i = 0;

    ERR_clear_error();

    // Use a randomized hash value.
    for (i = 0; i < 32; i++) hash[i] = rand() % 256;

    curve = EC_GROUP_new_by_curve_name(NID_X9_62_prime256v1);
    K = EC_KEY_new();
    EC_KEY_set_group(K, curve);
    if (EC_KEY_generate_key(K) <= 0) {
	rv = 0;
    }
    sig = FIPS_ecdsa_sign_digest (K, hash, sizeof(hash));
    if (!sig) {
	rv = 0;
    }

    if (FIPS_ecdsa_verify_digest (K, hash, sizeof(hash), sig) <= 0) {
	rv = 0;
    }

    ECDSA_SIG_free(sig);
    EC_KEY_free(K);
    EC_GROUP_free(curve);

    return rv;
}

/*
 * ECDH: compute a secret using ECDH keys.
 */
static int FIPS_ecdh_test()
{
    EC_GROUP *curve = NULL;
    EC_KEY *K = NULL;
    EC_KEY *K2 = NULL;
    int alen, aout;
    unsigned char *abuf;
    int rv = 1;

    ERR_clear_error();

    curve = EC_GROUP_new_by_curve_name(NID_X9_62_prime256v1);
    K = EC_KEY_new();
    K2 = EC_KEY_new();
    EC_KEY_set_group(K, curve);
    EC_KEY_set_group(K2, curve);
    if (EC_KEY_generate_key(K) <= 0) {
	rv = 0;
    }
    if (EC_KEY_generate_key(K2) <= 0) {
	rv = 0;
    }

    alen = 20; /* SHA-1 */
    abuf = (unsigned char *)OPENSSL_malloc(alen);
    aout = ECDH_compute_key(abuf,alen,EC_KEY_get0_public_key(K2),K,KDF1_SHA1);
    if (aout <= 0) {
	rv = 0;
    }

    free(abuf);
    EC_KEY_free(K);
    EC_KEY_free(K2);
    EC_GROUP_free(curve);
    return rv;
}

/* SHA1: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_sha1_test()
{
    unsigned char digest[SHA_DIGEST_LENGTH] =
    {0x11, 0xf1, 0x9a, 0x3a, 0xec, 0x1a, 0x1e, 0x8e, 0x65, 0xd4, 0x9a, 0x38, 0x0c, 0x8b, 0x1e, 0x2c, 0xe8, 0xb3, 0xc5, 0x18};
    unsigned char str[] = "etaonrishd";

    unsigned char md[SHA_DIGEST_LENGTH];

    ERR_clear_error();
    if (!FIPS_digest(str,sizeof(str) - 1,md, NULL, EVP_sha1())) return 0;
    if (memcmp(md,digest,sizeof(md))) return 0;
    return 1;
}

/* SHA256: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_sha256_test()
{
    unsigned char digest[SHA256_DIGEST_LENGTH] =
    {0xf5, 0x53, 0xcd, 0xb8, 0xcf, 0x1, 0xee, 0x17, 0x9b, 0x93, 0xc9, 0x68, 0xc0, 0xea, 0x40, 0x91,
     0x6, 0xec, 0x8e, 0x11, 0x96, 0xc8, 0x5d, 0x1c, 0xaf, 0x64, 0x22, 0xe6, 0x50, 0x4f, 0x47, 0x57};
    unsigned char str[] = "etaonrishd";

    unsigned char md[SHA256_DIGEST_LENGTH];

    ERR_clear_error();
    if (!FIPS_digest(str,sizeof(str) - 1,md, NULL, EVP_sha256())) return 0;
    if (memcmp(md,digest,sizeof(md))) return 0;
    return 1;
}

/* SHA512: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_sha512_test()
{
    unsigned char digest[SHA512_DIGEST_LENGTH] =
    {0x99, 0xc9, 0xe9, 0x5b, 0x88, 0xd4, 0x78, 0x88, 0xdf, 0x88, 0x5f, 0x94, 0x71, 0x64, 0x28, 0xca,
     0x16, 0x1f, 0x3d, 0xf4, 0x1f, 0xf3, 0x0f, 0xc5, 0x03, 0x99, 0xb2, 0xd0, 0xe7, 0x0b, 0x94, 0x4a,
     0x45, 0xd2, 0x6c, 0x4f, 0x20, 0x06, 0xef, 0x71, 0xa9, 0x25, 0x7f, 0x24, 0xb1, 0xd9, 0x40, 0x22,
     0x49, 0x54, 0x10, 0xc2, 0x22, 0x9d, 0x27, 0xfe, 0xbd, 0xd6, 0xd6, 0xeb, 0x2d, 0x42, 0x1d, 0xa3};
    unsigned char str[] = "etaonrishd";

    unsigned char md[SHA512_DIGEST_LENGTH];

    ERR_clear_error();
    if (!FIPS_digest(str,sizeof(str) - 1,md, NULL, EVP_sha512())) return 0;
    if (memcmp(md,digest,sizeof(md))) return 0;
    return 1;
}

/* HMAC-SHA1: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_hmac_sha1_test()
{
    unsigned char key[] = "etaonrishd1234";
    unsigned char iv[] = "Sample text";
    unsigned char kaval[EVP_MAX_MD_SIZE] =
    {0x3f, 0xf8, 0x05, 0x0b, 0x85, 0x7d, 0x51, 0xdd, 0x8f, 0x8a, 0x88, 0xac, 0x6a, 0xee, 0x4a, 0x9e,
     0x07, 0xaa, 0x70, 0x19};

    unsigned char out[EVP_MAX_MD_SIZE];
    unsigned int outlen;

    ERR_clear_error();
    if (!HMAC(EVP_sha1(),key,sizeof(key)-1,iv,sizeof(iv)-1,out,&outlen)) return 0;
    if (memcmp(out,kaval,outlen)) return 0;
    return 1;
}

/* HMAC-SHA224: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_hmac_sha224_test()
{
    unsigned char key[] = "etaonrishd1234";
    unsigned char iv[] = "Sample text";
    unsigned char kaval[EVP_MAX_MD_SIZE] =
    {0x4a,  0x56, 0xcd, 0x19, 0x13, 0x14, 0x7e, 0x2b, 0x0a, 0xeb, 0xa2, 0xac, 0x6b, 0x0f, 0x7c, 0x64,
     0xee, 0x14, 0x1c, 0x59, 0xd0, 0xc7, 0x2a, 0xbc, 0xd1, 0x71, 0xca, 0xbf};

    unsigned char out[EVP_MAX_MD_SIZE];
    unsigned int outlen;

    ERR_clear_error();
    if (!HMAC(EVP_sha224(),key,sizeof(key)-1,iv,sizeof(iv)-1,out,&outlen)) return 0;
    if (memcmp(out,kaval,outlen)) return 0;
    return 1;
}

/* HMAC-SHA256: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_hmac_sha256_test()
{
    unsigned char key[] = "etaonrishd1234";
    unsigned char iv[] = "Sample text";
    unsigned char kaval[EVP_MAX_MD_SIZE] =
    {0xc1, 0x12, 0x02, 0x15, 0x3d, 0xef, 0xde, 0xf9, 0x20, 0x46, 0x7f, 0x50, 0xfa, 0xeb, 0xa6, 0xec,
     0xd9, 0x08, 0x71, 0x99, 0x81, 0x8e, 0xfc, 0x7e, 0x2e, 0xec, 0x42, 0xcc, 0xdf, 0xcc, 0x9c, 0xba};

    unsigned char out[EVP_MAX_MD_SIZE];
    unsigned int outlen;

    ERR_clear_error();
    if (!HMAC(EVP_sha256(),key,sizeof(key)-1,iv,sizeof(iv)-1,out,&outlen)) return 0;
    if (memcmp(out,kaval,outlen)) return 0;
    return 1;
}

/* HMAC-SHA384: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_hmac_sha384_test()
{
    unsigned char key[] = "etaonrishd1234";
    unsigned char iv[] = "Sample text";
    unsigned char kaval[EVP_MAX_MD_SIZE] =
    {0x40, 0x52, 0x23, 0x86, 0x6b, 0xf8, 0x4c, 0xdb, 0x54, 0x36, 0x26, 0x3f, 0xaa, 0x77, 0xd4, 0xf6,
     0xf4, 0xc4, 0xf4, 0x57, 0xfd, 0x83, 0x66, 0xce, 0xf8, 0x9d, 0x12, 0x83, 0xcf, 0x64, 0x11, 0x45,
     0xce, 0xed, 0x89, 0xf0, 0x49, 0xba, 0x48, 0x8d, 0x24, 0x5d, 0xd7, 0x34, 0x87, 0x54, 0xea, 0xa3};

    unsigned char out[EVP_MAX_MD_SIZE];
    unsigned int outlen;

    ERR_clear_error();
    if (!HMAC(EVP_sha384(),key,sizeof(key)-1,iv,sizeof(iv)-1,out,&outlen)) return 0;
    if (memcmp(out,kaval,outlen)) return 0;
    return 1;
}

/* HMAC-SHA512: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_hmac_sha512_test()
{
    unsigned char key[] = "etaonrishd1234";
    unsigned char iv[] = "Sample text";
    unsigned char kaval[EVP_MAX_MD_SIZE] =
    {0x1a, 0xc1, 0x0b, 0x7e, 0x21, 0x64, 0xa7, 0x19, 0x25, 0x47, 0x5f, 0x46, 0xf7, 0xde, 0x4a, 0x3a,
     0x17, 0x1e, 0x6b, 0xd4, 0xca, 0x1b, 0xce, 0x01, 0x7c, 0x75, 0xef, 0x99, 0x2d, 0x18, 0x5f, 0x7d,
     0xec, 0x79, 0x27, 0xfd, 0x6b, 0x60, 0x87, 0xa9, 0xe0, 0xd6, 0x04, 0x32, 0x69, 0xf0, 0xbc, 0xd7,
     0x07, 0x7a, 0x91, 0xc1, 0xcc, 0xd6, 0xfa, 0x64, 0xd2, 0x41, 0x9a, 0xa6, 0xfe, 0x15, 0xa6, 0x34};

    unsigned char out[EVP_MAX_MD_SIZE];
    unsigned int outlen;

    ERR_clear_error();
    if (!HMAC(EVP_sha512(),key,sizeof(key)-1,iv,sizeof(iv)-1,out,&outlen)) return 0;
    if (memcmp(out,kaval,outlen)) return 0;
    return 1;
}

#define MAX_CMAC 128
/* CMAC-AES128: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_cmac_aes128_test()
{
    unsigned char key[16] = {0x2b,0x7e,0x15,0x16, 0x28,0xae,0xd2,0xa6,
                             0xab,0xf7,0x15,0x88, 0x09,0xcf,0x4f,0x3c, };
    unsigned char data[] = "Sample text";
    unsigned char kaval[EVP_MAX_MD_SIZE] =
    {0x16,0x83,0xfe,0xac, 0x52,0x9b,0xae,0x23,
     0xd7,0xd5,0x66,0xf5, 0xd2,0x8d,0xbd,0x2a, };

    unsigned char *out = NULL;
    size_t outlen;
    CMAC_CTX *ctx;
    int r = 0;

    out = OPENSSL_malloc(MAX_CMAC);

    ERR_clear_error();

    ctx = CMAC_CTX_new();
    if (!ctx) return 0;

    if (!CMAC_Init(ctx,key,sizeof(key),EVP_aes_128_cbc(),NULL)) {
	r = 0;
    }
    if (!CMAC_Update(ctx,data,sizeof(data)-1)) {
	r = 0;
    }
    if (!CMAC_Final(ctx, out, &outlen)) {
	r = 0;
    }
    if (!memcmp(out,kaval,outlen)) {
        r = 1;
    }

    CMAC_CTX_free(ctx);
    if (out) OPENSSL_free(out);
    return r;
}


/* CMAC-AES192: generate hash of known digest value and compare to known
   precomputed correct hash */
static int FIPS_cmac_aes192_test()
{
    unsigned char key[] = {0x8e,0x73,0xb0,0xf7, 0xda,0x0e,0x64,0x52,
                           0xc8,0x10,0xf3,0x2b, 0x80,0x90,0x79,0xe5,
                           0x62,0xf8,0xea,0xd2, 0x52,0x2c,0x6b,0x7b, };
    unsigned char data[] = "Sample text";
    unsigned char kaval[] =
    {0xd6,0x99,0x19,0x25, 0xe5,0x1d,0x95,0x48,
     0xb1,0x4a,0x0b,0xf2, 0xc6,0x3c,0x47,0x1f, };

    unsigned char *out = NULL;
    size_t outlen;
    CMAC_CTX *ctx;
    int r = 0;

    out = OPENSSL_malloc(MAX_CMAC);

    ERR_clear_error();

    ctx = CMAC_CTX_new();
    if (!ctx) return 0;

    if (!CMAC_Init(ctx,key,sizeof(key),EVP_aes_192_cbc(),NULL)) {
	r = 0;
    }
    if (!CMAC_Update(ctx,data,sizeof(data)-1)) {
	r = 0;
    }
    if (!CMAC_Final(ctx, out, &outlen)) {
	r = 0;
    }

    if (!memcmp(out,kaval,outlen)) {
        r = 1;
    }

    CMAC_CTX_free(ctx);
    if (out) OPENSSL_free(out);
    return r;
}

/* CMAC-AES256: generate hash of known digest value and compare to known
   precomputed correct hash
 */
static int FIPS_cmac_aes256_test()
{
    unsigned char key[] = {0x60,0x3d,0xeb,0x10, 0x15,0xca,0x71,0xbe,
                           0x2b,0x73,0xae,0xf0, 0x85,0x7d,0x77,0x81,
                           0x1f,0x35,0x2c,0x07, 0x3b,0x61,0x08,0xd7,
                           0x2d,0x98,0x10,0xa3, 0x09,0x14,0xdf,0xf4, };
    unsigned char data[] = "Sample text";
    unsigned char kaval[] =
    {0xec,0xc2,0xcf,0x63, 0xc7,0xce,0xfc,0xa4,
     0xb0,0x86,0x37,0x5f, 0x15,0x60,0xba,0x1f, };

    unsigned char *out = NULL;
    size_t outlen;
    CMAC_CTX *ctx;
    int r = 0;

    out = OPENSSL_malloc(MAX_CMAC);

    ERR_clear_error();

    ctx = CMAC_CTX_new();
    if (!ctx) return 0;
    if (!CMAC_Init(ctx,key,sizeof(key),EVP_aes_256_cbc(),NULL)) {
	r = 0;
    }
    if (!CMAC_Update(ctx,data,sizeof(data)-1)) {
	r = 0;
    }
    if (!CMAC_Final(ctx, out, &outlen)) {
	r = 0;
    }
    if (!memcmp(out,kaval,outlen)) {
        r = 1;
    }

    CMAC_CTX_free(ctx);
    if (out) OPENSSL_free(out);
    return r;
}

/* CMAC-TDEA3: generate hash of known digest value and compare to known
   precomputed correct hash
 */
static int FIPS_cmac_tdea3_test()
{
    unsigned char key[] = {0x8a,0xa8,0x3b,0xf8, 0xcb,0xda,0x10,0x62,
                           0x0b,0xc1,0xbf,0x19, 0xfb,0xb6,0xcd,0x58,
                           0xbc,0x31,0x3d,0x4a, 0x37,0x1c,0xa8,0xb5, };
    unsigned char data[] = "Sample text";
    unsigned char kaval[EVP_MAX_MD_SIZE] =
    {0xb4,0x06,0x4e,0xbf, 0x59,0x89,0xba,0x68, };

    unsigned char *out = NULL;
    size_t outlen;
    CMAC_CTX *ctx;
    int r = 0;

    out = OPENSSL_malloc(MAX_CMAC);
    ERR_clear_error();

    ctx = CMAC_CTX_new();
    if (!ctx) return 0;
    if (!CMAC_Init(ctx,key,sizeof(key),EVP_des_ede3_cbc(),NULL)) {
	r = 0;
    }
    if (!CMAC_Update(ctx,data,sizeof(data)-1)) {
	r = 0;
    }
    if (!CMAC_Final(ctx, out, &outlen)) {
	r = 0;
    }
    if (!memcmp(out,kaval,outlen)) {
        r = 1;
    }

    CMAC_CTX_free(ctx);
    if (out) OPENSSL_free(out);
    return r;
}

#ifdef OCTEON_SINGLE_PASS
static int FIPS_aes_cbc_hmac_sha1(void)
	{
	int ret = 0;
	unsigned char out[128];
	EVP_CIPHER_CTX ctx;
	const unsigned char aad[13] = {00,00,00,00,00,00,00,00,0x16,03,02,00,0x10};
	const unsigned char key[16] = {0x2B,0x7E,0x15,0x16,0x28,0xAE,0xD2,0xA6,0xAB,0xF7,0x15,0x88,0x09,0xCF,0x4F,0x3C};
	const unsigned char iv[16] = {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F};
	const unsigned char pt[16] = {0x6B,0xC1,0xBE,0xE2,0x2E,0x40,0x9F,0x96,0xE9,0x3D,0x7E,0x11,0x73,0x93,0x17,0x2A};
	unsigned char ct[16];


	FIPS_cipher_ctx_init(&ctx);
	memset(out, 0, sizeof(out));
	memset(ct, 0, sizeof(ct));
	if (!FIPS_cipherinit(&ctx, EVP_aes_128_cbc_hmac_sha1(), key, iv, 1))
		goto err;
	if (!FIPS_cipher_ctx_ctrl(&ctx, EVP_CTRL_CIPH_SET_MODE,
					EX_MODE_ENCRYPT, NULL))
		goto err;
        if(!FIPS_cipher_ctx_ctrl(&ctx, EVP_CTRL_AEAD_TLS1_AAD,
                                  EVP_AEAD_TLS1_AAD_LEN,(void *) aad))
		goto err;
	if (FIPS_cipher(&ctx, ct, pt, 16) < 0)
		goto err;


	if (!FIPS_cipherinit(&ctx, EVP_aes_128_cbc_hmac_sha1(), key, iv, 0))
		goto err;
	if (!FIPS_cipher_ctx_ctrl(&ctx, EVP_CTRL_CIPH_SET_MODE,
					EX_MODE_ENCRYPT, NULL))
		goto err;
        if(!FIPS_cipher_ctx_ctrl(&ctx, EVP_CTRL_AEAD_TLS1_AAD,
                                  EVP_AEAD_TLS1_AAD_LEN,(void *) aad))
		goto err;
	if (FIPS_cipher(&ctx, out, ct, 16) < 0)
		goto err;
	if (memcmp(out, pt, 16))
		goto err;

	ret = 1;

	err:
	FIPS_cipher_ctx_cleanup(&ctx);
	return ret;

	}
#endif

static int Error;
static const char * Fail(const char *msg)
{
    Error++;
    return msg;
}

static void test_msg(const char *msg, int result)
{
    printf("%s...%s\n", msg, result ? "successful" : Fail("Failed!"));
}

#ifdef FIPS_ALGVS
int fips_test_ig95_int(int argc, char **argv);
int fips_test_ig95_osx()
{
    char *dummy = "dummy";
    return fips_test_ig95_int(1, &dummy);
}
int fips_test_ig95_int(int argc, char **argv)
#else
int main(int argc, char **argv)
#endif
{

    printf("\tFIPS-mode test application for IG 9.5\n");

    printf("\t%s\n\n", FIPS_module_version_text());

    fips_algtest_init_nofips();

    /* Power-up self test */
    ERR_clear_error();

    /* AES encryption/decryption */
    test_msg("1. AES encryption/decryption", FIPS_aes_test());

    /* AES GCM encryption/decryption */
    test_msg("2. AES-GCM encryption/decryption", FIPS_aes_gcm_test());

    /* RSA key generation and encryption/decryption */
    test_msg("3. RSA key generation and encryption/decryption", FIPS_rsa_test());

    /* DES-CBC encryption/decryption */
    test_msg("4. DES-ECB encryption/decryption", FIPS_des3_test());

    /* DSA key generation and signature validation */
    test_msg("5. DSA key generation and signature validation", FIPS_dsa_test());

    /* SHA-1 hash */
    test_msg("6. SHA-1 hash", FIPS_sha1_test());

    /* SHA-256 hash */
    test_msg("7. SHA-256 hash", FIPS_sha256_test());

    /* SHA-512 hash */
    test_msg("8. SHA-512 hash", FIPS_sha512_test());

    /* HMAC-SHA-1 hash */
    test_msg("9. HMAC-SHA-1 hash", FIPS_hmac_sha1_test());

    /* HMAC-SHA-224 hash */
    test_msg("10. HMAC-SHA-224 hash", FIPS_hmac_sha224_test());

    /* HMAC-SHA-256 hash */
    test_msg("11. HMAC-SHA-256 hash", FIPS_hmac_sha256_test());

    /* HMAC-SHA-384 hash */
    test_msg("12. HMAC-SHA-384 hash", FIPS_hmac_sha384_test());

    /* HMAC-SHA-512 hash */
    test_msg("13. HMAC-SHA-512 hash", FIPS_hmac_sha512_test());

    /* CMAC-AES-128 hash */
    test_msg("14. CMAC-AES-128 hash", FIPS_cmac_aes128_test());

    /* CMAC-AES-192 hash */
    test_msg("15. CMAC-AES-192 hash", FIPS_cmac_aes192_test());

    /* CMAC-AES-256 hash */
    test_msg("16. CMAC-AES-256 hash", FIPS_cmac_aes256_test());

    /* CMAC-TDEA-3 hash */
    test_msg("17. CMAC-TDEA-3 hash", FIPS_cmac_tdea3_test());

    /* ECDSA key generation and encryption/decryption */
    test_msg("18. ECDSA key generation and encryption/decryption", FIPS_ecdsa_test());

    /* ECDDH computation */
    test_msg("19. ECDH computation", FIPS_ecdh_test());

#ifdef OCTEON_SINGLE_PASS
    /* Single Pass Octeon Ciphers computation */
    if (OCTEON_HWA_CAPABLE) {
        test_msg("20. Single Pass Ciphers", FIPS_aes_cbc_hmac_sha1());
    }
#endif

    //FIXME: need to add ECDH and ECDSA

    printf("\nAll tests completed with %d errors\n", Error);
    return Error ? 1 : 0;
}

#endif
