/*------------------------------------------------------------------
 * fips/kdf/fips_kwvs.c - Key Wrap vector tests
 *
 * February 2015
 *
 * Copyright (c) 2015 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#define OPENSSL_FIPSAPI
#include <openssl/opensslconf.h>

#ifndef OPENSSL_FIPS
#include <stdio.h>

int main(int argc, char *argv[])
{
    printf("No FIPS KDF support\n");
    return(0);
}

#else

#include <openssl/bn.h>
#include <openssl/fips.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <string.h>
#include <ctype.h>
#include "e_os.h"

#include "fips_utl.h"


#define LOOP_COUNTER 100   /* Sample vectors may be less than 100 */
#define VERBOSE 0
/*-----------------------------------------------*/
static int proc_kw_file (char *rqfile, char *rspfile, int dir, int padding)
    {
    char afn[256], rfn[256];
    FILE *afp = NULL, *rfp = NULL;
    unsigned char *out;
    char ibuf[2048];
    char tbuf[2048];
    int len, i;
    int err = 0, step = 0, preamble = 0, ready = 0;
    int plen = 0, klen = 0, clen = 0, plain_len = 0;
    char pt_len[8];
    unsigned char *key;
    unsigned char *plaintext;
    unsigned char *ciphertext;
    char *rp;
    int counter = 0;
    const EVP_CIPHER *cipher = NULL;
    EVP_CIPHER_CTX cctx;

    if (!rqfile || !(*rqfile))
	{
	printf("No req file\n");
	return -1;
	}
    strcpy(afn, rqfile);

    if ((afp = fopen(afn, "r")) == NULL)
	{
	printf("Cannot open file: %s, %s\n", 
	       afn, strerror(errno));
	return -1;
	}
    if (!rspfile)
	{
	strcpy(rfn,afn);
	rp=strstr(rfn,"req/");
#ifdef OPENSSL_SYS_WIN32
	if (!rp)
	    rp=strstr(rfn,"req\\");
#endif
	memcpy(rp,"rsp",3);
	rp = strstr(rfn, ".req");
	memcpy(rp, ".rsp", 4);
	rspfile = rfn;
	}
    if ((rfp = fopen(rspfile, "w")) == NULL)
	{
	printf("Cannot open file: %s, %s\n", 
	       rfn, strerror(errno));
	fclose(afp);
	afp = NULL;
	return -1;
	}

    /* Just allocate plenty here */
    key = malloc(1024);
    plaintext = malloc(1024*4);
    ciphertext = malloc(1024*4);
    if (!key || !plaintext || !ciphertext) {
	printf("\nFailed to allocate memory for cipher");
	return -1;
    }

    while (!err && (fgets(ibuf, sizeof(ibuf), afp)) != NULL)
	{
	tidy_line(tbuf, ibuf);
	if (VERBOSE)
	    printf("step=%d ibuf=%s",step,ibuf);

	switch (step)
	{
	case 0:  /* walk thru and write out preamble */
	    if (ibuf[0] == '\n')
	    {
	        if (preamble == 0)
		{ /* end of preamble */
		    printf("Missing parts of preamble \n");
		    err = 1;
		    break;
	        } else {
		    copy_line(ibuf, rfp);
		    step++;
		    break;
		}
            }
	    else if (ibuf[0] != '#')
		{
		printf("Invalid preamble item: %s\n", ibuf);
		err = 1;
		}
	    else
		{ /* process preamble */
		char *pp = ibuf+2;

		copy_line(ibuf, rfp);
		if (strncmp(pp, "CAVS ", 5) == 0)
		    {
		    preamble++;
		    }
		}
	    break;

	case 1:  /* read text length, not really used ? */
	    if (ibuf[0] == '[') 
	    {
                copy_line(ibuf, rfp);
		step++;
		if (fips_strncasecmp(ibuf+1, "PLAINTEXT LENGTH = ", 19) == 0) {
		    memset(pt_len, 0, 8);
		    /* parse the length value, its in bits */
		    strncpy(pt_len, (char*)ibuf+20, 4);
		    for (i=0; i<4; i++) {
		        if (ibuf[i+20] == ']') {
		            pt_len[i] = 0;
			    break;
                        }
		    }
		    plain_len = atoi(pt_len)/8;

		    if (VERBOSE) printf("\nFound plaintext length = 0x%08x", plain_len);
                } else {
		   printf("\nNo plaintext %s\n", ibuf);
		   return -1;
                }        
	    }
	    if (ibuf[0] == '\n') {
                copy_line(ibuf, rfp);
	    }
	    break;

	case 2:  /* read key, used for 100 tests */
	    if (strncmp(ibuf, "K = ", 4) == 0) {
	        fprintf(rfp, RESP_EOL);  /* add separator */
	        fprintf(rfp,"COUNT = %d" RESP_EOL ,counter);
	        copy_line(ibuf, rfp);
	        counter++;
	        step++;
		klen = hex2bin((char*)ibuf+4, key);
		/* test length is valid */
		if ((klen != 16)  && (klen != 24) && (klen != 32)) {
		    printf("\nFound bad key size = %d", klen);
		    return -1;
                }
	    }
   	    break;

	case 3:  /* read data in */
	    /* dir = 1, wrap = 2, unwrap */
	    if (dir) {
	        /* get the plaintext */
	        if (strncmp(ibuf, "P = ", 4) == 0) {
	            ready = 1;
		    copy_line(ibuf, rfp);
		    plen = hex2bin((char*)ibuf+4, plaintext);
		    /* should match the length parsed above */
		    if (plen != plain_len) {
		        printf("\nplaintext length does not match = 0x%08x 0x%08x", plen, plain_len);
		        return -1;
		    }
		    if (VERBOSE) printf("\nFound real plaintext length = 0x%08x", plen);
		}
	    } else {
	        /* get the ciphertext */
	        if (strncmp(ibuf, "C = ", 4) == 0) {
	            ready = 1;
		    copy_line(ibuf, rfp);
		    clen = hex2bin((char*)ibuf+4, ciphertext);
		    if (VERBOSE) printf("\nFound real ciphertext length = 0x%08x", clen);
                }
            }
	    if (ready) {
	        out = malloc(4096);  /* overkill for buffer overrun */
		if (!out) {
		    printf("\nFailed to malloc out");
		    return -1;
		}
	        if (VERBOSE) printf("\nProcessing with padding = %d, dir = %d", padding, dir);

		/* no padding uses EVP */
		switch (klen)
		{
		    case 16:
		        cipher = EVP_aes_128_wrap();
			break;
		    case 24:
		        cipher = EVP_aes_192_wrap();
			break;
		    case 32:
			cipher = EVP_aes_256_wrap();
			break;
		    default:
		        printf("\nIllegal key size %d", klen);
			return -1;
	        }
		/* Padding done here */
		if (padding) {
	            if (dir) {
		        FIPS_cipher_ctx_init(&cctx);

			if (FIPS_cipherinit(&cctx, cipher, key, NULL, dir) <= 0) {
			    printf("\nFIPS_cipherint failed");
		            return -1;
			}

			len = FIPS_cipher(&cctx, out, plaintext, plen);
			if (len <= 0) {
		            printf("\nFIPS_cipher failed on plaintext");
		            return -1;
			}

			if (VERBOSE) printf("\nWrap out = 0x%08x, len = %d\n", *(int *)out, len);
			OutputValue("C", out, len, rfp, 0);
			step = 2;
                    } else {

		        FIPS_cipher_ctx_init(&cctx);
		        if (FIPS_cipherinit(&cctx, cipher, key, NULL, dir) <= 0) {
		            printf("\nFIPS_cipherint failed");
		            return -1;
                        }
		        cctx.flags |= EVP_CIPHER_CTX_FLAG_UNWRAP_WITHPAD;
		        len = FIPS_cipher(&cctx, out, ciphertext, clen);

	                if (VERBOSE) printf("\nUnWrap out = 0x%08x, len = %d\n", *(int *)out, len);
		        /* some are supposed to fail */
		        if (len > 0) {
		            OutputValue("P", out, len, rfp, 0);
                        } else {
		            fprintf(rfp,"FAIL" RESP_EOL); 
		        }
		        step = 2;
                    }
		    FIPS_cipher_ctx_cleanup(&cctx);

	        } else {
	            /* No pading here */
	            if (dir) {
		        FIPS_cipher_ctx_init(&cctx);

		        if (FIPS_cipherinit(&cctx, cipher, key, NULL, dir) <= 0) {
		            printf("\nFIPS_cipherint failed");
		            return -1;
                        }

		        len = FIPS_cipher(&cctx, out, plaintext, plen);
		        if (len <= 0) {
		            printf("\nFIPS_cipher failed on plaintext");
		            return -1;
		        }

	                if (VERBOSE) printf("\nWrap out = 0x%08x, len = %d\n", *(int *)out, len);
		        OutputValue("C", out, len, rfp, 0);
		        step = 2;
                    } else {

		        FIPS_cipher_ctx_init(&cctx);
		        if (FIPS_cipherinit(&cctx, cipher, key, NULL, dir) <= 0) {
		            printf("\nFIPS_cipherint failed");
		            return -1;
                        }
		        len = FIPS_cipher(&cctx, out, ciphertext, clen);

	                if (VERBOSE) printf("\nUnWrap out = 0x%08x, len = %d\n", *(int *)out, len);
		        /* some are supposed to fail */
		        if (len > 0) {
		            OutputValue("P", out, len, rfp, 0);
                        } else {
		            fprintf(rfp,"FAIL" RESP_EOL); 
		        }
		        step = 2;
                    }
		    FIPS_cipher_ctx_cleanup(&cctx);
                }

	        if (out) {
	            free(out);
	        }
	        /* each length has 100 passes */
	        if (counter == LOOP_COUNTER) {
	            counter = 0;
		    ready = 0;
		    step = 1;
                }	    
	    }
	    break;
        }
    } 

    free(key);
    free(plaintext);
    free(ciphertext);

    if (rfp)
	fclose(rfp);
    if (afp)
	fclose(afp);

    return err;
}

/*
 * --------------------------------------------------
 * Processes a single file
 * --------------------------------------------------
 */

#ifdef FIPS_ALGVS
int fips_kwvs_main(int argc, char **argv)
#else
int main(int argc, char **argv)
#endif
    {
    char *rspfile = NULL;
    char fn[250] = "";
    int dir = -1;
    int padding = 0;
    fips_algtest_init();

    if (VERBOSE) printf("\nKWVS start: %s\n", argv[1]);
    if (argc > 1)
	{
	/* AES unwrap */
	if (strcasecmp(argv[1], "-d") == 0)
	    {
	    dir = 0;
	    padding = 0;
	    }
	/* AES wrap */
	else if (strcasecmp(argv[1], "-e") == 0)
	    {
	    dir = 1;
	    padding = 0;
	    }
	/* AES unwrap with padding */
	if (strcasecmp(argv[1], "-dp") == 0)
	    {
	    dir = 0;
	    padding = 1;
	    }
	/* AES wrap with padding */
	else if (strcasecmp(argv[1], "-ep") == 0)
	    {
	    dir = 1;
	    padding = 1;
	    }
	/* no argument */
	else if (dir == -1) {
    	    printf("Invalid parameter: %s\n", argv[1]);
	    return 0;
            }
	if (argc < 2)
	    {
	    printf("Missing parameter\n");
	    return 0;
	    }
	}

	if (VERBOSE)
	    printf("Processing: %s\n", argv[2]);

	strcpy(fn, argv[2]);
	rspfile = argv[3];
	if (proc_kw_file(fn, rspfile, dir, padding))
	    {
	    printf(">>> Processing failed for: %s <<<\n", fn);
	    }

    return 0;
    }
#endif
