/*------------------------------------------------------------------
 * fips/rsa/fips_rsa2gtest.c - RSA2 Key Generation algorithm processing
 *
 * July 2013 Barry Fussell
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */


#define OPENSSL_FIPSAPI

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>
#include <openssl/err.h>
#include <openssl/bn.h>
#include <openssl/rand.h>

#ifndef OPENSSL_FIPS

int main(int argc, char *argv[])
{
    printf("No FIPS RSA2 support\n");
    return(0);
}

#else

#include <openssl/rsa.h>
#include <openssl/fips.h>
#include "fips_utl.h"

int rsa2_test(FILE *out, FILE *in);


#ifdef FIPS_ALGVS
int fips_rsa2gtest_main(int argc, char **argv)
#else
int main(int argc, char **argv)
#endif
	{
	FILE *in = NULL, *out = NULL;

	int ret = 1;

	fips_algtest_init();

	if (argc == 1)
		in = stdin;
	else
		in = fopen(argv[1], "r");

	if (argc < 2)
		out = stdout;
	else
		out = fopen(argv[2], "w");

	if (!in)
		{
		fprintf(stderr, "FATAL input initialization error\n");
		goto end;
		}

	if (!out)
		{
		fprintf(stderr, "FATAL output initialization error\n");
		goto end;
		}

	if (!rsa2_test(out, in))
		{
		fprintf(stderr, "FATAL RSA2GTEST file processing error\n");
		goto end;
		}
	else
		ret = 0;

	end:

	if (in && (in != stdin))
		fclose(in);
	if (out && (out != stdout))
		fclose(out);

	return ret;

	}

#define RSA_TEST_MAXLINELEN	10240

typedef struct {
	int b1;
	int b2;
	int b3;
	int b4;
}BITLEN;

BITLEN b2048[] = {
{160, 272, 240, 184},
{352, 142, 280, 213},
{240, 243, 216, 223},
{344, 149, 232, 205},
{336, 147, 176, 299},
{328, 159, 176, 155},
{240, 147, 344, 150},
{184, 181, 168, 208},
{344, 144, 152, 289},
{296, 150, 160, 292},
{200, 164, 224, 219},
{312, 147, 296, 178},
{240, 159, 304, 142},
{264, 213, 160, 243},
{352, 143, 224, 143},
{280, 168, 160, 233},
{320, 151, 240, 242},
{152, 264, 320, 152},
{232, 192, 280, 151},
{216, 274, 152, 294},
{328, 165, 344, 141},
{344, 144, 144, 258},
{168, 188, 144, 166},
{168, 319, 272, 217},
{344, 148, 344, 150}
};

BITLEN b3072[] = {
{248, 418, 416, 188},
{488, 245, 216, 464},
{288, 234, 472, 191},
{224, 220, 552, 196},
{560, 182, 208, 462},
{528, 173, 480, 178},
{232, 359, 232, 442},
{232, 179, 408, 202},
{296, 210, 328, 281},
{200, 423, 256, 212},
{376, 200, 440, 178},
{352, 210, 352, 378},
{296, 361, 232, 385},
{176, 535, 520, 178},
{384, 245, 344, 323},
{376, 246, 256, 413},
{440, 307, 192, 191},
{336, 339, 400, 187},
{216, 407, 560, 186},
{256, 171, 368, 255},
{480, 265, 536, 202},
{456, 256, 232, 220},
{424, 238, 536, 173},
{568, 172, 368, 206},
{488, 232, 456, 206},
};

int rsa2_test(FILE *out, FILE *in)
	{
	char *linebuf, *olinebuf, *p, *q;
	char *keyword, *value;
	RSA *rsa = NULL;
	BIGNUM *e = NULL;
	int ret = 0, counter = 0, i;
	int lnum = 0;
	int bitlen1 = 0, bitlen2 = 0, bitlen3 = 0, bitlen4 = 0, keylen = 0;
	unsigned char seed[1024];
	unsigned int seed_len = 0;


	olinebuf = OPENSSL_malloc(RSA_TEST_MAXLINELEN);
	linebuf = OPENSSL_malloc(RSA_TEST_MAXLINELEN);

	if (!linebuf || !olinebuf)
		goto error;

	while (fgets(olinebuf, RSA_TEST_MAXLINELEN, in))
		{
		lnum++;
		strcpy(linebuf, olinebuf);
		keyword = linebuf;
		/* Skip leading space */
		while (isspace((unsigned char)*keyword))
			keyword++;

		/* Look for = sign */
		p = strchr(linebuf, '=');

		/* If no = or starts with [ (for [foo = bar] line) just copy */
		if (!p || *keyword=='[')
			{
			if (fputs(olinebuf, out) < 0)
				goto error;

			if (!p)
			    continue;

		        q = p - 1;

			/* Remove trailing space */
			while (isspace((unsigned char)*q))
			    *q-- = 0;

		        *p = 0;
			value = p + 1;

			/* Remove leading space from value */
			while (isspace((unsigned char)*value))
			    value++;

			/* Remove trailing space from value */
			p = value + strlen(value) - 1;

			while (*p == '\n' || isspace((unsigned char)*p))
			    *p-- = 0;

			/* Look for [mod = XXX] for key length */
			if (!strcmp(keyword, "[mod"))
			    {
			    p = value + strlen(value) - 1;
			    if (*p != ']')
				goto parse_error;
			    *p = 0;
			    keylen = atoi(value);
			    if (keylen < 0)
				goto parse_error;
			    }
			continue;
			}

		q = p - 1;

		/* Remove trailing space */
		while (isspace((unsigned char)*q))
			*q-- = 0;

		*p = 0;
		value = p + 1;

		/* Remove leading space from value */
		while (isspace((unsigned char)*value))
			value++;

		/* Remove trailing space from value */
		p = value + strlen(value) - 1;

		while (*p == '\n' || isspace((unsigned char)*p))
			*p-- = 0;

#if 0   /* keeping the old code around in case the format changes back */
		if (!strcmp(keyword, "e"))
			{
			if (e || !do_hex2bn(&e,value)) {
				goto parse_error;
				}
			}
		else if (!strcmp(keyword, "seed"))
			{
			seed_len = hex2bin(value, seed);
			}

		else if (!strcmp(keyword, "bitlen1"))
			{
			bitlen1 = atoi(value);
			}
		else if (!strcmp(keyword, "bitlen2"))
			{
			bitlen2 = atoi(value);
			}
		else if (!strcmp(keyword, "bitlen3"))
			{
			bitlen3 = atoi(value);
			}
		else if (!strcmp(keyword, "bitlen4"))
			{
			bitlen4 = atoi(value);
			}
		else    {
			goto parse_error;
		        }

	        fputs(olinebuf, out);

		if ((bitlen1 !=0) && (bitlen2 != 0) && 
		    (bitlen3 != 0) && (bitlen4 != 0) && 
		    (seed_len != 0) && (e != NULL) && (keylen != 0)) {
#else /* end old vector format, start new */


		if (!strcmp(keyword, "n"))
			{
			counter = atoi(value);
			}

		if (counter != 25) {
		    fprintf(stderr, "FATAL bad loop counter  %d\n", counter);
		    goto parse_error;
		}
	        fputs(olinebuf, out);

	        for (i=0; i<counter; i++) {
		    e = BN_new();
		    BN_set_word(e, 0x10001);

		    /* only support 2048 and 3072 presently */
		    if (keylen == 2048) {
		        bitlen1 = b2048[i].b1;
		        bitlen2 = b2048[i].b2;
		        bitlen3 = b2048[i].b3;
		        bitlen4 = b2048[i].b4;
			seed_len = 28;
		    } else {
		        bitlen1 = b3072[i].b1;
		        bitlen2 = b3072[i].b2;
		        bitlen3 = b3072[i].b3;
		        bitlen4 = b3072[i].b4;
			seed_len = 32;
		    }

		    /* generate a random seed */
		    if (RAND_bytes(seed, seed_len) <= 0) {
		        fprintf(stderr, "\nUnable to generate seed");
			goto error;
		    }
		    do_bn_print_name(out, "e", e);
		    OutputValue("seed", seed, seed_len, out, 0);
		    fprintf(out, "bitlen1 = %d\n", bitlen1);
		    fprintf(out, "bitlen2 = %d\n", bitlen2);
		    fprintf(out, "bitlen3 = %d\n", bitlen3);
		    fprintf(out, "bitlen4 = %d\n", bitlen4);
#endif
		    rsa = FIPS_rsa_new();
		    ret = rsa_generate_key_internal (&rsa->p, &rsa->q, 
						     &rsa->n, &rsa->d,
						     seed, seed_len,
						     bitlen1, bitlen2, 
						     bitlen3, bitlen4,
						     e, keylen, NULL);
		    if (ret != 1) {
		        fprintf(stderr, "\nERROR: during key generation\n");
			goto error;
                    }	

		    if ((rsa->q != NULL) && (rsa->p != NULL) && (rsa->n != NULL) && (rsa->d != NULL)) {
		        do_bn_print_name(out, "p", rsa->p);
		        do_bn_print_name(out, "q", rsa->q);
		        do_bn_print_name(out, "n", rsa->n);
		        do_bn_print_name(out, "d", rsa->d);
		        fprintf(out, "\n");
		    }

		    BN_free(e);
		    e = NULL;
		    FIPS_rsa_free(rsa);
		    rsa = NULL;
		    bitlen1 = bitlen2 = bitlen3 = bitlen4 = seed_len = 0;
		}
        }
   
	ret = 1;
 
	error:

	if (olinebuf)
		OPENSSL_free(olinebuf);
	if (linebuf)
		OPENSSL_free(linebuf);

	if (e)
		BN_free(e);
	if (rsa)
		FIPS_rsa_free(rsa);

	return ret;

	parse_error:

	fprintf(stderr, "FATAL parse error processing line %d\n", lnum);

	goto error;

	}

#endif
